# Generated by roxygen2: do not edit by hand

S3method(format,gargle_oauth_dat)
S3method(gargle_map_cli,"NULL")
S3method(gargle_map_cli,character)
S3method(gargle_map_cli,default)
S3method(print,gargle_oauth_client)
S3method(print,gargle_oauth_dat)
export(AuthState)
export(Gargle2.0)
export(GceToken)
export(WifToken)
export(bulletize)
export(check_is_service_account)
export(cred_funs_add)
export(cred_funs_clear)
export(cred_funs_list)
export(cred_funs_list_default)
export(cred_funs_set)
export(cred_funs_set_default)
export(credentials_app_default)
export(credentials_byo_oauth2)
export(credentials_external_account)
export(credentials_gce)
export(credentials_service_account)
export(credentials_user_oauth2)
export(field_mask)
export(gargle2.0_token)
export(gargle_api_key)
export(gargle_app)
export(gargle_client)
export(gargle_error_message)
export(gargle_last_content)
export(gargle_last_response)
export(gargle_map_cli)
export(gargle_oauth_cache)
export(gargle_oauth_client)
export(gargle_oauth_client_from_json)
export(gargle_oauth_client_type)
export(gargle_oauth_email)
export(gargle_oauth_sitrep)
export(gargle_oob_default)
export(gargle_verbosity)
export(gce_access_token)
export(gce_instance_service_accounts)
export(init_AuthState)
export(local_cred_funs)
export(local_gargle_verbosity)
export(oauth_app_from_json)
export(oauth_external_token)
export(request_build)
export(request_develop)
export(request_make)
export(request_retry)
export(response_as_json)
export(response_process)
export(secret_decrypt_json)
export(secret_encrypt_json)
export(secret_has_key)
export(secret_make_key)
export(secret_read_rds)
export(secret_write_rds)
export(tidyverse_api_key)
export(tidyverse_app)
export(tidyverse_client)
export(token_email)
export(token_fetch)
export(token_tokeninfo)
export(token_userinfo)
export(with_cred_funs)
export(with_gargle_verbosity)
import(fs)
import(rlang)
importFrom(glue,glue)
importFrom(glue,glue_collapse)
importFrom(glue,glue_data)
importFrom(lifecycle,deprecated)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# gargle

<!-- badges: start -->

[![CRAN
status](https://www.r-pkg.org/badges/version/gargle)](https://cran.r-project.org/package=gargle)
[![Codecov test
coverage](https://codecov.io/gh/r-lib/gargle/graph/badge.svg)](https://app.codecov.io/gh/r-lib/gargle)
[![R-CMD-check](https://github.com/r-lib/gargle/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/r-lib/gargle/actions/workflows/R-CMD-check.yaml)
<!-- badges: end -->

The goal of gargle is to take some of the agonizing pain out of working
with Google APIs. This includes functions and classes for handling
common credential types and for preparing, executing, and processing
HTTP requests.

The target user of gargle is an *R package author* who is wrapping one
of the ~250 Google APIs listed in the [APIs
Explorer](https://developers.google.com/apis-explorer). gargle aims to
play roughly the same role as [Google’s official client
libraries](https://developers.google.com/api-client-library/), but for
R. gargle may also be useful to useRs making direct calls to Google
APIs, who are prepared to navigate the details of low-level API access.

gargle’s functionality falls into two main domains:

- **Auth.** The `token_fetch()` function calls a series of concrete
  credential-fetching functions to obtain a valid access token (or it
  quietly dies trying).
  - This covers explicit service accounts, application default
    credentials, Google Compute Engine, (experimentally) workload
    identity federation, and the standard OAuth2 browser flow.
  - gargle offers the `Gargle2.0` class, which extends `httr::Token2.0`.
    It is the default class for user OAuth 2.0 credentials. There are
    two main differences from `httr::Token2.0`: greater emphasis on the
    user’s email (e.g. Google identity) and default token caching is at
    the user level.
- **Requests and responses**. A family of functions helps to prepare
  HTTP requests, (possibly with reference to an API spec derived from a
  Discovery Document), make requests, and process the response.

See the [articles](https://gargle.r-lib.org/articles/) for holistic
advice on how to use gargle.

## Installation

You can install the released version of gargle from
[CRAN](https://CRAN.R-project.org) with:

``` r
install.packages("gargle")
```

And the development version from [GitHub](https://github.com/) with:

``` r
# install.packages("pak")
pak::pak("r-lib/gargle")
```

## Basic usage

gargle is a low-level package and does not do anything visibly exciting
on its own. But here’s a bit of usage in an interactive scenario where a
user confirms they want to use a specific Google identity and loads an
OAuth2 token.

``` r
library(gargle)

token <- token_fetch()
#> The gargle package is requesting access to your Google account.
#> Enter '1' to start a new auth process or select a pre-authorized account.
#> 1: Send me to the browser for a new auth process.
#> 2: janedoe_personal@gmail.com
#> 3: janedoe@example.com
#> Selection: 2

token
#> ── <Token (via gargle)> ─────────────────────────────────────────────────────
#> oauth_endpoint: google
#>            app: gargle-clio
#>          email: janedoe_personal@gmail.com
#>         scopes: ...userinfo.email
#>    credentials: access_token, expires_in, refresh_token, scope, token_type, id_token
```

Here’s an example of using request and response helpers to make a
one-off request to the [Web Fonts Developer
API](https://developers.google.com/fonts/docs/developer_api). We show
the most popular web font families served by Google Fonts.

``` r
library(gargle)

req <- request_build(
  method = "GET",
  path = "webfonts/v1/webfonts",
  params = list(
    sort = "popularity"
  ),
  key = gargle_api_key(),
  base_url = "https://www.googleapis.com"
)
resp <- request_make(req)
out <- response_process(resp)

out <- out[["items"]][1:8]
sort(vapply(out, function(x) x[["family"]], character(1)))
#> [1] "Google Sans"  "Inter"        "Lato"         "Montserrat"   "Noto Sans JP"
#> [6] "Open Sans"    "Poppins"      "Roboto"
```

Please note that the ‘gargle’ project is released with a [Contributor
Code of Conduct](https://gargle.r-lib.org/CODE_OF_CONDUCT.html). By
contributing to this project, you agree to abide by its terms.

[Privacy policy](https://tidyverse.org/google_privacy_policy/)

## -----------------------------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## -----------------------------------------------------------------------------
library(gargle)

## -----------------------------------------------------------------------------
gargle_verbosity()

## -----------------------------------------------------------------------------
# save current value
op <- options(gargle_verbosity = "debug")

gargle_verbosity()

# restore original value
options(op)

## -----------------------------------------------------------------------------
gargle_verbosity()

with_gargle_verbosity(
  "debug",
  gargle_verbosity()
)

gargle_verbosity()

f <- function() {
  local_gargle_verbosity("debug")
  gargle_verbosity()
}

f()

gargle_verbosity()

## -----------------------------------------------------------------------------
# gargle_oauth_sitrep()
# #' > 14 tokens found in this gargle OAuth cache:
# #' '~/Library/Caches/gargle'
# #'
# #' email                         app         scope                          hash...
# #' ----------------------------- ----------- ------------------------------ ----------
# #' abcdefghijklm@gmail.com       thingy      ...bigquery, ...cloud-platform 128f9cc...
# #' buzzy@example.org             gargle-demo                                15acf95...
# #' stella@example.org            gargle-demo ...drive                       4281945...
# #' abcdefghijklm@gmail.com       gargle-demo ...drive                       48e7e76...
# #' abcdefghijklm@gmail.com       tidyverse                                  69a7353...
# #' nopqr@ABCDEFG.com             tidyverse   ...spreadsheets.readonly       86a70b9...
# #' abcdefghijklm@gmail.com       tidyverse   ...drive                       d9443db...
# #' nopqr@HIJKLMN.com             tidyverse   ...drive                       d9443db...
# #' nopqr@ABCDEFG.com             tidyverse   ...drive                       d9443db...
# #' stuvwzyzabcd@gmail.com        tidyverse   ...drive                       d9443db...
# #' efghijklmnopqrtsuvw@gmail.com tidyverse   ...drive                       d9443db...
# #' abcdefghijklm@gmail.com       tidyverse   ...drive.readonly              ecd11fa...
# #' abcdefghijklm@gmail.com       tidyverse   ...bigquery, ...cloud-platform ece63f4...
# #' nopqr@ABCDEFG.com             tidyverse   ...spreadsheets                f178dd8...

## -----------------------------------------------------------------------------
# install.packages("googlesheets4")

## -----------------------------------------------------------------------------
knitr::include_graphics("deleted_client.png")

## -----------------------------------------------------------------------------
# only run the chunk below in settings that are known to be safe, i.e. where
# occasional, incidental failure is OK
can_decrypt <- gargle::secret_has_key("GARGLE_KEY")

## -----------------------------------------------------------------------------
library(gargle)

req <- request_build(
  method = "GET",
  path = "webfonts/v1/webfonts",
  params = list(
    sort = "popularity"
  ),
  key = gargle_api_key(),
  base_url = "https://www.googleapis.com"
)
resp <- request_make(req)
out <- response_process(resp)

lr <- gargle_last_response()
tmp <- tempfile("gargle-last-response-")
saveRDS(lr, tmp)
# you could share this .rds file with a colleague or the gargle maintainer

# how it would look to complete the round trip, i.e. load this on the other end
rt_lr <- readRDS(tmp)

all.equal(lr, rt_lr)

# clean up
unlink("tmp")

## -----------------------------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## -----------------------------------------------------------------------------
library(gargle)

## -----------------------------------------------------------------------------
(path_to_installed_client <- system.file(
  "extdata", "client_secret_installed.googleusercontent.com.json",
  package = "gargle"
))
jsonlite::prettify(scan(path_to_installed_client, what = character()))
(client <- gargle_oauth_client_from_json(path_to_installed_client))
class(client)

(path_to_web_client <- system.file(
  "extdata", "client_secret_web.googleusercontent.com.json",
  package = "gargle"
))
jsonlite::prettify(scan(path_to_web_client, what = character()))
(client <- gargle_oauth_client_from_json(path_to_web_client))
class(client)

## -----------------------------------------------------------------------------
# # BEFORE
# drive_auth_configure <- function(app, path, api_key) {
#   # not showing this code
#   .auth$set_app(app)
#   # more code we're not showing
# }
# 
# drive_oauth_app <- function() .auth$app
# 
# # AFTER
# drive_auth_configure <- function(client, path, api_key, app = deprecated()) {
#   if (lifecycle::is_present(app)) {
#     lifecycle::deprecate_warn(
#       "2.1.0",
#       "drive_auth_configure(app)",
#       "drive_auth_configure(client)"
#     )
#     drive_auth_configure(client = app, path = path, api_key = api_key)
#   }
# 
#   # not showing this code
#   .auth$set_client(client)
#   # more code we're not showing
# }
# 
# drive_oauth_client <- function() .auth$client
# 
# drive_oauth_app <- function() {
#   lifecycle::deprecate_warn(
#     "2.1.0", "drive_oauth_app()", "drive_oauth_client()"
#   )
#   drive_oauth_client()
# }

## -----------------------------------------------------------------------------
# usethis::use_lifecycle()

## -----------------------------------------------------------------------------
# drive_auth <- function(...) {
#   # code not shown
#   cred <- gargle::token_fetch(
#     scopes = scopes,
#     # app = drive_oauth_client() %||% <BUILT_IN_DEFAULT_CLIENT>,   # BEFORE
#     client = drive_oauth_client() %||% <BUILT_IN_DEFAULT_CLIENT>,  # AFTER
#     email = email,
#     path = path,
#     package = "googledrive",
#     cache = cache,
#     use_oob = use_oob,
#     token = token
#   )
#   # code not shown
# }

library(tidyverse)

#' Get versioned IDs from API Discovery Service
#'
#' @return A character vector.
#' @keywords internal
#' @examples
#' get_discovery_ids()
#' grep("drive", get_discovery_ids(), value = TRUE)
#' grep("sheets", get_discovery_ids(), value = TRUE)
#' grep("gmail", get_discovery_ids(), value = TRUE)
#' grep("bigquery", get_discovery_ids(), value = TRUE)
get_discovery_ids <- function() {
  apis <- httr::content(
    httr::GET("https://www.googleapis.com/discovery/v1/apis")
  )
  map_chr(apis[["items"]], "id")
}

#' Form the URL for a Discovery Document
#'
#' @param id Versioned ID string for target API. Use [get_discovery_ids()] to
#'   see them all and find the one you want.
#' @return A URL to a JSON file.
#' @keywords internal
#' @examples
#' make_discovery_url("sheets:v4")
make_discovery_url <- function(id) {
  av <- set_names(as.list(strsplit(id, split = ":")[[1]]), c("api", "version"))
  ## https://developers.google.com/discovery/v1/reference/apis/getRest
  getRest_url <-
    "https://www.googleapis.com/discovery/v1/apis/{api}/{version}/rest"
  glue::glue_data(av, getRest_url)
}

#' List (likely) Discovery Documents in a local folder
#'
#' @param id Optional ID string, possibly versioned, for target API. Use
#'   [get_discovery_ids()] to see them all and find the one you want.
#' @param path Optional directory in which to look. Defaults to `data-raw`
#'   within the current project.
#'
#' @return Files whose names "look like" a Discovery Document
#' @keywords internal
#' @examples
#' list_discovery_documents()
#' list_discovery_documents("sheets")
#' list_discovery_documents("sheets:v4")
list_discovery_documents <- function(id = NULL, path = NULL) {
  path <- path %||% rprojroot::find_package_root_file("data-raw")
  if (!is.null(id)) {
    if (!grepl(":", id)) {
      id <- glue::glue("{id}:v[[:digit:]]+")
    }
    id <- sub(":", "-", id)
  }
  id <- id %||% "[[:alnum:]]+[-][[:alnum:]]+"
  date <- "[[:digit:]]{4}-[[:digit:]]{2}-[[:digit:]]{2}"
  regexp <- glue::glue("{id}_{date}[.]json")
  fs::dir_ls(path = path, regexp = regexp) |>
    fs::path_rel(path)
}

#' Download a Discovery Document
#'
#' @param id Versioned ID string for target API. Use [get_discovery_ids()] to
#'   see them all and find the one you want.
#' @param path Target filepath. Default filename is formed from the API's
#'   versioned ID and the Discovery Document's revision date. Default parent
#'   directory is the current package's `data-raw/` directory, if such exists,
#'   or current working directory, otherwise.
#'
#' @return Filepath
#' @keywords internal
#' @examples
#' download_discovery_document("drive:v3")
#' download_discovery_document("sheets:v4")
#' download_discovery_document("gmail:v1")
#' download_discovery_document("bigquery:v2")
#' download_discovery_document("docs:v1")
#' download_discovery_document("youtube:v3")
download_discovery_document <- function(id, path = NULL) {
  url <- make_discovery_url(id)
  dd <- httr::GET(url)
  httr::stop_for_status(dd, glue::glue("find Discovery Document for ID '{id}'"))

  if (is.null(path)) {
    dd_content <- httr::content(dd)
    api_date <- dd_content[c("revision", "id")]
    api_date <- c(
      id = sub(":", "-", api_date$id),
      revision = as.character(as.Date(api_date$revision, format = "%Y%m%d"))
    )
    json_filename <- fs::path(paste(api_date, collapse = "_"), ext = "json")
    data_raw <- rprojroot::find_package_root_file("data-raw")
    path <- if (fs::dir_exists(data_raw)) {
      fs::path(data_raw, json_filename)
    } else {
      json_filename
    }
  }

  writeLines(httr::content(dd, as = "text"), path)
  path
}

#' Read a Discovery Document
#'
#' @param path Path to a JSON Discovery Document
#'
#' @return A list
#' @examples
#' drive <- "data-raw/drive-v3_2019-02-07.json"
#' dd <- read_discovery_document(drive)
read_discovery_document <- function(path) {
  jsonlite::fromJSON(path)
}

#' Get raw methods
#'
#' https://developers.google.com/discovery/v1/using#discovery-doc-methods
#'
#' @param dd List representing a Discovery Document
#'
#' @return a named list with one element per method
#' @examples
#' drive <- "data-raw/drive-v3_2019-02-07.json"
#' dd <- read_discovery_document(drive)
#' ee <- get_raw_methods(dd)
get_raw_methods <- function(dd) {
  dd |>
    pluck("resources") |>
    map("methods") |>
    flatten() |>
    set_names(map_chr(., "id"))
}

#' Groom method properties
#'
#' Tweak raw method properties to make them more useful to us downstream:
#'
#'   * Prepend the API's `servicePath` to `path`s.
#'   * Remove the constant stem `"https://www.googleapis.com/auth/"` from
#'     scopes and collapse multiple scopes into one comma-separated string.
#'   * Elevate any `$ref` part of `request` or `response` to be the actual
#'     data for `request` or `response`.
#'   * Reorder the properties so they appear in a predictable order. However,
#'     we do not turn missing properties into explicitly missing properties,
#'     i.e. we don't guarantee all methods have the same properties.
#'
#' We don't touch the `parameters` list here, because it needs enough work to
#' justify having separate functions for that.
#'
#' @param methods A named list of raw methods, from [get_raw_methods()]
#' @param dd A Discovery Document as a list, from [read_discovery_document()]
#'
#' @return A named list of "less raw" methods
groom_properties <- function(method, dd) {
  method$path <- fs::path(dd$servicePath, method$path)

  condense_scopes <- function(scopes) {
    scopes |>
      str_remove("https://www.googleapis.com/auth/") |>
      str_c(collapse = ", ")
  }
  method$scopes <- condense_scopes(method$scopes)

  ## I am currently ignoring the fact that `request` sometimes has both
  ## a `$ref` and a `parameterName` part in the original JSON
  if (has_name(method, "request")) {
    method$request <- method$request$`$ref`
  }
  if (has_name(method, "response")) {
    method$response <- method$response$`$ref`
  }

  # all of the properties in the RestMethod schema, in order of usefulness
  property_names <- c(
    "id",
    "httpMethod",
    "path",
    "parameters",
    "scopes",
    "description",
    "request",
    "response",
    "mediaUpload",
    "supportsMediaDownload",
    "supportsMediaUpload",
    "useMediaDownloadService",
    "etagRequired",
    "parameterOrder",
    "supportsSubscription"
  )

  method[intersect(property_names, names(method))]
}

#' Expand schema placeholders
#'
#' Adds the properties associated with a `request` schema to a method's
#' parameter list.
#'
#' Some methods can send an instance of a API resource in the body of a request.
#' This is indicated by the presence of a schema in the method's `request`
#' property. For example, the `drive.files.copy` method permits a "Files
#' resource" in the request body. This is how you convey the desired `name` of
#' the new copy.
#'
#' In practice, this means you can drop such metadata in the body. That is, you
#' don't actually have to label this explicitly as having `kind = drive#file`
#' (although that would probably be more proper!), nor do you have to include
#' all the possible pieces of metadata that constitute a "Files resource". Just
#' specify the bits that you need to.
#'
#' https://developers.google.com/drive/api/v3/reference/files/copy
#' https://developers.google.com/drive/api/v3/reference/files#resource
#'
#' This function consults the method's `request` and, if it holds a schema, the
#' schema metadata is appended to the method's existing parameters. This way our
#' request building functions recognize the keys and know that such info belongs
#' in the body (vs. the url or the query).
#'
#' @param method A single method
#' @param dd A Discovery Document as a list, from [read_discovery_document()]
#'
#' @return The input method, but with a potentially expanded parameter list.
add_schema_params <- function(method, dd) {
  req <- pluck(method, "request")
  if (is.null(req)) {
    return(method)
  }

  id <- method$id
  schema_params <- dd[[c("schemas", req, "properties")]]
  schema_params <- modify(schema_params, ~ `[[<-`(.x, "location", "body"))

  message(glue::glue("{id} gains {req} schema params\n"))
  method$parameters <- c(method$parameters, schema_params)
  method
}

#' Add API-wide parameters
#'
#' Certain parameters are sensible for any request to a specific API and,
#' indeed, are usually common across APIs. Examples are "fields", "key", and
#' "oauth_token". This function appends these parameters to a method's parameter
#' list. Yes, this means some info is repeated in all methods, but this way our
#' methods are more self-contained and our request building functions can be
#' simpler.
#'
#' @param method A single method
#' @param dd A Discovery Document as a list, from [read_discovery_document()]
#'
#' @return The input method, but with an expanded parameter list.
add_global_params <- function(method, dd) {
  method[["parameters"]] <- c(method[["parameters"]], dd[["parameters"]])
  method
}

## -----------------------------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## -----------------------------------------------------------------------------
# # googledrive::
# drive_auth <- function(email = gargle::gargle_oauth_email(),
#                        path = NULL,
#                        scopes = "https://www.googleapis.com/auth/drive",
#                        cache = gargle::gargle_oauth_cache(),
#                        use_oob = gargle::gargle_oob_default(),
#                        token = NULL) {
#   # this catches a common error, where the user passes JSON for an OAuth client
#   # to the `path` argument, which only expects a service account token
#   gargle::check_is_service_account(path, hint = "drive_auth_configure")
# 
#   cred <- gargle::token_fetch(
#     scopes = scopes,
#     client = drive_oauth_client() %||% <BUILT_IN_DEFAULT_CLIENT>,
#     email = email,
#     path = path,
#     package = "googledrive",
#     cache = cache,
#     use_oob = use_oob,
#     token = token
#   )
#   if (!inherits(cred, "Token2.0")) {
#     # throw an informative error here
#   }
#   .auth$set_cred(cred)
#   .auth$set_auth_active(TRUE)
# 
#   invisible()
# }

## -----------------------------------------------------------------------------
# .auth <- NULL

## -----------------------------------------------------------------------------
.onLoad <- function(libname, pkgname) {
  utils::assignInMyNamespace(
    ".auth",
    gargle::init_AuthState(package = "googledrive", auth_active = TRUE)
  )
  
  # other stuff
}

## -----------------------------------------------------------------------------
# library(googledrive)
# 
# # first: download the OAuth client as a JSON file
# drive_auth_configure(
#   path = "/path/to/the/JSON/that/was/downloaded/from/gcp/console.json"
# )
# 
# drive_oauth_client()
# #> <gargle_oauth_client>
# #> name: acme-corp-google-client
# #> id: 123456789.apps.googleusercontent.com
# #> secret: <REDACTED>
# #> type: installed
# #> redirect_uris: http://localhost

## -----------------------------------------------------------------------------
# library(googledrive)
# 
# drive_auth_configure(api_key = "123456789")
# 
# drive_api_key()
# #> "123456789"

## -----------------------------------------------------------------------------
# # googledrive::
# drive_auth(email = "janedoe_work@gmail.com")

## -----------------------------------------------------------------------------
# # googledrive::
# drive_auth <- function(email = gargle::gargle_oauth_email(),
#                        path = NULL,
#                        scopes = "https://www.googleapis.com/auth/drive",
#                        cache = gargle::gargle_oauth_cache(),
#                        use_oob = gargle::gargle_oob_default(),
#                        token = NULL) { ... }

## -----------------------------------------------------------------------------
# # googledrive::
# drive_auth(scopes = "https://www.googleapis.com/auth/drive.readonly")

## -----------------------------------------------------------------------------
# # googledrive::
# request_generate <- function(endpoint = character(),
#                              params = list(),
#                              key = NULL,
#                              token = drive_token()) {
#   ept <- drive_endpoint(endpoint)
#   if (is.null(ept)) {
#     # throw error about unrecognized endpoint
#   }
# 
#   ## modifications specific to googledrive package
#   params$key <- key %||% params$key %||%
#     drive_api_key() %||% <BUILT_IN_DEFAULT_API_KEY>
#   if (!is.null(ept$parameters$supportsAllDrives)) {
#     params$supportsAllDrives <- TRUE
#   }
# 
#   req <- gargle::request_develop(endpoint = ept, params = params)
#   gargle::request_build(
#     path = req$path,
#     method = req$method,
#     params = req$params,
#     body = req$body,
#     token = token
#   )
# }

## -----------------------------------------------------------------------------
# # googledrive::
# drive_token <- function() {
#   if (isFALSE(.auth$auth_active)) {
#     return(NULL)
#   }
#   if (!drive_has_token()) {
#     drive_auth()
#   }
#   httr::config(token = .auth$cred)
# }

## -----------------------------------------------------------------------------
# # googledrive::
# drive_has_token <- function() {
#   inherits(.auth$cred, "Token2.0")
# }

## -----------------------------------------------------------------------------
# library(googledrive)
# library(googlesheets4)
# 
# drive_auth(email = "jane_doe@example.com") # gets a suitably scoped token
#                                            # and stashes for googledrive use
# 
# gs4_auth(token = drive_token())            # registers token with googlesheets4
# 
# # now work with both packages freely ...

## -----------------------------------------------------------------------------
# library(googledrive)
# 
# drive_auth(email = "janedoe_work@gmail.com")
# # do stuff with Google Drive here, with Jane Doe's "work" account
# 
# drive_auth(email = "janedoe_personal@gmail.com")
# # do other stuff with Google Drive here, with Jane Doe's "personal" account
# 
# drive_auth(path = "/path/to/a/service-account.json")
# # do other stuff with Google Drive here, using a service account

## -----------------------------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## -----------------------------------------------------------------------------
# library(googlesheets4)
# 
# gs4_auth_configure(api_key = "YOUR_API_KEY_GOES_HERE")
# gs4_deauth()
# 
# # now you can read public resources, such as official example Sheets,
# # without any need for auth
# gs4_example("gapminder") |>
#   read_sheet()

## -----------------------------------------------------------------------------
# library(googledrive)
# 
# google_client <- gargle::gargle_oauth_client_from_json(
#   path = "/path/to/the/JSON/that/was/downloaded/from/gcp/console.json",
#   name = "acme-corp-google-client"
# )
# drive_auth_configure(client = google_client)
# 
# # now any new OAuth tokens are obtained with the configured client

## -----------------------------------------------------------------------------
# # googledrive
# drive_auth(path = "/path/to/your/service-account-token.json")

## -----------------------------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## -----------------------------------------------------------------------------
library(gargle)

## -----------------------------------------------------------------------------
# token_fetch(scopes, ...)

## -----------------------------------------------------------------------------
writeLines(names(cred_funs_list()))

## -----------------------------------------------------------------------------
# token_fetch(token = <TOKEN2.0>)
# 
# credentials_byo_oauth2(
#   token = <TOKEN2.0>
# )

## -----------------------------------------------------------------------------
# token_fetch(scopes = <SCOPES>, path = "/path/to/your/service-account.json")
# 
# # credentials_byo_oauth2() fails because no `token`,
# # which leads to this call:
# credentials_service_account(
#   scopes = <SCOPES>,
#   path = "/path/to/your/service-account.json"
# )

## -----------------------------------------------------------------------------
# token_fetch(scopes = <SCOPES>, path = "/path/to/your/external-account.json")
# 
# # credentials_byo_oauth2() fails because no `token`,
# # credentials_service_account() fails because the JSON provided via
# #   `path` is not of type "service_account",
# # which leads to this call:
# credentials_external_account(
#   scopes = <SCOPES>,
#   path = "/path/to/your/external-account.json"
# )

## -----------------------------------------------------------------------------
# token_fetch(scopes = <SCOPES>)
# 
# # credentials_byo_oauth2() fails because no `token`,
# # credentials_service_account() fails because no `path`,
# # credentials_external_account() fails because no `path`,
# # which leads to this call:
# credentials_app_default(
#   scopes = <SCOPES>
# )

## -----------------------------------------------------------------------------
# ${GOOGLE_APPLICATION_CREDENTIALS}
# ${CLOUDSDK_CONFIG}/application_default_credentials.json
# 
# # on Windows:
# %APPDATA%\gcloud\application_default_credentials.json
# %SystemDrive%\gcloud\application_default_credentials.json
# C:\gcloud\application_default_credentials.json
# 
# # on not-Windows:
# ~/.config/gcloud/application_default_credentials.json

## -----------------------------------------------------------------------------
# token_fetch(scopes = <SCOPES>)
# # or perhaps
# token_fetch(scopes = <SCOPES>, service_account = <SERVICE_ACCOUNT>)
# 
# # credentials_byo_oauth2() fails because no `token`,
# # credentials_service_account() fails because no `path`,
# # credentials_external_account() fails because no `path`,
# # credentials_app_default() fails because no ADC found,
# # which leads to one of these calls:
# credentials_gce(
#   scopes = <SCOPES>,
#   service_account = "default"
# )
# # or
# credentials_gce(
#   scopes = <SCOPES>,
#   service_account = <SERVICE_ACCOUNT>
# )

## -----------------------------------------------------------------------------
# token_fetch(scopes = <SCOPES>)
# 
# # credentials_byo_oauth2() fails because no `token`,
# # credentials_service_account() fails because no `path`,
# # credentials_external_account() fails because no `path`,
# # credentials_app_default() fails because no ADC found,
# # credentials_gce() fails because not on GCE,
# # which leads to this call:
# credentials_user_oauth2(
#   scopes = <SCOPES>,
#   app = <OAUTH_APP>,
#   package = "<PACKAGE>"
# )

## -----------------------------------------------------------------------------
# # user initiates auth or does something that triggers it indirectly
# THINGY_auth()
# 
# # which then calls
# gargle::token_fetch(
#   scopes  = <SCOPES_NEEDED_FOR_THE_THINGY_API>,
#   app     = thingy_app(),
#   package = "thingyr"
# )
# 
# # which leads to this call:
# credentials_user_oauth2(
#   scopes  = <SCOPES_NEEDED_FOR_THE_THINGY_API>,
#   app     = thingy_app(),
#   package = "thingyr"
# )

## -----------------------------------------------------------------------------
# gargle2.0_token(
#   email   = gargle_oauth_email(),
#   app     = thingy_app(),
#   package = "thingyr",
#   scope   = <SCOPES_NEEDED_FOR_THE_THINGY_API>,
#   cache   = gargle_oauth_cache()
# )

## -----------------------------------------------------------------------------
# The thingyr package is requesting access to your Google account.
# Enter '1' to start a new auth process or select a pre-authorized account.
# 1: Send me to the browser for a new auth process.
# 2: janedoe_personal@gmail.com
# 3: janedoe@example.com
# 4: janedoe_work@gmail.com
# Selection:

## -----------------------------------------------------------------------------
# thingy_auth(email = "janedoe_work@gmail.com")

## -----------------------------------------------------------------------------
# gargle_oauth_sitrep()
# #> 14 tokens found in this gargle OAuth cache:
# #> ~/Library/Caches/gargle
# #'
# #' email                         app         scopes                         hash...
# #' ----------------------------- ----------- ------------------------------ ----------
# #' abcdefghijklm@gmail.com       thingy      ...bigquery, ...cloud-platform 128f9cc...
# #' buzzy@example.org             gargle-demo                                15acf95...
# #' stella@example.org            gargle-demo ...drive                       4281945...
# #' abcdefghijklm@gmail.com       gargle-demo ...drive                       48e7e76...
# #' abcdefghijklm@gmail.com       tidyverse                                  69a7353...
# #' nopqr@ABCDEFG.com             tidyverse   ...spreadsheets.readonly       86a70b9...
# #' abcdefghijklm@gmail.com       tidyverse   ...drive                       d9443db...
# #' nopqr@HIJKLMN.com             tidyverse   ...drive                       d9443db...
# #' nopqr@ABCDEFG.com             tidyverse   ...drive                       d9443db...
# #' stuvwzyzabcd@gmail.com        tidyverse   ...drive                       d9443db...
# #' efghijklmnopqrtsuvw@gmail.com tidyverse   ...drive                       d9443db...
# #' abcdefghijklm@gmail.com       tidyverse   ...drive.readonly              ecd11fa...
# #' abcdefghijklm@gmail.com       tidyverse   ...bigquery, ...cloud-platform ece63f4...
# #' nopqr@ABCDEFG.com             tidyverse   ...spreadsheets                f178dd8...

## -----------------------------------------------------------------------------
writeLines(names(cred_funs_list()))

## -----------------------------------------------------------------------------
# gargle::cred_funs_add(credentials_gce = NULL)

library(tidyverse)

## necessary only during gargle development to get devtools' shim for
## system.file()
load_all()
ddi_dir <- system.file("discovery-doc-ingest", package = "gargle")

source(fs::path(ddi_dir, "ingest-functions.R"))

x <- download_discovery_document("discovery:v1")
ee <- read_discovery_document(x)

# api-wide parameters
ap <- ee |>
  pluck("parameters") |>
  enframe(name = "property", value = "info") |>
  mutate(info = map(info, enframe)) |>
  unnest(cols = info) |>
  spread(key = name, value = value, convert = TRUE) |>
  select(property, type, description, everything()) |>
  write_csv(fs::path(ddi_dir, "api-wide-parameters.csv"))

# properties of a method
mp <- ee |>
  pluck("schemas", "RestMethod", "properties") |>
  enframe(name = "property", value = "info") |>
  mutate(info = map(info, enframe)) |>
  unnest(cols = info) |>
  spread(key = name, value = value, convert = TRUE) |>
  select(property, type, description, everything()) |>
  write_csv(fs::path(ddi_dir, "method-properties.csv"))

# properties of a parameter of a method
pp <- ee |>
  pluck("schemas", "JsonSchema", "properties") |>
  enframe(name = "property", value = "info") |>
  mutate(info = map(info, enframe)) |>
  unnest(cols = info) |>
  spread(key = name, value = value, convert = TRUE) |>
  select(property, type, description, everything()) |>
  write_csv(fs::path(ddi_dir, "parameter-properties.csv"))

make_humane_table <- function(df) {
  pad <- function(.x, .n) {
    length(.x) <- .n
    .x
  }
  df |>
    select(property, type, description) |>
    mutate(
      description = str_wrap(description, width = 45),
      description = str_split(description, pattern = "\n"),
      n = lengths(description),
      property = map2(property, n, pad),
      type = map2(type, n, pad),
      n = NULL
    ) |>
    unnest(cols = c(property, type, description)) |>
    replace_na(list(property = "", type = "")) |>
    modify_at(c("property", "type"), ~ format(.x, justify = "left")) |>
    glue::glue_data("{property} {type} {description}")
}

make_humane_table(ap) |>
  write_lines(fs::path(ddi_dir, "api-wide-parameters-humane.txt"))

make_humane_table(mp) |>
  write_lines(fs::path(ddi_dir, "method-properties-humane.txt"))

make_humane_table(pp) |>
  write_lines(fs::path(ddi_dir, "parameter-properties-humane.txt"))

write_lines(mp$property, fs::path(ddi_dir, "api-wide-parameter-names.txt"))

write_lines(mp$property, fs::path(ddi_dir, "method-property-names.txt"))

write_lines(pp$property, fs::path(ddi_dir, "parameter-property-names.txt"))

#' API key for demonstration purposes
#'
#' @description Some APIs accept requests for public resources, in which case
#'   the request must be sent with an API key in lieu of a token. This function
#'   provides an API key for limited use in prototyping and for testing and
#'   documentation of gargle itself. This key may be deleted or rotated at any
#'   time. There are no guarantees about which APIs are enabled. DO NOT USE THIS
#'   IN A PACKAGE or for anything other than interactive, small-scale
#'   experimentation.
#'
#'   You can get your own API key, without these limitations. See the [How to
#'   get your own API
#'   credentials](https://gargle.r-lib.org/articles/get-api-credentials.html)
#'   vignette for more details.
#'
#' @export
#' @keywords internal
#' @examples
#' gargle_api_key()
gargle_api_key <- function() {
  gak()
}

#' Assets for internal use
#'
#' Assets for use inside specific packages maintained by the tidyverse team.
#'
#' @name internal-assets
NULL

#' @export
#' @keywords internal
#' @rdname internal-assets
tidyverse_api_key <- function() {
  check_permitted_package(parent.frame())
  tak()
}

#' Fetch a token for the given scopes
#'
#' This is a rather magical function that calls a series of concrete
#' credential-fetching functions, each wrapped in a `tryCatch()`.
#' `token_fetch()` keeps trying until it succeeds or there are no more functions
#' to try. See the `vignette("how-gargle-gets-tokens")` for a full description
#' of `token_fetch()`.
#'
#' @seealso [cred_funs_list()] reveals the current registry of
#'   credential-fetching functions, in order.
#'
#' @inheritParams credentials_user_oauth2
#' @param ... Additional arguments passed to all credential functions.
#'
#' @return An [`httr::Token`][httr::Token-class] (often an instance of something
#'   that inherits from `httr::Token`) or `NULL`.
#' @family credential functions
#' @export
#' @examples
#' \dontrun{
#' token_fetch(scopes = "https://www.googleapis.com/auth/userinfo.email")
#' }
token_fetch <- function(scopes = NULL, ...) {
  gargle_debug("trying {.fun token_fetch}")
  for (f in gargle_env$cred_funs) {
    token <- NULL
    token <- tryCatch(
      f(scopes, ...),
      warning = function(e) {
        gargle_debug(c("Warning caught by {.fun token_fetch}:", e$message))
        NULL
      },
      error = function(e) {
        gargle_debug(c("Error caught by {.fun token_fetch}:", e$message))
        NULL
      }
    )
    if (!is.null(token)) {
      return(token)
    }
  }
  NULL
}

#' Generate a field mask
#'
#' Many Google API requests take a field mask, via a `fields` parameter, in the
#' URL and/or in the body. `field_mask()` generates such a field mask from an R
#' list, typically a list that is destined to be part of the body of a request
#' that writes or updates a resource. `field_mask()` is designed to help in the
#' common case where the attributes you wish to modify are exactly the ones
#' represented in the object. It is possible to use a "larger" field mask, that
#' is either less specific or that explicitly includes other attributes, in
#' which case the attributes covered by the mask but absent from the object are
#' reset to default values. This is not exactly the use case `field_mask()` is
#' designed for, but its output could still be useful as a first step in
#' constructing such a mask.
#'
#' @param x A named R list, where the requirement for names applies at all
#'   levels, i.e. recursively.
#'
#' @return A Google API field mask, as a string.
#' @export
#'
#' @seealso The documentation for the [JSON encoding of a Protocol Buffers
#'   FieldMask](https://developers.google.com/protocol-buffers/docs/reference/google.protobuf#json-encoding-of-field-masks).
#'
#' @examples
#' x <- list(sheetId = 1234, title = "my_favorite_worksheet")
#' field_mask(x)
#'
#' x <- list(
#'   userEnteredFormat = list(
#'     backgroundColor = list(
#'       red = 159 / 255, green = 183 / 255, blue = 196 / 255
#'     )
#'   )
#' )
#' field_mask(x)
#'
#' x <- list(
#'   sheetId = 1234,
#'   gridProperties = list(rowCount = 5, columnCount = 3)
#' )
#' field_mask(x)
field_mask <- function(x) {
  stopifnot(is_dictionaryish(x))
  explicit_mask <- imap(x, field_mask_impl_)
  as.character(glue_collapse(unname(unlist(explicit_mask)), ","))
}

field_mask_impl_ <- function(x, y = "") {
  if (!is_list(x)) {
    return(y)
  }
  stopifnot(is_dictionaryish(x))

  leafs <- !map_lgl(x, is_list)
  if (sum(leafs) <= 1) {
    leafs <- FALSE
  }
  names(x)[!leafs] <- glue(".{names(x)[!leafs]}")
  if (sum(leafs) > 1) {
    nm <- glue("({glue_collapse(names(x)[leafs], sep = ',')})")
    x <- list2(!!nm := NA, !!!x[!leafs])
  }
  map2(x, glue("{y}{names(x)}"), field_mask_impl_)
}

#' Credential function registry
#'
#' Functions to query or manipulate the registry of credential functions
#' consulted by [token_fetch()].
#'
#' @name cred_funs
#'
#' @seealso [token_fetch()], which is where the registry is actually used.
#' @return A list of credential functions or `NULL`.
#' @examples
#' names(cred_funs_list())
#'
#' creds_one <- function(scopes, ...) {}
#'
#' cred_funs_add(one = creds_one)
#' cred_funs_add(two = creds_one, three = creds_one)
#' names(cred_funs_list())
#'
#' cred_funs_add(two = NULL)
#' names(cred_funs_list())
#'
#' # restore the default list
#' cred_funs_set_default()
#'
#' # remove one specific credential fetcher
#' cred_funs_add(credentials_gce = NULL)
#' names(cred_funs_list())
#'
#' # force the use of one specific credential fetcher
#' cred_funs_set(list(credentials_user_oauth2 = credentials_user_oauth2))
#' names(cred_funs_list())
#'
#' # restore the default list
#' cred_funs_set_default()
#'
#' # run some code with a temporary change to the registry
#' # creds_one ONLY
#' with_cred_funs(
#'   list(one = creds_one),
#'   names(cred_funs_list())
#' )
#' # add creds_one to the list
#' with_cred_funs(
#'   list(one = creds_one),
#'   names(cred_funs_list()),
#'   action = "modify"
#' )
#' # remove credentials_gce
#' with_cred_funs(
#'   list(credentials_gce = NULL),
#'   names(cred_funs_list()),
#'   action = "modify"
#' )
NULL

#' @describeIn cred_funs Get the list of registered credential functions.
#' @export
cred_funs_list <- function() {
  gargle_env$cred_funs
}

#' @describeIn cred_funs Register one or more new credential fetching functions.
#'   Function(s) are added to the *front* of the list. So:
#'   * "First registered, last tried."
#'   * "Last registered, first tried."
#'
#' Can also be used to *remove* a function from the registry.
#'
#' @param ... <[`dynamic-dots`][rlang::dyn-dots]> One or more credential
#'   functions, in `name = value` form. Each credential function is subject to a
#'   superficial check that it at least "smells like" a credential function: its
#'   first argument must be named `scopes`, and its signature must include
#'   `...`. To remove a credential function, you can use a specification like
#'   `name = NULL`.
#' @export
cred_funs_add <- function(...) {
  dots <- dots_list(
    ...,
    .named = NULL,
    .ignore_empty = "all",
    .preserve_empty = FALSE,
    .homonyms = "error",
    .check_assign = TRUE
  )

  cred_funs_check(dots, allow_null = TRUE)

  nms_to_remove <- names(dots)[map_lgl(dots, is.null)]
  cf <- cred_funs_list()
  cf[nms_to_remove] <- NULL
  cred_funs_set(cf)
  dots <- dots[!map_lgl(dots, is.null)]

  dup_nm <- names(dots) %in% names(cred_funs_list())
  if (any(dup_nm)) {
    n_dup_nm <- sum(dup_nm)
    gargle_abort(c(
      "{cli::qty(n_dup_nm)}{?This/These} name{?s} already {?appears/appear} \\
      in the credential function registry:",
      x = "{.field {names(dots)[dup_nm]}}"
    ))
  }

  # add them in reverse order, to mimic what would happen if they'd been added
  # one-at-a-time
  cf <- cred_funs_list()
  cred_funs_set(c(rev(dots), cf))

  invisible(cred_funs_list())
}

#' @describeIn cred_funs Register a list of credential fetching functions.
#'
#' @param funs A named list of credential functions.
#' @param ls `r lifecycle::badge("deprecated")` This argument has been renamed
#'   to `funs`.
#' @export
cred_funs_set <- function(funs, ls = deprecated()) {
  if (lifecycle::is_present(ls)) {
    lifecycle::deprecate_warn(
      when = "1.3.0",
      what = "cred_funs_set(ls)",
      with = "cred_funs_set(funs)",
    )
    funs = ls
  }

  cred_funs_check(funs, allow_null = FALSE)
  gargle_env$cred_funs <- funs
  invisible(cred_funs_list())
}

#' @describeIn cred_funs Clear the credential function registry.
#' @export
cred_funs_clear <- function() {
  gargle_env$cred_funs <- list()
  invisible(cred_funs_list())
}

#' @describeIn cred_funs Return the default list of credential functions.
#' @export
cred_funs_list_default <- function() {
  list(
    credentials_byo_oauth2 = credentials_byo_oauth2,
    credentials_service_account = credentials_service_account,
    credentials_external_account = credentials_external_account,
    credentials_app_default = credentials_app_default,
    credentials_gce = credentials_gce,
    credentials_user_oauth2 = credentials_user_oauth2
  )
}

#' @describeIn cred_funs Reset the registry to the gargle default.
#' @export
cred_funs_set_default <- function() {
  cred_funs_set(cred_funs_list_default())
}

#' @describeIn cred_funs Modify the credential function registry in the current
#'   scope. It is an example of the `local_*()` functions in \pkg{withr}.
#' @param action Whether to use `funs` to replace or modify the registry with
#'   funs:
#'   * `"replace"` does `cred_funs_set(funs)`
#'   * `"modify"` does `cred_funs_add(!!!funs)`
#' @param .local_envir The environment to use for scoping. Defaults to current
#'   execution environment.
#' @export
local_cred_funs <- function(
  funs = cred_funs_list_default(),
  action = c("replace", "modify"),
  .local_envir = caller_env()
) {
  action <- arg_match(action)

  cred_funs_orig <- cred_funs_list()
  withr::defer(cred_funs_set(cred_funs_orig), envir = .local_envir)

  switch(
    action,
    replace = cred_funs_set(funs),
    modify = cred_funs_add(!!!funs)
  )
}

#' @describeIn cred_funs Evaluate `code` with a temporarily modified credential
#'   function registry. It is an example of the `with_*()` functions in
#'   \pkg{withr}.
#' @param code Code to run with temporary credential function registry.
#' @export
with_cred_funs <- function(
  funs = cred_funs_list_default(),
  code,
  action = c("replace", "modify")
) {
  local_cred_funs(funs = funs, action = action)
  force(code)
}

cred_funs_check <- function(ls, allow_null = FALSE) {
  if (allow_null) {
    not_cred_fun <- !map_lgl(ls, is.null) & !map_lgl(ls, is_cred_fun)
  } else {
    not_cred_fun <- !map_lgl(ls, is_cred_fun)
  }
  if (any(not_cred_fun)) {
    gargle_abort(c(
      "Not a valid credential function:",
      x = "Element{?s} {as.character(which(not_cred_fun))}"
    ))
  }

  if (!is_dictionaryish(ls)) {
    gargle_abort("Each credential function must have a unique name")
  }

  invisible()
}

#' Check that f is a viable credential fetching function
#'
#' In the abstract, a credential fetching function is any function which takes a
#' set of scopes and any number of additional arguments, and returns either a
#' valid [`httr::Token`][httr::Token-class] or `NULL`.
#'
#' Here we say that a function is valid if its first argument is named `scopes`,
#' and it includes `...` as an argument, since it's difficult to actually check
#' the behavior of the function.
#'
#' @param f A function to check.
#' @keywords internal
#' @noRd
#' @examples
#' f <- function(scopes, ...) {}
#' is_cred_fun(f)
is_cred_fun <- function(f) {
  if (!is.function(f)) {
    return(FALSE)
  }
  args <- names(formals(f))
  args[1] == "scopes" && "..." %in% args
}

test_that("is_cred_fun() recognizes a credential function signature", {
  creds_one <- function(scopes, ...) {}
  creds_two <- function(scopes, arg1, arg2 = "optional", ...) {}
  expect_true(is_cred_fun(creds_one))
  expect_true(is_cred_fun(creds_two))

  invalid_one <- function(scope, ...) {}
  invalid_two <- function(scopes, arg1, arg2 = "optional") {}
  invalid_three <- 17
  expect_false(is_cred_fun(invalid_one))
  expect_false(is_cred_fun(invalid_two))
  expect_false(is_cred_fun(invalid_three))
})

test_that("cred_funs_add() and _set() can add and set credential functions", {
  creds_one <- function(scopes, ...) {}
  creds_two <- function(scopes, arg1, arg2 = "optional", ...) {}

  withr::defer(cred_funs_set_default())
  cred_funs_clear()

  cred_funs_add(a = creds_one)
  expect_equal(names(cred_funs_list()), "a")

  cred_funs_add(b = creds_two)
  expect_equal(names(cred_funs_list()), c("b", "a"))

  cred_funs_clear()
  expect_equal(0, length(cred_funs_list()))

  for (i in 1:5) {
    nm <- glue("cf{i}")
    cred_funs_add(!!nm := creds_one)
    expect_equal(i, length(cred_funs_list()))
    expect_match(names(cred_funs_list()), "^cf[12345]")
  }

  cred_funs_set(list(first = creds_one, last = creds_two))
  expect_equal(names(cred_funs_list()), c("first", "last"))
})

test_that("We insist on valid credential function (or NULL)", {
  creds_one <- function(scopes, ...) {}

  withr::defer(cred_funs_set_default())
  cred_funs_clear()

  expect_snapshot(
    error = TRUE,
    cred_funs_add(a = mean)
  )
  expect_snapshot(
    error = TRUE,
    cred_funs_set(list(a = NULL))
  )
})

test_that("We insist on uniquely named credential functions", {
  creds_one <- function(scopes, ...) {}

  withr::defer(cred_funs_set_default())
  cred_funs_clear()

  cred_funs_add(a = creds_one)
  cred_funs_add(b = function(scopes, ...) {})
  expect_snapshot(
    error = TRUE,
    cred_funs_add(creds_one)
  )
  expect_snapshot(
    error = TRUE,
    cred_funs_add(a = creds_one)
  )
  expect_equal(names(cred_funs_list()), c("b", "a"))

  cred_funs_clear()

  expect_snapshot(
    error = TRUE,
    cred_funs_set(list(
      creds_one,
      a = function(scopes, ...) {}
    ))
  )
  expect_snapshot(
    error = TRUE,
    cred_funs_set(list(
      a = creds_one,
      a = function(scopes, ...) {}
    ))
  )
})

test_that("We can remove credential functions by name", {
  cred_fun <- function(scopes, ...) {}
  withr::defer(cred_funs_set_default())
  cred_funs_clear()

  cred_funs_add(a = cred_fun, b = cred_fun, c = cred_fun)
  cred_funs_add(b = NULL)
  expect_equal(names(cred_funs_list()), c("c", "a"))

  cred_funs_add(c = NULL, d = cred_fun)
  expect_equal(names(cred_funs_list()), c("d", "a"))
})

test_that("cred_funs_set() warns for use of `ls`", {
  withr::local_options(lifecycle_verbosity = "warning")
  withr::defer(cred_funs_set_default())
  expect_snapshot(
    out <- cred_funs_set(ls = list(a = function(scopes, ...) {}))
  )
  expect_equal(names(cred_funs_list()), "a")
})

test_that("local_cred_funs() works", {
  withr::defer(cred_funs_set_default())
  cred_fun <- function(scopes, ...) {}
  cred_funs_clear()

  cred_funs_add(a = cred_fun, b = cred_fun)

  local_cred_funs(funs = list(c = cred_fun), action = "modify")
  expect_equal(names(cred_funs_list()), c("c", "b", "a"))

  local_cred_funs(funs = list(d = cred_fun)) # action = "replace"
  expect_equal(names(cred_funs_list()), "d")

  local_cred_funs()
  expect_equal(names(cred_funs_list()), names(cred_funs_list_default()))
})

test_that("with_cred_funs() works", {
  withr::defer(cred_funs_set_default())
  cred_fun <- function(scopes, ...) {}
  cred_funs_clear()

  cred_funs_add(a = cred_fun, b = cred_fun)

  with_cred_funs(
    funs = list(c = cred_fun),
    action = "modify",
    expect_equal(names(cred_funs_list()), c("c", "b", "a"))
  )

  with_cred_funs(
    funs = list(d = cred_fun), # action = "replace"
    expect_equal(names(cred_funs_list()), "d")
  )

  with_cred_funs(
    code = expect_equal(
      names(cred_funs_list()),
      names(cred_funs_list_default())
    )
  )
})

#' Create an AuthState
#'
#' Constructor function for objects of class [AuthState].
#'
#' @param package Package name, an optional string. It is recommended to record
#'   the name of the package whose auth state is being managed. Ultimately, this
#'   may be used in some downstream messaging.
#' @param api_key Optional. API key (a string). Some APIs accept unauthorized,
#'   "token-free" requests for public resources, but only if the request
#'   includes an API key.
#' @param auth_active Logical. `TRUE` means requests should include a token (and
#'   probably not an API key). `FALSE` means requests should include an API key
#'   (and probably not a token).
#' @param cred Credentials. Typically populated indirectly via [token_fetch()].
#' @inheritParams gargle2.0_token
#'
#' @return An object of class [AuthState].
#' @export
#' @examples
#' my_client <- gargle_oauth_client(
#'   id = "some_long_client_id",
#'   secret = "ssshhhhh_its_a_secret",
#'   name = "my-nifty-oauth-client"
#' )
#'
#' init_AuthState(
#'   package = "my_package",
#'   client = my_client,
#'   api_key = "api_key_api_key_api_key",
#' )
init_AuthState <- function(
  package = NA_character_,
  client = NULL,
  api_key = NULL,
  auth_active = TRUE,
  cred = NULL,
  app = deprecated()
) {
  if (lifecycle::is_present(app)) {
    lifecycle::deprecate_soft(
      "1.5.0",
      "init_AuthState(app)",
      "init_AuthState(client)"
    )
    client <- app
  }

  AuthState$new(
    package = package,
    client = client,
    api_key = api_key,
    auth_active = auth_active,
    cred = cred
  )
}

#' Authorization state
#'
#' @description
#' An `AuthState` object manages an authorization state, typically on behalf of
#' a wrapper package that makes requests to a Google API.
#'

#' The `vignette("gargle-auth-in-client-package)` describes a design for wrapper
#' packages that relies on an `AuthState` object. This state can then be
#' incorporated into the package's requests for tokens and can control the
#' inclusion of tokens in requests to the target API.
#'
#'   * `api_key` is the simplest way to associate a request with a specific
#'     Google Cloud Platform [project](https://docs.cloud.google.com/resource-manager/docs/cloud-platform-resource-hierarchy).
#'     A few calls to certain APIs, e.g. reading a public Sheet, can succeed
#'     with an API key, but this is the exception.
#'   * `client` is an OAuth client ID (and secret) associated with a specific
#'     Google Cloud Platform [project](https://docs.cloud.google.com/resource-manager/docs/cloud-platform-resource-hierarchy).
#'     This is used in the OAuth flow, in which an authenticated user authorizes
#'     the client to access or manipulate data on their behalf.
#'   * `auth_active` reflects whether outgoing requests will be authorized by an
#'     authenticated user or are unauthorized requests for public resources.
#'     These two states correspond to sending a request with a token versus an
#'     API key, respectively.
#'   * `cred` is where the current token is cached within a session, once one
#'     has been fetched. It is generally assumed to be an instance of
#'     [`httr::TokenServiceAccount`][httr::Token-class] or
#'     [`httr::Token2.0`][httr::Token-class] (or a subclass thereof), probably
#'     obtained via [token_fetch()] (or one of its constituent credential
#'     fetching functions).
#'
#' An `AuthState` should be created through the constructor function
#' [init_AuthState()], which has more details on the arguments.
#'
#' @param package Package name.
#' @param client An OAuth client.
#' @param api_key An API key.
#' @param auth_active Logical, indicating whether auth is active.
#' @param cred Credentials.
#' @param app `r lifecycle::badge('deprecated')` Use `client` instead.
#'
#' @export
#' @name AuthState-class
AuthState <- R6::R6Class(
  "AuthState",
  list(
    #' @field package Package name.
    package = NULL,
    #' @field client An OAuth client.
    client = NULL,
    #' @field app `r lifecycle::badge('deprecated')` Use `client` instead.
    app = NULL,
    #' @field api_key An API key.
    api_key = NULL,
    #' @field auth_active Logical, indicating whether auth is active.
    auth_active = NULL,
    #' @field cred Credentials.
    cred = NULL,
    #' @description Create a new AuthState
    #' @details For more details on the parameters, see [init_AuthState()]
    initialize = function(
      package = NA_character_,
      client = NULL,
      api_key = NULL,
      auth_active = TRUE,
      cred = NULL,
      app = deprecated()
    ) {
      gargle_debug("initializing AuthState")
      if (lifecycle::is_present(app)) {
        # I'm using deprecate_warn() intentionally here. If I use
        # deprecate_soft(), you don't see the warning for a call to
        # AuthState$new(app). Most folks should be instantiating through
        # init_AuthState() anyway, so anyone who sees this warning probably needs
        # to see it.
        lifecycle::deprecate_warn(
          "1.5.0",
          "AuthState$initialize(app)",
          "AuthState$initialize(client)"
        )
        client <- app
      }
      stopifnot(
        is_scalar_character(package),
        is.null(client) || is.oauth_app(client),
        is.null(api_key) || is_string(api_key),
        is_bool(auth_active),
        is.null(cred) || inherits(cred, "Token2.0")
      )
      self$package <- package
      self$client <- client
      # for backwards compatibility; could eventually be removed
      self$app <- client
      self$api_key <- api_key
      self$auth_active <- auth_active
      self$cred <- cred
      self
    },
    #' @description Format an AuthState
    #' @param ... Not used.
    format = function(...) {
      x <- list(
        package = cli::format_inline("{.pkg {self$package}}"),
        client = self$client$name,
        api_key = obfuscate(self$api_key),
        auth_active = self$auth_active,
        credentials = cli::format_inline("{.cls {class(self$cred)[[1]]}}")
      )
      c(
        cli::cli_format_method(
          cli::cli_h1("<AuthState (via {.pkg gargle})>")
        ),
        glue("{fr(names(x))}: {fl(x)}")
      )
    },
    #' @description Set the OAuth client
    set_client = function(client) {
      stopifnot(is.null(client) || is.oauth_app(client))
      self$client <- client
      invisible(self)
    },
    #' @description `r lifecycle::badge('deprecated')` Deprecated method to set
    #'   the OAuth client
    set_app = function(app) {
      lifecycle::deprecate_soft(
        "1.5.0",
        "AuthState$set_app()",
        "AuthState$set_client()",
        details = make_package_hint(self$package)
      )
      # needed for backwards compatibility, as long as there are packages out
      # there consulting .auth$app
      self$app <- app
      self$set_client(client = app)
    },
    #' @description Set the API key
    #' @param value An API key.
    set_api_key = function(value) {
      stopifnot(is.null(value) || is_string(value))
      self$api_key <- value
      invisible(self)
    },
    #' @description Set whether auth is (in)active
    #' @param value Logical, indicating whether to send requests authorized with
    #'   user credentials.
    set_auth_active = function(value) {
      stopifnot(isTRUE(value) || isFALSE(value))
      self$auth_active <- value
      invisible(self)
    },
    #' @description Set credentials
    #' @param cred User credentials.
    set_cred = function(cred) {
      self$cred <- cred
      invisible(self)
    },
    #' @description Clear credentials
    clear_cred = function() {
      self$set_cred(NULL)
    },
    #' @description Get credentials
    get_cred = function() {
      self$cred
    },
    #' @description Report if we have credentials
    has_cred = function() {
      ## FIXME(jennybc): how should this interact with auth_active? should it?
      !is.null(self$cred)
    }
  )
)

make_package_hint <- function(pkg) {
  hint <- NULL
  if (is_string(pkg)) {
    hint <- glue(
      "
      This probably needs to be addressed in the {pkg} package."
    )
    url <- pkg_url_bug(pkg)
    if (!is.null(url)) {
      hint <- c(hint, glue("Please report the issue at <{url}>."))
    }
  }
  hint
}

test_that("inputs are checked when creating AuthState", {
  client <- gargle_oauth_client(id = "CLIENT_ID", secret = "SECRET")

  expect_snapshot(
    init_AuthState(
      package = NULL,
      client = client,
      api_key = "API_KEY",
      auth_active = TRUE
    ),
    error = TRUE
  )
  expect_snapshot(init_AuthState(client = "not_an_oauth_client"), error = TRUE)
  expect_snapshot(init_AuthState(client = client, api_key = 1234), error = TRUE)
  expect_snapshot(
    init_AuthState(client = client, api_key = "API_KEY", auth_active = NULL),
    error = TRUE
  )

  a <- init_AuthState(
    package = "PACKAGE",
    client = client,
    api_key = "API_KEY",
    auth_active = TRUE
  )
  expect_s3_class(a, "AuthState")
})

test_that("AuthState client can be modified and cleared", {
  client <- gargle_oauth_client(
    id = "CLIENT_ID",
    secret = "SECRET",
    name = "AAA"
  )
  a <- init_AuthState(client = client, api_key = "API_KEY", auth_active = TRUE)
  expect_equal(a$client$name, "AAA")

  client2 <- gargle_oauth_client(
    id = "CLIENT_ID",
    secret = "SECRET",
    name = "BBB"
  )
  a$set_client(client2)
  expect_equal(a$client$name, "BBB")

  a$set_client(NULL)
  expect_null(a$client)
})

test_that("AuthState api_key can be modified and cleared", {
  client <- gargle_oauth_client(
    id = "CLIENT_ID",
    secret = "SECRET",
    name = "AAA"
  )
  a <- init_AuthState(client = client, api_key = "AAA", auth_active = TRUE)
  expect_equal(a$api_key, "AAA")

  a$set_api_key("BBB")
  expect_equal(a$api_key, "BBB")

  a$set_api_key(NULL)
  expect_null(a$api_key)
})

test_that("AuthState auth_active can be toggled", {
  client <- gargle_oauth_client(
    id = "CLIENT_ID",
    secret = "SECRET",
    name = "AAA"
  )
  a <- init_AuthState(client = client, api_key = "AAA", auth_active = TRUE)
  expect_true(a$auth_active)

  a$set_auth_active(FALSE)
  expect_false(a$auth_active)
})

test_that("AuthState supports basic handling of cred", {
  client <- gargle_oauth_client(
    id = "CLIENT_ID",
    secret = "SECRET",
    name = "AAA"
  )
  a <- init_AuthState(client = client, api_key = "AAA", auth_active = TRUE)

  a$set_cred("hi")
  expect_true(a$has_cred())
  expect_equal(a$get_cred(), "hi")
  a$clear_cred()
  expect_false(a$has_cred())
  a$set_cred("bye")
  expect_equal(a$get_cred(), "bye")
})

test_that("AuthState prints nicely", {
  client <- gargle_oauth_client(
    id = "CLIENT_ID",
    secret = "SECRET",
    name = "AAA"
  )
  a <- init_AuthState(
    package = "PKG",
    client = client,
    api_key = "API_KEY",
    auth_active = TRUE
  )
  a$set_cred(structure("TOKEN", class = "some_sort_of_token"))
  expect_snapshot(print(a))
})

test_that("init_Authstate(app) argument is deprecated, but still works", {
  withr::local_options(lifecycle_verbosity = "warning")
  client <- gargle_oauth_client(id = "CLIENT_ID", secret = "SECRET")

  expect_snapshot(
    a <- init_AuthState(
      package = "PACKAGE",
      app = client,
      api_key = "API_KEY",
      auth_active = TRUE
    )
  )
  expect_s3_class(a, "AuthState")
  expect_s3_class(a$client, "gargle_oauth_client")
})

test_that("AuthState$new(app) is deprecated, but still works", {
  withr::local_options(lifecycle_verbosity = "warning")
  client <- gargle_oauth_client(id = "CLIENT_ID", secret = "SECRET")

  expect_snapshot(
    a <- AuthState$new(
      package = "PACKAGE",
      app = client,
      api_key = "API_KEY",
      auth_active = TRUE
    )
  )
  expect_s3_class(a, "AuthState")
  expect_s3_class(a$client, "gargle_oauth_client")
})

test_that("$set_app is deprecated, but still works", {
  withr::local_options(lifecycle_verbosity = "warning")

  client <- gargle_oauth_client(
    id = "CLIENT_ID",
    secret = "SECRET",
    name = "AAA"
  )
  a <- init_AuthState(
    client = client,
    # this just needs to be some package that is guaranteed to be installed, in
    # order to fully exercise the deprecation warning
    package = "rlang",
    api_key = "API_KEY",
    auth_active = TRUE
  )
  client2 <- gargle_oauth_client(
    id = "CLIENT_ID",
    secret = "SECRET",
    name = "BBB"
  )

  expect_snapshot(
    a$set_app(client2)
  )
  expect_equal(a$client$name, "BBB")
})

test_that("$app still returns the client", {
  client <- gargle_oauth_client(
    id = "CLIENT_ID",
    secret = "SECRET",
    name = "AAA"
  )
  a <- init_AuthState(client = client, api_key = "API_KEY", auth_active = TRUE)
  expect_equal(a$app, client)
})

# this file has its origins in oauth-refresh.R and oauth-error.R in httr
# I want to introduce behaviour to error informatively for a deleted OAuth client

# Refresh an OAuth 2.0 credential.
#
# Refreshes the given token, and returns a new credential with a
# valid access_token. Based on:
# https://developers.google.com/identity/protocols/oauth2/native-app#offline
refresh_oauth2.0 <- function(endpoint, client, credentials, package = NULL) {
  if (is.null(credentials$refresh_token)) {
    gargle_abort("Refresh token not available.")
  }

  refresh_url <- endpoint$access
  req_params <- list(
    refresh_token = credentials$refresh_token,
    client_id = client$key,
    client_secret = client$secret,
    grant_type = "refresh_token"
  )

  response <- httr::POST(refresh_url, body = req_params, encode = "form")

  err <- find_oauth2.0_error(response)
  if (!is.null(err)) {
    gargle_refresh_failure(err, client, package)
    return(NULL)
  }

  httr::stop_for_status(response)
  refresh_data <- httr::content(response)
  utils::modifyList(credentials, refresh_data)
}

oauth2.0_error_codes <- c(
  400,
  401
)

# This implements error checking according to the OAuth2.0
# specification: https://tools.ietf.org/html/rfc6749#section-5.2
find_oauth2.0_error <- function(response) {
  if (!httr::status_code(response) %in% oauth2.0_error_codes) {
    return(NULL)
  }

  content <- httr::content(response)
  if (is.null(content$error)) {
    return(NULL)
  }

  list(
    error = content$error,
    error_description = content$error_description,
    error_uri = content$error_uri
  )
}

gargle_refresh_failure <- function(err, client, package = NULL) {
  if (!identical(err$error, "deleted_client")) {
    # this is basically what httr does, except we don't have an explicit
    # whitelist of acceptable values of err$error, because we know Google does
    # not limit itself to these
    gargle_warn(c(
      "Unable to refresh token: {err$error}",
      "*" = err$error_description,
      "*" = err$error_uri
    ))
    return(invisible())
  }

  # special handling for 'deleted_client'
  client_name <- client$name %||% client$appname %||% ""
  is_legacy_app <- grepl(gargle_legacy_app_pattern(), client_name)

  # client looks like one of "ours"
  if (is_legacy_app) {
    main_pkg <- package %||% "gargle"
    all_pkgs <- if (main_pkg == "gargle") "gargle" else c(main_pkg, "gargle")
    gargle_warn(c(
      "Unable to refresh token, because the associated OAuth client \\
       has been deleted.",
      "i" = "You appear to be relying on the default client used by the \\
             {.pkg {main_pkg}} package.",
      " " = "Consider re-installing {.pkg {all_pkgs}}, \\
             in case the default client has been updated."
    ))
    return(invisible())
  }

  # deleted client doesn't seem to be one of "ours"
  gargle_warn(c(
    "Unable to refresh token, because the associated OAuth client \\
     has been deleted.",
    "*" = if (nzchar(client_name)) "Client name: {.field {client_name}}",
    if (!is.null(package)) {
      c(
        "i" = "If you did not configure this OAuth client, it may be built into \\
               the {.pkg {package}} package.",
        " " = "If so, consider re-installing {.pkg {package}} to get an updated \\
               client."
      )
    }
  ))
  invisible()
}

test_that("'deleted_client' causes extra special feedback", {
  err <- list(
    error = "deleted_client",
    error_description = "The OAuth client was deleted."
  )

  expect_snapshot(
    gargle_refresh_failure(
      err,
      httr::oauth_app(appname = NULL, key = "KEY", secret = "SECRET")
    )
  )

  expect_snapshot(
    gargle_refresh_failure(
      err,
      httr::oauth_app(appname = "APPNAME", key = "KEY", secret = "SECRET")
    )
  )

  expect_snapshot(
    gargle_refresh_failure(
      err,
      httr::oauth_app(appname = "APPNAME", key = "KEY", secret = "SECRET"),
      package = "PACKAGE"
    )
  )

  expect_snapshot(
    gargle_refresh_failure(
      err,
      httr::oauth_app(appname = "fake-calliope", key = "KEY", secret = "SECRET")
    )
  )

  expect_snapshot(
    gargle_refresh_failure(
      err,
      httr::oauth_app(
        appname = "fake-calliope",
        key = "KEY",
        secret = "SECRET"
      ),
      package = "PACKAGE"
    )
  )
})

# response provided by the user from
# https://github.com/tidyverse/googlesheets4/issues/317

# in theory, I should be able to reprex by creating a service account and NOT
# enabling the Sheets API, then trying to access even an example sheet via
# gs4_example()

# this script doesn't create the response, but records how I processed their
# original raw response to remove the user's project id

resp_before <- readRDS(testthat::test_path("fixtures", "response.rds"))

content_before <- httr::content(resp_before, type = "raw")
content_before <- rawToChar(content_before)
Encoding(content_before) <- "UTF-8"
content_before <- jsonlite::fromJSON(content_before, simplifyVector = FALSE)

id <- content_before$error$details[[1]]$metadata$containerInfo
redacted_content <- rapply(
  content_before,
  function(x) gsub(id, "1234567890", x),
  classes = "character",
  how = "replace"
)

content_after <- jsonlite::toJSON(redacted_content, auto_unbox = TRUE)
content_after <- charToRaw(content_after)

resp_after <- resp_before
resp_after$content <- content_after

# make sure this looks reasonable
response_process(resp_after)

saveRDS(
  gargle:::redact_response(resp_after),
  testthat::test_path(
    "fixtures",
    "sheets-spreadsheets-get-service-disabled_403.rds"
  ),
  version = 2
)

# ---- API key that does not have Sheets enabled
req <- gargle::request_develop(
  endpoint = googlesheets4::gs4_endpoints("sheets.spreadsheets.get")[[1]],
  params = list(
    spreadsheetId = "DOES_NOT_MATTER",
    fields = "spreadsheetId"
  ),
  base_url = "https://sheets.googleapis.com/"
)
req <- gargle::request_build(
  path = req$path,
  method = req$method,
  params = req$params,
  key = gargle::gargle_api_key(),
  base_url = req$base_url
)
resp <- gargle::request_make(req)

stopifnot(httr::status_code(resp) == 403)

saveRDS(
  gargle:::redact_response(resp),
  testthat::test_path(
    "fixtures",
    "sheets-spreadsheets-get-api-key-not-enabled_403.rds"
  ),
  version = 2
)

gargle::response_process(resp)

# Setup support for the NAME=PASSWORD envvar ----------------------------------

# secret_pw_name("gargle") --> "GARGLE_PASSWORD"
secret_pw_name <- function(package) {
  paste0(toupper(package), "_PASSWORD")
}

# secret_pw_gen() --> "9AkKLa50wf1zHNCnHiQWeFLDoch9MYJHmPNnIVYZgSUt0Emwgi"
secret_pw_gen <- function() {
  x <- sample(c(letters, LETTERS, 0:9), 50, replace = TRUE)
  paste0(x, collapse = "")
}

# secret_pw_exists("gargle") --> TRUE or FALSE
secret_pw_exists <- function(package) {
  pw_name <- secret_pw_name(package)
  pw <- Sys.getenv(pw_name, "")
  !identical(pw, "")
}

# secret_pw_get("gargle") --> error or key-ified PASSWORD =
#                             hash of charToRaw(PASSWORD)
secret_pw_get <- function(package) {
  pw_name <- secret_pw_name(package)
  pw <- Sys.getenv(pw_name, "")
  if (identical(pw, "")) {
    gargle_abort_secret(
      message = "Env var {.envvar {pw_name}} is not defined.",
      package = package
    )
  }

  sodium::sha256(charToRaw(pw))
}

# Store and retrieve encrypted data -------------------------------------------

secret_can_decrypt <- function(package) {
  requireNamespace("sodium", quietly = TRUE) && secret_pw_exists(package)
}

# input should either be a filepath or a raw vector
secret_write <- function(package, name, input) {
  if (is.character(input)) {
    input <- readBin(input, "raw", file.size(input))
  } else if (!is.raw(input)) {
    gargle_abort_bad_class(input, c("character", "raw"))
  }

  destdir <- fs::path("inst", "secret")
  fs::dir_create(destdir)
  destpath <- fs::path(destdir, name)

  enc <- sodium::data_encrypt(
    msg = input,
    key = secret_pw_get(package),
    nonce = secret_nonce()
  )
  attr(enc, "nonce") <- NULL
  writeBin(enc, destpath)

  invisible(destpath)
}

# Generated with sodium::bin2hex(sodium::random(24)). AFAICT nonces are
# primarily used to prevent replay attacks, which shouldn't be a concern here
secret_nonce <- function() {
  sodium::hex2bin("cb36bab652dec6ae9b1827c684a7b6d21d2ea31cd9f766ac")
}

secret_path <- function(package, name) {
  fs::path_package(package, "secret", name)
}

# Returns a raw vector
secret_read <- function(package, name) {
  if (!secret_can_decrypt(package)) {
    gargle_abort_secret(message = "Decryption not available.", package = package)
  }

  path <- secret_path(package, name)
  raw <- readBin(path, "raw", file.size(path))

  sodium::data_decrypt(
    bin = raw,
    key = secret_pw_get(package),
    nonce = secret_nonce()
  )
}

gargle_abort_secret <- function(message, package, call = caller_env()) {
  gargle_abort(
    class = "gargle_error_secret",
    call = call,
    message = message,
    package = package
  )
}

# See notes about the userinfo and tokeninfo endpoints below.

#' Get info from a token
#'
#' These functions send the `token` to Google endpoints that return info about a
#' token or a user.
#'
#' @param token A token with class [Token2.0][httr::Token-class] or an object of
#'   httr's class `request`, i.e. a token that has been prepared with
#'   [httr::config()] and has a [Token2.0][httr::Token-class] in the
#'   `auth_token` component.
#' @name token-info
#'
#' @return A list containing:
#'   * `token_userinfo()`: user info
#'   * `token_email()`: user's email (obtained from a call to `token_userinfo()`)
#'   * `token_tokeninfo()`: token info
#'
#' @examples
#' \dontrun{
#' # with service account token
#' t <- token_fetch(
#'   scopes = "https://www.googleapis.com/auth/drive",
#'   path   = "path/to/service/account/token/blah-blah-blah.json"
#' )
#' # or with an OAuth token
#' t <- token_fetch(
#'   scopes = "https://www.googleapis.com/auth/drive",
#'   email  = "janedoe@example.com"
#' )
#' token_userinfo(t)
#' token_email(t)
#' tokens_tokeninfo(t)
#' }
NULL

#' @rdname token-info
#' @export
#'
#' @details
#' It's hard to say exactly what info will be returned by the "userinfo"
#' endpoint targetted by `token_userinfo()`. It depends on the token's scopes.
#' Where possible, OAuth2 tokens obtained via the gargle package include the
#' `https://www.googleapis.com/auth/userinfo.email` scope, which guarantees we
#' can learn the email associated with the token. If the token has the
#' `https://www.googleapis.com/auth/userinfo.profile` scope, there will be even
#' more information available. But for a token with unknown or arbitrary scopes,
#' we can't make any promises about what information will be returned.
token_userinfo <- function(token) {
  if (inherits(token, "request")) {
    token <- token$auth_token
  }
  stopifnot(inherits(token, "Token2.0"))

  req <- request_build(
    method = "GET",
    path = "v1/userinfo",
    token = token,
    base_url = "https://openidconnect.googleapis.com"
  )
  resp <- request_make(req)
  response_process(resp)
}

#' @rdname token-info
#' @export
token_email <- function(token) {
  # Assumes the token was obtained with userinfo.email scope.
  # This is true of all gargle-mediated tokens, by definition.
  token_userinfo(token)$email
}

#' @rdname token-info
#' @export
token_tokeninfo <- function(token) {
  if (inherits(token, "request")) {
    token <- token$auth_token
  }
  stopifnot(inherits(token, "Token2.0"))
  # I only want to explicitly refresh a user token, which I identify in this
  # rather back-ass-wards way, i.e. by a process of elimination.
  if (!inherits(token, c("TokenServiceAccount", "WifToken", "GceToken"))) {
    # A stale token does not fail in a way that leads to auto refresh.
    # It results in: "Bad Request (HTTP 400)."
    # Hence, the explicit refresh here.
    token$refresh()
  }

  # https://www.googleapis.com/oauth2/v3/tokeninfo
  # This can still lead to a doomed attempt to refresh, for example, a GceToken,
  # where the problem is actually (lack of) scope. That happens inside of httr.
  req <- request_build(
    method = "GET",
    path = "oauth2/v3/tokeninfo",
    # also works
    # params = list(access_token = token$credentials$access_token),
    token = token,
    base_url = "https://www.googleapis.com"
  )
  resp <- request_make(req)
  response_process(resp)
}

# Good 3rd party overview re: learning about your Google user / token:
# https://www.oauth.com/oauth2-servers/signing-in-with-google/verifying-the-user-info/
#
# It suggests perhaps we could recover, e.g., the email during token initiation?
# But I'm not sure and it might require a change in httr, in any case.
#
# Therefore, I focus on the two post hoc methods for retrieving info based on
# the token you've already got:
#   * userinfo endpoint
#   * tokeninfo endpoint
# Both require an additional HTTP request, so are "not recommended for
# production applications".
#
# I had working code to access both the userinfo and tokeninfo endpoints before
# I read this and there are a few differences between what I do and what they
# say to do. The discrepancies are around which token to send (ID? access? etc.)
# and where/how to send it. I conclude that there must be a few variations that
# work.
#
# Google's docs re: how to get user info:
# https://developers.google.com/identity/protocols/OpenIDConnect#obtaininguserprofileinformation
# "Add your access token to the authorization header and make an HTTPS GET
# request to the userinfo endpoint, which you should retrieve from the Discovery
# document using the key userinfo_endpoint."
#
# Here's the OpenID Connect discovery document:
# https://accounts.google.com/.well-known/openid-configuration
#
# Here's the URL for userinfo_endpoint, at the time of writing:
# https://openidconnect.googleapis.com/v1/userinfo

test_that("token_*() functions work", {
  skip_if_offline()
  skip_if_no_auth()

  token <- credentials_service_account(
    scopes = "https://www.googleapis.com/auth/userinfo.email",
    path = rawToChar(secret_read("gargle", "gargle-testing.json"))
  )

  expect_error_free(
    # this implies a call to token_userinfo()
    email <- token_email(token)
  )
  expect_error_free(
    tokeninfo <- token_tokeninfo(token)
  )

  expect_match(email, "^gargle-testing@.*[.]iam[.]gserviceaccount[.]com")
  expect_equal(email, tokeninfo$email)
  expect_true(
    "https://www.googleapis.com/auth/userinfo.email" %in% tokeninfo$scope
  )
})

#' Get a token for an external account
#'
#' @description

#' `r lifecycle::badge('experimental')`

#' Workload identity federation is a new (as of April 2021) keyless
#' authentication mechanism that allows applications running on a non-Google
#' Cloud platform, such as AWS, to access Google Cloud resources without using a
#' conventional service account token. This eliminates the dilemma of how to
#' safely manage service account credential files.
#'

#'  Unlike service accounts, the configuration file for workload identity
#'  federation contains no secrets. Instead, it holds non-sensitive metadata.
#'  The external application obtains the needed sensitive data "on-the-fly" from
#'  the running instance. The combined data is then used to obtain a so-called
#'  subject token from the external identity provider, such as AWS. This is then
#'  sent to Google's Security Token Service API, in exchange for a very
#'  short-lived federated access token. Finally, the federated access token is
#'  sent to Google's Service Account Credentials API, in exchange for a
#'  short-lived GCP access token. This access token allows the external
#'  application to impersonate a service account and inherit the permissions of
#'  the service account to access GCP resources.

#'
#'  This feature is still experimental in gargle and **currently only supports
#'  AWS**. It also requires installation of the suggested packages
#'  \pkg{aws.signature} and \pkg{aws.ec2metadata}. Workload identity federation
#'  **can** be used with other platforms, such as Microsoft Azure or any
#'  identity provider that supports OpenID Connect. If you would like gargle to
#'  support this token flow for additional platforms, please [open an issue on
#'  GitHub](https://github.com/r-lib/gargle/issues) and describe your use case.

#'
#' @inheritParams token_fetch

#' @param path JSON containing the workload identity configuration for the
#'   external account, in one of the forms supported for the `txt` argument of
#'   [jsonlite::fromJSON()] (probably, a file path, although it could be a JSON
#'   string). The instructions for generating this configuration are given at
#'   [Configuring workload identity federation](https://cloud.google.com/iam/docs/configuring-workload-identity-federation).
#'
#'   Note that external account tokens are a natural fit for use as Application
#'   Default Credentials, so consider storing the configuration file in one of
#'   the standard locations consulted for ADC, instead of providing `path`
#'   explicitly. See [credentials_app_default()] for more.
#'

#' @seealso There is substantial setup necessary, both on the GCP and AWS side,
#'   to use this authentication method. These two links provide, respectively,
#'   a high-level overview and step-by-step instructions.

#' * <https://cloud.google.com/blog/products/identity-security/enable-keyless-access-to-gcp-with-workload-identity-federation/>

#' * <https://cloud.google.com/iam/docs/configuring-workload-identity-federation>

#' @return A [WifToken()] or `NULL`.
#' @family credential functions
#' @export
#' @examples
#' \dontrun{
#' credentials_external_account()
#' }
credentials_external_account <- function(
  scopes = "https://www.googleapis.com/auth/cloud-platform",
  path = "",
  ...
) {
  gargle_debug("trying {.fun credentials_external_account}")
  if (!detect_aws_ec2() || is.null(scopes)) {
    return(NULL)
  }

  scopes <- normalize_scopes(add_email_scope(scopes))

  token <- oauth_external_token(path = path, scopes = scopes)

  if (
    is.null(token$credentials$access_token) ||
      !nzchar(token$credentials$access_token)
  ) {
    NULL
  } else {
    gargle_debug("service account email: {.email {token_email(token)}}")
    token
  }
}

#' Generate OAuth token for an external account
#'
#' @inheritParams credentials_external_account
#'
#' @keywords internal
#' @export
oauth_external_token <- function(
  path = "",
  scopes = "https://www.googleapis.com/auth/cloud-platform"
) {
  info <- jsonlite::fromJSON(path, simplifyVector = FALSE)
  if (!identical(info[["type"]], "external_account")) {
    gargle_debug("JSON does not appear to represent an external account")
    return()
  }

  params <- c(
    list(scopes = scopes),
    info,
    # the most pragmatic way to get super$sign() to work
    # can't implement my own method without needing unexported httr functions
    # request() or build_request()
    as_header = TRUE
  )
  WifToken$new(params = params)
}

#' Token for use with workload identity federation
#'
#' Not intended for direct use. See [credentials_external_account()] instead.
#'
#' @keywords internal
#' @export
WifToken <- R6::R6Class(
  "WifToken",
  inherit = httr::Token2.0,
  list(
    #' @description Get a token via workload identity federation
    #' @param params A list of parameters for `init_oauth_external_account()`.
    #' @return A WifToken.
    initialize = function(params = list()) {
      gargle_debug("WifToken initialize")
      # TODO: any desired validity checks on contents of params

      # NOTE: the final token exchange with
      # https://cloud.google.com/iam/docs/reference/credentials/rest/v1/projects.serviceAccounts/generateAccessToken
      # takes scopes as an **array**, not a space delimited string
      # so we do NOT collapse scopes in this flow
      params$scope <- params$scopes
      self$params <- params

      self$init_credentials()
    },

    #' @description Enact the actual token exchange for workload identity
    #'   federation.
    init_credentials = function() {
      gargle_debug("WifToken init_credentials")
      creds <- init_oauth_external_account(params = self$params)

      # for some reason, the serviceAccounts.generateAccessToken method of
      # Google's Service Account Credentials API returns in camelCase, not
      # snake_case
      # as in, we get this:
      # "accessToken":"ya29.c.KsY..."
      # "expireTime":"2021-06-01T18:01:06Z"
      # instead of this:
      # "access_token": "ya29.a0A..."
      # "expires_in": 3599
      snake_case <- function(x) {
        gsub("([a-z0-9])([A-Z])", "\\1_\\L\\2", x, perl = TRUE)
      }
      names(creds) <- snake_case(names(creds))
      self$credentials <- creds
      self
    },

    #' @description Refreshes the token, which means re-doing the entire token
    #'   flow in this case.
    refresh = function() {
      gargle_debug("WifToken refresh")
      # There's something kind of wrong about this, because it's not a true
      # refresh. But this method is basically required by the way httr currently
      # works.
      # This means that some uses of $refresh() aren't really appropriate for a
      # WifToken.
      # For example, if I attempt token_userinfo(x) on a WifToken that lacks
      # appropriate scope, it fails with 401.
      # httr tries to "fix" things by refreshing the token. But this is
      # not a problem that refreshing can fix.
      # I've now prevented that particular phenomenon in token_userinfo().
      self$init_credentials()
    },

    #' @description Format a [WifToken()].
    #' @param ... Not used.
    format = function(...) {
      x <- list(
        scopes = commapse(base_scope(self$params$scope)),
        credentials = commapse(names(self$credentials))
      )
      c(
        cli::cli_format_method(
          cli::cli_h1("<WifToken (via {.pkg gargle})>")
        ),
        glue("{fr(names(x))}: {fl(x)}")
      )
    },
    #' @description Print a [WifToken()].
    #' @param ... Not used.
    print = function(...) {
      # a format method is not sufficient for WifToken because the parent class
      # has a print method
      cli::cat_line(self$format())
    },

    #' @description Placeholder implementation of required method. Returns `TRUE`.
    can_refresh = function() {
      # TODO: see above re: my ambivalence about the whole notion of refresh with
      # respect to this flow
      TRUE
    },

    # TODO: are cache and load_from_cache really required?
    # alternatively, what if calling them threw an error?
    #' @description Placeholder implementation of required method. Returns self.
    cache = function() self,
    #' @description Placeholder implementation of required method. Returns self.
    load_from_cache = function() self,

    # TODO: are these really required?
    #' @description Placeholder implementation of required method.
    validate = function() {},
    #' @description Placeholder implementation of required method.
    revoke = function() {}
  )
)

# https://docs.aws.amazon.com/AWSEC2/latest/UserGuide/identify_ec2_instances.html
# https://docs.aws.amazon.com/AWSEC2/latest/UserGuide/instance-identity-documents.html
detect_aws_ec2 <- function() {
  if (is_installed("aws.ec2metadata")) {
    return(aws.ec2metadata::is_ec2())
  }
  gargle_debug(
    "
    {.pkg aws.ec2metadata} not installed; can't detect whether running on \\
    EC2 instance"
  )
  FALSE
}

init_oauth_external_account <- function(params) {
  credential_source <- params$credential_source
  if (!identical(credential_source$environment_id, "aws1")) {
    gargle_abort(
      "
      {.pkg gargle}'s workload identity federation flow only supports AWS at \\
      this time."
    )
  }
  subject_token <- aws_subject_token(
    credential_source = credential_source,
    audience = params$audience
  )
  serialized_subject_token <- serialize_subject_token(subject_token)

  federated_access_token <- fetch_federated_access_token(
    params = params,
    subject_token = serialized_subject_token
  )

  fetch_wif_access_token(
    federated_access_token,
    impersonation_url = params[["service_account_impersonation_url"]],
    scope = params[["scope"]]
  )
}

# For AWS, the subject token isn't really a token, but rather the instructions
# necessary to get a token:
#
# From https://cloud.google.com/iam/docs/access-resources-aws#exchange-token
#
# "The GetCallerIdentity token contains the information that you would normally
# include in a request to the AWS GetCallerIdentity() method, as well as the
# signature that you would normally generate for the request.
#
# Also scroll down here, to see the AWS-specific content
# https://cloud.google.com/iam/docs/reference/sts/rest/v1/TopLevel/token
aws_subject_token <- function(credential_source, audience) {
  if (!is_installed(c("aws.ec2metadata", "aws.signature"))) {
    gargle_abort(
      "
      Packages {.pkg aws.ec2metadata} and {.pkg aws.signature} must be \\
      installed in order to use workload identity federation on AWS."
    )
  }

  region <- aws.ec2metadata::instance_document()$region

  regional_cred_verification_url <- glue(
    credential_source[["regional_cred_verification_url"]],
    region = region
  )
  parsed_url <- httr::parse_url(regional_cred_verification_url)

  headers_orig <- list(
    host = parsed_url$hostname,
    # for some reason, this is not included as a signed header unless I provide
    # it
    `x-amz-date` = format(Sys.time(), "%Y%m%dT%H%M%SZ", tz = "UTC"),
    # in contrast, session token IS automatically included if it exists, which
    # it should
    `x-goog-cloud-target-resource` = audience
  )

  verb <- "POST"
  # https://docs.aws.amazon.com/general/latest/gr/signature-version-4.html
  signed <- aws.signature::signature_v4_auth(
    region = region,
    service = "sts",
    verb = verb,
    action = "/",
    query_args = parsed_url$query,
    canonical_headers = headers_orig,
    request_body = ""
  )

  # unfortunately, the headers actually used to make the canonical request are
  # not returned in the signed object, so we dig them out of the canonical
  # request
  req_parts <- strsplit(signed[["CanonicalRequest"]], split = "\n")[[1]]
  f <- function(needle) {
    needle <- paste0("^", needle, ":")
    x <- grep(needle, req_parts, value = TRUE)
    sub(needle, "", x)
  }
  headers <- list(
    host = f("host"),
    `x-amz-date` = f("x-amz-date"),
    `x-amz-security-token` = f("x-amz-security-token"),
    `x-goog-cloud-target-resource` = f("x-goog-cloud-target-resource")
  )

  list(
    url = regional_cred_verification_url,
    method = verb,
    headers = c(
      Authorization = signed$SignatureHeader,
      headers
    )
  )
}

serialize_subject_token <- function(x) {
  # The GCP STS endpoint expects the headers to be formatted as:
  # [
  #   {key: 'Authorization', value: '...'},
  #   {key: 'x-amz-date', value: '...'},
  #   ...
  # ]
  # even though the headers were formatted differently, i.e. in the usual way,
  # when we generated the V4 signature.
  # we're using a purrr compat file, so must call with actual function
  kv <- function(val, nm) list(key = nm, value = val)
  headers_key_value <- unname(imap(x$headers, kv))
  x$headers <- headers_key_value

  # The GCP STS endpoint expects the prepared request to be serialized as a JSON
  # string, which is then URL-encoded.
  utils::URLencode(
    jsonlite::toJSON(x, auto_unbox = TRUE),
    reserved = TRUE
  )
}

# https://datatracker.ietf.org/doc/html/rfc8693
# https://cloud.google.com/iam/docs/reference/sts/rest/v1/TopLevel/token
# https://cloud.google.com/iam/docs/reference/credentials/rest/v1/projects.serviceAccounts/generateAccessToken#authorization-scopes
fetch_federated_access_token <- function(params, subject_token) {
  req <- list(
    method = "POST",
    url = params$token_url,
    body = list(
      audience = params[["audience"]],
      grantType = "urn:ietf:params:oauth:grant-type:token-exchange",
      requestedTokenType = "urn:ietf:params:oauth:token-type:access_token",
      # this request must have one of these scopes:
      # https://www.googleapis.com/auth/cloud-platform
      # https://www.googleapis.com/auth/iam
      # I am hard-wiring the iam scope, guided by the least privilege principle,
      # as it is the narrower of the 2 scopes
      scope = "https://www.googleapis.com/auth/iam",
      subjectTokenType = params[["subject_token_type"]],
      subjectToken = subject_token
    )
  )
  # rfc 8693 says to encode as "application/x-www-form-urlencoded"
  resp <- request_make(req, encode = "form")
  response_process(resp)
}

# https://cloud.google.com/iam/docs/creating-short-lived-service-account-credentials#sa-credentials-oauth
fetch_wif_access_token <- function(
  federated_access_token,
  impersonation_url,
  scope = "https://www.googleapis.com/auth/cloud-platform"
) {
  req <- list(
    method = "POST",
    url = impersonation_url,
    # https://cloud.google.com/iam/docs/reference/credentials/rest/v1/projects.serviceAccounts/generateAccessToken
    # takes scope as an **array**, not a space delimited string
    body = list(scope = scope),
    token = httr::add_headers(
      Authorization = paste("Bearer", federated_access_token$access_token)
    )
  )
  resp <- request_make(req)
  response_process(resp)
}

#' Make a Google API request
#'
#' @description
#' Intended primarily for internal use in client packages that provide
#' high-level wrappers for users. `request_make()` does relatively little:
#' * Calls an HTTP method.
#' * Adds a user agent.
#' * Enforces `"json"` as the default for `encode`. This differs from httr's
#'   default behaviour, but aligns better with Google APIs.
#'
#' Typically the input is created with [request_build()] and the output is
#' processed with [response_process()].
#'
#' @param x List. Holds the components for an HTTP request, presumably created
#'   with [request_develop()] or [request_build()]. Must contain a `method` and
#'   `url`. If present, `body` and `token` are used.
#' @param user_agent A user agent string, prepared by [httr::user_agent()]. When
#'   in doubt, a client package should have an internal function that extends
#'   `gargle_user_agent()` by prepending its return value with the client
#'   package's name and version.
#' @inheritParams httr::POST
#' @param ... Optional arguments passed through to the HTTP method. Currently
#'   neither gargle nor httr checks that all are used, so be aware that unused
#'   arguments may be silently ignored.
#'
#' @return Object of class `response` from [httr].
#' @export
#' @family requests and responses
#' @examples
#' \dontrun{
#' req <- gargle::request_build(
#'   method = "GET",
#'   path = "path/to/the/resource",
#'   token = "PRETEND_I_AM_TOKEN"
#' )
#' gargle::request_make(req)
#' }
request_make <- function(
  x,
  ...,
  encode = "json",
  user_agent = gargle_user_agent()
) {
  check_string(x$method, allow_empty = FALSE)
  method <- switch(
    x$method,
    GET = httr::GET,
    POST = httr::POST,
    PATCH = httr::PATCH,
    PUT = httr::PUT,
    DELETE = httr::DELETE,
    gargle_abort("Not a recognized HTTP method: {.code {x$method}}.")
  )
  method(
    url = x$url,
    body = x$body,
    x$token,
    encode = encode,
    user_agent,
    ...
  )
}

gargle_user_agent <- function() {
  httr::user_agent(paste0(
    "gargle/",
    utils::packageVersion("gargle"),
    " ",
    "(GPN:RStudio; )",
    " ",
    "httr/",
    utils::packageVersion("httr")
  ))
}

test_that("request_make() errors for invalid HTTP methods", {
  expect_error(
    request_make(list(method = httr::GET)),
    "is.character(x$method) is not TRUE",
    fixed = TRUE
  )
  expect_error(
    request_make(list(method = "PETCH")),
    "Not a recognized HTTP method"
  )
})

#' API key for demonstration purposes
#'
#' @description Some APIs accept requests for public resources, in which case
#'   the request must be sent with an API key in lieu of a token. This function
#'   provides an API key for limited use in prototyping and for testing and
#'   documentation of gargle itself. This key may be deleted or rotated at any
#'   time. There are no guarantees about which APIs are enabled. DO NOT USE THIS
#'   IN A PACKAGE or for anything other than interactive, small-scale
#'   experimentation.
#'
#'   You can get your own API key, without these limitations. See the [How to
#'   get your own API
#'   credentials](https://gargle.r-lib.org/articles/get-api-credentials.html)
#'   vignette for more details.
#'
#' @export
#' @keywords internal
#' @examples
#' gargle_api_key()
gargle_api_key <- function() {
  gak()
}

#' Assets for internal use
#'
#' Assets for use inside specific packages maintained by the tidyverse team.
#'
#' @name internal-assets
NULL

#' @export
#' @keywords internal
#' @rdname internal-assets
tidyverse_api_key <- function() {
  check_permitted_package(caller_env())
  tak()
}

#' Load a user-provided token
#'
#' @description
#' This function is designed to pass its `token` input through, after doing a
#' few checks and some light processing:
#'   * If `token` has class `request`, i.e. it is a token that has been prepared
#'     with [httr::config()], the `auth_token` component is extracted. For
#'     example, such input could be returned by `googledrive::drive_token()`
#'     or `bigrquery::bq_token()`.
#'   * If `token` is an instance of `Gargle2.0` (so: a gargle-obtained user
#'     token), checks that it appears to be a Google OAuth token, based on its
#'     embedded `oauth_endpoint`. Refreshes the token, if it's refreshable.
#'   * Returns the `token`.
#'
#' There is no point in providing `scopes`. They are ignored because the
#' `scopes` associated with the token have already been baked in to the token
#' itself and gargle does not support incremental authorization. The main point
#' of `credentials_byo_oauth2()` is to allow `token_fetch()` (and packages that
#' wrap it) to accommodate a "bring your own token" workflow.
#'
#' This also makes it possible to obtain a token with one package and then
#' register it for use with another package. For example, the default scope
#' requested by googledrive is also sufficient for operations available in
#' googlesheets4. You could use a shared token like so:
#' ```
#' library(googledrive)
#' library(googlesheets4)
#' drive_auth(email = "jane_doe@example.com")
#' gs4_auth(token = drive_token())
#' # work with both packages freely now, with the same identity
#' ```
#'
#' @inheritParams token_fetch
#' @inheritParams token-info
#'
#' @return An [Token2.0][httr::Token-class].
#' @family credential functions
#' @export
#' @examples
#' \dontrun{
#' # assume `my_token` is a Token2.0 object returned by a function such as
#' # credentials_user_oauth2()
#' credentials_byo_oauth2(token = my_token)
#' }
credentials_byo_oauth2 <- function(scopes = NULL, token, ...) {
  gargle_debug("Trying {.fun credentials_byo_oauth} ...")
  if (inherits(token, "request")) {
    token <- token$auth_token
  }
  stopifnot(inherits(token, "Token2.0"))

  if (!is.null(scopes)) {
    gargle_debug(c(
      "!" = "The {.arg scopes} cannot be specified when user brings their own \\
             OAuth token.",
      "i" = "The {.arg scopes} are already implicit in the token.",
      "i" = "Requested {.arg scopes} are effectively ignored."
    ))

    declared_scopes <- normalize_scopes(token$params$scope)
    requested_scopes <- normalize_scopes(scopes)

    if (!setequal(requested_scopes, declared_scopes)) {
      gargle_debug(c(
        "!" = "Token's declared scopes are not the same as the requested \\
               scopes.",
        "i" = "Scopes declared in token: {commapse(base_scope(declared_scopes))}",
        "i" = "Requested scopes: {commapse(base_scope(requested_scopes))}"
      ))
    }
  }

  if (inherits(token, "Gargle2.0")) {
    check_endpoint(token$endpoint)
    if (token$can_refresh()) {
      token$refresh()
    }
  }

  token
}

check_endpoint <- function(endpoint, call = caller_env()) {
  stopifnot(inherits(endpoint, "oauth_endpoint"))
  urls <- endpoint[c("authorize", "access", "validate", "revoke")]
  urls_ok <- all(grepl("google", urls))
  if (!urls_ok) {
    gargle_abort("Token doesn't use Google's OAuth endpoint.", call = call)
  }
  endpoint
}

test_that("credentials_byo_oauth2() demands a Token2.0", {
  expect_error(
    credentials_byo_oauth2(token = "a_naked_access_token"),
    'inherits(token, "Token2.0") is not TRUE',
    fixed = TRUE
  )
})

test_that("credentials_byo_oauth2() rejects a token that obviously not Google", {
  token <- httr::Token2.0$new(
    app = httr::oauth_app("x", "y", "z"),
    endpoint = httr::oauth_endpoints("github"),
    credentials = list(access_token = "ACCESS_TOKEN"),
    cache_path = FALSE
  )

  expect_error(
    credentials_byo_oauth2(token = token),
    "doesn't use Google's OAuth endpoint"
  )
})

test_that("credentials_byo_oauth2() just passes valid input through", {
  token <- httr::Token2.0$new(
    app = httr::oauth_app("x", "y", "z"),
    endpoint = httr::oauth_endpoints("google"),
    credentials = list(access_token = "ACCESS_TOKEN"),
    cache_path = FALSE
  )
  expect_equal(credentials_byo_oauth2(token = token), token)
})

test_that("credentials_byo_oauth2() extracts a token from a request", {
  token <- httr::Token2.0$new(
    app = httr::oauth_app("x", "y", "z"),
    endpoint = httr::oauth_endpoints("google"),
    credentials = list(access_token = "ACCESS_TOKEN"),
    cache_path = FALSE
  )
  configured_token <- httr::config(token = token)
  expect_equal(
    credentials_byo_oauth2(token = configured_token),
    token
  )
})

test_that("credentials_byo_oauth2() demands a Token2.0", {
  expect_snapshot(
    credentials_byo_oauth2(token = "a_naked_access_token"),
    error = TRUE
  )
})

test_that("credentials_byo_oauth2() rejects a token that obviously not Google", {
  fauxen <- gargle2.0_token(
    email = "a@example.org",
    client = gargle_oauth_client(
      id = "CLIENT_ID",
      secret = "SECRET",
      name = "CLIENT"
    ),
    credentials = list(access_token = "ACCESS_TOKEN_1"),
    cache = FALSE
  )
  fauxen$endpoint <- httr::oauth_endpoints("github")

  expect_snapshot(credentials_byo_oauth2(token = fauxen), error = TRUE)
})

test_that("credentials_byo_oauth2() just passes valid input through", {
  token <- httr::Token2.0$new(
    app = httr::oauth_app("x", "y", "z"),
    endpoint = httr::oauth_endpoints("google"),
    credentials = list(access_token = "ACCESS_TOKEN"),
    cache_path = FALSE
  )
  expect_equal(credentials_byo_oauth2(token = token), token)
})

test_that("credentials_byo_oauth2() extracts a token from a request", {
  token <- httr::Token2.0$new(
    app = httr::oauth_app("x", "y", "z"),
    endpoint = httr::oauth_endpoints("google"),
    credentials = list(access_token = "ACCESS_TOKEN"),
    cache_path = FALSE
  )
  configured_token <- httr::config(token = token)
  expect_equal(
    credentials_byo_oauth2(token = configured_token),
    token
  )
})

#' Generate a gargle token
#'
#' Constructor function for objects of class [Gargle2.0].
#'
#' @param email Optional. If specified, `email` can take several different
#'   forms:
#' * `"jane@gmail.com"`, i.e. an actual email address. This allows the user to
#' target a specific Google identity. If specified, this is used for token
#' lookup, i.e. to determine if a suitable token is already available in the
#' cache. If no such token is found, `email` is used to pre-select the targeted
#' Google identity in the OAuth chooser. (Note, however, that the email
#' associated with a token when it's cached is always determined from the token
#' itself, never from this argument).
#' * `"*@example.com"`, i.e. a domain-only glob pattern. This can be helpful if
#' you need code that "just works" for both `alice@example.com` and
#' `bob@example.com`.
#' * `TRUE` means that you are approving email auto-discovery. If exactly one
#'   matching token is found in the cache, it will be used.
#' * `FALSE` or `NA` mean that you want to ignore the token cache and force a
#' new OAuth dance in the browser.
#'
#' Defaults to the option named `"gargle_oauth_email"`, retrieved by
#' [gargle::gargle_oauth_email()] (unless a wrapper package implements different
#' default behavior).
#' @param client A Google OAuth client, preferably constructed via
#'   [gargle::gargle_oauth_client_from_json()], which returns an instance of
#'   `gargle_oauth_client`. For backwards compatibility, for a limited time,
#'   gargle will still accept an "OAuth app" created with [httr::oauth_app()].
#' @param package Name of the package requesting a token. Used in messages.
#' @param scope A character vector of scopes to request.
#' @param use_oob Whether to use out-of-band authentication (or, perhaps, a
#'   variant implemented by gargle and known as "pseudo-OOB") when first
#'   acquiring the token. Defaults to the value returned by
#'   [gargle::gargle_oob_default()]. Note that (pseudo-)OOB auth only affects
#'   the initial OAuth dance. If we retrieve (and possibly refresh) a
#'   cached token, `use_oob` has no effect.
#'
#'   If the OAuth client is provided implicitly by a wrapper package, its type
#'   probably defaults to the value returned by
#'   [gargle::gargle_oauth_client_type()]. You can take control of the client
#'   type by setting `options(gargle_oauth_client_type = "web")` or
#'   `options(gargle_oauth_client_type = "installed")`.
#' @param cache Specifies the OAuth token cache. Defaults to the option named
#'   `"gargle_oauth_cache"`, retrieved via [gargle::gargle_oauth_cache()].
#' @inheritParams httr::oauth2.0_token
#' @param ... Absorbs arguments intended for use by other credential functions.
#'   Not used.
#' @param app `r lifecycle::badge('deprecated')` Replaced by the `client`
#'   argument.
#' @return An object of class [Gargle2.0], either new or loaded from the cache.
#' @export
#' @examples
#' \dontrun{
#' gargle2.0_token()
#' }
gargle2.0_token <- function(
  email = gargle_oauth_email(),
  client = gargle_client(),
  package = "gargle",
  ## params start
  scope = NULL,
  use_oob = gargle_oob_default(),
  ## params end
  credentials = NULL,
  cache = if (is.null(credentials)) gargle_oauth_cache() else FALSE,
  ...,
  app = deprecated()
) {
  if (lifecycle::is_present(app)) {
    lifecycle::deprecate_soft(
      "1.5.0",
      "gargle2.0_token(app)",
      "gargle2.0_token(client)"
    )
    client <- app
  }

  params <- list(
    scope = scope,
    use_oob = use_oob,
    as_header = TRUE
  )

  # pseudo-OOB flow
  client_type <- if (inherits(client, "gargle_oauth_client")) {
    client$type
  } else {
    NA
  }
  if (use_oob && identical(client_type, "web")) {
    params$oob_value <- select_pseudo_oob_value(client$redirect_uris)
  }
  # params$oob_value is deliberately left unspecified for conventional OOB,
  # with the intent of falling back to urn:ietf:wg:oauth:2.0:oob

  # this allows pseudo-OOB auth to work on colab, because:
  # 1) gargle's attempts to communicate with the user route through readline()
  #    which is shimmed in Jupyter (and therefore Colab)
  # 2) httr >= 1.4.5 honors the "rlang_interactive" option when deciding whether
  #   it will try the oauth dance
  if (is_google_colab()) {
    withr::local_options(rlang_interactive = TRUE)
  }

  Gargle2.0$new(
    email = email,
    client = client,
    package = package,
    params = params,
    credentials = credentials,
    cache_path = cache
  )
}

#' OAuth2 token objects specific to Google APIs
#'
#' @description
#' `Gargle2.0` is based on the [`Token2.0`][httr::Token-class] class provided in
#' httr. The preferred way to create a `Gargle2.0` token is through the
#' constructor function [gargle2.0_token()]. Key differences with `Token2.0`:
#' * The key for a cached `Token2.0` comes from hashing the endpoint, client,
#' and scopes. For the `Gargle2.0` subclass, the identifier or key is expanded
#' to include the email address associated with the token. This makes it easier
#' to work with Google APIs with multiple identities.
#' * `Gargle2.0` tokens are cached, by default, at the user level, following the
#' XDG spec for storing user-specific data and cache files. In contrast, the
#' default location for `Token2.0` is `./.httr-oauth`, i.e. in current working
#' directory. `Gargle2.0` behaviour makes it easier to reuse tokens across
#' projects and makes it less likely that tokens are accidentally synced to a
#' remote location like GitHub or DropBox.
#' * Each `Gargle2.0` token is cached in its own file. The token cache is a
#' directory of such files. In contrast, `Token2.0` tokens are cached as
#' components of a list, which is typically serialized to `./.httr-oauth`.
#'

#' @param email Optional email address. See [gargle2.0_token()] for full
#'   details.
#' @param client An OAuth consumer application.
#' @param package Name of the package requesting a token. Used in messages.
#' @param credentials Exists largely for testing purposes.
#' @param params A list of parameters for the internal function
#'   `init_oauth2.0()`, which is a modified version of [httr::init_oauth2.0()].
#'   gargle actively uses `scope` and `use_oob`, but does not use `user_params`,
#'   `type`, `as_header` (hard-wired to `TRUE`), `use_basic_auth` (accept
#'   default of `use_basic_auth = FALSE`), `config_init`, or
#'   `client_credentials`.
#' @param cache_path Specifies the OAuth token cache. Read more in
#'   [gargle::gargle_oauth_cache()].
#' @param app `r lifecycle::badge('deprecated')` Use `client` instead.
#'
#' @keywords internal
#' @export
#' @name Gargle-class
Gargle2.0 <- R6::R6Class(
  "Gargle2.0",
  inherit = httr::Token2.0,
  list(
    #' @field email Email associated with the token.
    email = NULL,
    #' @field package Name of the package requesting a token. Used in messages.
    package = NULL,
    #' @field client An OAuth client.
    client = NULL,
    #' @description Create a Gargle2.0 token
    #' @return A Gargle2.0 token.
    initialize = function(
      email = gargle_oauth_email(),
      client = gargle_client(),
      package = "gargle",
      credentials = NULL,
      params = list(),
      cache_path = gargle_oauth_cache(),
      app = deprecated()
    ) {
      gargle_debug("Gargle2.0 initialize")
      # I'm using deprecate_warn() intentionally here. Most folks should be
      # instantiating through gargle2.0_token() anyway, so anyone who sees this
      # warning probably needs to see it.
      if (lifecycle::is_present(app)) {
        lifecycle::deprecate_warn(
          "1.5.0",
          "Gargle2.0$initialize(app)",
          "Gargle2.0$initialize(client)"
        )
        client <- app
      }
      stopifnot(
        is.null(email) ||
          is_scalar_character(email) ||
          isTRUE(email) ||
          isFALSE(email) ||
          is_na(email),
        is.oauth_app(client),
        is_string(package),
        is.list(params)
      )
      if (identical(email, "")) {
        gargle_abort(c(
          "{.arg email} must not be \"\" (the empty string).",
          "i" = "Do you intend to consult an env var, but it's unset?"
        ))
      }
      if (isTRUE(email)) {
        email <- "*"
      }
      if (isFALSE(email) || is_na(email)) {
        email <- NA_character_
      }
      # https://developers.google.com/identity/protocols/OpenIDConnect#login-hint
      # optional hint for the auth server to pre-fill the email box
      login_hint <- if (is_string(email) && !startsWith(email, "*")) email

      self$endpoint <- gargle_oauth_endpoint()
      self$email <- email
      self$client <- client
      # for backwards compatibility and also because the parent class has $app;
      # I can never remove it
      self$app <- client
      self$package <- package
      params$scope <- normalize_scopes(add_email_scope(params$scope))
      params$query_authorize_extra <- list(login_hint = login_hint)
      self$params <- params
      self$cache_path <- cache_establish(cache_path)

      if (!is.null(credentials)) {
        # Use credentials created elsewhere - usually for tests
        gargle_debug("credentials provided directly")
        self$credentials <- credentials
        return(self$cache())
      }

      # Are credentials cached already?
      if (self$load_from_cache()) {
        self
      } else {
        gargle_debug("no matching token in the cache")
        self$init_credentials()
        self$email <- token_email(self) %||% NA_character_
        self$cache()
      }
    },
    #' @description Format a Gargle2.0 token
    #' @param ... Not used.
    format = function(...) {
      x <- list(
        oauth_endpoint = "google",
        client = self$client$name,
        email = cli::format_inline("{.email {self$email}}"),
        scopes = commapse(base_scope(self$params$scope)),
        credentials = commapse(names(self$credentials))
      )
      c(
        cli::cli_format_method(
          cli::cli_h1("<Token (via {.pkg gargle})>")
        ),
        glue("{fr(names(x))}: {fl(x)}")
      )
    },
    #' @description Print a Gargle2.0 token
    #' @param ... Not used.
    print = function(...) {
      # a format method is not sufficient for Gargle2.0 because the parent class
      # has a print method
      cli::cat_line(self$format())
    },
    #' @description Generate the email-augmented hash of a Gargle2.0 token
    hash = function() {
      paste(super$hash(), self$email, sep = "_")
    },
    #' @description Put a Gargle2.0 token into the cache
    cache = function() {
      token_into_cache(self)
      self
    },
    #' @description (Attempt to) get a Gargle2.0 token from the cache
    load_from_cache = function() {
      gargle_debug("loading token from the cache")
      if (is.null(self$cache_path) || is_na(self$email)) {
        return(FALSE)
      }

      gargle_debug("email: {.email {self$email}}")
      gargle_debug("oauth client name: {self$client$name}")
      gargle_debug("oauth client name: {self$client$type}")
      gargle_debug("oauth client id: {self$client$id}")
      gargle_debug("scopes: {commapse(base_scope(self$params$scope))}")

      cached <- token_from_cache(self)
      if (is.null(cached)) {
        return(FALSE)
      }

      gargle_debug("matching token found in the cache")
      self$endpoint <- cached$endpoint
      self$email <- cached$email
      self$client <- cached$client
      self$app <- cached$client
      self$credentials <- cached$credentials
      self$params <- cached$params
      TRUE
    },
    #' @description (Attempt to) refresh a Gargle2.0 token
    refresh = function() {
      cred <- refresh_oauth2.0(
        self$endpoint,
        self$client,
        self$credentials,
        package = self$package
      )
      if (is.null(cred)) {
        token_remove_from_cache(self)
        # It's tricky to decide what to do here. Currently we return the current,
        # invalid, unrefreshed token, but we clear the refresh_token field, to
        # prevent subsequent refresh attempts.
        #
        # Analysis from a BYO token POV:
        # I've decided the status quo may be the best move, because it causes
        # token_fetch() to return instead of moving on to try other methods. If
        # someone provides token_fetch(token =), I think it's clear that they
        # want/hope to use that token and they don't want to end up doing the
        # OAuth browser dance. If we threw an error or returned NULL,
        # token_fetch() would just keep going. The refresh failure does throw a
        # visible warning:
        #
        # Warning message:
        # Unable to refresh token: invalid_grant
        # • Token has been expired or revoked.
        #
        # However, this does mean that functions like PKG_has_token() still return
        # TRUE and that some other method must be used to find out if we have a
        # *valid* token. gargle::token_tokeninfo() and API-specific functions for
        # "tell me about the current user" are good candidates, such as
        # gmailr::gm_profile() or googledrive::drive_user().
        self$credentials$refresh_token <- NULL
      } else {
        self$credentials <- cred
        self$cache()
      }
      self
    },
    #' @description Initiate a new Gargle2.0 token
    init_credentials = function() {
      gargle_debug("initiating new token")
      if (is_interactive()) {
        if (!isTRUE(self$params$use_oob) && !is_hosted_session()) {
          encourage_httpuv()
        }
        self$credentials <- init_oauth2.0(
          self$endpoint,
          self$client,
          scope = self$params$scope,
          use_oob = self$params$use_oob,
          oob_value = self$params$oob_value,
          query_authorize_extra = self$params$query_authorize_extra
        )
      } else {
        # TODO: good candidate for an eventual sub-classed gargle error
        # would be useful in testing to know that this is exactly where we aborted
        gargle_abort("OAuth2 flow requires an interactive session.")
      }
    }
  )
)

encourage_httpuv <- function() {
  if (!is_interactive() || isTRUE(is_installed("httpuv"))) {
    return(invisible())
  }
  choice <- cli_menu(
    "The {.pkg httpuv} package enables a nicer Google auth experience, in many \\
    cases, but it isn't installed.",
    "Would you like to install it now?",
    choices = c("Yes", "No")
  )
  if (choice == 1) {
    utils::install.packages("httpuv")
  }
  invisible()
}

# I want to encourage users to create an OAuth client (newer httr2-y language)
# directly from downloaded JSON, using gargle_oauth_client_from_json().
# Sometimes there are multiple URIs and I think we can usually figure out which
# one to use for the pseudo-OOB flow.
select_pseudo_oob_value <- function(redirect_uris) {
  # https://developers.google.com/identity/protocols/oauth2/resources/oob-migration#inspect-your-application-code
  bad_values <- c(
    "urn:ietf:wg:oauth:2.0:oob",
    "urn:ietf:wg:oauth:2.0:oob:auto",
    "oob"
  )
  redirect_uris <- setdiff(redirect_uris, bad_values)

  # https://developers.google.com/identity/protocols/oauth2/web-server#uri-validation
  bad_regex <- "^http[s]?://(localhost|127.0.0.1)"
  redirect_uris <- grep(bad_regex, redirect_uris, value = TRUE, invert = TRUE)
  redirect_uris <- grep("^https", redirect_uris, value = TRUE)

  # inspired by these guidelines re: URIs associated with URL shorteners:
  # 'redirect URI must either contain "/google-callback/" in its path or end
  # with "/google-callback"'
  m <- grep("/google-callback(/|$)", redirect_uris)
  if (length(m) > 0) {
    redirect_uris <- redirect_uris[m]
  }

  if (length(redirect_uris) == 0) {
    gargle_abort(
      '
      OAuth client does not have a redirect URI suitable for the pseudo-OOB \\
      flow.'
    )
  }

  if (length(redirect_uris) > 1) {
    msg <- c(
      "Can't determine which redirect URI to use for the pseudo-OOB flow:",
      set_names(redirect_uris, ~ rep_along(., "*"))
    )
    gargle_abort(msg)
  }

  redirect_uris
}

test_that("email is ingested correctly", {
  fauxen_email <- function(email = NULL) {
    gargle2.0_token(email = email, credentials = list(a = 1))$email
  }
  expect_null(fauxen_email())
  expect_null(fauxen_email(NULL))
  expect_equal(fauxen_email(NA), NA_character_)
  expect_equal(fauxen_email(FALSE), NA_character_)
  expect_equal(fauxen_email(TRUE), "*")
  expect_equal(fauxen_email("a@example.org"), "a@example.org")
})

test_that("email can be set in option", {
  fauxen_email <- function(email = NULL) {
    withr::with_options(
      list(gargle_oauth_email = email),
      gargle2.0_token(credentials = list(a = 1))$email
    )
  }
  expect_null(fauxen_email(NULL))
  expect_equal(fauxen_email(NA), NA_character_)
  expect_equal(fauxen_email(FALSE), NA_character_)
  expect_equal(fauxen_email(TRUE), "*")
  expect_equal(fauxen_email("a@example.org"), "a@example.org")
})

test_that("Attempt to initiate OAuth2 flow fails if non-interactive", {
  local_interactive(FALSE)
  expect_snapshot(gargle2.0_token(cache = FALSE), error = TRUE)
})

test_that("`email = NA`, `email = FALSE` means we don't consult the cache", {
  cache_folder <- path_temp("email-na-test")
  withr::defer(dir_delete(cache_folder))
  local_interactive(FALSE)

  # make sure there's one token in the cache and that, by default, we use it
  fauxen_in <- gargle2.0_token(
    email = "a@example.org",
    credentials = list(a = 1),
    cache = cache_folder
  )
  # can't use with_gargle_verbosity() here, because we use
  # local_gargle_verbosity("info") in token_match(), to force the user to
  # see messaging about auto-discovery
  suppressMessages(
    fauxen_out <- gargle2.0_token(cache = cache_folder)
  )
  expect_gargle2.0_token(fauxen_in, fauxen_out)

  # `email = NA` and `email = FALSE` prevent the cache from being consulted
  expect_snapshot(
    gargle2.0_token(email = NA, cache = cache_folder),
    error = TRUE
  )
  expect_snapshot(
    gargle2.0_token(email = FALSE, cache = cache_folder),
    error = TRUE
  )
})

test_that("Gargle2.0 prints nicely", {
  fauxen <- gargle2.0_token(
    email = "a@example.org",
    client = gargle_oauth_client(
      id = "CLIENT_ID",
      secret = "SECRET",
      name = "CLIENT"
    ),
    credentials = list(a = 1),
    cache = FALSE
  )
  expect_snapshot(print(fauxen))
})

test_that("we reject redirect URIs from conventional OOB for pseudo-OOB flow", {
  expect_snapshot(
    error = TRUE,
    select_pseudo_oob_value("urn:ietf:wg:oauth:2.0:oob")
  )
  expect_error(
    select_pseudo_oob_value("urn:ietf:wg:oauth:2.0:oob:auto"),
    class = "gargle_error"
  )
  expect_error(
    select_pseudo_oob_value("oob"),
    class = "gargle_error"
  )
})

test_that("we reject local web server redirect URIs for pseudo-OOB flow", {
  expect_snapshot(
    error = TRUE,
    select_pseudo_oob_value("http://localhost")
  )
  expect_error(
    select_pseudo_oob_value("http://localhost:4000"),
    class = "gargle_error"
  )
  expect_error(
    select_pseudo_oob_value("http://127.0.0.1:1410"),
    class = "gargle_error"
  )
})

test_that("we reject non-https redirect URIs for pseudo-OOB flow", {
  expect_error(
    select_pseudo_oob_value("http://example.com/google-callback/blah.html"),
    class = "gargle_error"
  )
})

test_that("we insist on finding exactly one redirect URI for pseudo-OOB flow", {
  redirect_uris <- c(
    "https://example.com/google-callback/one.html",
    "https://example.com/google-callback/two.html"
  )
  expect_snapshot(error = TRUE, select_pseudo_oob_value(redirect_uris))
})

test_that("we can identify the redirect URI suitable for pseudo-OOB flow", {
  redirect_uris <- c(
    "http://localhost:8111/",
    "http://localhost:8111",
    "http://127.0.0.1:8100/",
    "https://codepen.io/USER/full/abcdef123456"
  )
  expect_equal(select_pseudo_oob_value(redirect_uris), redirect_uris[4])
})

test_that("gargle2.0_token(app) is deprecated but still works", {
  withr::local_options(lifecycle_verbosity = "warning")
  client <- gargle_oauth_client(
    id = "CLIENT_ID",
    secret = "SECRET",
    name = "ABC"
  )

  expect_snapshot(
    t <- gargle2.0_token(email = NA, credentials = list(a = 1), app = client)
  )
  expect_equal(t$client$name, "ABC")
  expect_equal(t$app$appname, "ABC")
})

#' Credential function registry
#'
#' Functions to query or manipulate the registry of credential functions
#' consulted by [token_fetch()].
#'
#' @name cred_funs
#'
#' @seealso [token_fetch()], which is where the registry is actually used.
#' @return A list of credential functions or `NULL`.
#' @examples
#' names(cred_funs_list())
#'
#' creds_one <- function(scopes, ...) {}
#'
#' cred_funs_add(one = creds_one)
#' cred_funs_add(two = creds_one, three = creds_one)
#' names(cred_funs_list())
#'
#' cred_funs_add(two = NULL)
#' names(cred_funs_list())
#'
#' # restore the default list
#' cred_funs_set_default()
#'
#' # remove one specific credential fetcher
#' cred_funs_add(credentials_gce = NULL)
#' names(cred_funs_list())
#'
#' # force the use of one specific credential fetcher
#' cred_funs_set(list(credentials_user_oauth2 = credentials_user_oauth2))
#' names(cred_funs_list())
#'
#' # restore the default list
#' cred_funs_set_default()
#'
#' # run some code with a temporary change to the registry
#' # creds_one ONLY
#' with_cred_funs(
#'   list(one = creds_one),
#'   names(cred_funs_list())
#' )
#' # add creds_one to the list
#' with_cred_funs(
#'   list(one = creds_one),
#'   names(cred_funs_list()),
#'   action = "modify"
#' )
#' # remove credentials_gce
#' with_cred_funs(
#'   list(credentials_gce = NULL),
#'   names(cred_funs_list()),
#'   action = "modify"
#' )
NULL

#' @describeIn cred_funs Get the list of registered credential functions.
#' @export
cred_funs_list <- function() {
  gargle_env$cred_funs
}

#' @describeIn cred_funs Register one or more new credential fetching functions.
#'   Function(s) are added to the *front* of the list. So:
#'   * "First registered, last tried."
#'   * "Last registered, first tried."
#'
#' Can also be used to *remove* a function from the registry.
#'
#' @param ... <[`dynamic-dots`][rlang::dyn-dots]> One or more credential
#'   functions, in `name = value` form. Each credential function is subject to a
#'   superficial check that it at least "smells like" a credential function: its
#'   first argument must be named `scopes`, and its signature must include
#'   `...`. To remove a credential function, you can use a specification like
#'   `name = NULL`.
#' @export
cred_funs_add <- function(...) {
  dots <- dots_list(
    ...,
    .named = NULL,
    .ignore_empty = "all",
    .preserve_empty = FALSE,
    .homonyms = "error",
    .check_assign = TRUE
  )

  cred_funs_check(dots, allow_null = TRUE)

  nms_to_remove <- names(dots)[map_lgl(dots, is.null)]
  cf <- cred_funs_list()
  cf[nms_to_remove] <- NULL
  cred_funs_set(cf)
  dots <- dots[!map_lgl(dots, is.null)]

  dup_nm <- names(dots) %in% names(cred_funs_list())
  if (any(dup_nm)) {
    n_dup_nm <- sum(dup_nm)
    gargle_abort(c(
      "{cli::qty(n_dup_nm)}{?This/These} name{?s} already {?appears/appear} \\
      in the credential function registry:",
      x = "{.field {names(dots)[dup_nm]}}"
    ))
  }

  # add them in reverse order, to mimic what would happen if they'd been added
  # one-at-a-time
  cf <- cred_funs_list()
  cred_funs_set(c(rev(dots), cf))

  invisible(cred_funs_list())
}

#' @describeIn cred_funs Register a list of credential fetching functions.
#'
#' @param funs A named list of credential functions.
#' @param ls `r lifecycle::badge("deprecated")` This argument has been renamed
#'   to `funs`.
#' @export
cred_funs_set <- function(funs, ls = deprecated()) {
  if (lifecycle::is_present(ls)) {
    lifecycle::deprecate_warn(
      when = "1.3.0",
      what = "cred_funs_set(ls)",
      with = "cred_funs_set(funs)",
    )
    funs = ls
  }

  cred_funs_check(funs, allow_null = FALSE)
  gargle_env$cred_funs <- funs
  invisible(cred_funs_list())
}

#' @describeIn cred_funs Clear the credential function registry.
#' @export
cred_funs_clear <- function() {
  gargle_env$cred_funs <- list()
  invisible(cred_funs_list())
}

#' @describeIn cred_funs Return the default list of credential functions.
#' @export
cred_funs_list_default <- function() {
  list(
    credentials_byo_oauth2       = credentials_byo_oauth2,
    credentials_service_account  = credentials_service_account,
    credentials_external_account = credentials_external_account,
    credentials_app_default      = credentials_app_default,
    credentials_gce              = credentials_gce,
    credentials_user_oauth2      = credentials_user_oauth2
  )
}

#' @describeIn cred_funs Reset the registry to the gargle default.
#' @export
cred_funs_set_default <- function() {
  cred_funs_set(cred_funs_list_default())
}

#' @describeIn cred_funs Modify the credential function registry in the current
#'   scope. It is an example of the `local_*()` functions in \pkg{withr}.
#' @param action Whether to use `funs` to replace or modify the registry with
#'   funs:
#'   * `"replace"` does `cred_funs_set(funs)`
#'   * `"modify"` does `cred_funs_add(!!!funs)`
#' @param .local_envir The environment to use for scoping. Defaults to current
#'   execution environment.
#' @export
local_cred_funs <- function(funs = cred_funs_list_default(),
                            action = c("replace", "modify"),
                            .local_envir = parent.frame()) {
  action <- arg_match(action)

  cred_funs_orig <- cred_funs_list()
  withr::defer(cred_funs_set(cred_funs_orig), envir = .local_envir)

  switch(
    action,
    replace = cred_funs_set(funs),
    modify = cred_funs_add(!!!funs)
  )
}

#' @describeIn cred_funs Evaluate `code` with a temporarily modified credential
#'   function registry. It is an example of the `with_*()` functions in
#'   \pkg{withr}.
#' @param code Code to run with temporary credential function registry.
#' @export
with_cred_funs <- function(funs = cred_funs_list_default(),
                           code,
                           action = c("replace", "modify")) {
  local_cred_funs(funs = funs, action = action)
  force(code)
}

cred_funs_check <- function(ls, allow_null = FALSE) {
  if (allow_null) {
    not_cred_fun <- !map_lgl(ls, is.null) & !map_lgl(ls, is_cred_fun)
  } else {
    not_cred_fun <- !map_lgl(ls, is_cred_fun)
  }
  if (any(not_cred_fun)) {
    gargle_abort(c(
      "Not a valid credential function:",
      x = "Element{?s} {as.character(which(not_cred_fun))}"
    ))
  }

  if (!is_dictionaryish(ls)) {
    gargle_abort("Each credential function must have a unique name")
  }

  invisible()
}

#' Check that f is a viable credential fetching function
#'
#' In the abstract, a credential fetching function is any function which takes a
#' set of scopes and any number of additional arguments, and returns either a
#' valid [`httr::Token`][httr::Token-class] or `NULL`.
#'
#' Here we say that a function is valid if its first argument is named `scopes`,
#' and it includes `...` as an argument, since it's difficult to actually check
#' the behavior of the function.
#'
#' @param f A function to check.
#' @keywords internal
#' @noRd
#' @examples
#' f <- function(scopes, ...) {}
#' is_cred_fun(f)
is_cred_fun <- function(f) {
  if (!is.function(f)) {
    return(FALSE)
  }
  args <- names(formals(f))
  args[1] == "scopes" && "..." %in% args
}

# ---- file ID that does not exist
googledrive::drive_deauth()
req <- googledrive::request_generate(
  endpoint = "drive.files.get",
  params = list(
    fileId = "NOPE_NOT_A_GOOD_ID"
  )
)
resp <- googledrive::request_make(req)

stopifnot(httr::status_code(resp) == 404)
saveRDS(
  gargle:::redact_response(resp),
  testthat::test_path(
    "fixtures",
    "drive-files-get-nonexistent-file-id_404.rds"
  ),
  version = 2
)

gargle::response_process(resp)

#' Get an OAuth token for a user
#'
#' @description Consults the token cache for a suitable OAuth token and, if
#'   unsuccessful, gets a token via the browser flow. A cached token is suitable
#'   if it's compatible with the user's request in this sense:
#'   * OAuth client must be same.
#'   * Scopes must be same.
#'   * Email, if provided, must be same. If specified email is a glob pattern
#'     like `"*@example.com"`, email matching is done at the domain level.
#'

#' gargle is very conservative about using OAuth tokens discovered in the user's
#' cache and will generally seek interactive confirmation. Therefore, in a
#' non-interactive setting, it's important to explicitly specify the `"email"`
#' of the target account or to explicitly authorize automatic discovery. See
#' [gargle2.0_token()], which this function wraps, for more. Non-interactive use
#' also suggests it might be time to use a [service account
#' token][credentials_service_account] or [workload identity
#' federation][credentials_external_account].
#'

#' @param scopes A character vector of scopes to request. Pick from those listed
#'   at <https://developers.google.com/identity/protocols/oauth2/scopes>.
#'
#'   For certain token flows, the
#'   `"https://www.googleapis.com/auth/userinfo.email"` scope is unconditionally
#'   included. This grants permission to retrieve the email address associated
#'   with a token; gargle uses this to index cached OAuth tokens. This grants no
#'   permission to view or send email and is generally considered a low-value
#'   scope.
#' @inheritParams gargle2.0_token
#' @inheritDotParams gargle2.0_token -scope -client -package
#'
#' @return A [Gargle2.0] token.
#' @family credential functions
#' @export
#' @examples
#' \dontrun{
#' # Drive scope, built-in gargle demo client
#' scopes <- "https://www.googleapis.com/auth/drive"
#' credentials_user_oauth2(scopes, client = gargle_client())
#'
#' # bring your own client
#' client <- gargle_oauth_client_from_json(
#'   path = "/path/to/the/JSON/you/downloaded/from/gcp/console.json",
#'   name = "my-nifty-oauth-client"
#' )
#' credentials_user_oauth2(scopes, client)
#' }
credentials_user_oauth2 <- function(
  scopes = NULL,
  client = gargle_client(),
  package = "gargle",
  ...,
  app = deprecated()
) {
  gargle_debug("trying {.fun credentials_user_oauth2}")
  if (lifecycle::is_present(app)) {
    lifecycle::deprecate_soft(
      "1.5.0",
      "credentials_user_oauth2(app)",
      "credentials_user_oauth2(client)"
    )
    client <- app
  }
  gargle2.0_token(
    client = client,
    scope = scopes,
    package = package,
    ...
  )
}

#' Load a service account token
#'
#' @inheritParams token_fetch
#' @param path JSON identifying the service account, in one of the forms
#'   supported for the `txt` argument of [jsonlite::fromJSON()] (typically, a
#'   file path or JSON string).
#' @param subject An optional subject claim. Specify this if you wish to use the
#'   service account represented by `path` to impersonate the `subject`, who is
#'   a normal user. Before this can work, an administrator must grant the service
#'   account domain-wide authority. Identify the user to impersonate via their
#'   email, e.g. `subject = "user@example.com"`. Note that gargle automatically
#'   adds the non-sensitive `"https://www.googleapis.com/auth/userinfo.email"`
#'   scope, so this scope must be enabled for the service account, along with
#'   any other `scopes` being requested.
#'
#' @details Note that fetching a token for a service account requires a
#'   reasonably accurate system clock. For more information, see the
#'   `vignette("how-gargle-gets-tokens")`.
#' @seealso Additional reading on delegation of domain-wide authority:
#' * <https://developers.google.com/identity/protocols/oauth2/service-account#delegatingauthority>
#'
#' @return An [`httr::TokenServiceAccount`][httr::Token-class] or `NULL`.
#' @family credential functions
#' @export
#' @examples
#' \dontrun{
#' token <- credentials_service_account(
#'   scopes = "https://www.googleapis.com/auth/userinfo.email",
#'   path = "/path/to/your/service-account.json"
#' )
#' }
credentials_service_account <- function(
  scopes = NULL,
  path = "",
  ...,
  subject = NULL
) {
  gargle_debug("trying {.fun credentials_service_account}")
  info <- jsonlite::fromJSON(path, simplifyVector = FALSE)
  if (!identical(info[["type"]], "service_account")) {
    gargle_debug(c(
      "JSON does not appear to represent a service account",
      "Did you provide the JSON for an OAuth client instead of for a \\
       service account?"
    ))
    return()
  }

  # I add email scope explicitly, whereas I don't need to do so in
  # credentials_user_oauth2(), because it's done in Gargle2.0$new().
  scopes <- normalize_scopes(add_email_scope(scopes))
  token <- httr::oauth_service_token(
    ## FIXME: not sure endpoint is truly necessary, but httr thinks it is.
    ## https://github.com/r-lib/httr/issues/576
    endpoint = gargle_oauth_endpoint(),
    secrets = info,
    scope = scopes,
    sub = subject
  )
  if (
    is.null(token$credentials$access_token) ||
      !nzchar(token$credentials$access_token)
  ) {
    NULL
  } else {
    gargle_debug("service account email: {.email {token_email(token)}}")
    token
  }
}

#' Check for a service account
#'
#' This pre-checks information provided to a high-level, user-facing auth
#' function, such as `googledrive::drive_auth()`, before passing the user's
#' input along to [token_fetch()], which is designed to silently swallow errors.
#' Some users are confused about the difference between an OAuth client and a
#' service account and they provide the (path to the) JSON for one, when the
#' other is what's actually expected.
#'
#' @inheritParams credentials_service_account
#' @param hint The relevant function to call for configuring an OAuth client.
#' @inheritParams rlang::abort
#'
#' @return Nothing. Exists purely to throw an error.
#' @export
#' @keywords internal
check_is_service_account <- function(path, hint, call = caller_env()) {
  if (is.null(path)) {
    return(invisible())
  }

  info <- tryCatch(
    jsonlite::fromJSON(path, simplifyVector = FALSE),
    error = function(e) NULL
  )

  if (is.null(info) || !identical(info[["type"]], "service_account")) {
    cli::cli_abort(
      c(
        "{.arg path} does not represent a service account.",
        "i" = "If {.arg path} is meant to be a filepath, perhaps the file does \\
               not exist?",
        "i" = "Did you provide the JSON for an OAuth client instead of for a \\
               service account?",
        "i" = "To configure the OAuth client, use {.fun {hint}} instead."
      ),
      call = call
    )
  }
}

test_that("check_is_service_account() errors for OAuth client", {
  # call indirectly, so we can also check the caller is reported
  PKG_auth <- function(path) {
    check_is_service_account(path, hint = "PKG_auth_configure")
  }
  expect_snapshot(
    error = TRUE,
    PKG_auth(
      fs::path_package(
        "gargle",
        "extdata",
        "client_secret_installed.googleusercontent.com.json"
      )
    )
  )
})

test_that("check_is_service_account() errors for invalid input", {
  # call indirectly, so we can also check the caller is reported
  PKG_auth <- function(path) {
    check_is_service_account(path, hint = "PKG_auth_configure")
  }
  expect_snapshot(
    error = TRUE,
    PKG_auth("wut")
  )
})

#' Process a Google API response
#'
#' @description
#' `response_process()` is intended primarily for internal use in client
#' packages that provide high-level wrappers for users. Typically applied as the
#' final step in this sequence of calls:
#'   * Request prepared with [request_build()].
#'   * Request made with [request_make()].
#'   * Response processed with `response_process()`.
#'
#' All that's needed for a successful request is to parse the JSON extracted via
#' `httr::content()`. Therefore, the main point of `response_process()` is to
#' handle less happy outcomes:
#'   * Status codes in the 400s (client error) and 500s (server error). The
#'     structure of the error payload varies across Google APIs and we try to
#'     create a useful message for all variants we know about.
#'   * Non-JSON content type, such as HTML.
#'   * Status code in the 100s (information) or 300s (redirection). These are
#'     unexpected.
#'
#' If `response_process()` results in an error, a redacted version of the `resp`
#' input is returned in the condition (auth tokens are removed).
#'
#' @details
#' When `remember = TRUE` (the default), gargle stores the most recently seen
#' response internally, for *post hoc* examination. The stored response is
#' literally just the most recent `resp` input, but with auth tokens redacted.
#' It can be accessed via [gargle_last_response()]. A companion function
#' [gargle_last_content()] returns just the parsed content, which is probably
#' the most useful form for *post mortem* analysis.
#'
#' The `response_as_json()` helper is exported only as an aid to maintainers who
#' wish to use their own `error_message` function, instead of gargle's built-in
#' `gargle_error_message()`. When implementing a custom `error_message`
#' function, call `response_as_json()` immediately on the input in order to
#' inherit gargle's handling of non-JSON input.
#'
#' @param resp Object of class `response` from [httr].
#' @param error_message Function that produces an informative error message from
#'   the primary input, `resp`. It must return a character vector.
#' @param error_class Optional character vector of error classes to add to the
#'   condition object. These classes are prepended to gargle's default classes
#'   (`"gargle_error_request_failed"` and `"http_error_{status_code}"`).
#' @param remember Whether to remember the most recently processed response.
#' @inheritParams rlang::abort
#'
#' @return The content of the request, as a list. An HTTP status code of 204 (No
#'   content) is a special case returning `TRUE`.
#' @family requests and responses
#' @export
#' @examples
#' \dontrun{
#' # get an OAuth2 token with 'userinfo.email' scope
#' token <- token_fetch(scopes = "https://www.googleapis.com/auth/userinfo.email")
#'
#' # see the email associated with this token
#' req <- gargle::request_build(
#'   method = "GET",
#'   path = "v1/userinfo",
#'   token = token,
#'   base_url = "https://openidconnect.googleapis.com"
#' )
#' resp <- gargle::request_make(req)
#' response_process(resp)
#'
#' # make a bad request (this token has incorrect scope)
#' req <- gargle::request_build(
#'   method = "GET",
#'   path = "fitness/v1/users/{userId}/dataSources",
#'   token = token,
#'   params = list(userId = 12345)
#' )
#' resp <- gargle::request_make(req)
#' response_process(resp)
#' }
response_process <- function(
  resp,
  error_message = gargle_error_message,
  error_class = NULL,
  remember = TRUE,
  call = caller_env()
) {
  if (remember) {
    gargle_env$last_response <- redact_response(resp)
  }
  code <- httr::status_code(resp)

  if (code >= 200 && code < 300) {
    if (code == 204) {
      # HTTP status: No content
      TRUE
    } else {
      response_as_json(resp, call = call)
    }
  } else {
    gargle_abort_request_failed(
      error_message(resp, call = call),
      resp,
      class = error_class,
      call = call
    )
  }
}

#' @export
#' @rdname response_process
response_as_json <- function(resp, call = caller_env()) {
  check_for_json(resp, call = call)

  content <- httr::content(resp, type = "raw")
  content <- rawToChar(content)
  Encoding(content) <- "UTF-8"
  jsonlite::fromJSON(content, simplifyVector = FALSE)
}

check_for_json <- function(resp, call = caller_env()) {
  type <- httr::http_type(resp)
  if (grepl("^application/json", type)) {
    return(invisible(resp))
  }

  gargle_abort_request_failed(
    "Expected content type {.field application/json}, not {.field {type}}.",
    call = call,
    resp = resp
  )
}

gargle_abort_request_failed <- function(
  message,
  resp,
  class = NULL,
  .envir = caller_env(),
  call = caller_env()
) {
  gargle_abort(
    message,
    class = c(
      class,
      "gargle_error_request_failed",
      glue("http_error_{httr::status_code(resp)}")
    ),
    .envir = .envir,
    call = call,
    resp = redact_response(resp)
  )
}

#' @export
#' @rdname response_process
gargle_error_message <- function(resp, call = caller_env()) {
  type <- httr::http_type(resp)
  if (grepl("^text/html", type)) {
    return(gargle_html_error_message(resp))
  }

  content <- response_as_json(resp, call = call)
  error <- content[["error"]]

  # Handle variety of error messages returned by different google APIs

  # It would be fussy to employ cli styling here, as we would need to
  # pre-interpolate data from `resp`. So we either don't style or we use simple
  # "no color" styles.

  if (identical(names(content), c("error", "error_description"))) {
    # seen when calling userinfo endpoint with an access token obtained via
    # workload identity federation
    message <- c(
      httr::http_status(resp)$message,
      "*" = content$error,
      "*" = content$error_description
    )
    return(escape_braces(message))
  }

  if (is.null(error)) {
    # developed from test fixture from tokeninfo endpoint
    message <- c(
      httr::http_status(resp)$message,
      "*" = content$error_description
    )
    return(escape_braces(message))
  }
  errors <- error[["errors"]]

  if (is.null(errors)) {
    # developed from test fixtures from "sheets.spreadsheets.get" endpoint
    status <- httr::http_status(resp)

    message <- c(
      glue("{status$category}: ({error$code}) {error$status}"),
      "*" = rpc_description(error$status),
      "*" = error$message
    )

    error_details <- error$details
    if (is.null(error_details)) {
      return(escape_braces(message))
    }

    # https://github.com/googleapis/googleapis/blob/master/google/rpc/error_details.proto
    # Look for a localized error message
    # Below, we display it alongside the main message, if it's not identical to
    # the primary message
    # Remove the localized message from error details, regardless
    # Currently ignoring the locale info, because it seems hard to reliably
    # compare to something returned by Sys.getlocale()
    types <- map_chr(error_details, function(x) x[["@type"]])
    lm_index <- match("type.googleapis.com/google.rpc.LocalizedMessage", types)
    localized_message <- error_details[[lm_index]][["message"]]
    if (!is.na(lm_index)) {
      error_details <- error_details[-lm_index]
    }

    if (
      !is.null(localized_message) &&
        !identical(localized_message, error$message)
    ) {
      message <- c(message, "*" = localized_message)
    }

    message <- c(
      message,
      "",
      reveal_details(error_details)
    )
    return(escape_braces(message))
  }

  # developed from
  # - test fixture from "drive.files.get" endpoint
  # - response_process() example of under/mis-scoped token
  errors <- unlist(errors)
  message <- c(
    httr::http_status(resp)$message,
    error$message,
    error$status,
    bulletize(
      # format() has no effect when processed by cli; bring that back later?
      # glue("{format(names(errors), justify = 'right')}: {errors}"),
      glue("{names(errors)}: {errors}"),
      n_show = 10
    )
  )
  message
}

redact_response <- function(resp) {
  resp$request$auth_token <- "<REDACTED>"
  resp$request$headers["Authorization"] <- "<REDACTED>"
  resp$handle <- NULL
  resp
}

rpc_description <- function(rpc) {
  m <- match(rpc, oops$RPC)
  if (is_na(m)) {
    NULL
  } else {
    oops$Description[[m]]
  }
}

# https://cloud.google.com/apis/design/errors
# @craigcitro says:
# "... a published description of how new APIs do errors, which includes the
# canonical error codes and http mappings. This view of errors is ... what ...
# APIs will ultimately converge on"
# https://github.com/googleapis/googleapis/blob/master/google/rpc/error_details.proto
oops <- read.csv(
  text = trimws(c(
    '
  HTTP,                   RPC,  Description
   200,                  "OK", "No error."
   400,    "INVALID_ARGUMENT", "Client specified an invalid argument. Check error message and error details for more information."
   400, "FAILED_PRECONDITION", "Request can not be executed in the current system state, such as deleting a non-empty directory."
   400,        "OUT_OF_RANGE", "Client specified an invalid range."
   401,     "UNAUTHENTICATED", "Request not authenticated due to missing, invalid, or expired OAuth token."
   403,   "PERMISSION_DENIED", "Client does not have sufficient permission. This can happen because the OAuth token does not have the right scopes, the client doesn\'t have permission, or the API has not been enabled for the client project."
   404,           "NOT_FOUND", "A specified resource is not found, or the request is rejected by undisclosed reasons, such as whitelisting."
   409,             "ABORTED", "Concurrency conflict, such as read-modify-write conflict."
   409,      "ALREADY_EXISTS", "The resource that a client tried to create already exists."
   429,  "RESOURCE_EXHAUSTED", "Either out of resource quota or reaching rate limiting. The client should look for google.rpc.QuotaFailure error detail for more information."
   499,           "CANCELLED", "Request cancelled by the client."
   500,           "DATA_LOSS", "Unrecoverable data loss or data corruption. The client should report the error to the user."
   500,             "UNKNOWN", "Unknown server error. Typically a server bug."
   500,            "INTERNAL", "Internal server error. Typically a server bug."
   501,     "NOT_IMPLEMENTED", "API method not implemented by the server."
   503,         "UNAVAILABLE", "Service unavailable. Typically the server is down."
   504,   "DEADLINE_EXCEEDED", "Request deadline exceeded. This will happen only if the caller sets a deadline that is shorter than the method\'s default deadline (i.e. requested deadline is not enough for the server to process the request) and the request did not finish within the deadline."
                         '
  )),
  stringsAsFactors = FALSE,
  strip.white = TRUE
)

# https://github.com/googleapis/googleapis/blob/master/google/rpc/error_details.proto
reveal_details <- function(details) {
  c("Error details:", unlist(lapply(details, reveal_detail)))
}

# https://github.com/googleapis/googleapis/blob/master/google/rpc/rpc_publish.yaml
# https://github.com/googleapis/googleapis/blob/master/google/rpc/error_details.proto
reveal_detail <- function(x) {
  type <- sub("^type.googleapis.com/", "", x$`@type`)

  rpc_bad_request <- function(e) {
    f <- function(x) {
      c(
        "*" = glue("Field: {x$field}"),
        " " = glue("Description: {x$description}")
      )
    }
    bullets <- unlist(map(e[["fieldViolations"]], f))
    c("Field violations", bullets)
  }
  rpc_help <- function(e) {
    f <- function(x) {
      c(
        "*" = glue("Description: {x$description}"),
        " " = glue("URL: {x$url}")
      )
    }
    bullets <- unlist(map(e[["links"]], f))
    c("Links", bullets)
  }
  rpc_error_info <- function(e) {
    e <- unlist(e)
    e <- e[names(e) != "@type"]
    bulletize(glue_data(as.list(e), "{names(e)}: {e}"), n_show = 10)
  }

  switch(
    type,
    "google.rpc.BadRequest" = rpc_bad_request(x),
    "google.rpc.Help" = rpc_help(x),
    "google.rpc.ErrorInfo" = rpc_error_info(x),
    # must be an unimplemented type, such as RetryInfo, QuotaFailure, etc.
    bulletize(
      map_chr(
        c(
          "Error details of type '{type}' may not be fully revealed.",
          "Workaround: use `gargle_last_response()` or \\
           `gargle_last_content()` to inspect error payload \\
           yourself.",
          "Consider opening an issue at \\
           <https://github.com/r-lib/gargle/issues.>"
        ),
        # https://github.com/tidyverse/googlesheets4/issues/317
        # this needs to be an anonymous function (as opposed to naked `glue`)
        # in order to capture the value of `type`
        # good write-up in https://stackoverflow.com/a/75247635
        function(x) glue(x)
      )
    )
  )
}

gargle_html_error_message <- function(resp) {
  stopifnot(httr::http_type(resp) == "text/html")

  content <- httr::content(resp, as = "text")
  tmp <- tempfile("gargle-unexpected-html-error-", fileext = ".html")
  writeLines(content, tmp)
  browse_hint <- glue('browseURL("{tmp}")')

  # pre-interpolate, since `tmp` and `browse_hint` are only known here.
  c(
    httr::http_status(resp)$message,
    "x" = "Expected content type {.field application/json}, not \\
           {.field text/html}.",
    "i" = gargle_map_cli(tmp, "See {.file <<x>>} for the html error content."),
    "i" = gargle_map_cli(
      browse_hint,
      "Or execute {.code <<x>>} to view it in your browser."
    )
  )
}

# Google APIs might return error messages containing curly braces, e.g.
# "metadata.quota_unit: 1/min/{project}/{user}"
# This can cause problems when eventually process by cli::cli_abort(), e.g.
# Error:
# ! ! Could not evaluate cli `{}` expression: `project`.
# Caused by error in `eval(expr, envir = envir)`:
# ! object 'project' not found
# So we need to escape them by doubling them, so they are taken literally.
# Seen by me and by a user:
# https://github.com/tidyverse/googlesheets4/issues/319
escape_braces <- function(msg) {
  gsub("([{}])", "\\1\\1", msg)
}

expect_recorded_error <- function(filename, status_code) {
  rds_file <- test_path("fixtures", fs::path_ext_set(filename, "rds"))
  resp <- readRDS(rds_file)
  expect_error(response_process(resp), class = "gargle_error_request_failed")
  expect_error(response_process(resp), class = glue("http_error_{status_code}"))
  # HTML errors (as opposed to JSON) need this
  scrub_filepath <- function(x) {
    gsub(
      "([\"\'])\\S+gargle-unexpected-html-error-\\S+[.]html([\"\'])",
      "\\1VOLATILE_FILE_PATH\\2",
      x,
      perl = TRUE
    )
  }
  expect_snapshot(
    response_process(resp),
    error = TRUE,
    transform = scrub_filepath
  )
}

test_that("Resource exhausted (Sheets, ReadGroup)", {
  expect_recorded_error(
    "sheets-spreadsheets-get-quota-exceeded-readgroup_429",
    429
  )
})

test_that("Request for non-existent resource (Drive)", {
  expect_recorded_error(
    "drive-files-get-nonexistent-file-id_404",
    404
  )
})

# https://github.com/r-lib/gargle/issues/254
test_that("Too many requests (Drive, HTML content)", {
  expect_recorded_error(
    "drive-automated-queries_429",
    429
  )
})

# https://github.com/r-lib/gargle/issues/254
test_that("HTML error is offered as a file", {
  rds_file <- test_path("fixtures", "drive-automated-queries_429.rds")
  resp <- readRDS(rds_file)
  err <- tryCatch(
    response_process(resp),
    gargle_error_request_failed = function(e) e
  )
  regex <- "[^'\" \\t\\n\\r]+gargle-unexpected-html-error-\\S+[.]html"
  m <- gregexpr(regex, err$body, perl = TRUE)
  path_to_html_error <- unique(unlist(regmatches(err$body, m)))
  # the strwrap() result is a bit goofy, but seems least of all evils
  # this is mostly about making sure we excavate the HTML
  expect_snapshot(strwrap(readLines(path_to_html_error), width = 60))
  unlink(path_to_html_error)
})

test_that("Request for which we don't have scope (Fitness)", {
  expect_recorded_error(
    "fitness-get-wrong-scope_403",
    403
  )
})

test_that("Use key that's not enabled for the API (Sheets)", {
  expect_recorded_error(
    "sheets-spreadsheets-get-api-key-not-enabled_403",
    403
  )
})

test_that("Request with invalid argument (Sheets, bad range)", {
  expect_recorded_error(
    "sheets-spreadsheets-get-nonexistent-range_400",
    400
  )
})

test_that("Request with bad field mask (Sheets)", {
  expect_recorded_error(
    "sheets-spreadsheets-get-bad-field-mask_400",
    400
  )
})

test_that("Request for nonexistent resource (Sheets)", {
  expect_recorded_error(
    "sheets-spreadsheets-get-nonexistent-sheet-id_404",
    404
  )
})

# https://github.com/tidyverse/googlesheets4/issues/317
test_that("Use service account that's not enabled for the API (Sheets)", {
  expect_recorded_error(
    "sheets-spreadsheets-get-service-disabled_403",
    403
  )
})

test_that("Request with invalid value (tokeninfo, stale token)", {
  expect_recorded_error(
    "tokeninfo-stale_400",
    400
  )
})

test_that("Request to bad URL (tokeninfo, HTML content)", {
  expect_recorded_error(
    "tokeninfo-bad-path_404",
    404
  )
})

# error_class parameter ----
test_that("error_class parameter adds custom classes", {
  rds_file <- test_path(
    "fixtures",
    "drive-files-get-nonexistent-file-id_404.rds"
  )
  resp <- readRDS(rds_file)

  # Without error_class, we get default classes
  err <- tryCatch(
    response_process(resp),
    gargle_error_request_failed = function(e) e
  )
  klass <- class(err)
  expect_equal(
    klass[1:3],
    c("gargle_error_request_failed", "http_error_404", "gargle_error")
  )

  err_custom <- tryCatch(
    response_process(resp, error_class = c("custom_error1", "custom_error2")),
    custom_error1 = function(e) e
  )
  klass <- class(err_custom)
  expect_equal(
    klass[1:5],
    c(
      "custom_error1",
      "custom_error2",
      "gargle_error_request_failed",
      "http_error_404",
      "gargle_error"
    )
  )
})

# helpers ----
test_that("RPC codes can be looked up (or not)", {
  expect_match(
    rpc_description("ALREADY_EXISTS"),
    "resource .* already exists"
  )
  expect_null(rpc_description("MATCHES_NOTHING"))
})

# nocov start

# example of data client package needs to provide -----------------------------

# a client package should define a list like this and pass it to the functions
# below, to provide data to populate the templates

# see googledrive/R/drive_auth.R for an example
# this is an exhaustive list of the pieces of data required by the templates

# gargle_lookup_table <- list(
#   PACKAGE     = "googledrive",
#   YOUR_STUFF  = "your Drive files",
#   PRODUCT     = "Google Drive",
#   API         = "Drive API",
#   PREFIX      = "drive",
# )

# 2023-03 developments related to the 'app' -> 'client' transition:
# - `PREFIX_auth_configure_description()` crosslinks to
#   `PREFIX_oauth_client()` now, not `PREFIX_oauth_app()`
# - `PREFIX_auth_configure_params()` gains `client` argument
# - `PREFIX_auth_configure_params()` deprecates the `app` argument and uses a
#   lifecycle badge
# - `PREFIX_auth_configure_params() crosslinks to
#   `gargle::gargle_oauth_client_from_json()` which requires gargle (>= 1.3.0)

glue_data_lines <- function(.data, lines, ..., .envir = caller_env()) {
  # work around name collision of `.x` of map_chr() vs. of glue_data()
  # and confusion re: `...` of glue_data_lines() vs. `...` of map_chr()
  # plus: I've only got compat-purrr here, so I have to write a function
  gd <- function(line) glue_data(.x = .data, line, ..., .envir = .envir)
  map_chr(lines, gd)
}

# PREFIX_auth() ----------------------------------------------------------

PREFIX_auth_description <- function(
  .data = list(
    PACKAGE = "PACKAGE",
    YOUR_STUFF = "YOUR STUFF",
    PRODUCT = "A GOOGLE PRODUCT"
  )
) {
  glue_data_lines(
    c(
      "@description",
      "Authorize {PACKAGE} to view and manage {YOUR_STUFF}. This function is a",
      "wrapper around [gargle::token_fetch()].",
      "",
      "By default, you are directed to a web browser, asked to sign in to your",
      "Google account, and to grant {PACKAGE} permission to operate on your",
      "behalf with {PRODUCT}. By default, with your permission, these user",
      "credentials are cached in a folder below your home directory, from where",
      "they can be automatically refreshed, as necessary. Storage at the user",
      "level means the same token can be used across multiple projects and",
      "tokens are less likely to be synced to the cloud by accident."
    ),
    .data = .data
  )
}

PREFIX_auth_details <- function(
  .data = list(
    PACKAGE = "PACKAGE",
    PREFIX = "PREFIX"
  )
) {
  glue_data_lines(
    c(
      "@details",
      "Most users, most of the time, do not need to call `{PREFIX}_auth()`",
      "explicitly -- it is triggered by the first action that requires",
      "authorization. Even when called, the default arguments often suffice.",
      "",
      "However, when necessary, `{PREFIX}_auth()` allows the user to explicitly:",
      "  * Declare which Google identity to use, via an `email` specification.",
      "  * Use a service account token or workload identity federation via",
      "    `path`.",
      "  * Bring your own `token`.",
      "  * Customize `scopes`.",
      "  * Use a non-default `cache` folder or turn caching off.",
      "  * Explicitly request out-of-band (OOB) auth via `use_oob`.",
      "",
      "If you are interacting with R within a browser (applies to RStudio",
      "Server, Posit Workbench, Posit Cloud, and Google Colaboratory), you need",
      "OOB auth or the pseudo-OOB variant. If this does not happen",
      "automatically, you can request it explicitly with `use_oob = TRUE` or,",
      "more persistently, by setting an option via",
      "`options(gargle_oob_default = TRUE)`.",
      "",
      "The choice between conventional OOB or pseudo-OOB auth is determined",
      "by the type of OAuth client. If the client is of the \"installed\" type,",
      "`use_oob = TRUE` results in conventional OOB auth. If the client is of",
      "the \"web\" type, `use_oob = TRUE` results in pseudo-OOB auth. Packages",
      "that provide a built-in OAuth client can usually detect which type of",
      "client to use. But if you need to set this explicitly, use the",
      "`\"gargle_oauth_client_type\"` option:",
      "```r",
      "options(gargle_oauth_client_type = \"web\")       # pseudo-OOB",
      "# or, alternatively",
      "options(gargle_oauth_client_type = \"installed\") # conventional OOB",
      "```",
      "",
      "For details on the many ways to find a token, see",
      "[gargle::token_fetch()]. For deeper control over auth, use",
      "[{PREFIX}_auth_configure()] to bring your own OAuth client or API key.",
      "To learn more about gargle options, see [gargle::gargle_options]."
    ),
    .data = .data
  )
}

PREFIX_auth_params <- function() {
  c(
    "@inheritParams gargle::credentials_service_account",
    "@inheritParams gargle::credentials_external_account",
    "@inheritParams gargle::credentials_app_default",
    "@inheritParams gargle::credentials_gce",
    "@inheritParams gargle::credentials_byo_oauth2",
    "@inheritParams gargle::credentials_user_oauth2",
    "@inheritParams gargle::gargle2.0_token"
  )
}

# PREFIX_deauth() ----------------------------------------------------------

PREFIX_deauth_description_with_api_key <- function(
  .data = list(
    PACKAGE = "PACKAGE",
    PREFIX = "PREFIX"
  ),
  .fallback_api_key = TRUE
) {
  lines <- c(
    "@description",
    "Put {PACKAGE} into a de-authorized state. Instead of sending a token,",
    "{PACKAGE} will send an API key. This can be used to access public",
    "resources for which no Google sign-in is required. This is handy for using",
    "{PACKAGE} in a non-interactive setting to make requests that do not",
    "require a token. It will prevent the attempt to obtain a token",
    "interactively in the browser. The user can configure their own API key",
    "via [{PREFIX}_auth_configure()] and retrieve that key via",
    "[{PREFIX}_api_key()].",
    if (.fallback_api_key) {
      "In the absence of a user-configured key, a built-in default key is used."
    }
  )
  glue_data_lines(lines, .data = .data)
}

PREFIX_deauth_description_no_api_key <- function(
  .data = list(
    PACKAGE = "PACKAGE",
    PREFIX = "PREFIX"
  ),
  .fallback_api_key = TRUE
) {
  lines <- c(
    "@description",
    "Clears any currently stored token. The next time {PACKAGE} needs a token,",
    "the token acquisition process starts over, with a fresh call to",
    "[{PREFIX}_auth()] and, therefore, internally, a call to",
    "[gargle::token_fetch()]. Unlike some other packages that use gargle,",
    "{PACKAGE} is not usable in a de-authorized state. Therefore, calling",
    "`{PREFIX}_deauth()` only clears the token, i.e. it does NOT imply that",
    "subsequent requests are made with an API key in lieu of a token."
  )
  glue_data_lines(lines, .data = .data)
}

# PREFIX_token() ----------------------------------------------------------

PREFIX_token_description <- function(
  .data = list(
    API = "GOOGLE API",
    PREFIX = "PREFIX"
  ),
  .deauth_possible = TRUE
) {
  lines <- c(
    "@description",
    "For internal use or for those programming around the {API}.",
    "Returns a token pre-processed with [httr::config()]. Most users",
    "do not need to handle tokens \"by hand\" or, even if they need some",
    "control, [{PREFIX}_auth()] is what they need. If there is no current",
    "token, [{PREFIX}_auth()] is called to either load from cache or",
    "initiate OAuth2.0 flow.",
    if (.deauth_possible) {
      c(
        "If auth has been deactivated via [{PREFIX}_deauth()], `{PREFIX}_token()`",
        "returns `NULL`."
      )
    }
  )
  glue_data_lines(lines, .data = .data)
}

PREFIX_token_return <- function() {
  "@return A `request` object (an S3 class provided by [httr][httr::httr])."
}

# PREFIX_has_token() ----------------------------------------------------------

PREFIX_has_token_description <- function(.data = list(PACKAGE = "PACKAGE")) {
  glue_data_lines(
    c(
      "@description",
      "Reports whether {PACKAGE} has stored a token, ready for use in downstream",
      "requests."
    ),
    .data = .data
  )
}

PREFIX_has_token_return <- function() {
  "@return Logical."
}

# PREFIX_auth_configure() -------------------------------------------------------

PREFIX_auth_configure_description <- function(
  .data = list(
    PACKAGE = "PACKAGE",
    PREFIX = "PREFIX"
  ),
  .has_api_key = TRUE,
  .fallbacks = TRUE
) {
  lines <- c(
    "@description",
    "These functions give more control over and visibility into the auth",
    "configuration than [{PREFIX}_auth()] does. `{PREFIX}_auth_configure()`",
    "lets the user specify their own:",
    "  * OAuth client, which is used when obtaining a user token.",
    if (.has_api_key) {
      c(
        "  * API key. If {PACKAGE} is de-authorized via [{PREFIX}_deauth()], all",
        "    requests are sent with an API key in lieu of a token."
      )
    },
    "",
    'See the `vignette("get-api-credentials", package = "gargle")`',
    "for more.",
    if (.fallbacks) {
      c(
        "If the user does not configure these settings, internal defaults",
        "are used."
      )
    },
    "",
    if (.has_api_key) {
      c(
        "`{PREFIX}_oauth_client()` and `{PREFIX}_api_key()` retrieve the",
        "currently configured OAuth client and API key, respectively."
      )
    } else {
      "`{PREFIX}_oauth_client()` retrieves the currently configured OAuth client."
    }
  )
  glue_data_lines(lines, .data = .data)
}

PREFIX_auth_configure_params <- function(.has_api_key = TRUE) {
  c(
    "@param client A Google OAuth client, presumably constructed via",
    "[gargle::gargle_oauth_client_from_json()]. Note, however, that it is",
    "preferred to specify the client with JSON, using the `path` argument.",
    "@inheritParams gargle::gargle_oauth_client_from_json",
    if (.has_api_key) {
      "@param api_key API key."
    },
    "@param app `r lifecycle::badge('deprecated')` Replaced by the `client`",
    "argument."
  )
}

PREFIX_auth_configure_return <- function(
  .data = list(
    PREFIX = "PREFIX"
  ),
  .has_api_key = TRUE
) {
  lines <- c(
    "@return",
    "  * `{PREFIX}_auth_configure()`: An object of R6 class",
    "    [gargle::AuthState], invisibly.",
    "  * `{PREFIX}_oauth_client()`: the current user-configured OAuth client.",
    if (.has_api_key) {
      "  * `{PREFIX}_api_key()`: the current user-configured API key."
    }
  )
  glue_data_lines(lines, .data = .data)
}

# PREFIX_user() ----------------------------------------------------------

PREFIX_user_description <- function() {
  c(
    "@description",
    "Reveals the email address of the user associated with the current token.",
    "If no token has been loaded yet, this function does not initiate auth."
  )
}

PREFIX_user_seealso <- function() {
  c(
    "@seealso [gargle::token_userinfo()], [gargle::token_email()],",
    "[gargle::token_tokeninfo()]"
  )
}

PREFIX_user_return <- function() {
  "@return An email address or, if no token has been loaded, `NULL`."
}

# nocov end

#' Build a Google API request
#'
#' Intended primarily for internal use in client packages that provide
#' high-level wrappers for users. The vignette [Request helper
#' functions](https://gargle.r-lib.org/articles/request-helper-functions.html)
#' describes how one might use these functions inside a wrapper package.
#'
#' @param endpoint List of information about the target endpoint or, in
#'   Google's vocabulary, the target "method". Presumably prepared from the
#'   [Discovery
#'   Document](https://developers.google.com/discovery/v1/getting_started#background-resources)
#'   for the target API.
#' @param params Named list. Values destined for URL substitution, the query,
#'   or, for `request_develop()` only, the body. For `request_build()`, body
#'   parameters must be passed via the `body` argument.
#' @param base_url Character.
#' @param method Character. An HTTP verb, such as `GET` or `POST`.
#' @param path Character. Path to the resource, not including the API's
#'   `base_url`. Examples: `drive/v3/about` or `drive/v3/files/{fileId}`. The
#'   `path` can be a template, i.e. it can include variables inside curly
#'   brackets, such as `{fileId}` in the example. Such variables are substituted
#'   by `request_build()`, using named parameters found in `params`.
#' @param body List. Values to send in the API request body.
#' @param key API key. Needed for requests that don't contain a token. For more,
#'   see Google's document Credentials, access, security, and identity
#'   (`https://support.google.com/googleapi/answer/6158857?hl=en&ref_topic=7013279`).
#'    A key can be passed as a named component of `params`, but note that the
#'   formal argument `key` will clobber it, if non-`NULL`.
#' @param token Token, ready for inclusion in a request, i.e. prepared with
#'   [httr::config()].
#'
#' @section `request_develop()`:
#'
#' Combines user input (`params`) with information about an API endpoint.
#' `endpoint` should contain these components:
#'   - `path`: See documentation for argument.
#'   - `method`: See documentation for argument.
#'   - `parameters`: Compared with `params` supplied by user. An error is
#'      thrown if user-supplied `params` aren't named in
#'      `endpoint$parameters` or if user fails to supply all required
#'      parameters. In the return value, body parameters are separated from
#'      those destined for path substitution or the query.
#'
#' The return value is typically used as input to `request_build()`.
#'
#' @section `request_build()`:
#'
#' Builds a request, in a purely mechanical sense. This function does nothing
#' specific to any particular Google API or endpoint.
#'   - Use with the output of `request_develop()` or with hand-crafted input.
#'   - `params` are used for variable substitution in `path`. Leftover `params`
#'     that are not bound by the `path` template automatically become HTTP
#'     query parameters.
#'   - Adds an API key to the query iff `token = NULL` and removes the API key
#'   otherwise. Client packages should generally pass their own API key in, but
#'   note that [gargle_api_key()] is available for small-scale experimentation.
#'
#' See `googledrive::generate_request()` for an example of usage in a client
#' package. googledrive has an internal list of selected endpoints, derived from
#' the [Drive API Discovery
#' Document](https://www.googleapis.com/discovery/v1/apis/drive/v3/rest),
#' exposed via `googledrive::drive_endpoints()`. An element from such a list is
#' the expected input for `endpoint`. `googledrive::generate_request()` is a
#' wrapper around `request_develop()` and `request_build()` that inserts a
#' googledrive-managed API key and some logic about Team Drives. All user-facing
#' functions use `googledrive::generate_request()` under the hood.
#'
#' @return
#' `request_develop()`: `list()` with components `method`, `path`, `params`,
#' `body`, and `base_url`.
#'
#' `request_build()`: `list()` with components `method`, `path`
#' (post-substitution), `query` (the input `params` not used in URL
#' substitution), `body`, `token`, `url` (the full URL, post-substitution,
#' including the query).
#'
#' @export
#' @family requests and responses
#' @examples
#' \dontrun{
#' ## Example with a prepared endpoint
#' ept <- googledrive::drive_endpoints("drive.files.update")[[1]]
#' req <- request_develop(
#'   ept,
#'   params = list(
#'     fileId = "abc",
#'     addParents = "123",
#'     description = "Exciting File"
#'   )
#' )
#' req
#'
#' req <- request_build(
#'   method = req$method,
#'   path = req$path,
#'   params = req$params,
#'   body = req$body,
#'   token = "PRETEND_I_AM_A_TOKEN"
#' )
#' req
#'
#' ## Example with no previous knowledge of the endpoint
#' ## List a file's comments
#' ## https://developers.google.com/drive/v3/reference/comments/list
#' req <- request_build(
#'   method = "GET",
#'   path = "drive/v3/files/{fileId}/comments",
#'   params = list(
#'     fileId = "your-file-id-goes-here",
#'     fields = "*"
#'   ),
#'   token = "PRETEND_I_AM_A_TOKEN"
#' )
#' req
#'
#' # Example with no previous knowledge of the endpoint and no token
#' # use an API key for which the Places API is enabled!
#' API_KEY <- "1234567890"
#'
#' # get restaurants close to a location in Vancouver, BC
#' req <- request_build(
#'   method = "GET",
#'   path = "maps/api/place/nearbysearch/json",
#'   params = list(
#'     location = "49.268682,-123.167117",
#'     radius = 100,
#'     type = "restaurant"
#'   ),
#'   key = API_KEY,
#'   base_url = "https://maps.googleapis.com"
#' )
#' resp <- request_make(req)
#' out <- response_process(resp)
#' vapply(out$results, function(x) x$name, character(1))
#' }
request_develop <- function(endpoint,
                            params = list(),
                            base_url = "https://www.googleapis.com") {
  check_params(params, endpoint$parameters, endpoint$id)
  in_body <- vapply(
    endpoint$parameters,
    function(x) x$location == "body",
    logical(1)
  )
  body_param_names  <- names(endpoint$parameters[in_body])
  other_param_names <- names(endpoint$parameters[!in_body])
  list(
    method = endpoint$httpMethod,
    path = endpoint$path,
    params = params[intersect(other_param_names, names(params))],
    body = params[intersect(body_param_names, names(params))],
    base_url = base_url
  )
}

#' @rdname request_develop
#' @export
request_build <- function(method = "GET",
                          path = "",
                          params = list(),
                          body = list(),
                          token = NULL,
                          key = NULL,
                          base_url = "https://www.googleapis.com") {
  path_param_names <- extract_path_names(path)
  path_params <- params[path_param_names]
  query_param_names <- setdiff(names(params), path_param_names)
  query_params <- params[query_param_names]

  ## send a token or a key, but never both
  query_params$key <- if (is.null(token)) {
    key %||% query_params$key
  } else {
    NULL
  }

  out <- list(
    method = method,
    url = httr::modify_url(
      url = base_url,
      path = glue_data(path_params, path),
      query = query_params
    ),
    body = body,
    token = token
  )
  out
}

## check params provided by user against spec
##   * error if required params are missing
##   * error for unknown params
check_params <- function(provided, spec, id, call = caller_env()) {
  required <- Filter(function(x) isTRUE(x$required), spec)
  missing <- setdiff(names(required), names(provided))
  if (length(missing)) {
    gargle_abort_bad_params(missing, reason = "missing", id, call = call)
  }

  unknown <- setdiff(names(provided), names(spec))
  if (length(unknown)) {
    gargle_abort_bad_params(unknown, reason = "unknown", id, call = call)
  }

  invisible(provided)
}

##  input: /v4/spreadsheets/{spreadsheetId}/sheets/{sheetId}:copyTo
## output: spreadsheetId, sheetId
extract_path_names <- function(path) {
  m <- gregexpr("\\{[^/]*\\}", path)
  path_param_names <- regmatches(path, m)[[1]]
  gsub("[\\{\\}]", "", path_param_names)
}

test_that("request_develop() errors for unrecognized parameters", {
  expect_snapshot_error(
    request_develop(
      endpoint = list(parameters = list(a = list()), id = "some.api.endpoint"),
      params = list(b = list(), c = list())
    )
  )
})

test_that("request_develop() errors if required parameter is missing", {
  expect_snapshot_error(
    request_develop(
      endpoint = list(
        parameters = list(a = list(required = TRUE)),
        id = "some.api.endpoint"
      ),
      params = list(b = list())
    )
  )
})

test_that("request_develop() separates body params from query", {
  req <- request_develop(
    endpoint = list(
      parameters = list(
        a = list(location = "body", required = FALSE),
        b = list(location = "query", required = FALSE)
      )
    ),
    params = list(a = list(), b = list())
  )
  expect_equal(req$body, list(a = list()))
  expect_equal(req$params, list(b = list()))
})

# https://github.com/r-lib/gargle/issues/122
test_that("request_develop() copes with a param that goes to path and body", {
  req <- request_develop(
    endpoint = list(
      parameters = list(
        two_places = list(location = "path", required = FALSE),
        two_places = list(location = "body", required = FALSE),
        just_path = list(location = "path", required = FALSE),
        just_body = list(location = "body", required = FALSE),
        elsewhere = list(location = "????", required = FALSE)
      )
    ),
    params = list(two_places = list(), just_path = list(), just_body = list())
  )
  expect_equal(req$params, list(two_places = list(), just_path = list()))
  expect_equal(req$body, list(two_places = list(), just_body = list()))
})

test_that("request_build() does substitution and puts remainder in query", {
  req <- request_build(
    path = "/{a}/xx/{b}",
    params = list(a = "A", b = "B", c = "C")
  )
  expect_equal(req$url, "https://www.googleapis.com/A/xx/B?c=C")
})

test_that("request_build() suppresses API key if token is non-NULL", {
  req <- request_build(
    params = list(key = "key in params"),
    key = "explicit key",
    token = httr::config(token = "token!")
  )
  expect_false(grepl("key", req$url))
})

test_that("request_build() adds key, if available when token = NULL", {
  req <- request_build(key = "abc", token = NULL)
  expect_match(req$url, "key=abc")
  req <- request_build(params = list(key = "abc"), token = NULL)
  expect_match(req$url, "key=abc")
})

test_that("request_build(): explicit API key > key in params", {
  req <- request_build(key = "abc", params = list(key = "def"), token = NULL)
  expect_match(req$url, "key=abc")
})

#' @keywords internal
"_PACKAGE"

# The following block is used by usethis to automatically manage
# roxygen namespace tags. Modify with care!
## usethis namespace: start
#' @import fs
#' @import rlang
#' @importFrom glue glue
#' @importFrom glue glue_collapse
#' @importFrom glue glue_data
#' @importFrom lifecycle deprecated
## usethis namespace: end
NULL

#' Options consulted by gargle
#'
#' @description
#' Wrapper functions around options consulted by gargle, which provide:
#'   * A place to hang documentation.
#'   * The mechanism for setting a default.
#'
#' If the built-in defaults don't suit you, set one or more of these options.
#' Typically, this is done in the `.Rprofile` startup file, with code along
#' these lines:
#' ```
#' options(
#'   gargle_oauth_email = "jane@example.com",
#'   gargle_oauth_cache = "/path/to/folder/that/does/not/sync/to/cloud"
#' )
#' ```
#'
#' @name gargle_options
#' @examples
#' gargle_oauth_email()
#' gargle_oob_default()
#' gargle_oauth_cache()
#' gargle_oauth_client_type()
#' gargle_verbosity()
NULL

#' @rdname gargle_options
#' @export
#' @section `gargle_oauth_email`:
#' `gargle_oauth_email()` returns the option named "gargle_oauth_email", which
#' is undefined by default. If set, this option should be one of:
#'   * An actual email address corresponding to your preferred Google identity.
#'     Example:`janedoe@gmail.com`.
#'   * A glob pattern that indicates your preferred Google domain.
#'     Example:`*@example.com`.
#'   * `TRUE` to allow email and OAuth token auto-discovery, if exactly one
#'     suitable token is found in the cache.
#'   * `FALSE` or `NA` to force the OAuth dance in the browser.
gargle_oauth_email <- function() {
  getOption("gargle_oauth_email")
}

#' @rdname gargle_options
#' @export
#' @section `gargle_oob_default`:
#' `gargle_oob_default()` returns `TRUE` unconditionally on RStudio Server,
#' Posit Workbench, Posit Cloud, or Google Colaboratory, since it is not
#' possible to launch a local web server in these contexts. In this case, for
#' the final step of the OAuth dance, the user is redirected to a specific URL
#' where they must copy a code and paste it back into the R session.
#'
#' In all other contexts, `gargle_oob_default()` consults the option named
#' `"gargle_oob_default"`, then the option named `"httr_oob_default"`, and
#' eventually defaults to `FALSE`.
#'
#' "oob" stands for out-of-band. Read more about out-of-band authentication in
#' the vignette `vignette("auth-from-web")`.
gargle_oob_default <- function() {
  if (is_hosted_session()) {
    TRUE
  } else {
    getOption("gargle_oob_default") %||%
      getOption("httr_oob_default") %||%
      FALSE
  }
}

#' @rdname gargle_options
#' @export
#' @section `gargle_oauth_cache`:
#' `gargle_oauth_cache()` returns the option named "gargle_oauth_cache",
#' defaulting to `NA`. If defined, the option must be set to a logical value or
#' a string. `TRUE` means to cache using the default user-level cache file,
#' `~/.R/gargle/gargle-oauth`, `FALSE` means don't cache, and `NA` means to
#' guess using some sensible heuristics.
gargle_oauth_cache <- function() {
  getOption("gargle_oauth_cache", default = NA)
}

#' @rdname gargle_options
#' @export
#' @section `gargle_oauth_client_type`:
#' `gargle_oauth_client_type()` returns the option named
#' "gargle_oauth_client_type", if defined. If defined, the option must be either
#' "installed" or "web". If the option is not defined, the function returns:
#' * "web" on RStudio Server, Posit Workbench, Posit Cloud, or Google
#' Colaboratory
#' * "installed" otherwise
#'
#' Primarily intended to help infer the most suitable OAuth client type when a
#' user is relying on a built-in client, such as the tidyverse client used by
#' packages like bigrquery, googledrive, and googlesheets4.
gargle_oauth_client_type <- function() {
  opt <- getOption("gargle_oauth_client_type")
  if (is.null(opt)) {
    if (is_hosted_session()) "web" else "installed"
  } else {
    check_string(opt)
    arg_match(opt, values = c("installed", "web"))
  }
}

test_that("default options", {
  withr::local_options(list(
    gargle_oauth_cache = NULL,
    gargle_oob_default = NULL,
    httr_oob_default = NULL,
    gargle_oauth_client_type = NULL,
    gargle_oauth_email = NULL,
    gargle_verbosity = NULL,
    gargle_quiet = NULL
  ))
  expect_equal(gargle_oauth_cache(), NA)
  if (is_hosted_session()) {
    expect_true(gargle_oob_default())
    expect_equal(gargle_oauth_client_type(), "web")
  } else {
    expect_false(gargle_oob_default())
    expect_equal(gargle_oauth_client_type(), "installed")
  }
  expect_null(gargle_oauth_email())
  expect_equal(gargle_verbosity(), "info")
})

test_that("gargle_oob_default() consults gargle's option before httr's", {
  withr::local_options(list(
    gargle_oob_default = TRUE,
    httr_oob_default = FALSE
  ))
  expect_true(gargle_oob_default())
})

test_that("gargle_oob_default() consults httr's option", {
  withr::local_options(list(
    gargle_oob_default = NULL,
    httr_oob_default = TRUE
  ))
  expect_true(gargle_oob_default())
})

test_that("gargle_oauth_client_type() consults the option", {
  withr::local_options(list(gargle_oauth_client_type = "web"))
  expect_equal(gargle_oauth_client_type(), "web")
})

test_that("gargle API key", {
  key <- gargle_api_key()
  expect_true(is_string(key))
})

#' Make a Google API request
#'
#' @description
#' Intended primarily for internal use in client packages that provide
#' high-level wrappers for users. `request_make()` does relatively little:
#' * Calls an HTTP method.
#' * Adds a user agent.
#' * Enforces `"json"` as the default for `encode`. This differs from httr's
#'   default behaviour, but aligns better with Google APIs.
#'
#' Typically the input is created with [request_build()] and the output is
#' processed with [response_process()].
#'
#' @param x List. Holds the components for an HTTP request, presumably created
#'   with [request_develop()] or [request_build()]. Must contain a `method` and
#'   `url`. If present, `body` and `token` are used.
#' @param user_agent A user agent string, prepared by [httr::user_agent()]. When
#'   in doubt, a client package should have an internal function that extends
#'   `gargle_user_agent()` by prepending its return value with the client
#'   package's name and version.
#' @inheritParams httr::POST
#' @param ... Optional arguments passed through to the HTTP method. Currently
#'   neither gargle nor httr checks that all are used, so be aware that unused
#'   arguments may be silently ignored.
#'
#' @return Object of class `response` from [httr].
#' @export
#' @family requests and responses
#' @examples
#' \dontrun{
#' req <- gargle::request_build(
#'   method = "GET",
#'   path = "path/to/the/resource",
#'   token = "PRETEND_I_AM_TOKEN"
#' )
#' gargle::request_make(req)
#' }
request_make <- function(x,
                         ...,
                         encode = "json",
                         user_agent = gargle_user_agent()) {
  stopifnot(is.character(x$method))
  method <- switch(x$method,
    GET    = httr::GET,
    POST   = httr::POST,
    PATCH  = httr::PATCH,
    PUT    = httr::PUT,
    DELETE = httr::DELETE,
    gargle_abort("Not a recognized HTTP method: {.code {x$method}}.")
  )
  method(
    url = x$url,
    body = x$body,
    x$token,
    encode = encode,
    user_agent,
    ...
  )
}

gargle_user_agent <- function() {
  httr::user_agent(paste0(
    "gargle/", utils::packageVersion("gargle"), " ",
    "(GPN:RStudio; )", " ",
    "httr/", utils::packageVersion("httr")
  ))
}

#' Load Application Default Credentials
#'
#' @description

#' Loads credentials from a file identified via a search strategy known as
#' Application Default Credentials (ADC). The hope is to make auth "just work"
#' for someone working on Google-provided infrastructure or who has used Google
#' tooling to get started, such as the [`gcloud` command line
#' tool](https://docs.cloud.google.com/sdk/gcloud).
#'
#' A sequence of paths is consulted, which we describe here, with some abuse of
#' notation. ALL_CAPS represents the value of an environment variable and `%||%`
#' is used in the spirit of a [null coalescing
#' operator](https://en.wikipedia.org/wiki/Null_coalescing_operator).

#' ```
#' GOOGLE_APPLICATION_CREDENTIALS
#' CLOUDSDK_CONFIG/application_default_credentials.json
#' # on Windows:
#' (APPDATA %||% SystemDrive %||% C:)\gcloud\application_default_credentials.json
#' # on not-Windows:
#' ~/.config/gcloud/application_default_credentials.json
#' ```

#' If the above search successfully identifies a JSON file, it is parsed and
#' ingested as a service account, an external account ("workload identity
#' federation"), or a user account. Literally, if the JSON describes a service
#' account, we call [credentials_service_account()] and if it describes an
#' external account, we call [credentials_external_account()].
#'
#' @inheritParams credentials_service_account
#'
#' @seealso

#' * <https://docs.cloud.google.com/docs/authentication>

#' * <https://docs.cloud.google.com/sdk/docs>

#' @return An [`httr::TokenServiceAccount`][httr::Token-class], a [`WifToken`],
#'   an [`httr::Token2.0`][httr::Token-class] or `NULL`.

#' @family credential functions
#' @export
#' @examples
#' \dontrun{
#' credentials_app_default()
#' }
credentials_app_default <- function(scopes = NULL, ..., subject = NULL) {
  gargle_debug("trying {.fun credentials_app_default}")
  # In general, application default credentials only include the cloud-platform
  # scope.
  path <- credentials_app_default_path()
  if (!file_exists(path)) {
    return(NULL)
  }
  gargle_debug(c("file exists at ADC path:", "{.file {path}}"))

  info <- jsonlite::fromJSON(path, simplifyVector = FALSE)
  if (info$type == "authorized_user") {
    # In the case of *user* credentials stored as the application default, only
    # the cloud-platform scope will be included. This means we need our scopes to
    # be *implied* by the cloud-platform scope, which is hard to validate;
    # instead, we just approximate.
    valid_scopes <- c(
      "https://www.googleapis.com/auth/bigquery",
      "https://www.googleapis.com/auth/bigquery",
      "https://www.googleapis.com/auth/cloud-platform",
      "https://www.googleapis.com/auth/cloud-platform.readonly"
    )
    if (is.null(scopes) || !all(scopes %in% valid_scopes)) {
      return(NULL)
    }
    gargle_debug("ADC cred type: {.val authorized_user}")
    app <- httr::oauth_app(
      "google",
      info$client_id,
      secret = info$client_secret
    )
    scope <- "https://www.googleapis.com/auth/cloud.platform"
    token <- httr::Token2.0$new(
      endpoint = gargle_oauth_endpoint(),
      app = app,
      credentials = list(refresh_token = info$refresh_token),
      # ADC is already cached.
      cache_path = FALSE,
      params = list(scope = scope, as_header = TRUE)
    )
    token$refresh()
    token
  } else if (info$type == "service_account") {
    gargle_debug("ADC cred type: {.val service_account}")
    credentials_service_account(scopes, path = path, subject = subject)
  } else if (info$type == "external_account") {
    gargle_debug("ADC cred type: {.val external_account}")
    credentials_external_account(scopes, path = path)
  }
}

credentials_app_default_path <- function() {
  if (nzchar(Sys.getenv("GOOGLE_APPLICATION_CREDENTIALS"))) {
    return(path_expand(Sys.getenv("GOOGLE_APPLICATION_CREDENTIALS")))
  }

  pth <- "application_default_credentials.json"
  if (nzchar(Sys.getenv("CLOUDSDK_CONFIG"))) {
    pth <- c(Sys.getenv("CLOUDSDK_CONFIG"), pth)
  } else if (is_windows()) {
    appdata <- Sys.getenv("APPDATA", Sys.getenv("SystemDrive", "C:"))
    pth <- c(appdata, "gcloud", pth)
  } else {
    pth <- path_home(".config", "gcloud", pth)
  }
  path_join(pth)
}

test_that("credentials_app_default_path(), default, non-Windows", {
  withr::local_envvar(c(
    GOOGLE_APPLICATION_CREDENTIALS = NA,
    CLOUDSDK_CONFIG = NA,
    APPDATA = NA,
    SystemDrive = NA
  ))
  local_mocked_bindings(is_windows = function() FALSE)
  expect_equal(
    credentials_app_default_path(),
    path_home(".config", "gcloud", "application_default_credentials.json")
  )
})

test_that("credentials_app_default_path(), default, Windows", {
  withr::local_envvar(c(
    GOOGLE_APPLICATION_CREDENTIALS = NA,
    CLOUDSDK_CONFIG = NA,
    APPDATA = NA,
    SystemDrive = NA
  ))
  local_mocked_bindings(is_windows = function() TRUE)
  expect_equal(
    credentials_app_default_path(),
    path("C:", "gcloud", "application_default_credentials.json")
  )
})

test_that("credentials_app_default_path(), system drive, Windows", {
  withr::local_envvar(c(
    GOOGLE_APPLICATION_CREDENTIALS = NA,
    CLOUDSDK_CONFIG = NA,
    APPDATA = NA,
    SystemDrive = "D:"
  ))
  local_mocked_bindings(is_windows = function() TRUE)
  expect_equal(
    credentials_app_default_path(),
    path("D:", "gcloud", "application_default_credentials.json")
  )
})

test_that("credentials_app_default_path(), APPDATA env var, Windows", {
  withr::local_envvar(c(
    GOOGLE_APPLICATION_CREDENTIALS = NA,
    CLOUDSDK_CONFIG = NA,
    APPDATA = path("D:", "AppData"),
    SystemDrive = "D:"
  ))
  local_mocked_bindings(is_windows = function() TRUE)
  expect_equal(
    credentials_app_default_path(),
    path("D:", "AppData", "gcloud", "application_default_credentials.json")
  )
})

test_that("credentials_app_default_path(), CLOUDSDK_CONFIG env var", {
  withr::local_envvar(c(
    GOOGLE_APPLICATION_CREDENTIALS = NA,
    CLOUDSDK_CONFIG = path("CLOUDSDK", "path")
  ))
  expect_equal(
    credentials_app_default_path(),
    path("CLOUDSDK", "path", "application_default_credentials.json")
  )
})

test_that("credentials_app_default_path(), GOOGLE_APPLICATION_CREDENTIALS env var", {
  withr::local_envvar(c(
    GOOGLE_APPLICATION_CREDENTIALS = path("GAC", "path"),
    CLOUDSDK_CONFIG = path("CLOUDSDK", "path")
  ))
  expect_equal(
    credentials_app_default_path(),
    path("GAC", "path")
  )
})

empty_string <- function(x) {
  check_string(x)
  !nzchar(x)
}

is_windows <- function() {
  tolower(Sys.info()[["sysname"]]) == "windows"
}

is.oauth_app <- function(x) inherits(x, "oauth_app")

is.oauth_endpoint <- function(x) inherits(x, "oauth_endpoint")

# meant to evoke "RStudio, on a server", which includes RStudio Server and
# Posit Workbench
is_rstudio_server <- function() {
  Sys.getenv("RSTUDIO") == "1" &&
    Sys.getenv("RSTUDIO_PROGRAM_MODE") == "server"
}

is_positron <- function() {
  identical(Sys.getenv("POSITRON"), "1")
}

# meant to evoke "Positron, on a server", not Positron Server,
# which does not exist
is_positron_server <- function() {
  is_positron() &&
    # yes, it really is RSTUDIO_PROGRAM_MODE vs. POSITRON_MODE
    Sys.getenv("POSITRON_MODE") == "server"
}

# intended to detect execution on RStudio Server or Posit Workbench
# mostly aimed at detecting R use via VS Code on a server,
# since use via RStudio or Positron should be picked up by more specialized
# helpers: is_(rstudio|positron)_server()
is_workbench <- function() {
  # values seen on Workbench for RS_SERVER_URL in 2025-08-29 experimentation
  # VS Code, R in a terminal: "https://dev.palm.ptd.posit.it/"
  # Positron Pro, R Console and R in terminal: "https://dev.palm.ptd.posit.it/"
  # RStudio Pro, R Console and R in terminal: ""
  # yes, on RStudio Pro, RS_SERVER_URL is intentionally(?) set to the empty
  # string, but it IS set
  !is.na(Sys.getenv("RS_SERVER_URL", unset = NA_character_))
}

is_google_colab <- function() {
  # idea from https://stackoverflow.com/a/74930276
  # 2023-02-21 I created new notebook with
  # https://colab.research.google.com/#create=true&language=r
  # and I see:
  # Sys.getenv("COLAB_RELEASE_TAG") returns 'release-colab-20230216-060056-RC01'
  #
  # https://github.com/r-lib/gargle/issues/140#issuecomment-1439111627
  # via @craigcitro, the existence of this directory is another indicator:
  # /var/colab/hostname
  nzchar(Sys.getenv("COLAB_RELEASE_TAG"))
}

is_hosted_session <- function() {
  is_rstudio_server() ||
    is_positron_server() ||
    is_workbench() ||
    is_google_colab()
}

add_line <- function(path, line) {
  if (file_exists(path)) {
    lines <- readLines(path, warn = FALSE)
    lines <- lines[lines != ""]
  } else {
    lines <- character()
  }

  if (line %in% lines) {
    return(TRUE)
  }

  gargle_info("Adding {.val {line}} to {.file {path}}.")
  lines <- c(lines, line)
  writeLines(lines, path)
  TRUE
}

## in the spirit of basename(), but for Google scopes
## for printing purposes
base_scope <- function(x) {
  gsub("/$", "", gsub("(.*)/(.+$)", "...\\2", x))
}

normalize_scopes <- function(x) {
  stats::setNames(sort(unique(x)), NULL)
}

add_email_scope <- function(scopes = NULL) {
  gargle_debug("adding {.val userinfo.email} scope")
  url <- "https://www.googleapis.com/auth/userinfo.email"
  union(scopes %||% character(), url)
}

new_srcref <- function(lines) {
  n <- length(lines)

  srcref(
    srcfilecopy("HIDDEN", lines),
    c(1L, 1L, n, nchar(lines[[n]]))
  )
}

test_that("add_email_scope() works", {
  email_url <- add_email_scope()
  expect_length(email_url, 1)
  expect_equal(add_email_scope(email_url), email_url)
  expect_equal(
    add_email_scope("whatever"),
    c("whatever", email_url)
  )
  expect_equal(
    add_email_scope(c("whatever" = "whatever")),
    c("whatever", email_url)
  )
})

test_that("base_scope() extracts the last scope part", {
  scopes <- c(
    "https://www.googleapis.com/auth/userinfo.email",
    "https://www.googleapis.com/auth/drive",
    "https://www.googleapis.com/auth/gmail.readonly",
    "https://www.googleapis.com/auth/gmail.modify",
    "https://www.googleapis.com/auth/gmail.compose",
    "https://mail.google.com/"
  )
  out <- base_scope(scopes)
  expect_equal(
    out,
    c(
      "...userinfo.email",
      "...drive",
      "...gmail.readonly",
      "...gmail.modify",
      "...gmail.compose",
      "...mail.google.com"
    )
  )
})

test_that("obfuscate() works", {
  x <- c("123", "12345", "123456789")
  expect_equal(
    obfuscate(x, first = 1, last = 1),
    c("1...3", "1...5", "1...9")
  )
  expect_equal(
    obfuscate(x, first = 2, last = 1),
    c("123", "12...5", "12...9")
  )
  expect_equal(
    obfuscate(x, first = 4, last = 1),
    c("123", "12345", "1234...9")
  )
  expect_equal(
    obfuscate(x, first = 3, last = 3),
    c("123", "12345", "123...789")
  )
})

gargle_theme <- function() {
  list(
    span.field = list(transform = single_quote_if_no_color),
    # make the default bullet "regular" color, instead of explicitly colored
    # mostly motivated by consistency with googledrive, where the cumulative
    # use of color made me want to do this
    ".bullets .bullet-*" = list(
      "text-exdent" = 2,
      before = function(x) paste0(cli::symbol$bullet, " ")
    )
  )
}

single_quote_if_no_color <- function(x) quote_if_no_color(x, "'")
double_quote_if_no_color <- function(x) quote_if_no_color(x, '"')

quote_if_no_color <- function(x, quote = "'") {
  # TODO: if a better way appears in cli, use it
  # @gabor says: "if you want to have before and after for the no-color case
  # only, we can have a selector for that, such as:
  # span.field::no-color
  # (but, at the time I write this, cli does not support this yet)
  if (cli::num_ansi_colors() > 1) {
    x
  } else {
    paste0(quote, x, quote)
  }
}


#' @rdname gargle_options
#' @export
#' @section `gargle_verbosity`:
#' `gargle_verbosity()` returns the option named "gargle_verbosity", which
#' determines gargle's verbosity. There are three possible values, inspired by
#' the logging levels of log4j:
#' * "debug": Fine-grained information helpful when debugging, e.g. figuring out
#'   how `token_fetch()` is working through the registry of credential
#'   functions. Previously, this was activated by setting an option named
#'   "gargle_quiet" to `FALSE`.
#' * "info" (default): High-level information that a typical user needs to see.
#'   Since typical gargle usage is always indirect, i.e. gargle is called by
#'   another package, gargle itself is very quiet. There are very few messages
#'   emitted when `gargle_verbosity = "info"`.
#' * "silent": No messages at all. However, warnings or errors are still thrown
#'   normally.
gargle_verbosity <- function() {
  gv <- getOption("gargle_verbosity")

  # help people using the previous option
  if (is.null(gv)) {
    gq <- getOption("gargle_quiet")
    if (isFALSE(gq)) {
      options(gargle_verbosity = "debug")
      lifecycle::deprecate_warn(
        when = "1.1.0",
        what = I('The "gargle_quiet" option'),
        with = I('the "gargle_verbosity" option'),
        details = c(
          "x" = "Don't do this: `options(gargle_quiet = FALSE)`",
          "v" = 'Do this instead: `options(gargle_verbosity = "debug")`'
        ),
        always = TRUE
      )
    }
  }
  gv <- getOption("gargle_verbosity", "info")

  vals <- c("debug", "info", "silent")
  if (!is_string(gv) || !(gv %in% vals)) {
    gargle_abort(
      'Option "gargle_verbosity" must be one of: {.or {.field {vals}}}.'
    )
  }
  gv
}

#' @rdname gargle_options
#' @export
#' @param level Verbosity level: "debug" > "info" > "silent"
#' @param env The environment to use for scoping
local_gargle_verbosity <- function(level, env = caller_env()) {
  withr::local_options(list(gargle_verbosity = level), .local_envir = env)
}

#' @rdname gargle_options
#' @export
#' @param code Code to execute with specified verbosity level
with_gargle_verbosity <- function(level, code) {
  withr::with_options(list(gargle_verbosity = level), code = code)
}

gargle_debug <- function(text, .envir = caller_env()) {
  if (gargle_verbosity() == "debug") {
    cli::cli_div(theme = gargle_theme())
    cli::cli_bullets(text, .envir = .envir)
  }
}

gargle_info <- function(text, .envir = caller_env()) {
  if (gargle_verbosity() %in% c("debug", "info")) {
    cli::cli_div(theme = gargle_theme())
    cli::cli_bullets(text, .envir = .envir)
  }
}

commapse <- function(...) paste0(..., collapse = ", ")
fr <- function(x) format(x, justify = "right")
fl <- function(x) format(x, justify = "left")

## obscure part of (sensitive?) strings with '...'
## obfuscate("sensitive", first = 3, last = 2) = "sen...ve"
obfuscate <- function(x, first = 7, last = 0) {
  nc <- nchar(x)
  ellipsize <- nc > first + last
  out <- x
  out[ellipsize] <-
    paste0(
      substr(x[ellipsize], start = 1, stop = first),
      "...",
      substr(
        x[ellipsize],
        start = nc[ellipsize] - last + 1,
        stop = nc[ellipsize]
      )
    )
  out
}

message <- function(...) {
  gargle_abort(
    "
    Internal error: use {.pkg gargle}'s UI functions, not {.fun message}."
  )
}

#' Error conditions for the gargle package
#'
#' @param class Use only if you want to subclass beyond `gargle_error`
#'
#' @keywords internal
#' @name gargle-conditions
#' @noRd
NULL

gargle_abort <- function(
  message,
  ...,
  class = NULL,
  .envir = caller_env(),
  call = caller_env()
) {
  cli::cli_div(theme = gargle_theme())
  cli::cli_abort(
    message,
    class = c(class, "gargle_error"),
    .envir = .envir,
    call = call,
    ...
  )
}

# my heart's not totally in this because I'm not sure we should really be
# throwing any warnings, however we currently do re: token refresh
# so this wrapper makes the messaging more humane
# I am declining to add a class, e.g. gargle_warning
gargle_warn <- function(message, ..., class = NULL, .envir = caller_env()) {
  cli::cli_div(theme = gargle_theme())
  cli::cli_warn(message, .envir = .envir, ...)
}

gargle_abort_bad_params <- function(
  names,
  reason,
  endpoint_id,
  call = caller_env()
) {
  gargle_abort(
    c(
      "These parameters are {reason}:",
      bulletize(gargle_map_cli(names), bullet = "x"),
      "i" = gargle_map_cli(
        endpoint_id,
        template = "API endpoint: {.field <<x>>}"
      )
    ),
    class = "gargle_error_bad_params",
    call = call,
    names = names,
    reason = reason
  )
}

#' Map a cli-styled template over an object
#'
#' For internal use in gargle, googledrive, and googlesheets4 (for now).
#'
#' @keywords internal
#' @export
gargle_map_cli <- function(x, ...) UseMethod("gargle_map_cli")

#' @export
gargle_map_cli.default <- function(x, ...) {
  gargle_abort(
    "
    Don't know how to {.fun gargle_map_cli} an object of class \\
    {.cls {class(x)}}."
  )
}

#' @export
gargle_map_cli.NULL <- function(x, ...) NULL

#' @export
gargle_map_cli.character <- function(
  x,
  template = "{.field <<x>>}",
  .open = "<<",
  .close = ">>",
  ...
) {
  as.character(glue(template, .open = .open, .close = .close))
}

#' Abbreviate a bullet list neatly
#'
#' For internal use in gargle, googledrive, and googlesheets4 (for now).
#'
#' @keywords internal
#' @export
bulletize <- function(x, bullet = "*", n_show = 5, n_fudge = 2) {
  n <- length(x)
  n_show_actual <- compute_n_show(n, n_show, n_fudge)
  out <- utils::head(x, n_show_actual)
  n_not_shown <- n - n_show_actual

  out <- set_names(out, rep_along(out, bullet))

  if (n_not_shown == 0) {
    out
  } else {
    c(out, " " = glue("{cli::symbol$ellipsis} and {n_not_shown} more"))
  }
}

# I don't want to do "... and x more" if x is silly, i.e. 1 or 2
compute_n_show <- function(n, n_show_nominal = 5, n_fudge = 2) {
  if (n > n_show_nominal && n - n_show_nominal > n_fudge) {
    n_show_nominal
  } else {
    n
  }
}

# menu(), but based on readline() + cli and mockable ---------------------------
# https://github.com/r-lib/cli/issues/228
# https://github.com/rstudio/rsconnect/blob/main/R/utils-cli.R

cli_menu <- function(
  header,
  prompt,
  choices,
  not_interactive = choices,
  exit = integer(),
  .envir = caller_env(),
  error_call = caller_env()
) {
  if (!is_interactive()) {
    cli::cli_abort(
      c(header, not_interactive),
      .envir = .envir,
      call = error_call
    )
  }

  choices <- paste0(cli::style_bold(seq_along(choices)), ": ", choices)
  cli::cli_inform(
    c(header, prompt, choices),
    .envir = .envir
  )

  repeat {
    selected <- cli_readline("Selection: ")
    if (selected %in% c("0", seq_along(choices))) {
      break
    }
    cli::cli_inform(
      "Enter a number between 1 and {length(choices)}, or enter 0 to exit."
    )
  }

  selected <- as.integer(selected)
  if (selected %in% c(0, exit)) {
    if (is_testing()) {
      cli::cli_abort("Exiting...", call = NULL)
    } else {
      cli::cli_alert_danger("Exiting...")
      # simulate user pressing Ctrl + C
      invokeRestart("abort")
    }
  }

  selected
}

cli_readline <- function(prompt) {
  local_input <- getOption("cli_input", character())

  # not convinced that we need to plan for multiple mocked inputs, but leaving
  # this feature in for now
  if (length(local_input) > 0) {
    input <- local_input[[1]]
    cli::cli_inform(paste0(prompt, input))
    options(cli_input = local_input[-1])
    input
  } else {
    readline(prompt)
  }
}

local_user_input <- function(x, env = caller_env()) {
  withr::local_options(
    rlang_interactive = TRUE,
    # trailing 0 prevents infinite loop if x only contains invalid choices
    cli_input = c(x, "0"),
    .local_envir = env
  )
}

is_testing <- function() {
  identical(Sys.getenv("TESTTHAT"), "true")
}

# taken from lifecycle
# https://github.com/r-lib/lifecycle/blob/9417eca8f5091f95b4569fb6c388e4394e2b2157/R/utils.R#L30
pkg_url_bug <- function(pkg) {
  # First check that package is installed, e.g. in case of
  # runtime-only namespace created by pkgload
  if (nzchar(system.file(package = pkg))) {
    url <- utils::packageDescription(pkg)$BugReports

    # `url` can be NULL if not part of the description
    if (is_string(url) && grepl("^https://", url)) {
      return(url)
    }
  }

  NULL
}

test_that("gargle_verbosity() defaults to 'info'", {
  withr::local_options(list(
    gargle_verbosity = NULL,
    gargle_quiet = NULL
  ))
  expect_equal(gargle_verbosity(), "info")
})

test_that("gargle_verbosity() validates the value it finds", {
  withr::local_options(list(gargle_verbosity = TRUE))
  expect_snapshot_error(gargle_verbosity())
})

test_that("gargle_verbosity() accomodates people using the old option", {
  withr::local_options(list(
    gargle_verbosity = NULL,
    gargle_quiet = FALSE
  ))
  expect_snapshot(
    out <- gargle_verbosity()
  )
  expect_equal(out, "debug")
})

test_that("gargle_info() works", {
  blah <- "BLAH"

  local_gargle_verbosity("debug")
  expect_snapshot(gargle_info(c("aa {.field {blah}} bb", "cc {.emph xyz} dd")))

  local_gargle_verbosity("info")
  expect_snapshot(gargle_info(c("ee {.field {blah}} ff", "gg {.emph xyz} hh")))

  local_gargle_verbosity("silent")
  expect_snapshot(gargle_info(c("ii {.field {blah}} jj", "kk {.emph xyz} ll")))
})

test_that("gargle_debug() works", {
  foo <- "FOO"

  local_gargle_verbosity("debug")
  expect_snapshot(gargle_debug(c(
    "11 {.field {foo}} 22",
    "33 {.file a/b/c} 44"
  )))

  local_gargle_verbosity("info")
  expect_snapshot(gargle_debug(c(
    "55 {.field {foo}} 66",
    "77 {.file a/b/c} 88"
  )))

  local_gargle_verbosity("silent")
  expect_snapshot(gargle_debug(c(
    "99 {.field {foo}} 00",
    "11 {.file a/b/c} 22"
  )))
})

test_that("bulletize() works", {
  expect_snapshot(cli::cli_bullets(bulletize(letters)))
  expect_snapshot(cli::cli_bullets(bulletize(letters, bullet = "x")))
  expect_snapshot(cli::cli_bullets(bulletize(letters, n_show = 2)))
  expect_snapshot(cli::cli_bullets(bulletize(letters[1:6])))
  expect_snapshot(cli::cli_bullets(bulletize(letters[1:7])))
  expect_snapshot(cli::cli_bullets(bulletize(letters[1:8])))
  expect_snapshot(cli::cli_bullets(bulletize(letters[1:6], n_fudge = 0)))
  expect_snapshot(cli::cli_bullets(bulletize(letters[1:8], n_fudge = 3)))
})

# menu(), but based on readline() + cli and mockable ---------------------------

test_that("cli_menu() basic usage", {
  cli_menu_with_mock <- function(x) {
    local_user_input(x)
    cli_menu(
      "Found multiple thingies.",
      "Which one do you want to use?",
      glue("label {head(letters, 3)}")
    )
  }

  expect_snapshot(cli_menu_with_mock(1))
})

test_that("cli_menu() does not infinite loop with invalid mocked input", {
  cli_menu_with_mock <- function(x) {
    local_user_input(x)
    cli_menu(
      "Found multiple thingies.",
      "Which one do you want to use?",
      glue("label {head(letters, 3)}")
    )
  }

  expect_snapshot(cli_menu_with_mock("nope"), error = TRUE)
})

test_that("cli_menu() can work through multiple valid mocked inputs", {
  cli_menu_with_mock <- function(x) {
    local_user_input(x)
    header <- "Found multiple thingies."
    prompt <- "Which one do you want to use?"
    choices <- glue("label {1:3}")
    first <- cli_menu(header, prompt, choices)
    second <- cli_menu(header, prompt, choices)
    c(first, second)
  }

  expect_snapshot(
    out <- cli_menu_with_mock(c(1, 3))
  )
  expect_equal(out, c(1, 3))
})

test_that("cli_menu(), request exit via 0", {
  cli_menu_with_mock <- function(x) {
    local_user_input(x)
    cli_menu(
      "Found multiple thingies.",
      "Which one do you want to use?",
      glue("label {head(letters, 3)}")
    )
  }

  expect_snapshot(error = TRUE, cli_menu_with_mock(0))
})

test_that("cli_menu(exit =) works", {
  cli_menu_with_mock <- function(x) {
    local_user_input(x)
    cli_menu(
      header = "Hey we need to talk.",
      prompt = "What do you want to do?",
      choices = c(
        "Give up",
        "Some other thing"
      ),
      exit = 1
    )
  }

  expect_snapshot(error = TRUE, cli_menu_with_mock(1))
  expect_snapshot(cli_menu_with_mock(2))
})

test_that("cli_menu() inline markup and environment passing", {
  cli_menu_with_mock <- function(x) {
    local_user_input(x)
    verb <- "talk"
    action <- "do"
    pkg_name <- "nifty"
    cli_menu(
      header = "Hey we need to {.str {verb}}.",
      prompt = "What do you want to {.str {action}}?",
      choices = c(
        "Send email to {.email jane@example.com}",
        "Install the {.pkg {pkg_name}} package"
      )
    )
  }
  expect_snapshot(cli_menu_with_mock(1))
})

test_that("cli_menu() not_interactive, many strings, chained error", {
  wrapper_fun <- function() {
    local_interactive(FALSE)
    things <- glue("thing {1:3}")
    cli_menu(
      header = "Multiple things found.",
      prompt = "Which one do you want to use?",
      choices = things,
      not_interactive = c(
        i = "Use {.arg thingy} to specify one of {.str {things}}."
      )
    )
  }
  expect_snapshot(wrapper_fun(), error = TRUE)
})

# nocov start --- r-lib/rlang compat-types-check
#
# Dependencies
# ============
#
# - compat-obj-type.R
#
# Changelog
# =========
#
# 2022-09-28:
# - Removed `what` arguments.
# - Added `allow_na` and `allow_null` arguments.
# - Added `allow_decimal` and `allow_infinite` arguments.
# - Improved errors with absent arguments.
#
#
# 2022-09-16:
# - Unprefixed usage of rlang functions with `rlang::` to
#   avoid onLoad issues when called from rlang (#1482).
#
# 2022-08-11:
# - Added changelog.

# Scalars -----------------------------------------------------------------

check_bool <- function(x,
                       ...,
                       allow_na = FALSE,
                       allow_null = FALSE,
                       arg = caller_arg(x),
                       call = caller_env()) {
  if (!missing(x)) {
    if (is_bool(x)) {
      return(invisible(NULL))
    }
    if (allow_null && is_null(x)) {
      return(invisible(NULL))
    }
    if (allow_na && identical(x, NA)) {
      return(invisible(NULL))
    }
  }

  stop_input_type(
    x,
    c("`TRUE`", "`FALSE`"),
    ...,
    allow_na = allow_na,
    allow_null = allow_null,
    arg = arg,
    call = call
  )
}

check_string <- function(x,
                         ...,
                         allow_empty = FALSE,
                         allow_na = FALSE,
                         allow_null = FALSE,
                         arg = caller_arg(x),
                         call = caller_env()) {
  if (!missing(x)) {
    if (is_string(x)) {
      if (allow_empty || !is_string(x, "")) {
        return(invisible(NULL))
      }
    }
    if (allow_null && is_null(x)) {
      return(invisible(NULL))
    }
    if (allow_na && (identical(x, NA) ||
                     identical(x, na_chr))) {
      return(invisible(NULL))
    }
  }

  stop_input_type(
    x,
    "a single string",
    ...,
    allow_na = allow_na,
    allow_null = allow_null,
    arg = arg,
    call = call
  )

}

check_number_decimal <- function(x,
                                 ...,
                                 allow_infinite = TRUE,
                                 allow_na = FALSE,
                                 allow_null = FALSE,
                                 arg = caller_arg(x),
                                 call = caller_env()) {
  .rlang_types_check_number(
    x,
    ...,
    allow_decimal = TRUE,
    allow_infinite = allow_infinite,
    allow_na = allow_na,
    allow_null = allow_null,
    arg = arg,
    call = call
  )
}

check_number_whole <- function(x,
                               ...,
                               allow_na = FALSE,
                               allow_null = FALSE,
                               arg = caller_arg(x),
                               call = caller_env()) {
  .rlang_types_check_number(
    x,
    ...,
    allow_decimal = FALSE,
    allow_infinite = FALSE,
    allow_na = allow_na,
    allow_null = allow_null,
    arg = arg,
    call = call
  )
}

.rlang_types_check_number <- function(x,
                                      ...,
                                      allow_decimal = FALSE,
                                      allow_infinite = FALSE,
                                      allow_na = FALSE,
                                      allow_null = FALSE,
                                      arg = caller_arg(x),
                                      call = caller_env()) {
  if (!missing(x)) {
    is_number <- is_number(
      x,
      allow_decimal = allow_decimal,
      allow_infinite = allow_infinite
    )
    if (is_number) {
      return(invisible(NULL))
    }
    if (allow_null && is_null(x)) {
      return(invisible(NULL))
    }
    if (allow_na && (identical(x, NA) ||
                     identical(x, na_dbl) ||
                     identical(x, na_int))) {
      return(invisible(NULL))
    }
  }

  if (allow_decimal) {
    what <- "a number"
  } else {
    what <- "a whole number"
  }

  stop_input_type(
    x,
    what,
    ...,
    allow_na = allow_na,
    allow_null = allow_null,
    arg = arg,
    call = call
  )
}

is_number <- function(x,
                      allow_decimal = FALSE,
                      allow_infinite = FALSE) {
  if (!typeof(x) %in% c("integer", "double")) {
    return(FALSE)
  }
  if (length(x) != 1) {
    return(FALSE)
  }
  if (is.na(x)) {
    return(FALSE)
  }
  if (!allow_decimal && !is_integerish(x)) {
    return(FALSE)
  }
  if (!allow_infinite && is.infinite(x)) {
    return(FALSE)
  }
  TRUE
}

check_symbol <- function(x,
                         ...,
                         allow_null = FALSE,
                         arg = caller_arg(x),
                         call = caller_env()) {
  if (!missing(x)) {
    if (is_symbol(x)) {
      return(invisible(NULL))
    }
    if (allow_null && is_null(x)) {
      return(invisible(NULL))
    }
  }

  stop_input_type(
    x,
    "a symbol",
    ...,
    allow_null = allow_null,
    arg = arg,
    call = call
  )
}

check_arg <- function(x,
                      ...,
                      allow_null = FALSE,
                      arg = caller_arg(x),
                      call = caller_env()) {
  if (!missing(x)) {
    if (is_symbol(x)) {
      return(invisible(NULL))
    }
    if (allow_null && is_null(x)) {
      return(invisible(NULL))
    }
  }

  stop_input_type(
    x,
    "an argument name",
    ...,
    allow_null = allow_null,
    arg = arg,
    call = call
  )
}

check_call <- function(x,
                       ...,
                       allow_null = FALSE,
                       arg = caller_arg(x),
                       call = caller_env()) {
  if (!missing(x)) {
    if (is_call(x)) {
      return(invisible(NULL))
    }
    if (allow_null && is_null(x)) {
      return(invisible(NULL))
    }
  }

  stop_input_type(
    x,
    "a defused call",
    ...,
    allow_null = allow_null,
    arg = arg,
    call = call
  )
}

check_environment <- function(x,
                              ...,
                              allow_null = FALSE,
                              arg = caller_arg(x),
                              call = caller_env()) {
  if (!missing(x)) {
    if (is_environment(x)) {
      return(invisible(NULL))
    }
    if (allow_null && is_null(x)) {
      return(invisible(NULL))
    }
  }

  stop_input_type(
    x,
    "an environment",
    ...,
    allow_null = allow_null,
    arg = arg,
    call = call
  )
}

check_function <- function(x,
                           ...,
                           allow_null = FALSE,
                           arg = caller_arg(x),
                           call = caller_env()) {
  if (!missing(x)) {
    if (is_function(x)) {
      return(invisible(NULL))
    }
    if (allow_null && is_null(x)) {
      return(invisible(NULL))
    }
  }

  stop_input_type(
    x,
    "a function",
    ...,
    allow_null = allow_null,
    arg = arg,
    call = call
  )
}

check_closure <- function(x,
                          ...,
                          allow_null = FALSE,
                          arg = caller_arg(x),
                          call = caller_env()) {
  if (!missing(x)) {
    if (is_closure(x)) {
      return(invisible(NULL))
    }
    if (allow_null && is_null(x)) {
      return(invisible(NULL))
    }
  }

  stop_input_type(
    x,
    "an R function",
    ...,
    allow_null = allow_null,
    arg = arg,
    call = call
  )
}

check_formula <- function(x,
                          ...,
                          allow_null = FALSE,
                          arg = caller_arg(x),
                          call = caller_env()) {
  if (!missing(x)) {
    if (is_formula(x)) {
      return(invisible(NULL))
    }
    if (allow_null && is_null(x)) {
      return(invisible(NULL))
    }
  }

  stop_input_type(
    x,
    "a formula",
    ...,
    allow_null = allow_null,
    arg = arg,
    call = call
  )
}


# Vectors -----------------------------------------------------------------

check_character <- function(x,
                            ...,
                            allow_null = FALSE,
                            arg = caller_arg(x),
                            call = caller_env()) {
  if (!missing(x)) {
    if (is_character(x)) {
      return(invisible(NULL))
    }
    if (allow_null && is_null(x)) {
      return(invisible(NULL))
    }
  }

  stop_input_type(
    x,
    "a character vector",
    ...,
    allow_null = allow_null,
    arg = arg,
    call = call
  )
}

# nocov end

#' Make a Google API request, repeatedly
#'
#' Intended primarily for internal use in client packages that provide
#' high-level wrappers for users. It is a drop-in substitute for
#' [request_make()] that also has the ability to retry the request. Codes that
#' are considered retryable: 408, 429, 500, 502, 503.
#'
#' Consider an example where we are willing to make a request up to 5 times.
#'
#' ```
#' try  1  2    3        4                5
#'      |--|----|--------|----------------|
#' wait  1   2      3           4
#' ```
#'
#' There will be up to 5 - 1 = 4 waits and we generally want the waiting period
#' to get longer, in an exponential way. Such schemes are called exponential
#' backoff. `request_retry()` implements exponential backoff with "full jitter",
#' where each waiting time is generated from a uniform distribution, where the
#' interval of support grows exponentially. A common alternative is "equal
#' jitter", which adds some noise to fixed, exponentially increasing waiting
#' times.
#'
#' Either way our waiting times are based on a geometric series, which, by
#' convention, is usually written in terms of powers of 2:
#'
#' ```
#' b, 2b, 4b, 8b, ...
#'   = b * 2^0, b * 2^1, b * 2^2, b * 2^3, ...
#' ```
#'
#' The terms in this series require knowledge of `b`, the so-called exponential
#' base, and many retry functions and libraries require the user to specify
#' this. But most users find it easier to declare the total amount of waiting
#' time they can tolerate for one request. Therefore `request_retry()` asks for
#' that instead and solves for `b` internally. This is inspired by the Opnieuw
#' Python library for retries. Opnieuw's interface is designed to eliminate
#' uncertainty around:
#' * Units: Is this thing given in seconds? minutes? milliseconds?
#' * Ambiguity around how things are counted: Are we starting at 0 or 1?
#'   Are we counting tries or just the retries?
#' * Non-intuitive required inputs, e.g., the exponential base.
#'
#' Let *n* be the total number of tries we're willing to make (the argument
#' `max_tries_total`) and let *W* be the total amount of seconds we're willing
#' to dedicate to making and retrying this request (the argument
#' `max_total_wait_time_in_seconds`). Here's how we determine *b*:
#'
#' ```
#' sum_{i=0}^(n - 1) b * 2^i = W
#' b * sum_{i=0}^(n - 1) 2^i = W
#'        b * ( (2 ^ n) - 1) = W
#'                         b = W / ( (2 ^ n) - 1)
#' ```
#'
#' @section Special cases:
#' `request_retry()` departs from exponential backoff in three special cases:
#' * It actually implements *truncated* exponential backoff. There is a floor
#'   and a ceiling on random wait times.
#' * `Retry-After` header: If the response has a header named `Retry-After`
#'   (case-insensitive), it is assumed to provide a non-negative integer
#'   indicating the number of seconds to wait. If present, we wait this many
#'   seconds and do not generate a random waiting time. (In theory, this header
#'   can alternatively provide a datetime after which to retry, but we have no
#'   first-hand experience with this variant for a Google API.)
#' * Sheets API quota exhaustion: In the course of googlesheets4 development,
#'   we've grown very familiar with the `429 RESOURCE_EXHAUSTED` error. As of
#'   2023-04-15, the Sheets API v4 has a limit of 300 requests per minute per
#'   project and 60 requests per minute per user per project. Limits for reads
#'   and writes are tracked separately. In our experience, the "60 (read or
#'   write) requests per minute per user" limit is the one you hit most often.
#'   If we detect this specific failure, the first wait time is a bit more than
#'   one minute, then we revert to exponential backoff.
#'
#'
#' @param ... Passed along to [request_make()].
#' @param max_tries_total Maximum number of tries.
#' @param max_total_wait_time_in_seconds Total seconds we are willing to
#'   dedicate to waiting, summed across all tries. This is a technical upper
#'   bound and actual cumulative waiting will be less.
#'
#' @seealso
#' * <https://aws.amazon.com/blogs/architecture/exponential-backoff-and-jitter/>
#' * `https://www.channable.com/tech/opnieuw-a-simple-and-intuitive-retrying-library-for-python`
#' * <https://github.com/channable/opnieuw>
#' * <https://docs.cloud.google.com/storage/docs/retry-strategy>
#' * <https://www.rfc-editor.org/rfc/rfc7231#section-7.1.3>
#' * <https://developers.google.com/sheets/api/limits>
#' * <https://googleapis.dev/python/google-api-core/latest/retry.html>
#'
#' @inherit request_make return
#' @export
#'
#' @examples
#' \dontrun{
#' req <- gargle::request_build(
#'   method = "GET",
#'   path = "path/to/the/resource",
#'   token = "PRETEND_I_AM_TOKEN"
#' )
#' gargle::request_retry(req)
#' }
request_retry <- function(
  ...,
  max_tries_total = 5,
  max_total_wait_time_in_seconds = 100
) {
  resp <- request_make(...)

  tries_made <- 1
  per_user_failures <- 0

  b <- calculate_base_wait(
    n_waits = max_tries_total - 1,
    total_wait_time = max_total_wait_time_in_seconds
  )

  while (we_should_retry(tries_made, max_tries_total, resp)) {
    wait_info <- backoff(tries_made, resp, base = b, per_user_failures)
    wait_time <- wait_info$wait_time

    announce_retryable_failure(resp, tries_made, wait_info)

    # n = progress updates per second, which is really about the spinner
    n <- 20
    cli::cli_progress_bar(
      format = "{cli::pb_spin} Retry happens in {cli::pb_eta}",
      total = wait_time * n
    )
    for (i in seq_len(wait_time * n)) {
      Sys.sleep(1 / n)
      cli::cli_progress_update()
    }

    resp <- request_make(...)

    tries_made <- tries_made + 1
    if (sheets_per_user_quota_exhaustion(resp)) {
      per_user_failures <- per_user_failures + 1
    }
  }

  if (tries_made > 1) {
    code <- httr::status_code(resp)
    adjective <- if (code >= 200 && code < 300) "successful!" else "failed :("
    gargle_info(c("v" = "Request {tries_made} {adjective}"))
  }

  invisible(resp)
}

retryable_codes <- c("408", "429", "500", "502", "503")

we_should_retry <- function(tries_made, max_tries_total, resp) {
  if (tries_made >= max_tries_total) {
    FALSE
  } else if (httr::status_code(resp) %in% retryable_codes) {
    TRUE
  } else {
    FALSE
  }
}

backoff <- function(
  tries_made,
  resp,
  base = 1,
  per_user_failures = 0,
  min_wait = 1,
  max_wait = 64
) {
  wait_time <- stats::runif(1, 0, base * (2^(tries_made - 1)))
  wait_rationale <- "exponential backoff, full jitter"

  if (wait_time < min_wait) {
    wait_time <- min_wait + stats::runif(1)
    wait_rationale <- glue(
      "{wait_rationale}, clipped to floor of {min_wait} seconds"
    )
  }

  if (wait_time > max_wait) {
    wait_time <- max_wait + stats::runif(1)
    wait_rationale <- glue(
      "{wait_rationale}, clipped to ceiling of {max_wait} seconds"
    )
  }

  if (sheets_per_user_quota_exhaustion(resp) && per_user_failures < 1) {
    # 60s plus 1s and some jitter, for some wiggle
    wait_time <- 60 + 1 + stats::runif(1)
    wait_rationale <- "fixed 60 second wait for first per user quota exhaustion"
  }

  retry_after <- retry_after_header(resp)
  if (!is.null(retry_after)) {
    wait_time <- retry_after
    wait_rationale <- "'Retry-After' header"
  }

  list(wait_time = wait_time, wait_rationale = wait_rationale)
}

retry_after_header <- function(resp) {
  # TODO: consider honoring Retry-After with status codes besides 429
  if (!(httr::status_code(resp) == "429")) {
    return(NULL)
  }

  h <- httr::headers(resp)
  retry_after <- resp$headers[["retry-after"]]
  if (is.null(retry_after)) {
    NULL
  } else {
    as.numeric(retry_after)
  }
}

# targets the most common quota problem, which is the per user per minute quota
# from the Sheets API
sheets_per_user_quota_exhaustion <- function(resp) {
  msg <- gargle_error_message(resp)
  # the structure of this error and the wording of this message have changed
  # over time
  any(grepl("per user per 60 seconds", msg)) ||
    any(grepl("per minute per user", msg))
}

calculate_base_wait <- function(n_waits, total_wait_time) {
  stopifnot(is.numeric(n_waits), length(n_waits) == 1L, n_waits > 0)
  stopifnot(
    is.numeric(total_wait_time),
    length(total_wait_time) == 1L,
    total_wait_time > 0
  )
  b <- total_wait_time / (2^(n_waits) - 1)
  gargle_debug(c("i" = "Exponential base for retries is {round(b, 1)}s."))
  b
}

announce_retryable_failure <- function(resp, tries_made, wait_info) {
  status_code <- httr::status_code(resp)

  # add a bit more info, without doing full-blown error processing
  status_rpc <- tryCatch(
    {
      # oops is defined in response_process.R
      oops$RPC[[match(status_code, oops$HTTP)]]
    },
    error = function(e) NULL
  )
  status_extra <- glue(": {status_rpc}") %||% ""
  if (sheets_per_user_quota_exhaustion(resp)) {
    status_extra <- glue("{status_extra}, per user quota")
  }

  gargle_info(c(
    "x" = "Request {tries_made} failed [{status_code}{status_extra}].",
    "i" = "Will retry in {round(wait_info$wait_time, 1)}s."
  ))
  gargle_debug(c(
    "i" = "Wait time strategy: {wait_info$wait_rationale}",
    " " = gargle_error_message(resp)
  ))
}

test_that("request_retry() logic works as advertised", {
  faux_response <- function(status_code = 200, h = NULL) {
    structure(
      list(
        status_code = status_code,
        headers = if (!is.null(h)) httr::insensitive(h)
      ),
      class = "response"
    )
  }
  # allows us to replay a fixed set of responses
  faux_request_make <- function(responses = list(faux_response())) {
    i <- 0
    force(responses)
    function(...) {
      i <<- i + 1
      responses[[i]]
    }
  }

  # turn this: Will retry in 1.19769861013629s.
  # into this: Will retry in {WAIT_TIME}s.
  scrub_wait_time <- function(x) {
    sub(
      "(?<=retry in )[[:digit:]]+([.][[:digit:]]+)?(?=s)",
      "{WAIT_TIME}",
      x,
      perl = TRUE
    )
  }
  # turn this: (strategy: exponential backoff, full jitter, clipped to floor of 1 seconds)
  #            (strategy: exponential backoff, full jitter, clipped to ceiling of 45 seconds)
  # into this: (strategy: exponential backoff, full jitter)
  scrub_strategy <- function(x) {
    sub(
      ", clipped to (floor|ceiling) of [[:digit:]]+([.][[:digit:]]+)? seconds",
      "",
      x,
      perl = TRUE
    )
  }
  # turn this: base for retries is 2s
  # into this: base for retries is {BASE}s
  scrub_base <- function(x) {
    sub(
      "(?<=base for retries is )[[:digit:]]+([.][[:digit:]]+)?(?=s)",
      "{BASE}",
      x,
      perl = TRUE
    )
  }

  # get rid of the progress ticks like:
  # \ Retry 1 happens in  1s
  scrub_tick <- function(x) {
    grep("happens in", x, value = TRUE, invert = TRUE)
  }

  scrub <- function(x) {
    scrub_tick(scrub_base(scrub_strategy(scrub_wait_time(x))))
  }

  local_mocked_bindings(gargle_error_message = function(...) {
    "PLACEHOLDER FOR GOOGLE ERROR MESSAGE"
  })

  # succeed on first try
  local_mocked_bindings(request_make = faux_request_make())
  out <- request_retry()
  expect_equal(httr::status_code(out), 200)

  local_gargle_verbosity("debug")
  local_options(
    cli.progress_show_after = 0,
    cli.width = 200
  )

  # fail, then succeed (exponential backoff)
  r <- list(faux_response(429), faux_response())
  local_mocked_bindings(request_make = faux_request_make(r))
  expect_snapshot(
    fail_then_succeed <- request_retry(max_total_wait_time_in_seconds = 5),
    transform = scrub
  )
  expect_equal(httr::status_code(fail_then_succeed), 200)

  # fail, then succeed (Retry-After header)
  r <- list(
    faux_response(429, h = list(`Retry-After` = 1.4)),
    faux_response()
  )
  local_mocked_bindings(request_make = faux_request_make(r))
  expect_snapshot(
    fail_then_succeed <- request_retry(),
    transform = scrub
  )
  expect_equal(httr::status_code(fail_then_succeed), 200)

  # make sure max_tries_total is adjustable)
  r <- list(
    faux_response(429),
    faux_response(429),
    faux_response(429),
    faux_response()
  )
  local_mocked_bindings(request_make = faux_request_make(r[1:3]))
  expect_snapshot(
    fail_max_tries <- request_retry(
      max_tries_total = 3,
      max_total_wait_time_in_seconds = 6
    ),
    transform = scrub
  )
  expect_equal(httr::status_code(fail_max_tries), 429)
})

test_that("backoff() obeys obvious bounds from min_wait and max_wait", {
  faux_error <- function() {
    structure(list(status_code = 429), class = "response")
  }

  backoff_time_only <- function(...) backoff(...)$wait_time

  # raw wait_times in U[0,1], therefore all become min_wait + U[0,1]
  local_mocked_bindings(gargle_error_message = function(...) "oops")

  suppressMessages(
    wait_times <- vapply(
      rep.int(1, 100),
      backoff_time_only,
      FUN.VALUE = numeric(1),
      resp = faux_error(),
      min_wait = 3
    )
  )
  expect_true(all(wait_times > 3))
  expect_true(all(wait_times < 4))

  # raw wait_times in U[0,6], those that are < 1 become min_wait + U[0,1] and
  # those > 3 become max_wait + U[0,1]
  suppressMessages(
    wait_times <- vapply(
      rep.int(1, 100),
      backoff_time_only,
      FUN.VALUE = numeric(1),
      resp = faux_error(),
      base = 6,
      max_wait = 3
    )
  )
  expect_true(all(wait_times > 1))
  expect_true(all(wait_times < 3 + 1))
})

test_that("backoff() honors Retry-After header", {
  faux_429 <- function(h) {
    structure(
      list(
        status_code = 429,
        headers = httr::insensitive(h)
      ),
      class = "response"
    )
  }
  local_mocked_bindings(gargle_error_message = function(...) "oops")

  # play with capitalization and character vs numeric
  suppressMessages(
    out <- backoff(1, faux_429(list(`Retry-After` = "1.2")))
  )
  expect_equal(out$wait_time, 1.2)
  expect_equal(out$wait_rationale, "'Retry-After' header")

  suppressMessages(
    out <- backoff(1, faux_429(list(`retry-after` = 2.4)))
  )
  expect_equal(out$wait_time, 2.4)
  expect_equal(out$wait_rationale, "'Retry-After' header")

  # should work even when tries_made > 1
  suppressMessages(
    out <- backoff(3, faux_429(list(`reTry-aFteR` = 3.6)))
  )
  expect_equal(out$wait_time, 3.6)
  expect_equal(out$wait_rationale, "'Retry-After' header")
})

#' Fetch a token for the given scopes
#'
#' This is a rather magical function that calls a series of concrete
#' credential-fetching functions, each wrapped in a `tryCatch()`.
#' `token_fetch()` keeps trying until it succeeds or there are no more functions
#' to try. See the `vignette("how-gargle-gets-tokens")` for a full description
#' of `token_fetch()`.
#'
#' @seealso [cred_funs_list()] reveals the current registry of
#'   credential-fetching functions, in order.
#'
#' @inheritParams credentials_user_oauth2
#' @param ... Additional arguments passed to all credential functions.
#'
#' @return An [`httr::Token`][httr::Token-class] (often an instance of something
#'   that inherits from `httr::Token`) or `NULL`.
#' @family credential functions
#' @export
#' @examples
#' \dontrun{
#' token_fetch(scopes = "https://www.googleapis.com/auth/userinfo.email")
#' }
token_fetch <- function(scopes = NULL, ...) {
  gargle_debug("trying {.fun token_fetch}")
  for (f in gargle_env$cred_funs) {
    token <- NULL
    token <- tryCatch(
      error = function(e) {
        gargle_debug(c("Error caught by {.fun token_fetch}:", e$message))
        NULL
      },
      withCallingHandlers(
        f(scopes, ...),
        warning = function(e) {
          gargle_debug(c("Warning caught by {.fun token_fetch}:", e$message))
        }
      )
    )
    if (!is.null(token)) {
      return(token)
    }
  }
  NULL
}

# ---- make too many read requests in a time interval
googlesheets4::gs4_deauth()
deaths_id <- as.character(googlesheets4::gs4_example("deaths"))

f <- function(ssid = deaths_id, i = 0) {
  req <- googlesheets4::request_generate(
    "sheets.spreadsheets.get",
    params = list(spreadsheetId = ssid, fields = "spreadsheetId")
  )
  raw_resp <- gargle::request_make(req)
  code <- httr::status_code(raw_resp)
  cat(i, "code:", code, "\n")
  if (code >= 200 && code < 300) {
    invisible(TRUE)
  } else {
    invisible(raw_resp)
  }
}

n <- 300
i <- 0
resp <- TRUE
while (isTRUE(resp) && i < n) {
  i <- i + 1
  resp <- f(i = i)
}

resp

stopifnot(httr::status_code(resp) == 429)
saveRDS(
  gargle:::redact_response(resp),
  testthat::test_path(
    "fixtures",
    "sheets-spreadsheets-get-quota-exceeded-readgroup_429.rds"
  ),
  version = 2
)

gargle::response_process(resp)

#' Create an OAuth client for Google
#'
#' @description
#' `r lifecycle::badge("experimental")`

#' A `gargle_oauth_client` consists of:

#' * A type. gargle only supports the "Desktop app" and "Web application" client
#' types. Different types are associated with different OAuth flows.

#' * A client ID and secret.
#' * Optionally, one or more redirect URIs.

#' * A name. This is really a human-facing label. Or, rather, it can be used
#' that way, but the default is just a hash. We recommend using the same name
#' here as the name used to label the client ID in the [Google Cloud Platform
#' Console](https://console.cloud.google.com).
#'
#' A `gargle_oauth_client` is an adaptation of httr's [oauth_app()] (currently)
#' and httr2's `oauth_client()` (which gargle will migrate to in the future),
#' specialized for Google APIs. This function and class is marked "experimental"
#' since the details of this transition are necessarily uncertain.

#' @param path JSON downloaded from [Google Cloud
#'   Console](https://console.cloud.google.com), containing a client id and
#'   secret, in one of the forms supported for the `txt` argument of
#'   [jsonlite::fromJSON()] (typically, a file path or JSON string).

#' @param name A label for this specific client, presumably the same name used
#'   to label it in Google Cloud Console. Unfortunately there is no way to
#'   make that true programmatically, i.e. the JSON representation does not
#'   contain this information.

#' @param id Client ID
#' @param secret Client secret

#' @param redirect_uris Where your app listens for the response from Google's
#'   authorization server. If you didn't configure this specifically when
#'   creating the client (which is only possible for clients of the "web" type),
#'   you can leave this unspecified.

#' @param type Specifies the type of OAuth client. The valid values are a subset
#'   of possible Google client types and reflect the key used to describe the
#'   client in its JSON representation:

#'   * `"installed"` is associated with a "Desktop app"
#'   * `"web"` is associated with a "Web application"

#' @return An OAuth client: An S3 list with class `gargle_oauth_client`. For
#'   backwards compatibility reasons, this currently also inherits from the httr
#'   S3 class `oauth_app`, but that is a temporary measure. An instance of
#'   `gargle_oauth_client` stores more information than httr's `oauth_app`, such
#'   as the OAuth client's type ("web" or "installed").

#'
#'   There are some redundant fields in this object during the httr-to-httr2
#'   transition period. The legacy fields `appname` and `key` repeat the
#'   information in the future-facing fields `name` and (client) `id`. Prefer
#'   `name` and `id` to `appname` and `key` in downstream code. Prefer the
#'   constructors `gargle_oauth_client_from_json()` and `gargle_oauth_client()`
#'   to [httr::oauth_app()] and [oauth_app_from_json()].

#' @export
#'
#' @examples
#' \dontrun{
#' gargle_oauth_client_from_json(
#'   path = "/path/to/the/JSON/you/downloaded/from/gcp/console.json",
#'   name = "my-nifty-oauth-client"
#' )
#' }
#'
#' gargle_oauth_client(
#'   id = "some_long_id",
#'   secret = "ssshhhhh_its_a_secret",
#'   name = "my-nifty-oauth-client"
#' )
gargle_oauth_client_from_json <- function(path, name = NULL) {
  check_string(path)
  if (!is.null(name)) {
    check_string(name)
  }

  json <- jsonlite::fromJSON(path, simplifyVector = FALSE)
  if (length(json) != 1) {
    gargle_abort(c(
      "JSON has an unexpected form",
      "i" = "Are you sure this is the JSON downloaded for an OAuth client?",
      "i" = "It is easy to confuse the JSON for an OAuth client and a service account."
    ))
  }

  info <- json[[1]]

  gargle_oauth_client(
    id = info$client_id,
    secret = info$client_secret,
    redirect_uris = info$redirect_uris,
    type = names(json),
    name = name %||% glue("{info$project_id}_{hash(info$project_id)}")
  )
}


#' @export
#' @rdname gargle_oauth_client_from_json
gargle_oauth_client <- function(id,
                                secret,
                                redirect_uris = NULL,
                                type = c("installed", "web"),
                                name = hash(id)) {
  check_string(id)
  check_string(secret)
  check_string(name)
  type <- arg_match(type)

  if (!is.null(redirect_uris)) {
    # httr appears to assume that an OAuth app can have exactly 1 redirect_uri
    # (gargle has never used the `redirect_uri` field of httr::oauth_app)
    # httr2 seems to think it can usually construct the redirect_uri?
    # I think I have to accept multiple URIs, because that can be true in the
    # downloaded JSON
    # we'll just have to decide which one to use downstream, based on context
    redirect_uris <- unlist(redirect_uris)
    check_character(redirect_uris)
  }

  if (type == "web" && length(redirect_uris) == 0) {
    gargle_abort('
        A "web" type OAuth client must have one or more {.field redirect_uris}.')
  }

  structure(
    list(
      name = name,
      id = id,
      secret = secret,
      type = type,
      redirect_uris = redirect_uris,
      # needed for backwards compatibility; I need this class to quack like a
      # specialization of httr's oauth_app class, for now
      appname = name,
      key = id
    ),
    class = c("gargle_oauth_client", "oauth_app")
    # in the future, maybe:
    # class = c("gargle_oauth_client", "httr2_oauth_client")
  )
}
# adapted from httr2 ----
#' @export
print.gargle_oauth_client <- function(x, ...) {
  # this print method needs work, but not a high priority atm
  cli::cli_text(cli::style_bold("<gargle_oauth_client>"))
  redacted <- list_redact(compact(x), "secret")
  # quick fix for multiple URIs case
  if (length(redacted$redirect_uris) > 0) {
    redacted$redirect_uris <- commapse(redacted$redirect_uris)
  }
  # hide redundant fields that exist only for backwards compatibility with
  # httr's oauth_app
  redacted$appname <- redacted$key <- NULL
  cli::cli_dl(redacted)
  invisible(x)
}

list_redact <- function(x, names, case_sensitive = TRUE) {
  if (case_sensitive) {
    i <- match(names, names(x))
  } else {
    i <- match(tolower(names), tolower(names(x)))
  }
  x[i] <- cli::col_grey("<REDACTED>")
  x
}

#' OAuth client for demonstration purposes
#'
#' @description

#' Invisibly returns an instance of
#' [`gargle_oauth_client`][gargle_oauth_client()] that can be used to test drive
#' gargle before obtaining your own client ID and secret. This OAuth client may
#' be deleted or rotated at any time. There are no guarantees about which APIs
#' are enabled. DO NOT USE THIS IN A PACKAGE or for anything other than
#' interactive, small-scale experimentation.
#'
#' You can get your own OAuth client ID and secret, without these limitations.
#' See the `vignette("get-api-credentials")` for more details.
#'
#' @inheritParams gargle_oauth_client_from_json
#'
#' @return An OAuth client, produced by [gargle_oauth_client()], invisibly.
#' @export
#' @keywords internal
#' @examples
#' \dontrun{
#' gargle_client()
#' }
gargle_client <- function(type = NULL) {
  if (is.null(type) || is.na(type)) {
    type <- if(is_rstudio_server()) "web" else "installed"
  }
  check_string(type)
  type <- arg_match(type, values = c("installed", "web"))

  switch(
    type,
    web       = goc_web(),
    installed = goc_installed()
  )
}

#' @export
#' @keywords internal
#' @rdname internal-assets
tidyverse_client <- function(type = NULL) {
  check_permitted_package(parent.frame())

  if (is.null(type) || is.na(type)) {
    type <- if(is_rstudio_server()) "web" else "installed"
  }
  check_string(type)
  type <- arg_match(type, values = c("installed", "web"))

  switch(
    type,
    web       = toc_web(),
    installed = toc_installed()
  )
}

# deprecated functions ----

#' Create an OAuth app from JSON
#'
#' @description
#' `r lifecycle::badge("deprecated")`
#'

#' `oauth_app_from_json()` is being replaced with
#' [`gargle_oauth_client_from_json()`], in light of the new
#' `gargle_oauth_client` class. Now `oauth_app_from_json()` potentially warns
#' about this deprecation and immediately passes its inputs through to
#' [`gargle_oauth_client_from_json()`].
#'
#' `gargle_app()` is being replaced with [gargle_client()].
#'
#' @inheritParams gargle_oauth_client
#' @inheritParams httr::oauth_app
#' @keywords internal
#' @export
oauth_app_from_json <- function(path,
                                appname = NULL) {
  lifecycle::deprecate_soft(
    "1.3.0", "oauth_app_from_json()", "gargle_oauth_client_from_json()"
  )
  gargle_oauth_client_from_json(path = path, name = appname)
}

#' @export
#' @keywords internal
#' @rdname internal-assets
tidyverse_app <- function() {
  lifecycle::deprecate_soft(
    "1.3.0", "tidyverse_app()", "tidyverse_client()"
  )
  tidyverse_client()
}

#' @export
#' @keywords internal
#' @rdname oauth_app_from_json
gargle_app <- function() {
  lifecycle::deprecate_soft(
    "1.3.0", "gargle_app()", "gargle_client()"
  )
  gargle_client()
}

test_that("gargle_oauth_client() rejects bad input", {
  expect_snapshot(error = TRUE, gargle_oauth_client())
  expect_snapshot(error = TRUE, gargle_oauth_client(1234))
  expect_snapshot(error = TRUE, gargle_oauth_client(id = "ID"))
  expect_snapshot(error = TRUE, gargle_oauth_client(id = "ID", secret = 1234))
  expect_snapshot(
    error = TRUE,
    gargle_oauth_client("ID", "SECRET", type = "nope")
  )
})

test_that("gargle_oauth_client() has special handling for web clients", {
  expect_snapshot(
    error = TRUE,
    gargle_oauth_client("ID", "SECRET", type = "web")
  )
  expect_snapshot(
    gargle_oauth_client(
      "ID", "SECRET", type = "web",
      redirect_uris = c(
        "http://localhost:8111/",
        "http://127.0.0.1:8100/",
        "https://example.com/aaa/bbb/v"
      )
    )
  )
})

test_that("service account JSON throws an informative error", {
  expect_snapshot(
    error = TRUE,
    gargle_oauth_client_from_json(
      test_path("fixtures", "service-account-token.json")
    )
  )
})

# deprecated functions ----

test_that("oauth app from JSON", {
  withr::local_options(lifecycle_verbosity = "quiet")
  oa <- oauth_app_from_json(
    test_path(
      "fixtures", "client_secret_123.googleusercontent.com.json"
    )
  )
  expect_s3_class(oa, "oauth_app")
  expect_match(oa$appname, "^a_project")
  expect_equal(oa$secret, "ssshh-i-am-a-secret")
  expect_equal(oa$key, "abc.apps.googleusercontent.com")

  oa <- oauth_app_from_json(
    test_path(
      "fixtures", "client_secret_456.googleusercontent.com.json"
    )
  )
  expect_s3_class(oa, "oauth_app")
  expect_match(oa$appname, "^a_project")
  expect_equal(oa$secret, "ssshh-i-am-a-secret")
  expect_equal(oa$key, "abc.apps.googleusercontent.com")
})

test_that("gargle is 'inside the house'", {
  expect_true(from_permitted_package())
  expect_error_free(check_permitted_package())
})

test_that("it is possible to be 'outside the house'", {
  expect_false(local(gargle:::from_permitted_package(), envir = globalenv()))
  expect_snapshot(
    local(gargle:::check_permitted_package(), envir = globalenv()),
    error = TRUE
  )
})

test_that("gargle API key", {
  key <- gargle_api_key()
  expect_true(is_string(key))
})

test_that("tidyverse API key", {
  key <- tidyverse_api_key()
  expect_true(is_string(key))
  expect_snapshot(
    local(tidyverse_api_key(), envir = globalenv()),
    error = TRUE
  )
})

test_that("gargle oauth app (deprecated)", {
  expect_snapshot(
    oa <- gargle_app()
  )
  expect_s3_class(oa, "oauth_app")
  expect_match(oa$appname, "^gargle")
})

test_that("gargle oauth installed client", {
  oc <- gargle_client()
  expect_s3_class(oc, "gargle_oauth_client")
  expect_s3_class(oc, "oauth_app")
  expect_match(oc$name, "^gargle")
  expect_equal(oc$type, "installed")

  expect_equal(gargle_client("installed"), oc)
})

test_that("gargle oauth web client", {
  oc <- gargle_client("web")
  expect_s3_class(oc, "gargle_oauth_client")
  expect_s3_class(oc, "oauth_app")
  expect_match(oc$name, "^gargle")
  expect_equal(oc$type, "web")
  expect_equal(oc$redirect_uris, "https://www.tidyverse.org/google-callback/")
})

test_that("tidyverse oauth app (deprecated)", {
  expect_snapshot(
    oa <- tidyverse_app()
  )
  expect_s3_class(oa, "oauth_app")
  expect_match(oa$appname, "^tidyverse")
})

test_that("tidyverse oauth installed client", {
  oc <- tidyverse_client()
  expect_s3_class(oc, "gargle_oauth_client")
  expect_s3_class(oc, "oauth_app")
  expect_match(oc$name, "^tidyverse")
  expect_equal(oc$type, "installed")

  expect_equal(tidyverse_client("installed"), oc)
})

test_that("tidyverse oauth web client", {
  oc <- tidyverse_client("web")
  expect_s3_class(oc, "gargle_oauth_client")
  expect_s3_class(oc, "oauth_app")
  expect_match(oc$name, "^tidyverse")
  expect_equal(oc$type, "web")
  expect_equal(oc$redirect_uris, "https://www.tidyverse.org/google-callback/")
})

test_that("gargle_oauth_client() rejects bad input", {
  expect_snapshot(error = TRUE, gargle_oauth_client())
  expect_snapshot(error = TRUE, gargle_oauth_client(1234))
  expect_snapshot(error = TRUE, gargle_oauth_client(id = "ID"))
  expect_snapshot(error = TRUE, gargle_oauth_client(id = "ID", secret = 1234))
  expect_snapshot(
    error = TRUE,
    gargle_oauth_client("ID", "SECRET", type = "nope")
  )
})

test_that("gargle_oauth_client() has special handling for web clients", {
  expect_snapshot(
    error = TRUE,
    gargle_oauth_client("ID", "SECRET", type = "web")
  )
  expect_snapshot(
    gargle_oauth_client(
      "ID",
      "SECRET",
      type = "web",
      redirect_uris = c(
        "http://localhost:8111/",
        "http://127.0.0.1:8100/",
        "https://example.com/aaa/bbb/v"
      )
    )
  )
})

test_that("service account JSON throws an informative error", {
  expect_snapshot(
    error = TRUE,
    gargle_oauth_client_from_json(
      test_path("fixtures", "service-account-token.json")
    )
  )
})

# deprecated functions ----

test_that("oauth app from JSON", {
  withr::local_options(lifecycle_verbosity = "quiet")
  oa <- oauth_app_from_json(
    fs::path_package(
      "gargle",
      "extdata",
      "client_secret_installed.googleusercontent.com.json"
    )
  )
  expect_s3_class(oa, "oauth_app")
  expect_match(oa$appname, "^a_project")
  expect_equal(oa$secret, "ssshh-i-am-a-secret")
  expect_equal(oa$key, "abc.apps.googleusercontent.com")

  oa <- oauth_app_from_json(
    fs::path_package(
      "gargle",
      "extdata",
      "client_secret_web.googleusercontent.com.json"
    )
  )
  expect_s3_class(oa, "oauth_app")
  expect_match(oa$appname, "^a_project")
  expect_equal(oa$secret, "ssshh-i-am-a-secret")
  expect_equal(oa$key, "abc.apps.googleusercontent.com")
})

# This file has its origins in oauth-init.R in httr.
# Motivated by the need to support the pseudo-OOB flow.
#
# Affected functions:
# - Modified: init_oauth2.0(). This function is the workhorse for the
#   $init_credentials() method of the Token2.0/Gargle2.0 class. Previously,
#   Gargle2.0 ultimately delegated to the Token2.0 method, but now the method is
#   fully implemented for Gargle2.0.
# - Modified: oauth_authorize(). This function gains the ability to do code
#   exchange *with state*, by calling the new function
#   oauth_exchanger_with_state().
# - Added: oauth_exchanger_with_state()
# - Added: csrf_token(). Used to create the `state` token (example of a
#   cross-site request forgery token). Switched one existing use of
#   httr:::nonce() to this, now that I can.
# - The internal helper check_scope() got inlined (it was a mix of a checker
#   and a processor).
# - The internal helper check_oob() got modified to use gargle conventions.

#' Retrieve OAuth 2.0 access token, but specific to gargle
#'
#' @param endpoint An OAuth endpoint, presumably the one returned by
#'   `gargle_oauth_endpoint()`. The fact that this is even an argument is
#'   because this function is based on `httr::init_oauth2.0()`.
#' @param client An OAuth client, preferably an instance of `gargle_oauth_client`.
#' @param scope a character vector of scopes to request.
#' @param use_oob Whether to use out-of-band auth. Results in conventional OOB
#'   if the `client` is of type `"installed"` (or if type is unknown) and
#'   pseudo-OOB if the `client` is of type `"web"`.
#' @param oob_value if provided, specifies the value to use for the redirect_uri
#'   parameter when retrieving an authorization URL. For conventional OOB, this
#'   defaults to "urn:ietf:wg:oauth:2.0:oob". For pseudo-OOB, this should be the
#'   (or a) redirect URI configured for the OAuth client. Consulted only when
#'   `use_oob = TRUE`.
#' @param query_authorize_extra Named list of query parameters to include in the
#'   initial request to the authorization server.
#' @noRd
init_oauth2.0 <- function(
  endpoint = gargle_oauth_endpoint(),
  client = gargle_client(),
  scope = NULL,
  use_oob = gargle_oob_default(),
  oob_value = NULL,
  is_interactive = interactive(),
  query_authorize_extra = list()
) {
  check_character(scope, allow_null = TRUE)
  scope <- glue_collapse(scope, sep = " ")

  use_oob <- check_oob(use_oob, oob_value)

  client_type <- if (inherits(client, "gargle_oauth_client")) {
    client$type
  } else {
    NA
  }

  if (use_oob) {
    redirect_uri <- oob_value %||% "urn:ietf:wg:oauth:2.0:oob"

    if (identical(client_type, "web")) {
      # pseudo-OOB flow
      # https://developers.google.com/identity/protocols/oauth2/web-server#creatingclient

      # We need so-called "offline" access, so the access token can be
      # refreshed without re-prompting the user for permission.
      # Specifically, this is necessary (though not sufficient!) to make the
      # authorization server return a **refresh token** in addition to an
      # access token.
      # Offline access is the default for installed applications, but it is NOT
      # the default for web apps, so we must explicitly request it.
      query_authorize_extra[["access_type"]] <- "offline"

      # https://stackoverflow.com/questions/10827920/not-receiving-google-oauth-refresh-token
      # https://developers.google.com/identity/protocols/oauth2/openid-connect#re-consent

      # By default, for a web app, the user is only prompted for consent once
      # per project. And this is necessary in order to get a refresh token.
      # So we must explicitly ask for re-consent.
      query_authorize_extra[["prompt"]] <- "consent"

      state <- csrf_token()
    } else {
      # conventional oob
      state <- NULL
    }
  } else {
    redirect_uri <- httr::oauth_callback()
    state <- csrf_token()
  }

  authorize_url <- httr::oauth2.0_authorize_url(
    endpoint,
    client,
    scope = scope,
    redirect_uri = redirect_uri,
    state = state,
    query_extra = query_authorize_extra
  )
  code <- oauth_authorize(
    authorize_url,
    oob = use_oob,
    client_type = client_type,
    state = state
  )

  # Use authorisation code to get (temporary) access token
  httr::oauth2.0_access_token(
    endpoint,
    client,
    code = code,
    redirect_uri = redirect_uri
  )
}

# https://developers.google.com/identity/protocols/oauth2/openid-connect#createxsrftoken
# "These tokens are often referred to as cross-site request forgery (CSRF)
# tokens.
#
# One good choice for a state token is a string of 30 or so characters
# constructed using a high-quality random-number generator."
csrf_token <- function(n_bytes = 16) {
  paste0(as.character(openssl::rand_bytes(n_bytes)), collapse = "")
}

oauth_authorize <- function(url, oob = FALSE, client_type = NA, state = NULL) {
  if (oob) {
    if (identical(client_type, "web")) {
      # pseudo oob
      oauth_exchanger_with_state(url, state)$code
    } else {
      httr::oauth_exchanger(url)$code
    }
  } else {
    httr::oauth_listener(url)$code
  }
}

oauth_exchanger_with_state <- function(request_url, state) {
  httr::BROWSE(request_url)

  info_enc <- trimws(readline("Enter authorization code: "))
  info <- jsonlite::fromJSON(rawToChar(openssl::base64_decode(info_enc)))
  if (!identical(info$state, state)) {
    stop("state did not match")
  }
  list(code = info$code)
}

check_oob <- function(use_oob, oob_value = NULL) {
  check_bool(use_oob)

  if (!use_oob && !is_installed("httpuv")) {
    gargle_info(
      "The {.pkg httpuv} package is not installed; using out-of-band auth."
    )
    use_oob <- TRUE
  }

  if (use_oob && !is_interactive()) {
    gargle_abort("Out-of-band auth only works in an interactive session.")
  }

  if (!is.null(oob_value) && !use_oob) {
    gargle_abort(
      "
      The {.arg oob_value} argument can only be used when {.code use_oob = TRUE}."
    )
  }

  if (use_oob && !is.null(oob_value)) {
    check_string(oob_value)
  }

  use_oob
}

# ---- non-existent range
googlesheets4::gs4_deauth()
deaths_id <- as.character(googlesheets4::gs4_example("deaths"))
req <- googlesheets4::request_generate(
  endpoint = "sheets.spreadsheets.get",
  params = list(
    spreadsheetId = deaths_id,
    ranges = "NOPE!A5:F15"
  )
)
resp <- googlesheets4::request_make(req)

stopifnot(httr::status_code(resp) == 400)
saveRDS(
  gargle:::redact_response(resp),
  testthat::test_path(
    "fixtures",
    "sheets-spreadsheets-get-nonexistent-range_400.rds"
  ),
  version = 2
)

gargle::response_process(resp)

#' Environment used for gargle global state
#'
#' This environment contains:
#' * `$cred_funs` is the ordered list of credential functions to use when trying
#'   to fetch credentials. It is populated by a call to
#'   `cred_funs_set_default()` in `.onLoad()`.
#' * `$last_response` is the most recent response provided to
#'   `response_process()`.
#'
#' @noRd
#' @format An environment.
#' @keywords internal
gargle_env <- new.env(parent = emptyenv())
gargle_env$cred_funs <- list()
gargle_env$last_response <- list()

#' Access the last response
#'
#' These functions give access to the most recent response processed by
#' [`response_process(..., remember = TRUE)`][response_process()]. They can be
#' useful for *post mortem* analysis of puzzling or failed API interactions.
#'
#' @returns
#'   * `gargle_last_response()` returns the most recent [httr::response()]
#'     object.
#'   * `gargle_last_content()` returns the parsed JSON content from the most
#'     recent response or an empty list if unavailable.
#'
#' @keywords internal
#' @export
gargle_last_response <- function() {
  gargle_env$last_response
}

#' @export
#' @rdname gargle_last_response
gargle_last_content <- function() {
  resp <- gargle_last_response()
  if (inherits(resp, "response")) {
    tryCatch(
      response_as_json(resp),
      gargle_error_request_failed = function(e) e$message
    )
  } else {
    list()
  }
}

test_that("token works", {
  skip_if_offline()

  expect_no_error(
    token <- credentials_service_account(
      scopes = "https://www.googleapis.com/auth/userinfo.email",
      path = secret_decrypt_json(
        fs::path_package("gargle", "secret", "gargle-testing.json"),
        key = "GARGLE_KEY"
      )
    )
  )
  email <- token_email(token)
  expect_match(email, "^gargle-testing@.*[.]iam[.]gserviceaccount[.]com")
})

test_that("gargle_last_response() captures error response", {
  rds_file <- test_path("fixtures", "fitness-get-wrong-scope_403.rds")
  resp <- readRDS(rds_file)

  expect_error(
    response_process(resp),
    class = "gargle_error_request_failed"
  )

  last_resp <- gargle_last_response()
  expect_s3_class(last_resp, "response")
  expect_equal(httr::status_code(last_resp), 403)
  expect_equal(
    last_resp$url,
    "https://www.googleapis.com/fitness/v1/users/12345/dataSources"
  )

  last_content <- gargle_last_content()
  expect_equal(
    last_content$error$message,
    "Request had insufficient authentication scopes."
  )
})

test_that("remember = FALSE prevents capturing response", {
  rds_file <- test_path("fixtures", "fitness-get-wrong-scope_403.rds")
  resp <- readRDS(rds_file)

  gargle_env$last_response <- list(marker = "old_value")

  expect_error(
    response_process(resp, remember = FALSE),
    class = "gargle_error_request_failed"
  )

  expect_equal(gargle_env$last_response$marker, "old_value")
})

#' Generate a field mask
#'
#' Many Google API requests take a field mask, via a `fields` parameter, in the
#' URL and/or in the body. `field_mask()` generates such a field mask from an R
#' list, typically a list that is destined to be part of the body of a request
#' that writes or updates a resource. `field_mask()` is designed to help in the
#' common case where the attributes you wish to modify are exactly the ones
#' represented in the object. It is possible to use a "larger" field mask, that
#' is either less specific or that explicitly includes other attributes, in
#' which case the attributes covered by the mask but absent from the object are
#' reset to default values. This is not exactly the use case `field_mask()` is
#' designed for, but its output could still be useful as a first step in
#' constructing such a mask.
#'
#' @param x A named R list, where the requirement for names applies at all
#'   levels, i.e. recursively.
#'
#' @return A Google API field mask, as a string.
#' @export
#'
#' @seealso The documentation for the [JSON encoding of a Protocol Buffers
#'   FieldMask](https://protobuf.dev/reference/protobuf/google.protobuf/#json-encoding-of-field-masks).
#'
#' @examples
#' x <- list(sheetId = 1234, title = "my_favorite_worksheet")
#' field_mask(x)
#'
#' x <- list(
#'   userEnteredFormat = list(
#'     backgroundColor = list(
#'       red = 159 / 255, green = 183 / 255, blue = 196 / 255
#'     )
#'   )
#' )
#' field_mask(x)
#'
#' x <- list(
#'   sheetId = 1234,
#'   gridProperties = list(rowCount = 5, columnCount = 3)
#' )
#' field_mask(x)
field_mask <- function(x) {
  stopifnot(is_dictionaryish(x))
  explicit_mask <- imap(x, field_mask_impl_)
  as.character(glue_collapse(unname(unlist(explicit_mask)), ","))
}

field_mask_impl_ <- function(x, y = "") {
  if (!is_list(x)) {
    return(y)
  }
  stopifnot(is_dictionaryish(x))

  leafs <- !map_lgl(x, is_list)
  if (sum(leafs) <= 1) {
    leafs <- FALSE
  }
  names(x)[!leafs] <- glue(".{names(x)[!leafs]}")
  if (sum(leafs) > 1) {
    nm <- glue("({glue_collapse(names(x)[leafs], sep = ',')})")
    x <- list2(!!nm := NA, !!!x[!leafs])
  }
  map2(x, glue("{y}{names(x)}"), field_mask_impl_)
}

## this file has its origins in oauth-cache.R in httr
## nothing there is exported, so copied over, then it evolved

# cache setup, loading, validation --------------------------------------------

gargle_default_oauth_cache_path <- function() {
  path_tidy(rappdirs::user_cache_dir("gargle"))
}

## this is the cache setup interface for the Gargle2.0 class
## returns NULL or cache path
cache_establish <- function(cache = NULL) {
  cache <- cache %||% gargle_oauth_cache()
  if (length(cache) != 1) {
    gargle_abort("{.arg cache} must have length 1, not {length(cache)}.")
  }
  # the inherits() call is so we accept 'fs_path'
  if (
    !is.logical(cache) && !is.character(cache) && !inherits(cache, "character")
  ) {
    stop_input_type(cache, what = c("logical", "character"))
  }

  # If NA, consider default cache folder
  # Request user's permission to create it, if doesn't exist yet
  # Store outcome of this mission (TRUE or FALSE) in the option for the session
  if (is_na(cache)) {
    cache <- cache_available(gargle_default_oauth_cache_path())
    options("gargle_oauth_cache" = cache)
  }
  ## cache is now TRUE, FALSE or path

  if (isFALSE(cache)) {
    return()
  }

  if (isTRUE(cache)) {
    cache <- gargle_default_oauth_cache_path()
  }
  ## cache is now a path

  if (dir_exists(cache)) {
    cache_clean(cache)
  } else {
    cache_create(cache)
  }

  return(cache)
}

cache_available <- function(path) {
  dir_exists(path) || cache_allowed(path)
}

cache_allowed <- function(path) {
  if (!is_interactive()) {
    return(FALSE)
  }

  choice <- cli_menu(
    header = character(),
    prompt = "Is it OK to cache OAuth access credentials in the folder \\
              {.path {path}} between R sessions?",
    choices = c("Yes", "No")
  )
  choice == 1
}

cache_create <- function(path) {
  ## owner (and only owner) can read, write, execute
  dir_create(path, recurse = TRUE, mode = "0700")

  cache_parent <- path_dir(path)
  desc <- path(cache_parent, "DESCRIPTION")
  if (file_exists(desc)) {
    add_line(
      path(cache_parent, ".Rbuildignore"),
      paste0("^", gsub("\\.", "\\\\.", path), "$")
    )
  }
  git <- path(cache_parent, c(".gitignore", ".git"))
  if (any(file_exists(git))) {
    add_line(
      path(cache_parent, ".gitignore"),
      path
    )
  }

  TRUE
}

cache_ls <- function(path) {
  names(cache_load(path))
}

cache_load <- function(path) {
  files <- as.character(dir_ls(path))
  files <- keep_hash_paths(files)
  names(files) <- path_file(files)
  tokens <- map(files, readRDS)

  hashes <- map_chr(tokens, function(t) t$hash())

  mismatch <- names(hashes) != hashes
  if (any(mismatch)) {
    # original motivation:
    # we've seen this with tokens cached on R 3.5 but reloaded on 3.6
    # because $hash() calls serialize() and default version changed
    #
    # later observation: I suppose this could also get triggered if someone
    # caches on, e.g., Windows, then moves/deploys the project to *nix
    n <- sum(mismatch)
    mismatch_name <- names(hashes)[mismatch]
    mismatch_hash <- hashes[mismatch]
    mismatch_name_fmt <- gargle_map_cli(
      mismatch_name,
      template = "{.val <<x>>} (name)"
    )
    mismatch_hash_fmt <- gargle_map_cli(
      mismatch_hash,
      template = "{.field <<x>>} (hash)"
    )
    msg <- c(
      "!" = "Cache contains {cli::qty(n)}{?a /}token{?s} with {?a /}name{?s} \\
             that do{?es/} not match {?its/their} hash:",
      bulletize(
        as.vector(rbind(mismatch_name_fmt, mismatch_hash_fmt)),
        n_show = 100
      ),
      " " = "Will attempt to repair by renaming"
    )
    gargle_debug(msg)
    file_move(files[mismatch], path(path, hashes[mismatch]))
    Recall(path)
  } else {
    tokens
  }
}

cache_clean <- function(cache, pattern = gargle_legacy_app_pattern()) {
  dat_tokens <- gargle_oauth_dat(cache)
  dat_tokens$legacy <- grepl(pattern, dat_tokens$client)
  n <- sum(dat_tokens$legacy)
  if (n == 0) {
    return(FALSE)
  }

  gargle_info(c(
    "v" = "Deleting {n} token{?s} obtained with an old tidyverse OAuth client.",
    "i" = "Expect interactive prompts to re-auth with the new client.",
    "!" = "Is this rolling of credentials highly disruptive to your \\
           workflow?",
    " " = "That means you should rely on your own OAuth client \\
           (or switch to a service account token).",
    " " = "Learn more these in these articles:",
    " " = "{.url https://gargle.r-lib.org/articles/get-api-credentials.html}",
    " " = "{.url https://gargle.r-lib.org/articles/non-interactive-auth.html}"
  ))
  file_delete(dat_tokens$filepath[dat_tokens$legacy])
  TRUE
}

gargle_legacy_app_pattern <- function() "-calliope$"

# retrieve and insert tokens from cache -----------------------------------

## these two functions provide the "current token <--> token cache" interface
## for the Gargle2.0 class
token_from_cache <- function(candidate) {
  cache_path <- candidate$cache_path

  if (is.null(cache_path)) {
    return()
  }

  existing <- cache_ls(cache_path)
  this_one <- token_match(
    candidate$hash(),
    existing,
    package = candidate$package
  )

  gargle_debug("token(s) found in cache:")
  gargle_debug(existing)
  gargle_debug("token we are looking for:")
  gargle_debug(this_one)

  if (is.null(this_one)) {
    NULL
  } else {
    readRDS(path(cache_path, this_one))
  }
}

token_into_cache <- function(candidate) {
  cache_path <- candidate$cache_path
  if (is.null(cache_path)) {
    gargle_debug("not caching token")
    return()
  }
  gargle_debug(c("putting token into the cache:", "{.file {cache_path}}"))
  saveRDS(candidate, path(cache_path, candidate$hash()))
}

token_remove_from_cache <- function(candidate) {
  cache_path <- candidate$cache_path
  if (is.null(cache_path)) {
    return()
  }
  token_path <- path(cache_path, candidate$hash())
  # when does token_path not exist?
  # the first time a token fails to refresh it is removed on disk,
  # but a package may still have it stored in its auth state
  if (file_exists(token_path)) {
    gargle_debug(c("Removing token from the cache:", "{.file {token_path}}"))
    file_delete(token_path)
  }
}

# helpers to compare tokens based on SHORTHASH_EMAIL ------------------------
token_match <- function(candidate, existing, package = "gargle") {
  if (length(existing) == 0) {
    return()
  }

  m <- match2(candidate, existing)
  if (!is_na(m)) {
    stopifnot(length(m) == 1)
    return(existing[[m]])
  }
  # there is no full match

  candidate_email <- extract_email(candidate)
  # possible values    what they mean
  # ------------------ ---------------------------------------------------------
  # 'blah@example.org' user specified an email
  # '*@example.org'    user specified only the domain
  #                    (we still scold for multiple matches)
  # '*'                `email = TRUE`, i.e. permission to use *one* that we find
  #                    (we still scold for multiple matches)
  # ''                 user gave no email and no instructions

  # if email was specified, we're done
  if (!empty_string(candidate_email) && !startsWith(candidate_email, "*")) {
    return()
  }
  # candidate_email is '*' or '' or domain-only, e.g. '*@example.org'

  # match on the short hash
  m <- match2(mask_email(candidate), mask_email(existing))

  # if no match on short hash, we're done
  if (is_na(m)) {
    return()
  }
  existing <- existing[m]
  # existing holds at least one short hash match

  # filter on domain, if provided
  if (!empty_string(candidate_email) && startsWith(candidate_email, "*@")) {
    domain_part <- function(x) sub(".+@(.+)$", "\\1", x)
    m <- match2(domain_part(candidate_email), domain_part(existing))
    if (is_na(m)) {
      return()
    }
    existing <- existing[m]
  }

  # if candidate_email is '*' or domain-only, e.g. '*@example.org' AND
  # we're down to 1 match, we're done
  if (!empty_string(candidate_email) && length(existing) == 1) {
    gargle_info(c(
      "i" = "The {.pkg {package}} package is using a cached token for \\
             {.email {extract_email(existing)}}."
    ))
    return(existing)
  }
  # if we're still here, one of these is true:
  # - email was partially specified and there are multiple matches
  # - email was unspecified and there is at least one match

  if (!is_interactive()) {
    # proceed, but make sure user sees messaging about how to do
    # non-interactive auth more properly
    # https://github.com/r-lib/gargle/issues/92
    local_gargle_verbosity("info")
    if (length(existing) > 1) {
      emails <- extract_email(existing)
      emails_fmt <- lapply(
        emails,
        function(x) cli::format_inline("{.email {x}}")
      )
      msg <- c(
        "i" = "Suitable tokens found in the cache, associated with these \\
               emails:",
        set_names(emails_fmt, ~ rep_along(., "*")),
        " " = "Defaulting to the first email."
      )
      gargle_info(msg)
      existing <- existing[[1]]
    }
    msg <- c(
      "!" = "Using an auto-discovered, cached token.",
      " " = "To suppress this message, modify your code or options \\
             to clearly consent to the use of a cached token.",
      " " = "See gargle's \"Non-interactive auth\" vignette for more details:",
      " " = "{.url https://gargle.r-lib.org/articles/non-interactive-auth.html}"
    )
    # morally, I'd like to throw a warning but current design of token_fetch()
    # means warnings are caught
    gargle_info(msg)
    gargle_info(c(
      "i" = "The {.pkg {package}} package is using a cached token for \\
             {.email {extract_email(existing)}}."
    ))
    return(existing)
  }

  # we need user to OK our discovery or pick from multiple emails
  choices <- c(
    "Send me to the browser for a new auth process.",
    extract_email(existing)
  )
  choice <- cli_menu(
    "The {.pkg {package}} package is requesting access to your Google account.",
    "Enter '1' to start a new auth process or select a pre-authorized account.",
    choices = choices
  )

  if (choice == 1) {
    NULL
  } else {
    existing[[choice - 1]]
  }
}

## for this token hash:
## 2a46e6750476326f7085ebdab4ad103d_jenny@example.org
## ^  mask_email() returns this   ^ ^ extract_email() returns this ^
hash_regex <- "^([0-9a-f]+)_(.*?)$"
mask_email <- function(x) sub(hash_regex, "\\1", x)
extract_email <- function(x) sub(hash_regex, "\\2", x)
keep_hash_paths <- function(x) x[grep(hash_regex, path_file(x))]

## match() but return location of all matches
match2 <- function(needle, haystack) {
  matches <- which(haystack == needle)
  if (length(matches) == 0) {
    matches <- NA
  }
  matches
}


# gargle situation report -------------------------------------------------

#' OAuth token situation report
#'
#' Get a human-oriented overview of the existing gargle OAuth tokens:
#'   * Filepath of the current cache
#'   * Number of tokens found there
#'   * Compact summary of the associated
#'     - Email = Google identity
#'     - OAuth client (actually, just its nickname)
#'     - Scopes
#'     - Hash (actually, just the first 7 characters)
#' Mostly useful for the development of gargle and client packages.
#'
#' @inheritParams gargle2.0_token
#'
#' @return A data frame with one row per cached token, invisibly. Note this data
#'   frame may contain more columns than it seems, e.g. the `filepath` column
#'   isn't printed by default.
#' @export
#'
#' @examples
#' gargle_oauth_sitrep()
gargle_oauth_sitrep <- function(cache = NULL) {
  cache <- cache %||% cache_locate()
  if (!is_dir(cache)) {
    gargle_info("No gargle OAuth cache found at {.path {cache}}.")
    return(invisible())
  }

  dat <- gargle_oauth_dat(cache)
  gargle_info(c(
    "{nrow(dat)} token{?s} found in this gargle OAuth cache:",
    "{.path {cache}}",
    ""
  ))

  if (gargle_verbosity() %in% c("debug", "info")) {
    cli::cli_verbatim(format(dat))
  }
  invisible(dat)
}

gargle_oauth_dat <- function(cache = NULL) {
  cache <- cache %||% gargle_default_oauth_cache_path()
  if (is_dir(cache)) {
    tokens <- cache_load(cache)
  } else {
    tokens <- list()
  }

  nms <- names(tokens)
  hash <- mask_email(nms)
  email <- extract_email(nms)
  client <- map_chr(tokens, function(t) t$client$name %||% t$app$appname)
  scopes <- map(tokens, function(t) t$params$scope)
  email_scope <- "https://www.googleapis.com/auth/userinfo.email"
  scopes <- map(scopes, function(s) s[s != email_scope])
  scopes <- map_chr(scopes, function(s) commapse(base_scope(s)))

  structure(
    data.frame(
      email,
      client,
      scopes,
      hash,
      filepath = path(cache, nms),
      stringsAsFactors = FALSE,
      row.names = NULL
    ),
    class = c("gargle_oauth_dat", "data.frame")
  )
}

#' @export
format.gargle_oauth_dat <- function(x, ...) {
  format_transformer <- function(text, envir) {
    res <- eval(parse(text = text, keep.source = FALSE), envir)
    res <- format(c(text, res))
    c(
      res[1],
      strrep("_", nchar(res[1])),
      res[-1]
    )
  }

  # obfuscate the hash for brevity
  hash_column <- which(names(x) == "hash")
  x[[hash_column]] <- obfuscate(x[[hash_column]], first = 7, last = 0)
  names(x)[hash_column] <- "hash..."

  # NOTE: the filepath variable is absent from the formatted data frame

  glue_data(
    x,
    "{email} {client} {scopes} {hash...}",
    .transformer = format_transformer
  )
}

#' @export
print.gargle_oauth_dat <- function(x, ...) {
  cli::cat_line(format(x))
  invisible(x)
}

# main point of this is **passive** cache discovery
cache_locate <- function() {
  option_cache <- gargle_oauth_cache()
  if (is_scalar_character(option_cache)) {
    gargle_info(c(
      "i" = 'Taking cache location from the {.code "gargle_oauth_cache"} option.'
    ))
    return(option_cache)
  }

  default_cache <- gargle_default_oauth_cache_path()
  gargle_info(c(
    "i" = 'Reporting the default cache location.'
  ))
  default_cache
}

# cache_establish ------------------------------------------------------------

test_that("cache_establish() insists on sensible input", {
  expect_snapshot(cache_establish(letters[1:2]), error = TRUE)
  expect_snapshot(cache_establish(1), error = TRUE)
  expect_snapshot(cache_establish(list(1)), error = TRUE)
})

test_that("`cache = TRUE` uses default cache path", {
  local_mocked_bindings(
    ## we don't want to actually initialize a cache
    cache_create = function(path) NULL
  )
  expect_equal(cache_establish(TRUE), gargle_default_oauth_cache_path())
})

test_that("`cache = FALSE` does nothing", {
  expect_null(cache_establish(FALSE))
})

test_that("`cache = NA` is like `cache = FALSE` if cache not available", {
  local_mocked_bindings(
    # we want no existing cache to be found, be it current or legacy
    gargle_default_oauth_cache_path = function() file_temp(),
    cache_allowed = function(path) FALSE
  )
  expect_equal(cache_establish(NA), cache_establish(FALSE))
})

test_that("`cache = <filepath>` creates cache folder, recursively", {
  tmpfolder <- path_temp("foo", "bar")
  withr::defer(dir_delete(tmpfolder))

  cache_establish(tmpfolder)
  expect_true(dir_exists(tmpfolder))
})

test_that("`cache = <filepath>` adds new cache folder to relevant 'ignores'", {
  tmpproj <- file_temp()
  withr::defer(dir_delete(tmpproj))
  local_gargle_verbosity("silent")

  dir_create(tmpproj)
  writeLines("", path(tmpproj, "DESCRIPTION"))
  writeLines("", path(tmpproj, ".gitignore"))
  cache_establish(path(tmpproj, "oauth-cache"))
  expect_match(readLines(path(tmpproj, ".gitignore")), "oauth-cache$")
  expect_match(
    readLines(path(tmpproj, ".Rbuildignore")),
    "oauth-cache$",
    fixed = TRUE
  )
})

test_that("default is to consult and set the oauth cache option", {
  withr::local_options(list(gargle_oauth_cache = NA))
  local_mocked_bindings(
    # we want no existing cache to be found, be it current or legacy
    gargle_default_oauth_cache_path = function() file_temp(),
    cache_allowed = function(path) FALSE
  )

  expect_equal(getOption("gargle_oauth_cache"), NA)
  cache_establish()
  expect_false(getOption("gargle_oauth_cache"))
})

# cache_allowed() --------------------------------------------------------------

test_that("cache_allowed() returns false when non-interactive (or testing)", {
  local_interactive(FALSE)
  expect_false(cache_allowed(getwd()))
})

# cache_clean() --------------------------------------------------------------

test_that("cache_clean() works", {
  cache_folder <- path_temp("cache_clean-test")
  withr::defer(dir_delete(cache_folder))

  fauxen_a <- gargle2.0_token(
    email = "a@example.org",
    client = gargle_oauth_client(
      id = "CLIENT_ID",
      secret = "SECRET",
      name = "apple"
    ),
    credentials = list(a = 1),
    cache = cache_folder
  )
  fauxen_b <- gargle2.0_token(
    email = "b@example.org",
    client = gargle_oauth_client(
      id = "CLIENT_ID",
      secret = "SECRET",
      name = "banana"
    ),
    credentials = list(b = 1),
    cache = cache_folder
  )
  dat <- gargle_oauth_dat(cache_folder)
  expect_equal(nrow(dat), 2)

  expect_snapshot(
    cache_clean(cache_folder, "apple")
  )
  dat <- gargle_oauth_dat(cache_folder)
  expect_equal(dat$client, "banana")

  local_gargle_verbosity("silent")
  cache_clean(cache_folder, "banana")
  dat <- gargle_oauth_dat(cache_folder)
  expect_equal(nrow(dat), 0)
})

# validate_token_list() ------------------------------------------------------
test_that("cache_load() repairs tokens stored with names != their hash", {
  cache_folder <- file_temp()
  withr::defer(dir_delete(cache_folder))

  fauxen_a <- gargle2.0_token(
    email = "a@example.org",
    credentials = list(a = 1),
    cache = cache_folder
  )
  fauxen_b <- gargle2.0_token(
    email = "b@example.org",
    credentials = list(b = 1),
    cache = cache_folder
  )
  file_move(
    dir_ls(cache_folder),
    path(cache_folder, c("abc123_c@example.org", "def456_d@example.org"))
  )
  local_gargle_verbosity("debug")

  # bit of fiddliness to deal with hashes that can vary by OS
  out <- capture.output(
    tokens <- cache_load(cache_folder),
    type = "message"
  )
  out <- sub("[[:xdigit:]]+(?=.+\\(hash\\)$)", "{TOKEN_HASH}", out, perl = TRUE)
  expect_snapshot(
    writeLines(out)
  )
  expect_gargle2.0_token(tokens[[1]], fauxen_a)
  expect_gargle2.0_token(tokens[[2]], fauxen_b)
})

# token into and out of cache ---------------------------------------------
test_that("token_from_cache() returns NULL when caching turned off", {
  fauxen <- gargle2.0_token(
    email = "a@example.org",
    credentials = list(a = 1),
    cache = FALSE
  )
  expect_null(token_from_cache(fauxen))
})

test_that("token_into_cache(), token_from_cache() roundtrip", {
  cache_folder <- file_temp()
  withr::defer(dir_delete(cache_folder))

  ## this calls token_into_cache()
  token_in <- gargle2.0_token(
    email = "a@example.org",
    credentials = list(a = 1),
    cache = cache_folder
  )

  ## this calls token_from_cache()
  token_out <- gargle2.0_token(
    email = "a@example.org",
    cache = cache_folder
  )

  expect_gargle2.0_token(token_in, token_out)
  expect_equal(token_out$credentials, list(a = 1))
})

# token_match() ----------------------------------------
test_that("token_match() returns NULL if nothing to match against", {
  expect_null(token_match("whatever", character()))
})

test_that("token_match() returns the full match", {
  one_existing <- "abc_a@example.com"
  two_existing <- c(one_existing, "def_b@example.com")

  expect_equal(
    one_existing,
    token_match(one_existing, one_existing)
  )
  expect_equal(
    one_existing,
    token_match(one_existing, two_existing)
  )
})

test_that("token_match() returns NULL if email given, but no full match", {
  candidate <- "abc_a@example.org"
  expect_null(token_match(candidate, "def_a@example.org"))
  expect_null(token_match(candidate, "abc_b@example.org"))
  expect_null(token_match(candidate, "a@example.org"))
  expect_null(token_match(candidate, "abc"))
  expect_null(token_match(candidate, "abc_"))
})

test_that("token_match() returns NULL if no email and no short hash match", {
  expect_null(token_match("abc_", "def_a@example.org"))
  expect_null(token_match("abc_*", "def_a@example.org"))
})

test_that("token_match() finds a match based on domain", {
  one_match_of_two <- c("abc_jane@example.org", "abc_jane@gmail.com")
  expect_snapshot(
    m <- token_match("abc_*@example.org", one_match_of_two)
  )
  expect_equal(m, one_match_of_two[[1]])
})

test_that("token_match() scolds but returns short hash match when non-interactive", {
  local_interactive(FALSE)

  one_existing <- "abc_a@example.com"
  two_existing <- c(one_existing, "abc_b@example.com")

  expect_snapshot(
    m <- token_match("abc_", one_existing)
  )
  expect_equal(m, one_existing)

  expect_snapshot(
    m <- token_match("abc_*", one_existing)
  )
  expect_equal(m, one_existing)

  expect_snapshot(
    m <- token_match("abc_", two_existing)
  )
  expect_equal(m, one_existing)

  expect_snapshot(
    m <- token_match("abc_*", two_existing)
  )
  expect_equal(m, one_existing)
})

# situation report ----------------------------------------------------------

test_that("gargle_oauth_dat() reports on specified cache", {
  tmp_cache <- file_temp()
  withr::defer(dir_delete(tmp_cache))

  gargle2.0_token(
    email = "a@example.org",
    credentials = list(a = 1),
    cache = tmp_cache
  )
  gargle2.0_token(
    email = "b@example.org",
    credentials = list(b = 2),
    cache = tmp_cache
  )

  dat <- gargle_oauth_dat(tmp_cache)
  expect_s3_class(dat, "gargle_oauth_dat")
  expect_equal(nrow(dat), 2)
  expect_equal(dat$email, c("a@example.org", "b@example.org"))
})

test_that("gargle_oauth_dat() is OK with nonexistent or empty cache", {
  tmp_cache <- file_temp()
  withr::defer(dir_delete(tmp_cache))

  columns <- c("email", "client", "scopes", "hash", "filepath")

  dat <- gargle_oauth_dat(tmp_cache)
  expect_s3_class(dat, "gargle_oauth_dat")
  expect_equal(nrow(dat), 0)
  expect_setequal(names(dat), columns)

  gargle2.0_token(
    email = "a@example.org",
    credentials = list(a = 1),
    cache = tmp_cache
  )
  file_delete(dir_ls(tmp_cache))

  dat <- gargle_oauth_dat(tmp_cache)
  expect_s3_class(dat, "gargle_oauth_dat")
  expect_equal(nrow(dat), 0)
  expect_setequal(names(dat), columns)
})

test_that("gargle_oauth_sitrep() works with a cache", {
  tmp_cache <- file_temp()
  withr::defer(dir_delete(tmp_cache))

  gargle2.0_token(
    email = "a@example.org",
    credentials = list(a = 1),
    cache = tmp_cache
  )
  gargle2.0_token(
    email = "b@example.org",
    credentials = list(b = 2),
    cache = tmp_cache
  )

  # bit of fiddliness to remove the volatile path and the hashes that can vary
  # by OS
  out <- capture.output(
    gargle_oauth_sitrep(tmp_cache),
    type = "message"
  )
  out <- sub(tmp_cache, "{path to gargle oauth cache}", out, fixed = TRUE)
  out <- sub("[[:xdigit:]]{7}[.]{3}", "{hash...}", out)
  expect_snapshot(
    writeLines(out)
  )
})

test_that("gargle_oauth_sitrep() consults the option for cache location", {
  tmp_cache <- file_temp()
  withr::defer(dir_delete(tmp_cache))

  gargle2.0_token(
    email = "a@example.org",
    credentials = list(a = 1),
    cache = tmp_cache
  )

  withr::local_options(list(gargle_oauth_cache = tmp_cache))
  local_gargle_verbosity("debug")
  out <- capture.output(
    gargle_oauth_sitrep(),
    type = "message"
  )
  # bit of fiddliness to remove the volatile path and the hashes that can vary
  # by OS
  out <- sub(tmp_cache, "{path to gargle oauth cache}", out, fixed = TRUE)
  out <- sub("[[:xdigit:]]{7}[.]{3}", "{hash...}", out)
  expect_snapshot(
    writeLines(out)
  )
})

# helpers -----------------------------------------------------------

test_that("match2() works", {
  expect_equal(match2("a", c("a", "b", "a")), c(1L, 3L))
  expect_equal(match2("b", c("a", "b", "a")), 2L)
  expect_true(is.na(match2("c", c("a", "b", "a"))))
})

test_that("mask_email() works", {
  hash <- "2a46e6750476326f7085ebdab4ad103d"
  expect_equal(
    mask_email(c(
      "2a46e6750476326f7085ebdab4ad103d_jenny@example.com",
      "2a46e6750476326f7085ebdab4ad103d_NA",
      "2a46e6750476326f7085ebdab4ad103d_",
      "2a46e6750476326f7085ebdab4ad103d_FIRST_LAST@example.com"
    )),
    rep_len(hash, 4)
  )
})

test_that("extract_email() works", {
  expect_equal(extract_email("abc123_a"), "a")
  expect_equal(extract_email("abc123_b@example.com"), "b@example.com")
  expect_equal(extract_email("abc123_"), "")
  expect_equal(extract_email("abc123_FIRST_LAST@a.com"), "FIRST_LAST@a.com")
})

test_that("keep_hash_paths() works", {
  x <- c("aa_bb_cc", "a.md", "b.rds", "c.txt", "dd123_e@example.org")
  expect_equal(keep_hash_paths(x), x[c(1, 5)])
})

from_permitted_package <- function(env = caller_env()) {
  env <- topenv(env, globalenv())
  if (!isNamespace(env)) {
    return(FALSE)
  }

  nm <- getNamespaceName(env)
  gargle_debug("attempt to access internal gargle data from: {.pkg {nm}}")
  nm %in% c("gargle", "googledrive", "bigrquery", "googlesheets4", "gmailr")
}

check_permitted_package <- function(env = caller_env(), call = caller_env()) {
  if (!from_permitted_package(env)) {
    msg <- c(
      "Attempt to directly access a credential that can only be used within \\
       tidyverse packages.",
      "This error may mean that you need to:",
      "*" = "Create a new project on Google Cloud Platform.",
      "*" = "Enable relevant APIs for your project.",
      "*" = "Create an API key and/or an OAuth client ID.",
      "*" = "Configure your requests to use your API key and OAuth client ID.",
      "i" = "See gargle's \"How to get your own API credentials\" vignette for more details:",
      "i" = "{.url https://gargle.r-lib.org/articles/get-api-credentials.html}"
    )
    gargle_abort(msg, call = call)
  }
  invisible(env)
}

#' Build a Google API request
#'
#' Intended primarily for internal use in client packages that provide
#' high-level wrappers for users. The `vignette("request-helper-functions")`
#' describes how one might use these functions inside a wrapper package.
#'
#' @param endpoint List of information about the target endpoint or, in
#'   Google's vocabulary, the target "method". Presumably prepared from the
#'   [Discovery
#'   Document](https://developers.google.com/discovery/v1/getting_started#background-resources)
#'   for the target API.
#' @param params Named list. Values destined for URL substitution, the query,
#'   or, for `request_develop()` only, the body. For `request_build()`, body
#'   parameters must be passed via the `body` argument.
#' @param base_url Character.
#' @param method Character. An HTTP verb, such as `GET` or `POST`.
#' @param path Character. Path to the resource, not including the API's
#'   `base_url`. Examples: `drive/v3/about` or `drive/v3/files/{fileId}`. The
#'   `path` can be a template, i.e. it can include variables inside curly
#'   brackets, such as `{fileId}` in the example. Such variables are substituted
#'   by `request_build()`, using named parameters found in `params`.
#' @param body List. Values to send in the API request body.
#' @param key API key. Needed for requests that don't contain a token. For more,
#'   see Google's document Credentials, access, security, and identity
#'   (`https://support.google.com/googleapi/answer/6158857?hl=en&ref_topic=7013279`).
#'    A key can be passed as a named component of `params`, but note that the
#'   formal argument `key` will clobber it, if non-`NULL`.
#' @param token Token, ready for inclusion in a request, i.e. prepared with
#'   [httr::config()].
#'
#' @section `request_develop()`:
#'
#' Combines user input (`params`) with information about an API endpoint.
#' `endpoint` should contain these components:
#'   - `path`: See documentation for argument.
#'   - `method`: See documentation for argument.
#'   - `parameters`: Compared with `params` supplied by user. An error is
#'      thrown if user-supplied `params` aren't named in
#'      `endpoint$parameters` or if user fails to supply all required
#'      parameters. In the return value, body parameters are separated from
#'      those destined for path substitution or the query.
#'
#' The return value is typically used as input to `request_build()`.
#'
#' @section `request_build()`:
#'
#' Builds a request, in a purely mechanical sense. This function does nothing
#' specific to any particular Google API or endpoint.
#'   - Use with the output of `request_develop()` or with hand-crafted input.
#'   - `params` are used for variable substitution in `path`. Leftover `params`
#'     that are not bound by the `path` template automatically become HTTP
#'     query parameters.
#'   - Adds an API key to the query iff `token = NULL` and removes the API key
#'   otherwise. Client packages should generally pass their own API key in, but
#'   note that [gargle_api_key()] is available for small-scale experimentation.
#'
#' See `googledrive::generate_request()` for an example of usage in a client
#' package. googledrive has an internal list of selected endpoints, derived from
#' the Drive API Discovery Document
#' (`https://www.googleapis.com/discovery/v1/apis/drive/v3/rest`),
#' exposed via `googledrive::drive_endpoints()`. An element from such a list is
#' the expected input for `endpoint`. `googledrive::generate_request()` is a
#' wrapper around `request_develop()` and `request_build()` that inserts a
#' googledrive-managed API key and some logic about Team Drives. All user-facing
#' functions use `googledrive::generate_request()` under the hood.
#'
#' @return
#' `request_develop()`: `list()` with components `method`, `path`, `params`,
#' `body`, and `base_url`.
#'
#' `request_build()`: `list()` with components `method`, `path`
#' (post-substitution), `query` (the input `params` not used in URL
#' substitution), `body`, `token`, `url` (the full URL, post-substitution,
#' including the query).
#'
#' @export
#' @family requests and responses
#' @examples
#' \dontrun{
#' ## Example with a prepared endpoint
#' ept <- googledrive::drive_endpoints("drive.files.update")[[1]]
#' req <- request_develop(
#'   ept,
#'   params = list(
#'     fileId = "abc",
#'     addParents = "123",
#'     description = "Exciting File"
#'   )
#' )
#' req
#'
#' req <- request_build(
#'   method = req$method,
#'   path = req$path,
#'   params = req$params,
#'   body = req$body,
#'   token = "PRETEND_I_AM_A_TOKEN"
#' )
#' req
#'
#' ## Example with no previous knowledge of the endpoint
#' ## List a file's comments
#' ## https://developers.google.com/drive/v3/reference/comments/list
#' req <- request_build(
#'   method = "GET",
#'   path = "drive/v3/files/{fileId}/comments",
#'   params = list(
#'     fileId = "your-file-id-goes-here",
#'     fields = "*"
#'   ),
#'   token = "PRETEND_I_AM_A_TOKEN"
#' )
#' req
#'
#' # Example with no previous knowledge of the endpoint and no token
#' # use an API key for which the Places API is enabled!
#' API_KEY <- "1234567890"
#'
#' # get restaurants close to a location in Vancouver, BC
#' req <- request_build(
#'   method = "GET",
#'   path = "maps/api/place/nearbysearch/json",
#'   params = list(
#'     location = "49.268682,-123.167117",
#'     radius = 100,
#'     type = "restaurant"
#'   ),
#'   key = API_KEY,
#'   base_url = "https://maps.googleapis.com"
#' )
#' resp <- request_make(req)
#' out <- response_process(resp)
#' vapply(out$results, function(x) x$name, character(1))
#' }
request_develop <- function(
  endpoint,
  params = list(),
  base_url = "https://www.googleapis.com"
) {
  check_params(params, endpoint$parameters, endpoint$id)
  in_body <- vapply(
    endpoint$parameters,
    function(x) x$location == "body",
    logical(1)
  )
  body_param_names <- names(endpoint$parameters[in_body])
  other_param_names <- names(endpoint$parameters[!in_body])
  list(
    method = endpoint$httpMethod,
    path = endpoint$path,
    params = params[intersect(other_param_names, names(params))],
    body = params[intersect(body_param_names, names(params))],
    base_url = base_url
  )
}

#' @rdname request_develop
#' @export
request_build <- function(
  method = "GET",
  path = "",
  params = list(),
  body = list(),
  token = NULL,
  key = NULL,
  base_url = "https://www.googleapis.com"
) {
  path_param_names <- extract_path_names(path)
  path_params <- params[path_param_names]
  query_param_names <- setdiff(names(params), path_param_names)
  query_params <- params[query_param_names]

  ## send a token or a key, but never both
  query_params$key <- if (is.null(token)) {
    key %||% query_params$key
  } else {
    NULL
  }

  out <- list(
    method = method,
    url = httr::modify_url(
      url = base_url,
      path = glue_data(path_params, path),
      query = query_params
    ),
    body = body,
    token = token
  )
  out
}

## check params provided by user against spec
##   * error if required params are missing
##   * error for unknown params
check_params <- function(provided, spec, id, call = caller_env()) {
  required <- Filter(function(x) isTRUE(x$required), spec)
  missing <- setdiff(names(required), names(provided))
  if (length(missing)) {
    gargle_abort_bad_params(missing, reason = "missing", id, call = call)
  }

  unknown <- setdiff(names(provided), names(spec))
  if (length(unknown)) {
    gargle_abort_bad_params(unknown, reason = "unknown", id, call = call)
  }

  invisible(provided)
}

##  input: /v4/spreadsheets/{spreadsheetId}/sheets/{sheetId}:copyTo
## output: spreadsheetId, sheetId
extract_path_names <- function(path) {
  m <- gregexpr("\\{[^/]*\\}", path)
  path_param_names <- regmatches(path, m)[[1]]
  gsub("[\\{\\}]", "", path_param_names)
}

#' Create an OAuth client for Google
#'
#' @description
#' A `gargle_oauth_client` consists of:

#' * A type. gargle only supports the "Desktop app" and "Web application" client
#' types. Different types are associated with different OAuth flows.

#' * A client ID and secret.
#' * Optionally, one or more redirect URIs.

#' * A name. This is really a human-facing label. Or, rather, it can be used
#' that way, but the default is just a hash. We recommend using the same name
#' here as the name used to label the client ID in the [Google Cloud Platform
#' Console](https://console.cloud.google.com).
#'
#' A `gargle_oauth_client` is an adaptation of httr's [oauth_app()] (currently)
#' and httr2's `oauth_client()` (which gargle will migrate to in the future).

#' @param path JSON downloaded from [Google Cloud
#'   Console](https://console.cloud.google.com), containing a client id and
#'   secret, in one of the forms supported for the `txt` argument of
#'   [jsonlite::fromJSON()] (typically, a file path or JSON string).

#' @param name A label for this specific client, presumably the same name used
#'   to label it in Google Cloud Console. Unfortunately there is no way to
#'   make that true programmatically, i.e. the JSON representation does not
#'   contain this information.

#' @param id Client ID
#' @param secret Client secret

#' @param redirect_uris Where your application listens for the response from
#'   Google's authorization server. If you didn't configure this specifically
#'   when creating the client (which is only possible for clients of the "web"
#'   type), you can leave this unspecified.

#' @param type Specifies the type of OAuth client. The valid values are a subset
#'   of possible Google client types and reflect the key used to describe the
#'   client in its JSON representation:

#'   * `"installed"` is associated with a "Desktop app"
#'   * `"web"` is associated with a "Web application"

#' @return An OAuth client: An S3 list with class `gargle_oauth_client`. For
#'   backwards compatibility reasons, this currently also inherits from the httr
#'   S3 class `oauth_app`, but that is a temporary measure. An instance of
#'   `gargle_oauth_client` stores more information than httr's `oauth_app`, such
#'   as the OAuth client's type ("web" or "installed").

#'
#'   There are some redundant fields in this object during the httr-to-httr2
#'   transition period. The legacy fields `appname` and `key` repeat the
#'   information in the future-facing fields `name` and (client) `id`. Prefer
#'   `name` and `id` to `appname` and `key` in downstream code. Prefer the
#'   constructors `gargle_oauth_client_from_json()` and `gargle_oauth_client()`
#'   to [httr::oauth_app()] and [oauth_app_from_json()].

#' @export
#'
#' @examples
#' \dontrun{
#' gargle_oauth_client_from_json(
#'   path = "/path/to/the/JSON/you/downloaded/from/gcp/console.json",
#'   name = "my-nifty-oauth-client"
#' )
#' }
#'
#' gargle_oauth_client(
#'   id = "some_long_id",
#'   secret = "ssshhhhh_its_a_secret",
#'   name = "my-nifty-oauth-client"
#' )
gargle_oauth_client_from_json <- function(path, name = NULL) {
  check_string(path)
  if (!is.null(name)) {
    check_string(name)
  }

  json <- jsonlite::fromJSON(path, simplifyVector = FALSE)
  if (length(json) != 1) {
    gargle_abort(c(
      "JSON has an unexpected form",
      "i" = "Are you sure this is the JSON downloaded for an OAuth client?",
      "i" = "It is easy to confuse the JSON for an OAuth client and a service account."
    ))
  }

  info <- json[[1]]

  gargle_oauth_client(
    id = info$client_id,
    secret = info$client_secret,
    redirect_uris = info$redirect_uris,
    type = names(json),
    name = name %||% glue("{info$project_id}_{hash(info$project_id)}")
  )
}


#' @export
#' @rdname gargle_oauth_client_from_json
gargle_oauth_client <- function(
  id,
  secret,
  redirect_uris = NULL,
  type = c("installed", "web"),
  name = hash(id)
) {
  check_string(id)
  check_string(secret)
  check_string(name)
  type <- arg_match(type)

  if (!is.null(redirect_uris)) {
    # httr appears to assume that an OAuth app can have exactly 1 redirect_uri
    # (gargle has never used the `redirect_uri` field of httr::oauth_app)
    # httr2 seems to think it can usually construct the redirect_uri?
    # I think I have to accept multiple URIs, because that can be true in the
    # downloaded JSON
    # we'll just have to decide which one to use downstream, based on context
    redirect_uris <- unlist(redirect_uris)
    check_character(redirect_uris)
  }

  if (type == "web" && length(redirect_uris) == 0) {
    gargle_abort(
      '
      A "web" type OAuth client must have one or more {.field redirect_uris}.'
    )
  }

  structure(
    list(
      name = name,
      id = id,
      secret = secret,
      type = type,
      redirect_uris = redirect_uris,
      # needed for backwards compatibility; I need this class to quack like a
      # specialization of httr's oauth_app class, for now
      appname = name,
      key = id
    ),
    class = c("gargle_oauth_client", "oauth_app")
    # in the future, maybe:
    # class = c("gargle_oauth_client", "httr2_oauth_client")
  )
}
# adapted from httr2 ----
#' @export
print.gargle_oauth_client <- function(x, ...) {
  # this print method needs work, but not a high priority atm
  cli::cli_text(cli::style_bold("<gargle_oauth_client>"))
  redacted <- list_redact(compact(x), "secret")
  # quick fix for multiple URIs case
  if (length(redacted$redirect_uris) > 0) {
    redacted$redirect_uris <- commapse(redacted$redirect_uris)
  }
  # hide redundant fields that exist only for backwards compatibility with
  # httr's oauth_app
  redacted$appname <- redacted$key <- NULL
  cli::cli_dl(redacted)
  invisible(x)
}

list_redact <- function(x, names, case_sensitive = TRUE) {
  if (case_sensitive) {
    i <- match(names, names(x))
  } else {
    i <- match(tolower(names), tolower(names(x)))
  }
  x[i] <- cli::col_grey("<REDACTED>")
  x
}

#' OAuth client for demonstration purposes
#'
#' @description

#' Invisibly returns an instance of
#' [`gargle_oauth_client`][gargle_oauth_client()] that can be used to test drive
#' gargle before obtaining your own client ID and secret. This OAuth client may
#' be deleted or rotated at any time. There are no guarantees about which APIs
#' are enabled. DO NOT USE THIS IN A PACKAGE or for anything other than
#' interactive, small-scale experimentation.
#'
#' You can get your own OAuth client ID and secret, without these limitations.
#' See the `vignette("get-api-credentials")` for more details.
#'
#' @inheritParams gargle_oauth_client_from_json
#'
#' @return An OAuth client, produced by [gargle_oauth_client()], invisibly.
#' @export
#' @keywords internal
#' @examples
#' \dontrun{
#' gargle_client()
#' }
gargle_client <- function(type = NULL) {
  if (is.null(type) || is.na(type)) {
    type <- gargle_oauth_client_type()
  }
  check_string(type)
  type <- arg_match(type, values = c("installed", "web"))

  switch(
    type,
    web = goc_web(),
    installed = goc_installed()
  )
}

#' @export
#' @keywords internal
#' @rdname internal-assets
tidyverse_client <- function(type = NULL) {
  check_permitted_package(caller_env())

  if (is.null(type) || is.na(type)) {
    type <- gargle_oauth_client_type()
  }
  check_string(type)
  type <- arg_match(type, values = c("installed", "web"))

  switch(
    type,
    web = toc_web(),
    installed = toc_installed()
  )
}

# deprecated functions ----

#' Create an OAuth app from JSON
#'
#' @description
#' `r lifecycle::badge("deprecated")`
#'

#' `oauth_app_from_json()` is being replaced with
#' [`gargle_oauth_client_from_json()`], in light of the new
#' `gargle_oauth_client` class. Now `oauth_app_from_json()` potentially warns
#' about this deprecation and immediately passes its inputs through to
#' [`gargle_oauth_client_from_json()`].
#'
#' `gargle_app()` is being replaced with [gargle_client()].
#'
#' @inheritParams gargle_oauth_client
#' @inheritParams httr::oauth_app
#' @keywords internal
#' @export
oauth_app_from_json <- function(path, appname = NULL) {
  lifecycle::deprecate_soft(
    "1.3.0",
    "oauth_app_from_json()",
    "gargle_oauth_client_from_json()"
  )
  gargle_oauth_client_from_json(path = path, name = appname)
}

#' @export
#' @keywords internal
#' @rdname internal-assets
tidyverse_app <- function() {
  lifecycle::deprecate_soft(
    "1.3.0",
    "tidyverse_app()",
    "tidyverse_client()"
  )
  tidyverse_client()
}

#' @export
#' @keywords internal
#' @rdname oauth_app_from_json
gargle_app <- function() {
  lifecycle::deprecate_soft(
    "1.3.0",
    "gargle_app()",
    "gargle_client()"
  )
  gargle_client()
}

#' OAuth endpoint for Google APIs
#'
#' Inlined from httr in case it changes. Internal function to centralize the
#' info.
#'
#' @return An OAuth endpoint, produced by [httr::oauth_endpoint()].
#' @keywords internal
#' @noRd
#' @examples
#' gargle_oauth_endpoint()
gargle_oauth_endpoint <- function() {
  out <- httr::oauth_endpoint(
    base_url = "https://oauth2.googleapis.com",
    authorize = "",
    access = "token",
    validate = "tokeninfo",
    revoke = "revoke"
  )
  out$authorize <- "https://accounts.google.com/o/oauth2/v2/auth"
  out
}

#' Get a token from the Google metadata server
#'
#' @description
#'
#' If your code is running on Google Cloud, we can often obtain a token for an
#' attached service account directly from a metadata server. This is more secure
#' than working with an explicit a service account key, as
#' [credentials_service_account()] does, and is the preferred method of auth for
#' workloads running on Google Cloud.

#'
#' The most straightforward scenario is when you are working in a VM on Google
#' Compute Engine and it's OK to use the default service account. This should
#' "just work" automatically.

#'
#' `credentials_gce()` supports other use cases (such as GKE Workload Identity),
#' but may require some explicit setup, such as:

#' * Create a service account, grant it appropriate scopes(s) and IAM roles,
#' attach it to the target resource. This prep work happens outside of R, e.g.,
#' in the Google Cloud Console. On the R side, provide the email address of this
#' appropriately configured service account via `service_account`.

#' * Specify details for constructing the root URL of the metadata service:
#'   - The logical option `"gargle.gce.use_ip"`. If undefined, this defaults to
#'     `FALSE`.
#'   - The environment variable `GCE_METADATA_URL` is consulted when
#'     `"gargle.gce.use_ip"` is `FALSE`. If undefined, the default is
#'     `metadata.google.internal`.
#'   - The environment variable `GCE_METADATA_IP` is consulted when
#'     `"gargle.gce.use_ip"` is `TRUE`. If undefined, the default is
#'     `169.254.169.254`.

#'
#' * Change (presumably increase) the timeout for requests to the metadata
#' server via the `"gargle.gce.timeout"` global option. This timeout is given in
#' seconds and is set to a value (strategy, really) that often works well in
#' practice. However, in some cases it may be necessary to increase the timeout
#' with code such as:
#' ``` r
#' options(gargle.gce.timeout = 3)
#' ```

#' For details on specific use cases, such as Google Kubernetes Engine (GKE),
#' see `vignette("non-interactive-auth")`.
#'
#' @inheritParams token_fetch
#' @param service_account Name of the GCE service account to use.
#'
#' @seealso A related auth flow that can be used on certain non-Google cloud
#' providers is workload identity federation, which is implemented in
#' [credentials_external_account()].
#'
#' <https://docs.cloud.google.com/compute/docs/access/service-accounts>
#'
#' <https://docs.cloud.google.com/iam/docs/best-practices-service-accounts>
#'
#' How to attach a service account to a resource:
#' <https://cloud.google.com/iam/docs/impersonating-service-accounts#attaching-to-resources>
#'
#' <https://docs.cloud.google.com/kubernetes-engine/docs/concepts/workload-identity>
#'
#' <https://docs.cloud.google.com/kubernetes-engine/docs/how-to/workload-identity>
#'
#' <https://docs.cloud.google.com/compute/docs/metadata/overview>
#'
#' @return A [GceToken()] or `NULL`.
#' @family credential functions
#' @export
#' @examples
#' \dontrun{
#' credentials_gce()
#' }
credentials_gce <- function(
  scopes = "https://www.googleapis.com/auth/cloud-platform",
  service_account = "default",
  ...
) {
  gargle_debug("Trying {.fun credentials_gce} ...")
  if (!is_gce()) {
    gargle_debug(c("x" = "We don't seem to be on GCE."))
    return(NULL)
  }

  scopes <- scopes %||% "https://www.googleapis.com/auth/cloud-platform"
  requested_scopes <- normalize_scopes(scopes)
  dat <- gce_instance_service_accounts()
  service_account_details <- as.list(dat[dat$name == service_account, ])

  account_scopes <- service_account_details$scopes
  account_scopes <- normalize_scopes(strsplit(account_scopes, split = ",")[[1]])
  missing <- setdiff(requested_scopes, account_scopes)
  if (length(missing) > 0) {
    gargle_debug(c(
      "!" = "{cli::qty(length(missing))}{?This/These} requested \\
             scope{?s} {?is/are} not among the scopes for the \\
             {.val {service_account}} service account:",
      bulletize(missing, bullet = "x"),
      "i" = "If there are problems downstream, this might be the root cause."
    ))
  }

  token <- gce_access_token(scopes, service_account = service_account)

  if (
    is.null(token$credentials$access_token) ||
      !nzchar(token$credentials$access_token)
  ) {
    NULL
  } else {
    gargle_debug(
      "GCE service account email: {.email {service_account_details$email}}"
    )
    gargle_debug(
      "GCE service account name: {.val {token$params$service_account}}"
    )
    gargle_debug(
      "GCE access token scopes: {.val {commapse(base_scope(token$params$scope))}}"
    )
    token
  }
}

#' Fetch access token for a service account on GCE
#'
#' @inheritParams credentials_gce
#'
#' @keywords internal
#' @export
gce_access_token <- function(
  scopes = "https://www.googleapis.com/auth/cloud-platform",
  service_account = "default"
) {
  params <- list(
    scope = scopes,
    service_account = service_account,
    as_header = TRUE
  )
  GceToken$new(
    params = params
  )
}

#' Token for use on Google Compute Engine instances
#'
#' This class uses the metadata service available on GCE VMs to fetch access
#' tokens. Not intended for direct use. See [credentials_gce()] instead.
#'
#' @param ... Not used.
#'
#' @keywords internal
#' @export
GceToken <- R6::R6Class(
  "GceToken",
  inherit = httr::Token2.0,
  list(
    #' @description Get an access for a GCE service account.
    #' @param params A list of parameters for `fetch_gce_access_token()`.
    #' @return A GceToken.
    initialize = function(params) {
      gargle_debug("GceToken initialize")
      self$params <- params
      self$init_credentials()
    },
    #' @description Request an access token.
    init_credentials = function() {
      gargle_debug("GceToken init_credentials")
      token <- fetch_gce_access_token(
        self$params$scope,
        service_account = self$params$service_account
      )

      # find out the scopes actually obtained
      # https://www.googleapis.com/oauth2/v3/tokeninfo
      req <- request_build(
        method = "GET",
        path = "oauth2/v3/tokeninfo",
        params = list(access_token = token$access_token),
        base_url = "https://www.googleapis.com"
      )
      resp <- request_make(req)
      info <- response_process(resp)
      actual_scopes <- normalize_scopes(strsplit(info$scope, split = "\\s+")[[
        1
      ]])

      missing <- setdiff(self$params$scope, actual_scopes)
      if (length(missing) > 0) {
        gargle_debug(c(
          "!" = "{cli::qty(length(missing))}{?This/These} requested \\
             scope{?s} {?is/are} not among the scopes for the \\
             access token returned by the metadata server:",
          bulletize(missing, bullet = "x"),
          "i" = "If there are problems downstream, this might be the root cause."
        ))
      }

      if (!setequal(self$params$scope, actual_scopes)) {
        gargle_debug(c(
          "!" = "Updating token scopes to reflect its actual scopes:",
          bulletize(actual_scopes)
        ))
        self$params$scope <- actual_scopes
      }

      self$credentials <- token
      self
    },
    #' @description Refreshes the token. In this case, that just means "ask again
    #'   for an access token".
    refresh = function() {
      gargle_debug("GceToken refresh")
      # There's something kind of wrong about this, because it's not a true
      # refresh. But this method is basically required by the way httr currently
      # works.
      # This means that some uses of $refresh() aren't really appropriate for a
      # GceToken.
      # For example, if I attempt token_userinfo(x) on a GceToken that lacks
      # appropriate scope, it fails with 401.
      # httr tries to "fix" things by refreshing the token. But this is
      # not a problem that refreshing can fix.
      # I've now prevented an explicit refresh in token_userinfo(), but an
      # implicit one still eventually happens in httr:::request_perform().
      self$init_credentials()
    },
    #' @description Placeholder implementation of required method. Returns `TRUE`.
    can_refresh = function() {
      TRUE
    },

    #' @description Format a [GceToken()].
    #' @param ... Not used.
    format = function(...) {
      x <- list(
        scopes = commapse(base_scope(self$params$scope)),
        credentials = commapse(names(self$credentials))
      )
      c(
        cli::cli_format_method(
          cli::cli_h1("<GceToken (via {.pkg gargle})>")
        ),
        glue("{fr(names(x))}: {fl(x)}")
      )
    },
    #' @description Print a [GceToken()].
    #' @param ... Not used.
    print = function(...) {
      # a format method is not sufficient for GceToken because the parent class
      # has a print method
      cli::cat_line(self$format())
    },

    # Never cache
    #' @description Placeholder implementation of required method.
    cache = function() self,
    #' @description Placeholder implementation of required method.
    load_from_cache = function() self,

    # These methods don't really make sense for GCE access tokens
    #' @description Placeholder implementation of required method.
    revoke = function() {
      gargle_abort("{.fun $revoke} is not implemented for {.cls GceToken}")
    },
    #' @description Placeholder implementation of required method
    validate = function() {
      gargle_abort("{.fun $validate} is not implemented for {.cls GceToken}")
    }
  )
)

gce_metadata_hostname <- function() {
  use_ip <- getOption("gargle.gce.use_ip", FALSE)
  if (isTRUE(use_ip)) {
    Sys.getenv("GCE_METADATA_IP", "169.254.169.254")
  } else {
    Sys.getenv("GCE_METADATA_URL", "metadata.google.internal")
  }
}

gce_metadata_request <- function(
  path = "",
  query = NULL,
  stop_on_error = TRUE
) {
  # TODO(craigcitro): Add options to ignore proxies.
  if (grepl("^/", path)) {
    path <- substring(path, 2)
  }
  url_parts <- structure(
    list(
      scheme = "http",
      hostname = gce_metadata_hostname(),
      path = path,
      query = query
    ),
    class = "url"
  )
  url <- httr::build_url(url_parts)
  response <- try(
    {
      httr::with_config(httr::timeout(gce_timeout()), {
        httr::GET(url, httr::add_headers("Metadata-Flavor" = "Google"))
      })
    },
    silent = TRUE
  )

  if (stop_on_error) {
    if (inherits(response, "try-error")) {
      gargle_abort(
        "
        Error fetching GCE metadata: {attr(response, 'condition')$message}"
      )
    } else if (httr::http_error(response)) {
      gargle_abort(
        "
        Error fetching GCE metadata: {httr::http_status(response)$message}"
      )
    }
    if (response$headers$`metadata-flavor` != "Google") {
      gargle_abort(
        "
        Error fetching GCE metadata: missing/invalid metadata-flavor header"
      )
    }
  }
  response
}

# https://cloud.google.com/compute/docs/instances/detect-compute-engine
is_gce <- function() {
  response <- gce_metadata_request(stop_on_error = FALSE)
  !(inherits(response, "try-error") || httr::http_error(response))
}

#' List all service accounts available on this GCE instance
#'
#' @returns A data frame, where each row is a service account. Due to aliasing,
#'   there is no guarantee that each row represents a distinct service account.
#'
#' @seealso The return value is built from a recursive query of the so-called
#'   "directory" of the instance's service accounts as documented in
#'   <https://cloud.google.com/compute/docs/metadata/default-metadata-values#vm_instance_metadata>.
#'
#' @export
#' @examplesIf gargle:::is_gce()
#' credentials_gce()
gce_instance_service_accounts <- function() {
  response <- gce_metadata_request(
    "computeMetadata/v1/instance/service-accounts",
    query = list(recursive = "true")
  )
  raw <- transpose(response_as_json(response))
  data.frame(
    name = names(raw$email),
    email = unlist(raw$email),
    aliases = map_chr(raw$aliases, function(x) glue_collapse(x, sep = ",")),
    scopes = map_chr(raw$scopes, function(x) glue_collapse(x, sep = ",")),
    stringsAsFactors = FALSE,
    row.names = NULL
  )
}

# TODO: why isn't scopes used here at all?
# the python auth library definitely passes scopes:
# https://github.com/googleapis/google-auth-library-python/blob/a83af399fe98764ee851997bf3078ec45a9b51c9/google/auth/compute_engine/_metadata.py#L237
# perhaps there are use cases where it would be helpful it we did same:
# https://github.com/r-lib/gargle/issues/216
fetch_gce_access_token <- function(scopes, service_account) {
  path <- glue(
    "computeMetadata/v1/instance/service-accounts/{service_account}/token"
  )
  scope_string <- glue_collapse(scopes, sep = ",")
  response <- gce_metadata_request(path, query = list(scopes = scope_string))
  httr::content(response, as = "parsed", type = "application/json")
}

# wrapper to access the "gargle.gce.timeout" option
# https://github.com/r-lib/gargle/issues/186
# https://github.com/r-lib/gargle/pull/195
# if called with no argument:
#   if option is set, return that value
#   if unset: return a short default, suitable for initial ping of
#     the metadata server (and not too burdensome for non-GCE users) and set the
#     option to a longer default, suitable for a subsequent request for all
#     service accounts or a specific token
# if called with an argument:
#   set the option to that value (and return the old value)
gce_timeout <- function(v) {
  opt <- getOption("gargle.gce.timeout")
  if (missing(v)) {
    if (is.null(opt)) {
      ret <- 0.8 # short default timeout
      options(gargle.gce.timeout = 2) # long default timeout
    } else {
      ret <- opt
    }
  } else {
    ret <- options(gargle.gce.timeout = v)[["gargle.gce.timeout"]]
  }
  ret
}

expect_gargle2.0_token <- function(object, expected) {
  expect_equal(object$cache_path, expected$cache_path)
  expect_equal(object$endpoint, expected$endpoint)
  expect_equal(object$email, expected$email)
  expect_equal(object$app, expected$app)
  expect_equal(object$params, expected$params)
}

---
title: "Managing tokens securely"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Managing tokens securely}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

Testing presents special challenges for packages that wrap an API. Here we tackle one of those problems: how to deal with auth in a non-interactive setting on a remote machine. This affects gargle itself and will affect any client package that relies on gargle for auth.

This article documents the token management approach taken in gargle. We wanted it to be relatively easy to have a secret, such as an auth token, that we can:

  * Use locally
  * Use with continuous integration (CI) services, such as GitHub Actions
  * Use with [R-hub](https://docs.r-hub.io)

all while keeping the secret secure.

The approach uses symmetric encryption, where the shared key is stored in an environment variable. Why? This works well with existing conventions for local R usage. Most CI services offer support for secure environment variables. And R-hub accepts environment variables via the `env_vars` argument of [`rhub::check()`](https://r-hub.github.io/rhub/reference/check.html).

This is based on an approach originally worked out in [bigrquery](https://bigrquery.r-dbi.org).

## Accessing the `secret_*()` functions

gargle's approach to managing test tokens is implemented through several functions that all start with the `secret_` prefix. These functions are not (currently?) exported. This may seem odd, since others might want to use these functions. But note they are only needed during setup or at test time. This sort of usage is compatible with others calling internal gargle functions and possibly inlining a version of a couple test helpers.

One way to make the `secret_*()` functions available for local experimentation is to call `devtools::load_all()`, which exposes all internal objects in a package:

```{r eval = FALSE}
devtools::load_all("path/to/source/of/gargle/")
```

The approach I'll take in this article is to call these functions via `:::`.

## Overview of the approach

  1. Add the [sodium package](https://cran.r-project.org/package=sodium) to
     Suggests in your DESCRIPTION, via
     `usethis::use_package("sodium", "Suggests")` if you like.
  1. Generate a random PASSWORD and give it a self-documenting name, e.g.
     `GARGLE_PASSWORD`. Store as an environment variable.
  1. Identify a secret file of interest, such as the JSON representing a
     service account token. This is presumably stored *outside* your package.
  1. Use the PASSWORD to apply a method for symmetric encryption to the target
     file. Store the resulting encrypted file in a designated location *within*
     your package.
  1. Store or pass the PASSWORD as an environment variable everywhere you'll
     need to decrypt the secret.
     - Check that the platform has support for keeping the PASSWORD concealed.
     - Make sure you don't do anything in your own code that would dump it to,
       e.g., a log file.
  1. Rig your tests to determine if the key is available and, therefore,
     whether decryption is going to be possible.
     - If "no", carry on gracefully with any tests that don't require auth.
     - If "yes", decrypt the secret and put the associated token into force
       globally for the test run or on an "as needed" basis in individual tests.

## Annotated code-through

### Generate a name for the PASSWORD

`secret_pw_name()` creates a name of the form "PACKAGE_PASSWORD", a convention
baked into the `secret_*()` family of functions.

```{r}
(pw_name <- gargle:::secret_pw_name("gargle"))
```

### Generate a random PASSWORD

In real life, you should keep the output of `secret_pw_gen()` to yourself! We reveal it here as part of the exposition.

```{r}
(pw <- gargle:::secret_pw_gen())
```

### Define environment variable in `.Renviron`

Combine the name and value to form a line like this in your user-level `.Renviron` file:

```{r, echo = FALSE, comment = NA}
cat(paste0(pw_name, "=", pw), sep = "\n")
```

[`usethis::edit_r_environ()`](https://usethis.r-lib.org/reference/edit.html) can help create or open this file. We **strongly recommend** using the user-level `.Renviron`, as opposed to project-level, because this makes it less likely you will share sensitive information by mistake. If you don't take our advice and choose to store the PASSWORD in a file inside a Git repo, you must make sure that file is listed in `.gitignore`. This still would not prevent leaking your secret if, for example, that project is in a directory that syncs to DropBox.

Make sure `.Renviron` ends in a newline; the lack of this is a notorious cause of silent failure. Remember you'll need to restart R or call `readRenviron("~/.Renviron")` for the newly defined environment variable to take effect.

### Provide environment variable to other services

#### GitHub Actions:

Define the environment variable as an encrypted secret in your repo:

<https://help.github.com/en/actions/configuring-and-managing-workflows/creating-and-storing-encrypted-secrets>

Use the secrets context to expose a secret as an environment variable in your workflows.     That will look like like so, in some appropriate place in your workflow file:

```
env:
  PACKAGE_PASSWORD: ${{ secrets.PACKAGE_PASSWORD }}
```

Remember that the secret, and therefore the associated environment variable, is not available when workflows are triggered via an external pull request. This is another reason to carry on gracefully when token decryption is not possible (see below).

#### Travis-CI

Define the environment variable in your repo settings via the browser UI:

<https://docs.travis-ci.com/user/environment-variables/#defining-variables-in-repository-settings>

Alternatively, you can use the Travis command line interface to configure the environment variable or even define an encrypted environment variable in `.travis.yml`.

Regardless of how you define it, remember that private environment variables are not available to external pull requests, which is another reason to carry on gracefully when token decryption is not possible (see below).

You may also need something like this in `.travis.yml` so that the sodium R package can be installed:
    
``` yaml
addons:
  apt:
    sources:
    - sourceline: 'ppa:chris-lea/libsodium'
    packages:
    - libsodium-dev
```

#### AppVeyor

Define the environment variable in the Environment page of your repo's Settings. Make sure to request variable encryption and to click "Save" at the bottom. In the General page, you probably want to check "Enable secure variables in Pull Requests from the same repository only" and, again, explicitly "Save".

As with Travis, it is also possible to encrypt the password using your AppVeyor account's public key and inline the value in `appveyor.yml`. There is a helpful web UI for that does the encryption and generates the lines to add to your config:

<https://ci.appveyor.com/tools/encrypt>

This can also be found via *Settings > Encrypt YAML*.

#### R-hub

Send the environment variable in your calls to [`rhub::check()`](https://r-hub.github.io/rhub/reference/check.html) and [friends](https://r-hub.github.io/rhub/reference/index.html#section-check-shortcuts):

```
rhub::check(env_vars = Sys.getenv(gargle:::secret_pw_name("gargle"), names = TRUE))
```

### Encrypt the secret file

`secret_write()` takes 3 arguments:

  * `package` name. Processed through `secret_pw_name()` in order to retrieve
    the PASSWORD from an appropriately named environment variable.
  * `name` of the encrypted file to write. The location is below `inst/secret`
    in the source of `package`.
  * `data`, either a file path to the unencrypted secret file or the data to
    be encrypted as a raw vector. In the case of a secret file, we **strongly
    recommend** that its primary home on your local computer is outside
    your package and, generally, outside of any folder that syncs regularly to
    a remote, e.g. GitHub or DropBox. This decreases the chance of accidental
    leakage.

Example of a call to `secret_write()`, where `gargle-testing.json` is a JSON
file downloaded for a service account managed via the [Google API / Cloud Platform console](https://console.cloud.google.com/project):

```{r eval = FALSE}
gargle:::secret_write(
  package = "gargle",
  name = "gargle-testing.json",
  input = "a/very/private/local/folder/gargle-testing.json"
)
```

This writes an encrypted version of `gargle-testing.json` to `inst/secret/gargle-testing.json` relative to the current working directory, which is presumably the top-level directory of gargle's source. This encrypted file *should* be committed and pushed.

### Test setup

Now you need to rig your tests or their setup around this encrypted token. You need to plan for two scenarios:

  * Decryption is going to work. This is where you actually get to test package functionality against the target API, with auth.
  * Decryption is not going to work. Either because the Suggested [sodium](https://cran.r-project.org/package=sodium) package is not available or (much more likely) because the environment variable that represents the key is not available.
    - This will be the case on CRAN, by definition, because there is no way to share an encrypted secret.
    - This will be the case for external contributors, on their personal machines and when their GitHub pull requests are checked via CI services, such as GitHub Actions, Travis-CI, or AppVeyor. 

#### CI configuration

We recommend that you actively check your package under the "no decryption, no token" scenario, so that you discover problems before CRAN or your contributors do. In fact, this should probably be the default situation for your CI jobs and you only supply the secret to a single, flagship job -- probably the check with the current R release and your favorite operating system.

Here's the simplified build matrix from the `R CMD check` GitHub Actions workflow file used by gargle (note: we redacted a very long `rspm` URL that's irrelevant here):

``` yaml
strategy:
  matrix:
    config:
      - {os: macOS-latest,   r: 'devel', gargle_auth: GARGLE_NOAUTH}
      - {os: macOS-latest,   r: '4.0',   gargle_auth: GARGLE_PASSWORD}
      - {os: windows-latest, r: '4.0',   gargle_auth: GARGLE_NOAUTH}
      - {os: ubuntu-16.04,   r: '4.0',   gargle_auth: GARGLE_NOAUTH, rspm: "..."}
      - {os: ubuntu-16.04,   r: '3.6',   gargle_auth: GARGLE_NOAUTH, rspm: "..."}
      - {os: ubuntu-16.04,   r: '3.5',   gargle_auth: GARGLE_NOAUTH, rspm: "..."}
      - {os: ubuntu-16.04,   r: '3.4',   gargle_auth: GARGLE_NOAUTH, rspm: "..."}
      - {os: ubuntu-16.04,   r: '3.3',   gargle_auth: GARGLE_NOAUTH, rspm: "..."}

env:
  GARGLE_PASSWORD: ${{ secrets[matrix.config.gargle_auth] }}
```

Notice how `GARGLE_PASSWORD` will only be available when checking against the released version of R on macOS-latest.

bigrquery implements the same idea with a different approach: it does not provide `BIGRQUERY_PASSWORD` to the main `R CMD check` GitHub Actions workflow at all. Instead there is a separate "live api" workflow that only has one job and that accesses the secret.

The tidyverse / r-lib team is transitioning from Travis/AppVeyor to GitHub Actions. But in case it is helpful, here's a simplified excerpt from a `.travis.yml` file used by gargle in the past. The main `r: release` build accesses `GARGLE_PASSWORD` implicitly as an encrypted environment variable, but `R CMD check` runs for the other builds with `GARGLE_PASSWORD` explicitly unset:
      
``` yaml
matrix:
  include:
    - r: release
      # <stuff about code coverage, pkgdown build & deploy, etc.>
    - r: release
      env: GARGLE_PASSWORD=''
    - r: devel
      env: GARGLE_PASSWORD=''
    - r: oldrel
      env: GARGLE_PASSWORD=''
```

Regardless of your CI platform, your absolute best bet for writing configuration files is to look at what other R package developers are doing in their public source repos. All of the above is static, simplified, and (probably) stale and will never reflect the current state-of-the-art.

#### Testthat configuration

In a wrapper package, you could determine decrypt-ability at the start of the test run. Here's representative code from googledrive's `tests/testthat/helper.R` file, but something similar can be seen in bigrquery and googlesheets4:

```{r eval = FALSE}
if (gargle:::secret_can_decrypt("googledrive")) {
  json <- gargle:::secret_read("googledrive", "googledrive-testing.json")
  drive_auth(path = rawToChar(json))
}
```

Versions of `secret_can_decrypt()` and `secret_read()` are defined here in gargle. `drive_auth()` is a function specific to googledrive that loads a token for use downstream (in multiple tests, in this case). Note that it can clearly accept a JSON string, as an alternative to a filepath, and that's very favorable for our workflow. We'll come back to this below.

But what if `secret_can_decrypt()` returns `FALSE` and no token is loaded? That's where you rely on a custom test skipper. Here is the [test skipper from googledrive](https://github.com/tidyverse/googledrive/blob/main/tests/testthat/helper.R):

```{r eval = FALSE}
# googledrive
skip_if_no_token <- function() {
  testthat::skip_if_not(drive_has_token(), "No Drive token")
}
```

`googledrive::drive_has_token()` returns `TRUE` if a token is available and `FALSE` otherwise. By calling the skipper at the start of tests that require auth, you arrange for your package to cope gracefully when the token cannot be decrypted, e.g., on CRAN and in pull requests. It is typical to define such a skipper in `tests/testthat/helper.R` or similar.

*gargle's usage of the testing token is a bit different, still evolving, and less relevant to the maintainers of wrapper packages. Therefore it's not featured here.*

### Known sources of friction

Once you dig into the `secret_*()` family, you will notice there are two recurring sources of friction:

  * File or object? You almost certainly store your secrets in files. But the sodium functions for data encrypt and decrypt work with R objects. So, for example, it is convenient if token ingest can accept an R object as opposed to only a file path.
  * Raw vectors. You might think of the PASSWORD or even the secret file itself (e.g., JSON) in terms of plain text. But the sodium functions for data encrypt and decrypt work with *raw vectors*, not character vectors. Be prepared to see related conversions in the `secret_*()` functions.
  
Functions useful for these conversions:

  * `writeBin()` / `readBin()`
  * `charToRaw()` / `rawToChar()`
  * `sodium::data_encrypt()` / `sodium::data_decrypt`
  * `sodium::bin2hex()` / `sodium::hex2bin()`

## Resources

bigrquery and googledrive, which both use this approach.

  * [`bigrquery/tests/testthat/helper-auth.R`](https://github.com/r-dbi/bigrquery/blob/main/tests/testthat/helper-auth.R)
  * [`googledrive/tests/testthat/helper.R`](https://github.com/tidyverse/googledrive/blob/main/tests/testthat/helper.R)
  * Setup chunk of a pkgdown article that is rendered and deployed via CI: [`googledrive/vignettes/articles/multiple-files.Rmd#L10-L29`](https://github.com/tidyverse/googledrive/blob/f08c545284eadd4e69a14d1e843d228e8166e896/vignettes/articles/multiple-files.Rmd#L10-L29)

"Managing secrets" vignette of httr:

  * <https://httr.r-lib.org/articles/secrets.html>
  
Vignettes of the sodium package, especially the parts relating to symmetric encryption:

  * <https://cran.r-project.org/web/packages/sodium/vignettes/crypto101.html>
  * <https://cran.r-project.org/web/packages/sodium/vignettes/intro.html>
  
The [cyphr](https://ropensci.github.io/cyphr/) package, which smooths over frictions like those identified above relating to "file vs. object?" and "character vs. raw?":

  * <https://docs.ropensci.org/cyphr/>

---
title: "Non-interactive auth"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Non-interactive auth}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r}
#| label: setup
#| include: false
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

Here we describe how to do auth with a package that uses gargle, without requiring any user interaction.
This comes up in a wide array of contexts, ranging from simple rendering of a local R Markdown document to deploying a data product on a remote server.

We assume the wrapper package uses the design described in `vignette("gargle-auth-in-client-package")`.
Examples include:

  * [bigrquery](https://bigrquery.r-dbi.org)
  * [googledrive](https://googledrive.tidyverse.org)
  * [googlesheets4](https://googlesheets4.tidyverse.org)
  * [gmailr](https://gmailr.r-lib.org) *note: gmailr does not use the built-in
    tidyverse OAuth client*

Full details on `gargle::token_fetch()`, which powers this strategy, are given in `vignette("how-gargle-gets-tokens")`.

## Provide a token or pre-authorize token discovery

The main principle for auth that does not require user interaction:

> Provide a token directly or take advance measures that indicate you want a token to be discovered.

We present several ways to achieve this, basically in order of preference.

## Sidebar 1: Deployment

First, a word about deployed environments.
If this doesn't apply to you, skip this section.

Let's identify a specific type of project: it is developed in one place, with interactivity -- such as your local computer -- and then deployed elsewhere, where it must run without further interaction -- such as on [Posit Connect](https://posit.co/products/enterprise/connect/) or [shinyapps.io](https://www.shinyapps.io).
In this context, it may make sense to depart from gargle's default behaviour, which is to store tokens outside the project, and to embed them in the project instead.
An example at the end of this vignette demonstrates the use of a project-level OAuth cache.
A service account token could also be stored in the project.

When you embed tokens in the project and deploy, remember, that, by default, they are no more secure or hidden than the other source files in the project.
`vignette("managing-tokens-securely")` describes a method for embedding an encrypted token in the project, which is an extra level of care needed if you want to access credentials within, e.g., a continuous integration service, such as GitHub Actions.

## Sidebar 2: I just want my `.Rmd` to render

TL;DR is that you need to successfully authenticate *once* in an interactive session and then, in your code, give gargle permission to use a token it finds in the cache.
These sorts of commands achieve that:

```r
# Approach #1: use an option.
# Either specify the user:
options(gargle_oauth_email = "jenny@example.com")
# Or, if you only use one Google identity, you can be more vague:
options(gargle_oauth_email = TRUE)
# Or, you can specify the identity to use at the domain level:
options(gargle_oauth_email = "*@example.com")

# Approach #2: call PACKAGE_auth() proactively.
library(googledrive)
# Either specify the user:
drive_auth(email = "jenny@example.com")
# Or, if you only use one Google identity, you can be more vague:
drive_auth(email = TRUE)
# Or, you can specify the identity to use at the domain level:
drive_auth(email = "*@example.com")
```

At the end of this article, this scenario is explained in detail, if you want to understand why this works.

## Embrace credentials available in certain cloud settings

In certain cloud computing contexts, a service account token may be ambiently available (or you can arrange for that to be true).
Think about it: if your workload is running on Google Compute Engine (GCE), it's already "inside the Google house".
It seems like there should be a way to avoid another round of auth and that is indeed the case.

Another advantage of these cloud auth workflows is that there is never any need to download and carefully manage a file that contains sensitive information.
This is why they are often described as "keyless".
If you *can* use one of these methods, you should seriously consider doing so.

### Google Compute Engine

This section applies to code running on a GCE instance, either literally, or on another Google Cloud product built on top of GCE.
You should consider Google's own documentation to be definitive, but we'll try to give a useful summary here and to explain how gargle works with GCE:

<https://docs.cloud.google.com/compute/docs/access/create-enable-service-accounts-for-instances>

A Google Cloud Platform (GCP) project generally has a GCE default service account and, by default, a new GCE instance runs as that service account.
(If you wish, you can use a *different* service account by taking explicit steps when you create an instance or by modifying it while it's stopped.)
The main point is that, for an application running on GCE, a service account identity is generally available.

GCE allows applications to get an OAuth access token from its metadata server and this is what `gargle::credentials_gce()` does (which is one of functions tried by `gargle::token_fetch()`, which is called by wrapper packages).
This token request can be made for specific scopes and, in general, most wrapper packages will indeed be asking for specific scopes relevant to the API they access.
Consider the signature of `googledrive::drive_auth()`:

```r
drive_auth <- function(
  email = gargle::gargle_oauth_email(),
  path = NULL,
  scopes = "https://www.googleapis.com/auth/drive",
  cache = gargle::gargle_oauth_cache(),
  use_oob = gargle::gargle_oob_default(),
  token = NULL
) {
  ...
}
```

The googledrive package asks for a token with `"drive"` scope, by default.
This brings up one big gotcha when using packages like googledrive or googlesheets4 on GCE.

By default, a GCE instance will be running as the default service account, with the `"cloud-platform"` scope and this will, generally speaking, allow the service account to work with various Cloud products.
However, the `"cloud-platform"` scope does not permit operations with non-Cloud APIs, such as Drive and Sheets.
If you want the service account identity for your GCE instance to be able to get an access token for use with Drive and Sheets, you will need to explicitly add, e.g., the `"drive"` scope when you create the instance (or stop the instance and add that scope).
(Note that, in contrast, BigQuery is considered a Cloud product and therefore bigrquery can operate with the `"cloud-platform"` scope.)

Be aware that you might also need to explicitly grant the service account an appropriate level of access (e.g. read or write) to any Drive files you intend to work on.

Finally, if you want to opt-out of using the default service account and, instead, auth as a normal user, even though you are on GCE, that is also possible.
One way to achieve that is to remove `credentials_gce()` from the set of auth functions tried by `gargle::token_fetch()` by executing this command before any explicit or implicit auth happens:

```r
# removes `credentials_gce()` from gargle's registry
gargle::cred_funs_add(credentials_gce = NULL)
```

You can make a similar change in more scoped way with the helpers `gargle::with_cred_funs()` or `gargle::local_cred_funs()`.

### Workload Identity on Google Kubernetes Engine (GKE)

Here we discuss how gargle's GCE auth can work for a related service, Google Kubernetes Engine (GKE), using [Workload Identity](https://docs.cloud.google.com/kubernetes-engine/docs/how-to/workload-identity).
This is more complicated that direct usage of GCE and some extra configuration is needed to make a service account's metadata available for the GKE instance to discover.
GKE is the underlying technology behind Google's managed Airflow service, [Cloud Composer](https://cloud.google.com/composer), so this also applies to R docker files being called in that environment.

Workload Identity is the recommended way to do authentication on GKE and other places, if possible, since it eliminates the use of a file that holds the service key, which is a potential security risk.

1. Following the [Workload Identity docs](https://docs.cloud.google.com/kubernetes-engine/docs/how-to/workload-identity), you create a service account as normal and give it permissions and scopes needed to, say, upload to BigQuery.
Imagine that `my-service-key@my-project.iam.gserviceaccount.com` has the `https://www.googleapis.com/auth/bigquery` scope.
2.  Instead of downloading a JSON key, you instead migrate that permission by adding a policy binding to another service account within Kubernetes.
3. Create the service account within Kubernetes, ideally within a new namespace:

```sh
# create namespace
kubectl create namespace my-namespace
# Create Kubernetes service account
kubectl create serviceaccount --namespace my-namespace bq-service-account
```

4. Bind that Kubernetes service account to the service account outside of Kubernetes you created in step 1, and assign it an annotation:

```sh
# Create IAM policy binding betwwen k8s SA and GSA
gcloud iam service-accounts add-iam-policy-binding my-service-key@my-project.iam.gserviceaccount.com \
    --role roles/iam.workloadIdentityUser \
    --member "serviceAccount:my-project.svc.id.goog[my-namespace/bq-service-account]"
# Annotate k8s SA
kubectl annotate serviceaccount bq-service-account \
    --namespace my-namespace \
    iam.gke.io/gcp-service-account=my-service-key@my-project.iam.gserviceaccount.com
```

This key will now be available to add to pods within the cluster.
For Airflow, you can pass them in using the Python code `GKEStartPodOperator(...., namespace='my-namespace', service_account_name='bq-service-account')`.
Documentation around `GKEStartPodOperator()` within Cloud Composer can be found [here](https://docs.cloud.google.com/composer/docs/composer-2/use-gke-operator).

5. In order for the R function `gargle::gce_credentials()` do the right thing, you need to do two things:
  - Set `"gargle.gce.use_ip"` option to `TRUE`, in order to use the metadata
    server that's relevant on GKE.
  - Specify the target service account, i.e. you can't just passively accept the
    default, which is to use the `"default"` service account.
    `gce_instance_service_accounts()` can be helpful, e.g., if you want to
    know which service accounts your Docker container can see.

Here is example code that you might execute in your Docker container:

```r
options(gargle.gce.use_ip = TRUE)
t <- gargle::credentials_gce(
  "my-service-key@my-project.iam.gserviceaccount.com"
)
# ... do authenticated stuff with the token t ...
```

Let's assume that PKG is an R package that implements gargle auth in the standard way, such as bigrquery or googledrive.
At the time of writing the `service_account` argument is not exposed in the usual, high-level `PKG_auth()` function (<https://github.com/r-lib/gargle/issues/249>.
So if you need to use a non-`default` service account, you need to call `credentials_gce()` directly and pass that token to `PKG_auth()`:
Here's an example of how that might look:

```r
library(PKG)

options(gargle.gce.use_ip = TRUE)
t <- gargle::credentials_gce(
  "my-service-key@my-project.iam.gserviceaccount.com", # use YOUR service account
  scopes = "https://www.googleapis.com/auth/PKG"       # use REAL scopes
)
PKG_auth(token = t)
# ... do authenticated stuff...
```

### AWS

Keyless auth is even possible from non-Google cloud platforms, using [Workload identity federation](https://cloud.google.com/iam/docs/configuring-workload-identity-federation).

This is implemented in the experimental function `credentials_external_account()`, which currently only supports AWS.

## Provide a service account token directly

When two computers are talking to each other, possibly with no human involvement, the most appropriate type of token to use is a service account token.
If you're not working in cloud context with automatic access to a service account (see previous section), you can still use a service account, but it will require more explicit effort.

1. Create a service account and then download its credentials as a JSON file.
   This is described in `vignette("get-api-credentials")`, specifically in the
   *Service account token* section.
1. Call the wrapper package's main auth function proactively and provide the
   path to this JSON file.

Example using googledrive:

```r
library(googledrive)

drive_auth(path = "/path/to/your/service-account-token.json")
```

If this code is running on, e.g., a continuous integration service and you need to use an encrypted token, see `vignette("managing-tokens-securely")`.

For certain APIs, service accounts are inherently awkward, because you often want to do things *on behalf of a specific user*.
Gmail is a good example.
If you are sending email programmatically, you probably want to send it as yourself (or from some other specific email account) instead of from `zestybus-geosyogl@fuffapster-654321.iam.gserviceaccount.com`.
This is, in fact, possible but is described as "impersonation", which should tip you off that Google does not exactly encourage this workflow.
Some details:

* This requires "delegating domain-wide authority " to the service account.
* It is only possible in the context of a G Suite domain and only an
  administrator of the domain can set this up.
* The domain-wide authority is granted only for specific scopes, so those can be
  as narrow as possible. This may make a domain administrator more receptive
  to the idea.
* This is documented in a few different places, such as:
  - [Delegating domain-wide authority to the service account](https://developers.google.com/identity/protocols/oauth2/service-account#delegatingauthority) from Google Identity Platform docs
  - [Perform G Suite Domain-Wide Delegation of Authority](https://developers.google.com/admin-sdk/directory/v1/guides/delegation) from G Suite Admin SDK docs
* The `subject` argument of `credentials_service_account()` and
  `credentials_app_default()` is available to specify which user to
  impersonate, e.g. `subject = "user@example.com"`.
  This argument first appeared in gargle 0.5.0, so it may not necessarily be
  exposed yet in user-facing auth functions like `drive_auth()`.
  If you need `subject` in a client package, that is a reasonable feature
  request.
  It is also possible to get a token with an explicit call to, e.g.,
  `credentials_service_account()` and then pass that token to the auth function:

```r
t <- gargle::credentials_service_account(
  path = "/path/to/your/service-account-token.json",
  scopes = ...,
  subject = "user@example.com"
)
googledrive::drive_auth(token = t)
```

If delegation of domain-wide authority is impossible or unappealing, you must use an OAuth user token, as described below.

## Rig a service or external account for use with Application Default Credentials

Wrapper packages that use `gargle::token_fetch()` in the recommended way have access to the token search strategy known as **Application Default Credentials**.

You need to put the JSON corresponding to your service or external account in a very specific location or, alternatively, record the location of this JSON file in a specific environment variable.

Full details are in the `credentials_app_default()` section of `vignette("how-gargle-gets-tokens")`.

If you have your token rigged properly, you **do not** need to do anything else, i.e. you do not need to call `PACKAGE_auth()` explicitly.
Your token should just get discovered upon first need.

For troubleshooting purposes, you can set a gargle option to see verbose output about the execution of `gargle::token_fetch()`:

```r
options(gargle_verbosity = "debug")
```

withr-style convenience helpers also exist: `with_gargle_verbosity()` and `local_gargle_verbosity()`.

## Provide an OAuth token directly

If you somehow have the OAuth token you want to use as an R object, you can provide it directly to the `token` argument of the main auth function. Example using googledrive:

```r
library(googledrive)

my_oauth_token <- # some process that results in the token you want to use
drive_auth(token = my_oauth_token)
```

gargle caches each OAuth user token it obtains to an `.rds` file, by default.
If you know the filepath to the token you want to use, you could use `readRDS()` to read it and provide as the `token` argument to the wrapper's auth function.
Example using googledrive:

```r
# googledrive
drive_auth(token = readRDS("/path/to/your/oauth-token.rds"))
```

How would you know this filepath?
That requires some attention to the location of gargle's OAuth token cache folder, which is described in the next section.

Full details are in the `credentials_byo_oauth2()` section of `vignette("how-gargle-gets-tokens")`.

## Arrange for an OAuth user token to be re-discovered

This is the least recommended strategy, but it appeals to many users, because it doesn't require creating a service account.
Just remember that the perceived ease of using the token you already have (an OAuth user token) is quickly cancelled out by the greater difficulty of managing such tokens for non-interactive use.
You might be forced to use this strategy with certain APIs, such as Gmail, that are difficult to use with a service account.

Two main principles:

  1. Take charge of -- or at least notice -- the folder where OAuth tokens are
     being cached.
  2. Make sure exactly one cached token will be identified and pre-authorize
     its use.

There are many ways to do this. We'll work several examples using that convey the range of what's possible.

### I just want my `.Rmd` to render

**Step 1**: Get that first token. You must run your code at least once, interactively, do the auth dance, and allow gargle to store the token in its cache.

```r
library(googledrive)

# do anything that triggers auth
drive_find(n_max = 5)
```

**Step 2**: Revise your code to pre-authorize the use of that token next time.
Now your `.Rmd` can be rendered or your `.R` script can run, without further interaction.

You have two choices to make:

  * Set the `gargle_oauth_email` option or call `PACKAGE_auth(email = ...)`.
    - The option-based approach can be implemented in each `.Rmd` or `.R` or
      in a user-level or project level `.Rprofile` startup file.
  * Authorize the use of the "matching token":
    - `email = TRUE` works if we're only going to find, at most, 1 token, i.e.
      you always auth with the same identity
    - `email = "jane@example.com"` pre-authorizes use of a token associated with
      a specific identity
    - `email = "*@example.com"` pre-authorizes use of a token associated with an
      identity from a specific domain; good for code that might be executed on
      the machines of both `alice@example.com` and `bob@example.com`

This sets an option that allows gargle to use cached tokens whenever there's a unique match:

```r
options(gargle_oauth_email = TRUE)
```

This sets an option to use tokens associated with a specific email address:

```r
options(gargle_oauth_email = "jenny@example.com")
```

This sets an option to use tokens associated with an email address with a specific domain:

```r
options(gargle_oauth_email = "*@example.com")
```

This gets a token *right now* and allows the use of a matching token, using googledrive as an example:

```r
drive_auth(email = TRUE)
```

This gets a token *right now*, for the user with a specific email address:

```r
drive_auth(email = "jenny@example.com")
```

This gets a token *right now*, first checking the cache for a token associated with a specific domain:

```r
drive_auth(email = "*@example.com")
```

### Project-level OAuth cache

This is like the previous example, but with an added twist: we use a project-level OAuth cache. This is good for deployed data products.

**Step 1**: Obtain the token intended for non-interactive use and make sure it's cached in a (hidden) directory of the current project.
Using googledrive as an example:

```r
library(googledrive)

# designate project-specific cache
options(gargle_oauth_cache = ".secrets")

# check the value of the option, if you like
gargle::gargle_oauth_cache()

# trigger auth on purpose --> store a token in the specified cache
drive_auth()

# see your token file in the cache, if you like
list.files(".secrets/")
```

Do this setup once per project.

Another way to accomplish the same setup is to specify the desired cache location directly in the call to the auth function:

```r
library(googledrive)

# trigger auth on purpose --> store a token in the specified cache
drive_auth(cache = ".secrets")
```

**Step 2**: In all downstream use, announce the location of the cache and pre-authorize the use of a suitable token discovered there. Continuing the googledrive example:

```r
library(googledrive)

options(
  gargle_oauth_cache = ".secrets",
  gargle_oauth_email = TRUE
)

# now use googledrive with no need for explicit auth
drive_find(n_max = 5)
```

Setting the option `gargle_oauth_email = TRUE` says that googledrive is allowed to use a token that it finds in the cache, without interacting with a user, as long as it discovers EXACTLY one matching token.
This option-setting code needs to appear in each script, `.Rmd`, or app that needs to use this token non-interactively.
Depending on the context, it might be suitable to accomplish this in a startup file, e.g. project-level `.Rprofile`.

Here's a variation where we say which token to use by explicitly specifying the associated email.
This is handy if there's a reason to have more than one token in the cache.

```r
library(googledrive)

options(
  gargle_oauth_cache = ".secrets",
  gargle_oauth_email = "jenny@example.com"
)

# now use googledrive with no need for explicit auth
drive_find(n_max = 5)
```

Here's another variation where we specify the necessary info directly in an auth call, instead of in options:

```r
library(googledrive)

drive_auth(cache = ".secrets", email = TRUE)

# now use googledrive with no need for explicit auth
drive_find(n_max = 5)
```

Here's one last variation that's applicable when the local cache could contain multiple tokens:

```r
library(googledrive)

drive_auth(cache = ".secrets", email = "jenny@example.com")

# now use googledrive with no need for explicit auth
drive_find(n_max = 5)
```

Be very intentional about paths and working directory.
Personally I would use `here::here(".secrets)"` everywhere above, to make things more robust.

For troubleshooting purposes, you can set a gargle option to see verbose output about the execution of `gargle::token_fetch()`:

```r
options(gargle_verbosity = "debug")
```

withr-style convenience helpers also exist: `with_gargle_verbosity()` and `local_gargle_verbosity()`.

For a cached token to be considered a "match", it must match the current request with respect to user's email, scopes, and OAuth client (client ID or key and secret).
By design, these settings have very low visibility, because we usually want to use the defaults. If your token is not being discovered, consider if any of these fields might explain the mismatch.

---
title: "Auth when using R from the browser"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Auth when using R from the browser}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r}
#| include: false
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

If you are working with R in a web-based context, such as [RStudio Server](https://posit.co/download/rstudio-server/), [Posit Cloud](https://posit.cloud/), [Posit Workbench](https://posit.co/products/enterprise/workbench/), or [Google Colaboratory](https://colab.research.google.com/), your experience of browser-based auth flows will be different from those using R on their local machine.
You need to use **out-of-band authentication**, sometimes denoted "oob" or "OOB".
After the usual auth dance, instead of seeing "authentication successful, return to R!", you are presented with an authorization code to copy and paste back into your R session.
For folks who are running R on their local machine, this final exchange can be done automagically, using a temporary local webserver, but that is not possible for those accessing a remote R session through the browser.

On February 16, 2022, Google announced the (partial) deprecation of the OAuth out-of-band (OOB) flow, to be enacted no later than February 1, 2023.
The deprecation applies to Google Cloud Platform (GCP) projects that are in production mode.
OOB still works for projects that are in testing mode.

The built-in tidyverse client (used by googledrive, googlesheets4, and bigrquery) is associated with a GCP project that is in production mode.
Therefore, conventional OOB auth stopped working for the built-in client in February 2023.
In anticipation of this, gargle gained a new auth flow in version 1.3.0 that we call "pseudo-OOB", which should allow casual users to continue to enjoy a low-friction auth experience, even from RStudio Server, Posit Cloud, Posit Workbench, and Google Colaboratory.

If you attempt to do conventional OOB auth with a client that no longer supports it, you'll see something like this:

```{r}
#| echo: false
#| out-width: 400px
#| fig-cap: >
#|   Access blocked: Tidyverse API Packages's request is invalid.
#|   Error 400: invalid_request
#| fig-alt: >
#|   Screenshot with the following text: 'Access blocked: Tidyverse API
#|   Packages's request is invalid', 'You can't sign in because Tidyverse API
#|   Packages sent an invalid request. You can try again later, or contact the
#|   developer about this issue. Learn more about this error', 'If you are a
#|   developer of Tidyverse API Packages, see error details.', 'Error 400:
#|   invalid_request'.
knitr::include_graphics("invalid_request.png")
```

If you work on any of the affected platforms and are experiencing new auth problems, your first move should be to update all packages involved (gargle and one or more of googledrive, googlesheets4, bigrquery).
**Restart R.**
Re-execute your code in an interactive context that will allow you to re-auth.

This vignette documents various matters around OOB auth, both conventional and pseudo-OOB, for users who want to understand this more deeply.

Some of the packages that use gargle for auth and for which this article applies:

  * [bigrquery](https://bigrquery.r-dbi.org)
  * [googledrive](https://googledrive.tidyverse.org)
  * [googlesheets4](https://googlesheets4.tidyverse.org)
  * [gmailr](https://gmailr.r-lib.org) *note: gmailr does not use the built-in
    tidyverse OAuth client*

## Consider using a service account token (or no token!)

If you have concerns about using OOB auth, consider whether your task truly requires auth as a specific, normal user.

Can the task be completed with _no auth_, i.e. you are accessing something that is world readable or readable for "anyone with a link"?
In that case, the wrapper package probably provides a function to go into a de-authorized state, such as `googledrive::drive_deauth()` or `googlesheets4::gs4_deauth()`.

If the task requires auth, consider whether it really must be as a specific user.
You may be able to accomplish the task with a service account, which you create for this specific purpose.
A service account token is much easier to work with on a server and in non-interactive contexts than a user token.
A service account can also be given much more selective permissions than a user account and can be more easily deleted, once it is no longer needed.
Remember that the service account will need to be explicitly given permission to access any necessary resources (e.g. permission to read or write a specific Drive file or Sheet).
A service account doesn't somehow inherit permissions indirectly from the user who owns the GCP project in which it lives.
To learn more about using a service account, see `vignette("non-interactive-auth")`.

## When and how to use OOB

In the absence of any user instructions, the function `gargle::gargle_oob_default()` is used to decide whether to use OOB auth.
By default, OOB auth is used on RStudio Server, Posit Cloud, Posit Workbench, and Google Colaboratory, or if the option `"gargle_oob_default"` is set to `TRUE`.
(Note that we use the term "OOB auth" here to include both the existing, conventional form of OOB and gargle's new pseudo-OOB.)

Wrapper packages generally also allow the user to opt-in to OOB auth when making a direct call to an auth function.
For example, the functions `googledrive::drive_auth()`, `googlesheets4::gs4_auth()`, `bigrquery::bq_auth()`, and `gmailr::gm_auth()` all have a `use_oob` argument.
Notably, all of these `use_oob` arguments default to `gargle::gargle_oob_default()`.

gargle usually automatically detects when it should use OOB auth, but here is what it could look like if we are not using OOB, but should be. During auth, you are redirected to localhost on port 1410 and receive an error along these lines:

```
Chrome: This site can't be reached; localhost refused to connect.
Firefox: Unable to connect; can't establish a connection.
```

If this happens you might need to explicitly request OOB.
Below we review two different methods.

## Request OOB auth in the `PKG_auth()` call

Packages like googledrive and bigrquery aim to make auth "just work" for most users, i.e. it's automatically triggered upon first need.
However, it is always possible to initiate auth yourself, which gives you the opportunity to specify non-default values of certain parameters.
Here's how you could request OOB auth, using googledrive as an example:

```{r}
#| eval: false
library(googledrive)

drive_auth(use_oob = TRUE)

# now carry on with your work
drive_find(n_max = 5)
```

## Set the `"gargle_oob_default"` option

If you know that you *always* want to use OOB, as a user or within a project, the best way to express this is to set the `"gargle_oob_default"` option.

```{r}
#| eval: false
options(gargle_oob_default = TRUE)
```

This code could appear at the top of a script, in a setup chunk for `.Rmd`, or in a Shiny app.
But it probably makes even more sense in a `.Rprofile` startup file, at the user or project level.

Once the `"gargle_oob_default"` option has been set, it is honored by downstream calls to `PKG_auth()`, explicit or implicit, because the default value of `use_oob` is `gargle::gargle_oob_default()`, which consults the option.

## Conventional vs. pseudo-OOB auth

gargle now supports two OOB flows, which we call "conventional OOB" (the existing, legacy OOB flow) and "pseudo-OOB" (the new flow introduced in response to the partial deprecation of conventional OOB).
If we are using OOB auth, the decision between conventional or pseudo-OOB is made based on the currently configured OAuth client.

* If the OAuth client is of type `"installed"` (shows as "Desktop" in Google
  Cloud Console) or is of unknown type, gargle uses conventional OOB. Note that
  this will not necessarily succeed, due to the deprecation process described
  above.
* If the OAuth client is of type `"web"` (shows as "Web application" in Google
  Cloud Console), gargle uses the new pseudo-OOB flow.

```{=html}
<style>
  table {
    border-collapse: collapse;
    border-style: none;
  }

  .empty-cell {
    border: none;
    background-color: transparent;
  }

  table thead {
    background-color: transparent;
  }

  table th, table td {
    border-style: solid;
  }
</style>


<table>
  <thead>
    <tr>
      <th class="empty-cell"></th>
      <th class="empty-cell"></th>
      <th colspan="2"><verb>use_oob</verb></th>
    </tr>
    <tr>
      <th class="empty-cell"></th>
      <th class="empty-cell"></th>
      <th>FALSE</th>
      <th>TRUE</th>
    </tr>
  </thead>
  <tbody>
    <tr>
      <th rowspan="2">client type</th>
      <th>installed</th>
      <td>use httpuv to spin up<br> a temporary web server</td>
      <td>conventional OOB</td>
    </tr>
    <tr>
      <th>web</th>
      <td>--not possible--</td>
      <td>pseudo-OOB</td>
    </tr>
  </tbody>
</table>
```

Packages that use a built-in tidyverse OAuth client (googledrive, googlesheets4, and bigrquery) should automatically select a "web" client on RStudio Server, Posit Cloud, Posit Workbench, and Google Colaboratory and an "installed" client otherwise.
If you need to explicitly request a "web" client in some other setting, you can use the global option `"gargle_oauth_client_type"`:

```{r}
#| eval: false
options(gargle_oauth_client_type = "web")
```

Users who configure their own OAuth client will need to be intentional when choosing the client type, depending on where the code is running.

On the R side, it is recommended to setup an OAuth client using `gargle_oauth_client_from_json()`, which allows the client type (`"installed"` vs. `"web"`) to be detected programmatically from the downloaded JSON.
The less-preferred approach is to use `gargle_oauth_client()` and provide the information yourself.

## How pseudo-OOB works

Pseudo-OOB works just like non-OOB and conventional OOB in terms of the user's interactions with Google authorization server.
This is where the user authenticates themselves with Google and consents to the type of access being requested by the R code.

These flows differ in how they handle a successful response from the authorization server.
Specifically, the flows use different redirect URIs.

* A (temporary) local webserver is used to listen for this response at, e.g.,
  `http://localhost:1410/` if R is running locally and the httpuv package is
  available (i.e. a non-OOB flow).
* In conventional OOB, a special redirect value is used, typically
  `urn:ietf:wg:oauth:2.0:oob`, and the authorization code is provided to the
  user via a browser window for manual copy/paste. This page is served by
  Google. Google has deprecated conventional OOB for projects in production
  mode (but it is still allowed for projects in testing mode).
* In gargle's pseudo-OOB, a redirect URI from the configured OAuth client is
  used to receive the response. This page is responsible for exposing a code
  that the user can copy/paste, similar to conventional OOB (except the page is
  *not* served by Google). Unlike conventional OOB, this is not the
  authorization code itself, but is something from which the code can be
  extracted, along with a state token to mitigate cross-site request forgery.
  This is actually implemented using an [OAuth flow for web server applications](https://developers.google.com/identity/protocols/oauth2/web-server).
  Note that we (gargle) call this pseudo-OOB, but it is not technically OOB from
  Google's point-of-view.

The built-in OAuth client used for pseudo-OOB by tidyverse packages redirects to <https://tidyverse.org/google-callback/>.
This is a static landing page that does not collect any data and exists solely to give the interactive R user a way to convey the authorization token back to the waiting R process and thereby complete the auth process.

### More details about the deprecation of conventional OOB

Key links:

* Blog post: [Making Google OAuth interactions safer by using more secure OAuth flows](https://developers.googleblog.com/2022/02/making-oauth-flows-safer.html)
* [Out-Of-Band (OOB) flow Migration Guide](https://developers.google.com/identity/protocols/oauth2/resources/oob-migration#web-application)
* [Using OAuth 2.0 to Access Google APIs](https://developers.google.com/identity/protocols/oauth2)

## But I didn't need OOB yesterday!

Sometimes the usual OAuth web flow suddenly stops working for people working directly with R (so NOT via the browser) and they use OOB auth to get unstuck again.
What's going on in this case?

The initial error looks something like this:

```
createTcpServer: address already in use
Error in httpuv::startServer(use$host, use$port, list(call = listen)) :
  Failed to create server
```

It's characteristic of some other process sitting on port 1410, which is what gargle is trying to use for auth.

It's true that using OOB auth is a workaround.
But OOB auth is, frankly, more clunky, so why use it if you don't have to?
Here are ways to fix this.

  * Restart your system. This will almost certainly kill the offending process,
    which is usually a zombie process.
  * Hunt down the offending process, verify it looks expendable, and kill it.

On *nix-y systems, use `lsof` to get the process ID:

```
sudo lsof -i :1410
```

The output will look something like this:

```
COMMAND   PID  USER   FD   TYPE            DEVICE SIZE/OFF NODE NAME
R       16664 jenny   20u  IPv4 0x63761a50856c65f      0t0  TCP localhost:hiq (LISTEN)
```

In this case, as is typical, this is a zombie R process and I feel confident killing it.
The process ID is listed there as PID.
Note that and kill the process, like so, filling in the PID you found:

```
kill -9 <PID>
```

So, to be clear, in this example, the command would be:

```
kill -9 16664
```

The normal, non-OOB auth web flow should work again now.

## Further reading

If you're working on a data product that will be deployed (for example on [shinyapps.io](https://www.shinyapps.io) or [Posit Connect](https://posit.co/products/enterprise/connect/)), you will also need to consider how the deployed content will authenticate non-interactively, which is covered in `vignette("non-interactive-auth")`.

---
title: "Request helper functions"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Request helper functions}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r}
#| include: false
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

This vignette explains the purpose and usage of:

  * `request_develop(endpoint, params, base_url)`
  * `request_build(method, path, params, body, token, key, base_url)`
  * `request_make(x, ..., user_agent)`
  
The target audience is someone writing an R package to wrap a Google API.

```{r}
#| label: setup
library(gargle)
```

## Why use gargle's request helpers?

Why would the developer of a Google-API-wrapping package care about the request helpers in gargle?

You can write less code and safer code, in return for a modest investment in studying your target API. That is done by ingesting the API's so-called Discovery Document.

Hundreds of Google APIs -- the ones addressed by the [API Discovery Service](https://developers.google.com/discovery/) -- share a great deal of behaviour. By ingesting the metadata provided by this service, you can use gargle's request helpers to exploit this shared data and logic, while also decreasing the chance that you and your users will submit ill-formed requests.

The request helpers in gargle check the combined inputs from user and developer against suitably prepared API metadata:

  * If required parameters are missing, an error is thrown.
  * If unrecognized parameters are submitted, an error is thrown.
  * Parameters are automatically placed in their correct location: URL substitution, query, or body.
  * *Is there something else you care about? It is possible to do more, but it would help to have concrete requests.*
  
Google provides [API libraries for several languages](https://developers.google.com/api-client-library/), including Java, Go, Python, JavaScript, Ruby and more (but not R). All of these libraries are machine-generated from the metadata provided by the API Discovery Service. It is the [official recommendation](https://developers.google.com/discovery/v1/using#build) to use the Discovery Document when building client libraries. The gargle package aims to implement key parts of this strategy, in a way that is also idiomatic for R and its developers.

## High-level design pattern

gargle facilitates this design for API-wrapping packages:

  * A machine-assisted low-level interface driven by the Discovery Document:
    - Your package exports thin wrapper functions around gargle's helpers to form and make HTTP requests, that inject package-specific logic and data, such as an API key and user agent. This is for power users and yourself.
  * High-level, task-oriented, user-facing functions that constitute the main interface of your package.
    - These functions convert user input into the form required by the API and pass it along to your low-level interface functions.
    
Later, specific examples are given, using the googledrive package.

## gargle's HTTP request helpers

gargle provides support for creating and sending HTTP requests via these functions:

`request_develop(endpoint, params, base_url)`: a.k.a. The Smart One.

  * Processes the info in `params` relative to detailed knowledge about the
    `endpoint`, derived from an API Discovery Document.
  * Checks for required and unrecognized parameters.
  * Peels off `params` destined for the body into their own part.
  * Returns request data in a form that anticipates the `httr::VERB()` call
    that is on the horizon.
  
`request_build(method, path, params, body, token, key, base_url)`: a.k.a. The Dumb One.

  * Typically consumes the output of `request_develop()`, although that is not
    required. It can be called directly to enjoy a few luxuries even when making
    one-off API calls in the absence of an ingested Discovery Document.
  * Integrates `params` into a URL via substitution and the query string.
  * Sends either an API key or an OAuth token, but it provides no default values
    or logic for either.

`request_make(x, ..., user_agent)`: actually makes the HTTP request.

  * Typically consumes the output of `request_build()`, although that is not
    required. However, if you have enough info to form a `request_make()`
    request, you would probably just make the `httr::VERB()` call yourself.
  * Consults `x$method` to determine which `httr::VERB()` to call, then calls it
    with the rest of `x`, `...`, and `user_agent` passed as arguments.
  
They are usually called in the above order, though they don't have to be used that way. It is also fine to ignore this part of gargle and use it only for help with auth. They are separate parts of the package.
  
## Discovery Documents

Google's [API Discovery Service](https://developers.google.com/discovery/) "provides a lightweight, JSON-based API that exposes machine-readable metadata about Google APIs". We recommend ingesting this metadata into an R list, stored as internal data in an API-wrapping client package. Then, HTTP requests inside high-level functions can be made concisely and safely, by referring to this metadata. The combined use of this data structure and gargle's request helpers can eliminate a lot of boilerplate data and logic that are shared across Google APIs and across endpoints within an API.

The gargle package ships with some functions and scripts to facilitate the ingest of a Discovery Document. You can find these files in the gargle installation like so:

```{r}
ddi_dir <- system.file("discovery-doc-ingest", package = "gargle")
list.files(ddi_dir)
```

Main files of interest to the developer of a client package:

  * `ingest-functions.R` is a collection of functions for downloading and ingesting a Discovery Document.
  * `drive-example.R` uses those functions to ingest metadata on the Drive v3 API and store it as an internal data object for use in [googledrive](https://googledrive.tidyverse.org).
  
The remaining files present an analysis of the Discovery Document for the Discovery API itself (very meta!) and write files that are useful for reference. Several are included at the end of this vignette.
  
Why aren't the ingest functions exported by gargle? First, we regard this as functionality that is needed at development time, not install or run time. This is something you'll do every few months, probably associated with preparing a release of a wrapper package. Second, the packages that are useful for wrangling JSON and lists are not existing dependencies of gargle, so putting these function in gargle would require some unappealing compromises.

## Method (or endpoint) data

Our Discovery Document ingest process leaves you with an R list. Let's assume it's available in your package's namespace as an internal object named `.endpoints`. Each item represents one method of the API (Google's vocabulary) or an endpoint (gargle's vocabulary).

Each endpoint has an `id`. These `id`s are also used as names for the list. Examples of some `id`s from the Drive and Sheets APIs:

```
drive.about.get
drive.files.create
drive.teamdrives.list
sheets.spreadsheets.create
sheets.spreadsheets.values.clear
sheets.spreadsheets.sheets.copyTo
```

Retrieve the metadata for one endpoint by name, e.g.:

```{r}
#| eval: false
.endpoints[["drive.files.create"]]
```

That info can be passed along to `request_develop(endpoint, params, base_url)`, which conducts sanity checks and combines this external knowledge with the data coming from the user and developer via `params`.

## Design suggestion: forming requests

Here's the model used in googledrive. There is a low-level request helper, `googledrive::request_generate()`, that is used to form every request in the package. It is exported as part of a low-level API for expert use, but most users will never know it exists.

```{r}
#| eval: false
# googledrive::
request_generate <- function(endpoint = character(),
                             params = list(),
                             key = NULL,
                             token = drive_token()) {
  ept <- .endpoints[[endpoint]]
  if (is.null(ept)) {
    stop_glue("\nEndpoint not recognized:\n  * {endpoint}")
  }

  ## modifications specific to googledrive package
  params$key <- key %||% params$key %||% drive_api_key()
  if (!is.null(ept$parameters$supportsTeamDrives)) {
    params$supportsTeamDrives <- TRUE
  }

  req <- gargle::request_develop(endpoint = ept, params = params)
  gargle::request_build(
    path = req$path,
    method = req$method,
    params = req$params,
    body = req$body,
    token = token
  )
}
```

The `endpoint` argument specifies an endpoint by its name, a.k.a. its `id`.

`params` is where the processed user input goes.

`key` and `token` refer to an API key and OAuth2 token, respectively. Both can be populated by default, but it is possible to pass them explicitly. If your package ships with a default API key, you should append it above as the final fallback value for `params$key`.

Do not "borrow" an API key from gargle or another package; always send a key associated with your package or provided by your user. Per the Google User Data Policy <https://developers.google.com/terms/api-services-user-data-policy>, your
application must accurately represent itself when authenticating to Google API services.

After `googledrive::request_generate()` takes care of everything specific to the Drive API and the user's input and task, we call `gargle::request_develop()`. We finish preparing the request with `gargle::request_build()`, which enforces the rule that we always send exactly **one** of `key` and `token`.

## Design suggestion: making requests

The output of `gargle::request_build()` specifies an HTTP request.

`gargle::request_make()` can be used to actually execute it.

```{r}
#| eval: false
# gargle::
request_make <- function(x, ..., user_agent = gargle_user_agent()) {
  stopifnot(is.character(x$method))
  method <- switch(
    x$method,
    GET    = httr::GET,
    POST   = httr::POST,
    PATCH  = httr::PATCH,
    PUT    = httr::PUT,
    DELETE = httr::DELETE,
    abort(glue("Not a recognized HTTP method: {bt(x$method)}"))
  )
  method(
    url = x$url,
    body = x$body,
    x$token,
    user_agent,
    ...
  )
}
```

`request_make()` consults `x$method` to identify the `httr::VERB()` and then calls it with the remainder of `x`, `...` and the `user_agent`.

In googledrive we have a thin wrapper around this that injects the googledrive user agent:

```{r}
#| eval: false
# googledrive::
request_make <- function(x, ...) {
  gargle::request_make(x, ..., user_agent = drive_ua())
}
```

## Reference

*derived from the Discovery Document for the Discovery Service*

Properties of an endpoint

```{r}
#| asis: true
#| echo: false
#| comment: NA
cat(readLines(fs::path(ddi_dir, "method-properties-humane.txt")), sep = "\n")
```

API-wide endpoint parameters (taken from Discovery API but, empirically, are shared with other APIs):

```{r}
#| asis: true
#| echo: false
#| comment: NA
cat(readLines(fs::path(ddi_dir, "api-wide-parameters-humane.txt")), sep = "\n")
```

Properties of an endpoint parameters:

```{r}
#| asis: true
#| echo: false
#| comment: NA
cat(readLines(fs::path(ddi_dir, "parameter-properties-humane.txt")), sep = "\n")
```

# use_oob must be TRUE or FALSE

    Code
      check_oob("a")
    Condition
      Error in `check_oob()`:
      ! `use_oob` must be `TRUE` or `FALSE`, not a string.

---

    Code
      check_oob(c(FALSE, FALSE))
    Condition
      Error in `check_oob()`:
      ! `use_oob` must be `TRUE` or `FALSE`, not a logical vector.

# OOB requires an interactive session

    Code
      check_oob(TRUE)
    Condition
      Error in `check_oob()`:
      ! Out-of-band auth only works in an interactive session.

# makes no sense to pass oob_value if not OOB

    Code
      check_oob(FALSE, "custom_value")
    Condition
      Error in `check_oob()`:
      ! The `oob_value` argument can only be used when `use_oob = TRUE`.

# oob_value has to be a string

    Code
      check_oob(TRUE, c("a", "b"))
    Condition
      Error in `check_oob()`:
      ! Out-of-band auth only works in an interactive session.

# Resource exhausted (Sheets, ReadGroup)

    Code
      response_process(resp)
    Condition
      Error in `expect_recorded_error()`:
      ! Client error: (429) RESOURCE_EXHAUSTED
      * Either out of resource quota or reaching rate limiting. The client should look for google.rpc.QuotaFailure error detail for more information.
      * Quota exceeded for quota metric 'Read requests' and limit 'Read requests per minute per user' of service 'sheets.googleapis.com' for consumer 'project_number:603366585132'.
      
      Error details:
      * reason: RATE_LIMIT_EXCEEDED
      * domain: googleapis.com
      * metadata.quota_limit_value: 60
      * metadata.service: sheets.googleapis.com
      * metadata.consumer: projects/603366585132
      * metadata.quota_unit: 1/min/{project}/{user}
      * metadata.quota_location: global
      * metadata.quota_metric: sheets.googleapis.com/read_requests
      * metadata.quota_limit: ReadRequestsPerMinutePerUser
      Links
      * Description: Request a higher quota limit.
        URL: https://cloud.google.com/docs/quotas/help/request_increase

# Request for non-existent resource (Drive)

    Code
      response_process(resp)
    Condition
      Error in `expect_recorded_error()`:
      ! Client error: (404) Not Found
      File not found: NOPE_NOT_A_GOOD_ID.
      * domain: global
      * reason: notFound
      * message: File not found: NOPE_NOT_A_GOOD_ID.
      * locationType: parameter
      * location: fileId

# Too many requests (Drive, HTML content)

    Code
      response_process(resp)
    Condition
      Error in `expect_recorded_error()`:
      ! Client error: (429) Too Many Requests (RFC 6585)
      x Expected content type 'application/json', not 'text/html'.
      i See 'VOLATILE_FILE_PATH' for the html error content.
      i Or execute `browseURL("VOLATILE_FILE_PATH")` to view it in your browser.

# HTML error is offered as a file

    Code
      strwrap(readLines(path_to_html_error), width = 60)
    Output
       [1] "<html><head><meta http-equiv=\"content-type\""                    
       [2] "content=\"text/html;"                                             
       [3] "charset=utf-8\"/><title>Sorry...</title><style> body {"           
       [4] "font-family: verdana, arial, sans-serif; background-color:"       
       [5] "#fff; color: #000;"                                               
       [6] "}</style></head><body><div><table><tr><td><b><font"               
       [7] "face=sans-serif size=10><font color=#4285f4>G</font><font"        
       [8] "color=#ea4335>o</font><font color=#fbbc05>o</font><font"          
       [9] "color=#4285f4>g</font><font color=#34a853>l</font><font"          
      [10] "color=#ea4335>e</font></font></b></td><td"                        
      [11] "style=\"text-align: left; vertical-align: bottom;"                
      [12] "padding-bottom: 15px; width: 50%\"><div"                          
      [13] "style=\"border-bottom: 1px solid"                                 
      [14] "#dfdfdf;\">Sorry...</div></td></tr></table></div><div"            
      [15] "style=\"margin-left: 4em;\"><h1>We're sorry...</h1><p>... but"    
      [16] "your computer or network may be sending automated queries."       
      [17] "To protect our users, we can't process your request right"        
      [18] "now.</p></div><div style=\"margin-left: 4em;\">See <a"            
      [19] "href=\"https://support.google.com/websearch/answer/86640\">Google"
      [20] "Help</a> for more information.<br/><br/></div><div"               
      [21] "style=\"text-align: center; border-top: 1px solid"                
      [22] "#dfdfdf;\"><a href=\"https://www.google.com\">Google"             
      [23] "Home</a></div></body></html>"                                     

# Request for which we don't have scope (Fitness)

    Code
      response_process(resp)
    Condition
      Error in `expect_recorded_error()`:
      ! Client error: (403) Forbidden
      Request had insufficient authentication scopes.
      PERMISSION_DENIED
      * message: Insufficient Permission
      * domain: global
      * reason: insufficientPermissions

# Use key that's not enabled for the API (Sheets)

    Code
      response_process(resp)
    Condition
      Error in `expect_recorded_error()`:
      ! Client error: (403) PERMISSION_DENIED
      * Client does not have sufficient permission. This can happen because the OAuth token does not have the right scopes, the client doesn't have permission, or the API has not been enabled for the client project.
      * Google Sheets API has not been used in project 977449744253 before or it is disabled. Enable it by visiting https://console.developers.google.com/apis/api/sheets.googleapis.com/overview?project=977449744253 then retry. If you enabled this API recently, wait a few minutes for the action to propagate to our systems and retry.
      
      Error details:
      Links
      * Description: Google developers console API activation
        URL: https://console.developers.google.com/apis/api/sheets.googleapis.com/overview?project=977449744253
      * reason: SERVICE_DISABLED
      * domain: googleapis.com
      * metadata.consumer: projects/977449744253
      * metadata.service: sheets.googleapis.com

# Request with invalid argument (Sheets, bad range)

    Code
      response_process(resp)
    Condition
      Error in `expect_recorded_error()`:
      ! Client error: (400) INVALID_ARGUMENT
      * Client specified an invalid argument. Check error message and error details for more information.
      * Unable to parse range: NOPE!A5:F15

# Request with bad field mask (Sheets)

    Code
      response_process(resp)
    Condition
      Error in `expect_recorded_error()`:
      ! Client error: (400) INVALID_ARGUMENT
      * Client specified an invalid argument. Check error message and error details for more information.
      * Request contains an invalid argument.
      
      Error details:
      Field violations
      * Field: sheets.sheetProperties
        Description: Error expanding 'fields' parameter. Cannot find matching fields for path 'sheets.sheetProperties'.

# Request for nonexistent resource (Sheets)

    Code
      response_process(resp)
    Condition
      Error in `expect_recorded_error()`:
      ! Client error: (404) NOT_FOUND
      * A specified resource is not found, or the request is rejected by undisclosed reasons, such as whitelisting.
      * Requested entity was not found.

# Use service account that's not enabled for the API (Sheets)

    Code
      response_process(resp)
    Condition
      Error in `expect_recorded_error()`:
      ! Client error: (403) PERMISSION_DENIED
      * Client does not have sufficient permission. This can happen because the OAuth token does not have the right scopes, the client doesn't have permission, or the API has not been enabled for the client project.
      * Google Sheets API has not been used in project 1234567890 before or it is disabled. Enable it by visiting https://console.developers.google.com/apis/api/sheets.googleapis.com/overview?project=1234567890 then retry. If you enabled this API recently, wait a few minutes for the action to propagate to our systems and retry.
      
      Error details:
      * reason: SERVICE_DISABLED
      * domain: googleapis.com
      * metadata.serviceTitle: Google Sheets API
      * metadata.consumer: projects/1234567890
      * metadata.activationUrl: https://console.developers.google.com/apis/api/sheets.googleapis.com/overview?project=1234567890
      * metadata.containerInfo: 1234567890
      * metadata.service: sheets.googleapis.com
      Links
      * Description: Google developers console API activation
        URL: https://console.developers.google.com/apis/api/sheets.googleapis.com/overview?project=1234567890

# Request with invalid value (tokeninfo, stale token)

    Code
      response_process(resp)
    Condition
      Error in `expect_recorded_error()`:
      ! Client error: (400) Bad Request
      * Invalid Value

# Request to bad URL (tokeninfo, HTML content)

    Code
      response_process(resp)
    Condition
      Error in `expect_recorded_error()`:
      ! Client error: (404) Not Found
      x Expected content type 'application/json', not 'text/html'.
      i See 'VOLATILE_FILE_PATH' for the html error content.
      i Or execute `browseURL("VOLATILE_FILE_PATH")` to view it in your browser.

# request_retry() logic works as advertised

    Code
      fail_then_succeed <- request_retry(max_total_wait_time_in_seconds = 5)
    Message
      i Exponential base for retries is {BASE}s.
      x Request 1 failed [429: RESOURCE_EXHAUSTED].
      i Will retry in {WAIT_TIME}s.
      i Wait time strategy: exponential backoff, full jitter
        PLACEHOLDER FOR GOOGLE ERROR MESSAGE
      v Request 2 successful!

---

    Code
      fail_then_succeed <- request_retry()
    Message
      i Exponential base for retries is {BASE}s.
      x Request 1 failed [429: RESOURCE_EXHAUSTED].
      i Will retry in {WAIT_TIME}s.
      i Wait time strategy: 'Retry-After' header
        PLACEHOLDER FOR GOOGLE ERROR MESSAGE
      v Request 2 successful!

---

    Code
      fail_max_tries <- request_retry(max_tries_total = 3,
        max_total_wait_time_in_seconds = 6)
    Message
      i Exponential base for retries is {BASE}s.
      x Request 1 failed [429: RESOURCE_EXHAUSTED].
      i Will retry in {WAIT_TIME}s.
      i Wait time strategy: exponential backoff, full jitter
        PLACEHOLDER FOR GOOGLE ERROR MESSAGE
      x Request 2 failed [429: RESOURCE_EXHAUSTED].
      i Will retry in {WAIT_TIME}s.
      i Wait time strategy: exponential backoff, full jitter
        PLACEHOLDER FOR GOOGLE ERROR MESSAGE
      v Request 3 failed :(

# cache_clean() works

    Code
      cache_clean(cache_folder, "apple")
    Message
      v Deleting 1 token obtained with an old tidyverse OAuth client.
      i Expect interactive prompts to re-auth with the new client.
      ! Is this rolling of credentials highly disruptive to your workflow?
        That means you should rely on your own OAuth client (or switch to a service
        account token).
        Learn more these in these articles:
        <https://gargle.r-lib.org/articles/get-api-credentials.html>
        <https://gargle.r-lib.org/articles/non-interactive-auth.html>
    Output
      [1] TRUE

# cache_load() repairs tokens stored with names != their hash

    Code
      writeLines(out)
    Output
      ! Cache contains tokens with names that do not match their hash:
      * "abc123_c@example.org" (name)
      * '{TOKEN_HASH}_a@example.org' (hash)
      * "def456_d@example.org" (name)
      * '{TOKEN_HASH}_b@example.org' (hash)
        Will attempt to repair by renaming

# token_match() finds a match based on domain

    Code
      m <- token_match("abc_*@example.org", one_match_of_two)
    Message
      i The gargle package is using a cached token for 'jane@example.org'.

# token_match() scolds but returns short hash match when non-interactive

    Code
      m <- token_match("abc_", one_existing)
    Message
      ! Using an auto-discovered, cached token.
        To suppress this message, modify your code or options to clearly consent to
        the use of a cached token.
        See gargle's "Non-interactive auth" vignette for more details:
        <https://gargle.r-lib.org/articles/non-interactive-auth.html>
      i The gargle package is using a cached token for 'a@example.com'.

---

    Code
      m <- token_match("abc_*", one_existing)
    Message
      ! Using an auto-discovered, cached token.
        To suppress this message, modify your code or options to clearly consent to
        the use of a cached token.
        See gargle's "Non-interactive auth" vignette for more details:
        <https://gargle.r-lib.org/articles/non-interactive-auth.html>
      i The gargle package is using a cached token for 'a@example.com'.

---

    Code
      m <- token_match("abc_", two_existing)
    Message
      i Suitable tokens found in the cache, associated with these emails:
      * 'a@example.com'
      * 'b@example.com'
        Defaulting to the first email.
      ! Using an auto-discovered, cached token.
        To suppress this message, modify your code or options to clearly consent to
        the use of a cached token.
        See gargle's "Non-interactive auth" vignette for more details:
        <https://gargle.r-lib.org/articles/non-interactive-auth.html>
      i The gargle package is using a cached token for 'a@example.com'.

---

    Code
      m <- token_match("abc_*", two_existing)
    Message
      i Suitable tokens found in the cache, associated with these emails:
      * 'a@example.com'
      * 'b@example.com'
        Defaulting to the first email.
      ! Using an auto-discovered, cached token.
        To suppress this message, modify your code or options to clearly consent to
        the use of a cached token.
        See gargle's "Non-interactive auth" vignette for more details:
        <https://gargle.r-lib.org/articles/non-interactive-auth.html>
      i The gargle package is using a cached token for 'a@example.com'.

# gargle_oauth_sitrep() works with a cache

    Code
      writeLines(out)
    Output
      2 tokens found in this gargle OAuth cache:
      '{path to gargle oauth cache}'
      
      email         app         scopes hash...   
      _____________ ___________ ______ __________
      a@example.org gargle-clio        {hash...}
      b@example.org gargle-clio        {hash...}

# gargle_oauth_sitrep() consults the option for cache location

    Code
      writeLines(out)
    Output
      i Taking cache location from the `"gargle_oauth_cache"` option.
      1 token found in this gargle OAuth cache:
      '{path to gargle oauth cache}'
      
      email         app         scopes hash...   
      _____________ ___________ ______ __________
      a@example.org gargle-clio        {hash...}

# gargle_verbosity() validates the value it finds

    Option "gargle_verbosity" must be one of: 'debug', 'info', and 'silent'

# gargle_verbosity() accomodates people using the old option

    Code
      out <- gargle_verbosity()
    Message
      ! Option "gargle_quiet" is deprecated in favor of "gargle_verbosity"
      i Instead of: `options(gargle_quiet = FALSE)`
        Now do: `options(gargle_verbosity = "debug")`

# gargle_info() works

    Code
      gargle_info(c("aa {.field {blah}} bb", "cc {.emph xyz} dd"))
    Message
      aa 'BLAH' bb
      cc xyz dd

---

    Code
      gargle_info(c("ee {.field {blah}} ff", "gg {.emph xyz} hh"))
    Message
      ee 'BLAH' ff
      gg xyz hh

---

    Code
      gargle_info(c("ii {.field {blah}} jj", "kk {.emph xyz} ll"))

# gargle_debug() works

    Code
      gargle_debug(c("11 {.field {foo}} 22", "33 {.file a/b/c} 44"))
    Message
      11 'FOO' 22
      33 'a/b/c' 44

---

    Code
      gargle_debug(c("55 {.field {foo}} 66", "77 {.file a/b/c} 88"))

---

    Code
      gargle_debug(c("99 {.field {foo}} 00", "11 {.file a/b/c} 22"))

# bulletize() works

    Code
      cli::cli_bullets(bulletize(letters))
    Message
      * a
      * b
      * c
      * d
      * e
        ... and 21 more

---

    Code
      cli::cli_bullets(bulletize(letters, bullet = "x"))
    Message
      x a
      x b
      x c
      x d
      x e
        ... and 21 more

---

    Code
      cli::cli_bullets(bulletize(letters, n_show = 2))
    Message
      * a
      * b
        ... and 24 more

---

    Code
      cli::cli_bullets(bulletize(letters[1:6]))
    Message
      * a
      * b
      * c
      * d
      * e
      * f

---

    Code
      cli::cli_bullets(bulletize(letters[1:7]))
    Message
      * a
      * b
      * c
      * d
      * e
      * f
      * g

---

    Code
      cli::cli_bullets(bulletize(letters[1:8]))
    Message
      * a
      * b
      * c
      * d
      * e
        ... and 3 more

---

    Code
      cli::cli_bullets(bulletize(letters[1:6], n_fudge = 0))
    Message
      * a
      * b
      * c
      * d
      * e
        ... and 1 more

---

    Code
      cli::cli_bullets(bulletize(letters[1:8], n_fudge = 3))
    Message
      * a
      * b
      * c
      * d
      * e
      * f
      * g
      * h

# check_is_service_account() errors for OAuth client

    Code
      PKG_auth(fs::path_package("gargle", "extdata",
        "client_secret_installed.googleusercontent.com.json"))
    Condition
      Error in `PKG_auth()`:
      ! `path` does not represent a service account.
      i If `path` is meant to be a filepath, perhaps the file does not exist?
      i Did you provide the JSON for an OAuth client instead of for a service account?
      i To configure the OAuth client, use `PKG_auth_configure()` instead.

# check_is_service_account() errors for invalid input

    Code
      PKG_auth("wut")
    Condition
      Error in `PKG_auth()`:
      ! `path` does not represent a service account.
      i If `path` is meant to be a filepath, perhaps the file does not exist?
      i Did you provide the JSON for an OAuth client instead of for a service account?
      i To configure the OAuth client, use `PKG_auth_configure()` instead.

# AuthState prints nicely

    Code
      print(a)
    Output
      
      -- <AuthState (via gargle)> ----------------------------------------------------
          package: PKG
              app: APPNAME
          api_key: API_KEY
      auth_active: TRUE
      credentials: <some_sort_of_token>

# Gargle2.0 prints nicely

    Code
      print(fauxen)
    Output
      
      -- <Token (via gargle)> --------------------------------------------------------
      oauth_endpoint: google
                 app: APPNAME
               email: 'a@example.org'
              scopes: ...userinfo.email
         credentials: a

# we reject redirect URIs from conventional OOB for pseudo-OOB flow

    Code
      select_pseudo_oob_value("urn:ietf:wg:oauth:2.0:oob")
    Condition
      Error in `select_pseudo_oob_value()`:
      ! OAuth client (a.k.a "app") does not have a redirect URI suitable for the pseudo-OOB flow.

# we reject local web server redirect URIs for pseudo-OOB flow

    Code
      select_pseudo_oob_value("http://localhost")
    Condition
      Error in `select_pseudo_oob_value()`:
      ! OAuth client (a.k.a "app") does not have a redirect URI suitable for the pseudo-OOB flow.

# we insist on finding exactly one redirect URI for pseudo-OOB flow

    Code
      select_pseudo_oob_value(redirect_uris)
    Condition
      Error in `select_pseudo_oob_value()`:
      ! Can't determine which redirect URI to use for the pseudo-OOB flow:
      * https://example.com/google-callback/one.html
      * https://example.com/google-callback/two.html

# gargle_oauth_client() rejects bad input

    Code
      gargle_oauth_client()
    Condition
      Error in `gargle_oauth_client()`:
      ! `id` must be a single string, not absent.

---

    Code
      gargle_oauth_client(1234)
    Condition
      Error in `gargle_oauth_client()`:
      ! `id` must be a single string, not a number.

---

    Code
      gargle_oauth_client(id = "ID")
    Condition
      Error in `gargle_oauth_client()`:
      ! `secret` must be a single string, not absent.

---

    Code
      gargle_oauth_client(id = "ID", secret = 1234)
    Condition
      Error in `gargle_oauth_client()`:
      ! `secret` must be a single string, not a number.

---

    Code
      gargle_oauth_client("ID", "SECRET", type = "nope")
    Condition
      Error in `gargle_oauth_client()`:
      ! `type` must be one of "installed" or "web", not "nope".

# gargle_oauth_client() has special handling for web clients

    Code
      gargle_oauth_client("ID", "SECRET", type = "web")
    Condition
      Error in `gargle_oauth_client()`:
      ! A "web" type OAuth client must have one or more 'redirect_uris'.

---

    Code
      gargle_oauth_client("ID", "SECRET", type = "web", redirect_uris = c(
        "http://localhost:8111/", "http://127.0.0.1:8100/",
        "https://example.com/aaa/bbb/v"))
    Message
      <gargle_oauth_client>
      name: 7f82e05dfbeb26a264621f1482a14e25
      id: ID
      secret: <REDACTED>
      type: web
      redirect_uris: http://localhost:8111/, http://127.0.0.1:8100/,
      https://example.com/aaa/bbb/v

# service account JSON throws an informative error

    Code
      gargle_oauth_client_from_json(test_path("fixtures",
        "service-account-token.json"))
    Condition
      Error in `gargle_oauth_client_from_json()`:
      ! JSON has an unexpected form
      i Are you sure this is the JSON downloaded for an OAuth client?
      i It is easy to confuse the JSON for an OAuth client and a service account.

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/credentials_external_account.R
\name{credentials_external_account}
\alias{credentials_external_account}
\title{Get a token for an external account}
\usage{
credentials_external_account(
  scopes = "https://www.googleapis.com/auth/cloud-platform",
  path = "",
  ...
)
}
\arguments{
\item{scopes}{A character vector of scopes to request. Pick from those listed
at \url{https://developers.google.com/identity/protocols/oauth2/scopes}.

For certain token flows, the
\code{"https://www.googleapis.com/auth/userinfo.email"} scope is unconditionally
included. This grants permission to retrieve the email address associated
with a token; gargle uses this to index cached OAuth tokens. This grants no
permission to view or send email and is generally considered a low-value
scope.}

\item{path}{JSON containing the workload identity configuration for the
external account, in one of the forms supported for the \code{txt} argument of
\code{\link[jsonlite:fromJSON]{jsonlite::fromJSON()}} (probably, a file path, although it could be a JSON
string). The instructions for generating this configuration are given at
\href{https://cloud.google.com/iam/docs/configuring-workload-identity-federation}{Configuring workload identity federation}.

Note that external account tokens are a natural fit for use as Application
Default Credentials, so consider storing the configuration file in one of
the standard locations consulted for ADC, instead of providing \code{path}
explicitly. See \code{\link[=credentials_app_default]{credentials_app_default()}} for more.}

\item{...}{Additional arguments passed to all credential functions.}
}
\value{
A \code{\link[=WifToken]{WifToken()}} or \code{NULL}.
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}}
Workload identity federation is a new (as of April 2021) keyless
authentication mechanism that allows applications running on a non-Google
Cloud platform, such as AWS, to access Google Cloud resources without using a
conventional service account token. This eliminates the dilemma of how to
safely manage service account credential files.

Unlike service accounts, the configuration file for workload identity
federation contains no secrets. Instead, it holds non-sensitive metadata.
The external application obtains the needed sensitive data "on-the-fly" from
the running instance. The combined data is then used to obtain a so-called
subject token from the external identity provider, such as AWS. This is then
sent to Google's Security Token Service API, in exchange for a very
short-lived federated access token. Finally, the federated access token is
sent to Google's Service Account Credentials API, in exchange for a
short-lived GCP access token. This access token allows the external
application to impersonate a service account and inherit the permissions of
the service account to access GCP resources.

This feature is still experimental in gargle and \strong{currently only supports
AWS}. It also requires installation of the suggested packages
\pkg{aws.signature} and \pkg{aws.ec2metadata}. Workload identity federation
\strong{can} be used with other platforms, such as Microsoft Azure or any
identity provider that supports OpenID Connect. If you would like gargle to
support this token flow for additional platforms, please \href{https://github.com/r-lib/gargle/issues}{open an issue on GitHub} and describe your use case.
}
\examples{
\dontrun{
credentials_external_account()
}
}
\seealso{
There is substantial setup necessary, both on the GCP and AWS side,
to use this authentication method. These two links provide, respectively,
a high-level overview and step-by-step instructions.
\itemize{
\item \url{https://cloud.google.com/blog/products/identity-security/enable-keyless-access-to-gcp-with-workload-identity-federation/}
\item \url{https://cloud.google.com/iam/docs/configuring-workload-identity-federation}
}

Other credential functions: 
\code{\link{credentials_app_default}()},
\code{\link{credentials_byo_oauth2}()},
\code{\link{credentials_gce}()},
\code{\link{credentials_service_account}()},
\code{\link{credentials_user_oauth2}()},
\code{\link{token_fetch}()}
}
\concept{credential functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/credentials_user_oauth2.R
\name{credentials_user_oauth2}
\alias{credentials_user_oauth2}
\title{Get an OAuth token for a user}
\usage{
credentials_user_oauth2(
  scopes = NULL,
  client = gargle_client(),
  package = "gargle",
  ...,
  app = deprecated()
)
}
\arguments{
\item{scopes}{A character vector of scopes to request. Pick from those listed
at \url{https://developers.google.com/identity/protocols/oauth2/scopes}.

For certain token flows, the
\code{"https://www.googleapis.com/auth/userinfo.email"} scope is unconditionally
included. This grants permission to retrieve the email address associated
with a token; gargle uses this to index cached OAuth tokens. This grants no
permission to view or send email and is generally considered a low-value
scope.}

\item{client}{A Google OAuth client, preferably constructed via
\code{\link[=gargle_oauth_client_from_json]{gargle_oauth_client_from_json()}}, which returns an instance of
\code{gargle_oauth_client}. For backwards compatibility, for a limited time,
gargle will still accept an "OAuth app" created with \code{\link[httr:oauth_app]{httr::oauth_app()}}.}

\item{package}{Name of the package requesting a token. Used in messages.}

\item{...}{
  Arguments passed on to \code{\link[=gargle2.0_token]{gargle2.0_token}}
  \describe{
    \item{\code{email}}{Optional. If specified, \code{email} can take several different
forms:
\itemize{
\item \code{"jane@gmail.com"}, i.e. an actual email address. This allows the user to
target a specific Google identity. If specified, this is used for token
lookup, i.e. to determine if a suitable token is already available in the
cache. If no such token is found, \code{email} is used to pre-select the targeted
Google identity in the OAuth chooser. (Note, however, that the email
associated with a token when it's cached is always determined from the token
itself, never from this argument).
\item \code{"*@example.com"}, i.e. a domain-only glob pattern. This can be helpful if
you need code that "just works" for both \code{alice@example.com} and
\code{bob@example.com}.
\item \code{TRUE} means that you are approving email auto-discovery. If exactly one
matching token is found in the cache, it will be used.
\item \code{FALSE} or \code{NA} mean that you want to ignore the token cache and force a
new OAuth dance in the browser.
}

Defaults to the option named \code{"gargle_oauth_email"}, retrieved by
\code{\link[=gargle_oauth_email]{gargle_oauth_email()}} (unless a wrapper package implements different
default behavior).}
    \item{\code{use_oob}}{Whether to use out-of-band authentication (or, perhaps, a
variant implemented by gargle and known as "pseudo-OOB") when first
acquiring the token. Defaults to the value returned by
\code{\link[=gargle_oob_default]{gargle_oob_default()}}. Note that (pseudo-)OOB auth only affects
the initial OAuth dance. If we retrieve (and possibly refresh) a
cached token, \code{use_oob} has no effect.

If the OAuth client is provided implicitly by a wrapper package, its type
probably defaults to the value returned by
\code{\link[=gargle_oauth_client_type]{gargle_oauth_client_type()}}. You can take control of the client
type by setting \code{options(gargle_oauth_client_type = "web")} or
\code{options(gargle_oauth_client_type = "installed")}.}
    \item{\code{cache}}{Specifies the OAuth token cache. Defaults to the option named
\code{"gargle_oauth_cache"}, retrieved via \code{\link[=gargle_oauth_cache]{gargle_oauth_cache()}}.}
    \item{\code{credentials}}{Advanced use only: allows you to completely customise
token generation.}
  }}

\item{app}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Replaced by the \code{client}
argument.}
}
\value{
A \link{Gargle2.0} token.
}
\description{
Consults the token cache for a suitable OAuth token and, if
unsuccessful, gets a token via the browser flow. A cached token is suitable
if it's compatible with the user's request in this sense:
\itemize{
\item OAuth client must be same.
\item Scopes must be same.
\item Email, if provided, must be same. If specified email is a glob pattern
like \code{"*@example.com"}, email matching is done at the domain level.
}

gargle is very conservative about using OAuth tokens discovered in the user's
cache and will generally seek interactive confirmation. Therefore, in a
non-interactive setting, it's important to explicitly specify the \code{"email"}
of the target account or to explicitly authorize automatic discovery. See
\code{\link[=gargle2.0_token]{gargle2.0_token()}}, which this function wraps, for more. Non-interactive use
also suggests it might be time to use a \link[=credentials_service_account]{service account token} or \link[=credentials_external_account]{workload identity federation}.
}
\examples{
\dontrun{
# Drive scope, built-in gargle demo client
scopes <- "https://www.googleapis.com/auth/drive"
credentials_user_oauth2(scopes, client = gargle_client())

# bring your own client
client <- gargle_oauth_client_from_json(
  path = "/path/to/the/JSON/you/downloaded/from/gcp/console.json",
  name = "my-nifty-oauth-client"
)
credentials_user_oauth2(scopes, client)
}
}
\seealso{
Other credential functions: 
\code{\link{credentials_app_default}()},
\code{\link{credentials_byo_oauth2}()},
\code{\link{credentials_external_account}()},
\code{\link{credentials_gce}()},
\code{\link{credentials_service_account}()},
\code{\link{token_fetch}()}
}
\concept{credential functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/AuthState-class.R
\name{AuthState-class}
\alias{AuthState-class}
\alias{AuthState}
\title{Authorization state}
\description{
An \code{AuthState} object manages an authorization state, typically on behalf of
a wrapper package that makes requests to a Google API.

The \verb{vignette("gargle-auth-in-client-package)} describes a design for wrapper
packages that relies on an \code{AuthState} object. This state can then be
incorporated into the package's requests for tokens and can control the
inclusion of tokens in requests to the target API.
\itemize{
\item \code{api_key} is the simplest way to associate a request with a specific
Google Cloud Platform \href{https://docs.cloud.google.com/resource-manager/docs/cloud-platform-resource-hierarchy}{project}.
A few calls to certain APIs, e.g. reading a public Sheet, can succeed
with an API key, but this is the exception.
\item \code{client} is an OAuth client ID (and secret) associated with a specific
Google Cloud Platform \href{https://docs.cloud.google.com/resource-manager/docs/cloud-platform-resource-hierarchy}{project}.
This is used in the OAuth flow, in which an authenticated user authorizes
the client to access or manipulate data on their behalf.
\item \code{auth_active} reflects whether outgoing requests will be authorized by an
authenticated user or are unauthorized requests for public resources.
These two states correspond to sending a request with a token versus an
API key, respectively.
\item \code{cred} is where the current token is cached within a session, once one
has been fetched. It is generally assumed to be an instance of
\code{\link[httr:Token-class]{httr::TokenServiceAccount}} or
\code{\link[httr:Token-class]{httr::Token2.0}} (or a subclass thereof), probably
obtained via \code{\link[=token_fetch]{token_fetch()}} (or one of its constituent credential
fetching functions).
}

An \code{AuthState} should be created through the constructor function
\code{\link[=init_AuthState]{init_AuthState()}}, which has more details on the arguments.
}
\section{Public fields}{
\if{html}{\out{<div class="r6-fields">}}
\describe{
\item{\code{package}}{Package name.}

\item{\code{client}}{An OAuth client.}

\item{\code{app}}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use \code{client} instead.}

\item{\code{api_key}}{An API key.}

\item{\code{auth_active}}{Logical, indicating whether auth is active.}

\item{\code{cred}}{Credentials.}
}
\if{html}{\out{</div>}}
}
\section{Methods}{
\subsection{Public methods}{
\itemize{
\item \href{#method-AuthState-new}{\code{AuthState$new()}}
\item \href{#method-AuthState-format}{\code{AuthState$format()}}
\item \href{#method-AuthState-set_client}{\code{AuthState$set_client()}}
\item \href{#method-AuthState-set_app}{\code{AuthState$set_app()}}
\item \href{#method-AuthState-set_api_key}{\code{AuthState$set_api_key()}}
\item \href{#method-AuthState-set_auth_active}{\code{AuthState$set_auth_active()}}
\item \href{#method-AuthState-set_cred}{\code{AuthState$set_cred()}}
\item \href{#method-AuthState-clear_cred}{\code{AuthState$clear_cred()}}
\item \href{#method-AuthState-get_cred}{\code{AuthState$get_cred()}}
\item \href{#method-AuthState-has_cred}{\code{AuthState$has_cred()}}
\item \href{#method-AuthState-clone}{\code{AuthState$clone()}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-AuthState-new"></a>}}
\if{latex}{\out{\hypertarget{method-AuthState-new}{}}}
\subsection{Method \code{new()}}{
Create a new AuthState
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{AuthState$new(
  package = NA_character_,
  client = NULL,
  api_key = NULL,
  auth_active = TRUE,
  cred = NULL,
  app = deprecated()
)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{package}}{Package name.}

\item{\code{client}}{An OAuth client.}

\item{\code{api_key}}{An API key.}

\item{\code{auth_active}}{Logical, indicating whether auth is active.}

\item{\code{cred}}{Credentials.}

\item{\code{app}}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use \code{client} instead.}
}
\if{html}{\out{</div>}}
}
\subsection{Details}{
For more details on the parameters, see \code{\link[=init_AuthState]{init_AuthState()}}
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-AuthState-format"></a>}}
\if{latex}{\out{\hypertarget{method-AuthState-format}{}}}
\subsection{Method \code{format()}}{
Format an AuthState
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{AuthState$format(...)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{...}}{Not used.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-AuthState-set_client"></a>}}
\if{latex}{\out{\hypertarget{method-AuthState-set_client}{}}}
\subsection{Method \code{set_client()}}{
Set the OAuth client
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{AuthState$set_client(client)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{client}}{An OAuth client.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-AuthState-set_app"></a>}}
\if{latex}{\out{\hypertarget{method-AuthState-set_app}{}}}
\subsection{Method \code{set_app()}}{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Deprecated method to set
the OAuth client
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{AuthState$set_app(app)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{app}}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use \code{client} instead.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-AuthState-set_api_key"></a>}}
\if{latex}{\out{\hypertarget{method-AuthState-set_api_key}{}}}
\subsection{Method \code{set_api_key()}}{
Set the API key
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{AuthState$set_api_key(value)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{value}}{An API key.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-AuthState-set_auth_active"></a>}}
\if{latex}{\out{\hypertarget{method-AuthState-set_auth_active}{}}}
\subsection{Method \code{set_auth_active()}}{
Set whether auth is (in)active
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{AuthState$set_auth_active(value)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{value}}{Logical, indicating whether to send requests authorized with
user credentials.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-AuthState-set_cred"></a>}}
\if{latex}{\out{\hypertarget{method-AuthState-set_cred}{}}}
\subsection{Method \code{set_cred()}}{
Set credentials
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{AuthState$set_cred(cred)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{cred}}{User credentials.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-AuthState-clear_cred"></a>}}
\if{latex}{\out{\hypertarget{method-AuthState-clear_cred}{}}}
\subsection{Method \code{clear_cred()}}{
Clear credentials
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{AuthState$clear_cred()}\if{html}{\out{</div>}}
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-AuthState-get_cred"></a>}}
\if{latex}{\out{\hypertarget{method-AuthState-get_cred}{}}}
\subsection{Method \code{get_cred()}}{
Get credentials
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{AuthState$get_cred()}\if{html}{\out{</div>}}
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-AuthState-has_cred"></a>}}
\if{latex}{\out{\hypertarget{method-AuthState-has_cred}{}}}
\subsection{Method \code{has_cred()}}{
Report if we have credentials
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{AuthState$has_cred()}\if{html}{\out{</div>}}
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-AuthState-clone"></a>}}
\if{latex}{\out{\hypertarget{method-AuthState-clone}{}}}
\subsection{Method \code{clone()}}{
The objects of this class are cloneable with this method.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{AuthState$clone(deep = FALSE)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{deep}}{Whether to make a deep clone.}
}
\if{html}{\out{</div>}}
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/secret.R
\name{gargle_secret}
\alias{gargle_secret}
\alias{secret_encrypt_json}
\alias{secret_decrypt_json}
\alias{secret_make_key}
\alias{secret_write_rds}
\alias{secret_read_rds}
\alias{secret_has_key}
\title{Encrypt/decrypt JSON or an R object}
\usage{
secret_encrypt_json(json, path = NULL, key)

secret_decrypt_json(path, key)

secret_make_key()

secret_write_rds(x, path, key)

secret_read_rds(path, key)

secret_has_key(key)
}
\arguments{
\item{json}{A JSON file (or string).}

\item{path}{The path to write to (\code{secret_encrypt_json()},
\code{secret_write_rds()}) or to read from (\code{secret_decrypt_json()},
\code{secret_read_rds()}).}

\item{key}{Encryption key, as implemented by httr2's \href{https://httr2.r-lib.org/reference/secrets.html}{secret functions}. This should
almost always be the name of an environment variable whose value was
generated with \code{secret_make_key()} (which is an inlined copy of
\code{httr2::secret_make_key()}).}

\item{x}{An R object.}
}
\value{
\itemize{
\item \code{secret_encrypt_json()}: The encrypted JSON string, invisibly. In typical
use, this function is mainly called for its side effect, which is to write an
encrypted file.
\item \code{secret_decrypt_json()}: The decrypted JSON string, invisibly.
\item \code{secret_write_rds()}: \code{x}, invisibly
\item \code{secret_read_rds()}: the decrypted object.
\item \code{secret_make_key()}: a random string to use as an encryption key.
\item \code{secret_has_key()} returns \code{TRUE} if the key is available and \code{FALSE}
otherwise.
}
}
\description{
These functions help to encrypt and decrypt confidential information that you
might need when deploying gargle-using projects or in CI/CD. They basically
rely on inlined copies of the \href{https://httr2.r-lib.org/reference/secrets.html}{secret functions in the httr2 package}. The awkwardness of
inlining code from httr2 can be removed if/when gargle starts to depend on
httr2.
\itemize{
\item The \code{secret_encrypt_json()} + \code{secret_decrypt_json()} pair is unique to
gargle, given how frequently Google auth relies on JSON files, e.g., service
account tokens and OAuth clients.
\item The \code{secret_write_rds()} + \code{secret_read_rds()} pair is just a copy of
functions from httr2. They are handy if you need to secure a user token.
\item \code{secret_make_key()} and \code{secret_has_key()} are also copies of functions
from httr2. Use \code{secret_make_key} to generate a key. Use \code{secret_has_key()}
to condition on key availability in, e.g., examples, tests, or apps.
}
}
\examples{
\dontshow{if (secret_has_key("GARGLE_KEY")) withAutoprint(\{ # examplesIf}
# gargle ships with JSON for a fake service account
# here we put the encrypted JSON into a new file
tmp <- tempfile()
secret_encrypt_json(
  fs::path_package("gargle", "extdata", "fake_service_account.json"),
  tmp,
  key = "GARGLE_KEY"
)

# complete the round trip by providing the decrypted JSON to a credential
# function
credentials_service_account(
 scopes = "https://www.googleapis.com/auth/userinfo.email",
 path = secret_decrypt_json(
   fs::path_package("gargle", "secret", "gargle-testing.json"),
   key = "GARGLE_KEY"
 )
)

file.remove(tmp)

# make an artificial Gargle2.0 token
fauxen <- gargle2.0_token(
  email = "jane@example.org",
  client = gargle_oauth_client(
    id = "CLIENT_ID", secret = "SECRET", name = "CLIENT"
  ),
  credentials = list(token = "fauxen"),
  cache = FALSE
)
fauxen

# store the fake token in an encrypted file
tmp2 <- tempfile()
secret_write_rds(fauxen, path = tmp2, key = "GARGLE_KEY")

# complete the round trip by providing the decrypted token to the "BYO token"
# credential function
rt_fauxen <- credentials_byo_oauth2(
  token  = secret_read_rds(tmp2, key = "GARGLE_KEY")
)
rt_fauxen

file.remove(tmp2)
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Gargle-class.R
\name{Gargle-class}
\alias{Gargle-class}
\alias{Gargle2.0}
\title{OAuth2 token objects specific to Google APIs}
\description{
\code{Gargle2.0} is based on the \code{\link[httr:Token-class]{Token2.0}} class provided in
httr. The preferred way to create a \code{Gargle2.0} token is through the
constructor function \code{\link[=gargle2.0_token]{gargle2.0_token()}}. Key differences with \code{Token2.0}:
\itemize{
\item The key for a cached \code{Token2.0} comes from hashing the endpoint, client,
and scopes. For the \code{Gargle2.0} subclass, the identifier or key is expanded
to include the email address associated with the token. This makes it easier
to work with Google APIs with multiple identities.
\item \code{Gargle2.0} tokens are cached, by default, at the user level, following the
XDG spec for storing user-specific data and cache files. In contrast, the
default location for \code{Token2.0} is \code{./.httr-oauth}, i.e. in current working
directory. \code{Gargle2.0} behaviour makes it easier to reuse tokens across
projects and makes it less likely that tokens are accidentally synced to a
remote location like GitHub or DropBox.
\item Each \code{Gargle2.0} token is cached in its own file. The token cache is a
directory of such files. In contrast, \code{Token2.0} tokens are cached as
components of a list, which is typically serialized to \code{./.httr-oauth}.
}
}
\keyword{internal}
\section{Super classes}{
\code{\link[httr:Token]{httr::Token}} -> \code{\link[httr:Token2.0]{httr::Token2.0}} -> \code{Gargle2.0}
}
\section{Public fields}{
\if{html}{\out{<div class="r6-fields">}}
\describe{
\item{\code{email}}{Email associated with the token.}

\item{\code{package}}{Name of the package requesting a token. Used in messages.}

\item{\code{client}}{An OAuth client.}
}
\if{html}{\out{</div>}}
}
\section{Methods}{
\subsection{Public methods}{
\itemize{
\item \href{#method-Gargle2.0-new}{\code{Gargle2.0$new()}}
\item \href{#method-Gargle2.0-format}{\code{Gargle2.0$format()}}
\item \href{#method-Gargle2.0-print}{\code{Gargle2.0$print()}}
\item \href{#method-Gargle2.0-hash}{\code{Gargle2.0$hash()}}
\item \href{#method-Gargle2.0-cache}{\code{Gargle2.0$cache()}}
\item \href{#method-Gargle2.0-load_from_cache}{\code{Gargle2.0$load_from_cache()}}
\item \href{#method-Gargle2.0-refresh}{\code{Gargle2.0$refresh()}}
\item \href{#method-Gargle2.0-init_credentials}{\code{Gargle2.0$init_credentials()}}
\item \href{#method-Gargle2.0-clone}{\code{Gargle2.0$clone()}}
}
}
\if{html}{\out{
<details open><summary>Inherited methods</summary>
<ul>
<li><span class="pkg-link" data-pkg="httr" data-topic="Token2.0" data-id="can_refresh"><a href='../../httr/html/Token2.0.html#method-Token2.0-can_refresh'><code>httr::Token2.0$can_refresh()</code></a></span></li>
<li><span class="pkg-link" data-pkg="httr" data-topic="Token2.0" data-id="revoke"><a href='../../httr/html/Token2.0.html#method-Token2.0-revoke'><code>httr::Token2.0$revoke()</code></a></span></li>
<li><span class="pkg-link" data-pkg="httr" data-topic="Token2.0" data-id="sign"><a href='../../httr/html/Token2.0.html#method-Token2.0-sign'><code>httr::Token2.0$sign()</code></a></span></li>
<li><span class="pkg-link" data-pkg="httr" data-topic="Token2.0" data-id="validate"><a href='../../httr/html/Token2.0.html#method-Token2.0-validate'><code>httr::Token2.0$validate()</code></a></span></li>
</ul>
</details>
}}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Gargle2.0-new"></a>}}
\if{latex}{\out{\hypertarget{method-Gargle2.0-new}{}}}
\subsection{Method \code{new()}}{
Create a Gargle2.0 token
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Gargle2.0$new(
  email = gargle_oauth_email(),
  client = gargle_client(),
  package = "gargle",
  credentials = NULL,
  params = list(),
  cache_path = gargle_oauth_cache(),
  app = deprecated()
)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{email}}{Optional email address. See \code{\link[=gargle2.0_token]{gargle2.0_token()}} for full
details.}

\item{\code{client}}{An OAuth consumer application.}

\item{\code{package}}{Name of the package requesting a token. Used in messages.}

\item{\code{credentials}}{Exists largely for testing purposes.}

\item{\code{params}}{A list of parameters for the internal function
\code{init_oauth2.0()}, which is a modified version of \code{\link[httr:init_oauth2.0]{httr::init_oauth2.0()}}.
gargle actively uses \code{scope} and \code{use_oob}, but does not use \code{user_params},
\code{type}, \code{as_header} (hard-wired to \code{TRUE}), \code{use_basic_auth} (accept
default of \code{use_basic_auth = FALSE}), \code{config_init}, or
\code{client_credentials}.}

\item{\code{cache_path}}{Specifies the OAuth token cache. Read more in
\code{\link[=gargle_oauth_cache]{gargle_oauth_cache()}}.}

\item{\code{app}}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use \code{client} instead.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
A Gargle2.0 token.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Gargle2.0-format"></a>}}
\if{latex}{\out{\hypertarget{method-Gargle2.0-format}{}}}
\subsection{Method \code{format()}}{
Format a Gargle2.0 token
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Gargle2.0$format(...)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{...}}{Not used.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Gargle2.0-print"></a>}}
\if{latex}{\out{\hypertarget{method-Gargle2.0-print}{}}}
\subsection{Method \code{print()}}{
Print a Gargle2.0 token
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Gargle2.0$print(...)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{...}}{Not used.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Gargle2.0-hash"></a>}}
\if{latex}{\out{\hypertarget{method-Gargle2.0-hash}{}}}
\subsection{Method \code{hash()}}{
Generate the email-augmented hash of a Gargle2.0 token
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Gargle2.0$hash()}\if{html}{\out{</div>}}
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Gargle2.0-cache"></a>}}
\if{latex}{\out{\hypertarget{method-Gargle2.0-cache}{}}}
\subsection{Method \code{cache()}}{
Put a Gargle2.0 token into the cache
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Gargle2.0$cache()}\if{html}{\out{</div>}}
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Gargle2.0-load_from_cache"></a>}}
\if{latex}{\out{\hypertarget{method-Gargle2.0-load_from_cache}{}}}
\subsection{Method \code{load_from_cache()}}{
(Attempt to) get a Gargle2.0 token from the cache
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Gargle2.0$load_from_cache()}\if{html}{\out{</div>}}
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Gargle2.0-refresh"></a>}}
\if{latex}{\out{\hypertarget{method-Gargle2.0-refresh}{}}}
\subsection{Method \code{refresh()}}{
(Attempt to) refresh a Gargle2.0 token
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Gargle2.0$refresh()}\if{html}{\out{</div>}}
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Gargle2.0-init_credentials"></a>}}
\if{latex}{\out{\hypertarget{method-Gargle2.0-init_credentials}{}}}
\subsection{Method \code{init_credentials()}}{
Initiate a new Gargle2.0 token
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Gargle2.0$init_credentials()}\if{html}{\out{</div>}}
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Gargle2.0-clone"></a>}}
\if{latex}{\out{\hypertarget{method-Gargle2.0-clone}{}}}
\subsection{Method \code{clone()}}{
The objects of this class are cloneable with this method.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Gargle2.0$clone(deep = FALSE)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{deep}}{Whether to make a deep clone.}
}
\if{html}{\out{</div>}}
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/credentials_external_account.R
\name{oauth_external_token}
\alias{oauth_external_token}
\title{Generate OAuth token for an external account}
\usage{
oauth_external_token(
  path = "",
  scopes = "https://www.googleapis.com/auth/cloud-platform"
)
}
\arguments{
\item{path}{JSON containing the workload identity configuration for the
external account, in one of the forms supported for the \code{txt} argument of
\code{\link[jsonlite:fromJSON]{jsonlite::fromJSON()}} (probably, a file path, although it could be a JSON
string). The instructions for generating this configuration are given at
\href{https://cloud.google.com/iam/docs/configuring-workload-identity-federation}{Configuring workload identity federation}.

Note that external account tokens are a natural fit for use as Application
Default Credentials, so consider storing the configuration file in one of
the standard locations consulted for ADC, instead of providing \code{path}
explicitly. See \code{\link[=credentials_app_default]{credentials_app_default()}} for more.}

\item{scopes}{A character vector of scopes to request. Pick from those listed
at \url{https://developers.google.com/identity/protocols/oauth2/scopes}.

For certain token flows, the
\code{"https://www.googleapis.com/auth/userinfo.email"} scope is unconditionally
included. This grants permission to retrieve the email address associated
with a token; gargle uses this to index cached OAuth tokens. This grants no
permission to view or send email and is generally considered a low-value
scope.}
}
\description{
Generate OAuth token for an external account
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/request_make.R
\name{request_make}
\alias{request_make}
\title{Make a Google API request}
\usage{
request_make(x, ..., encode = "json", user_agent = gargle_user_agent())
}
\arguments{
\item{x}{List. Holds the components for an HTTP request, presumably created
with \code{\link[=request_develop]{request_develop()}} or \code{\link[=request_build]{request_build()}}. Must contain a \code{method} and
\code{url}. If present, \code{body} and \code{token} are used.}

\item{...}{Optional arguments passed through to the HTTP method. Currently
neither gargle nor httr checks that all are used, so be aware that unused
arguments may be silently ignored.}

\item{encode}{If the body is a named list, how should it be encoded? Can be
one of form (application/x-www-form-urlencoded), multipart,
(multipart/form-data), or json (application/json).

For "multipart", list elements can be strings or objects created by
\code{\link[httr:upload_file]{upload_file()}}. For "form", elements are coerced to strings
and escaped, use \code{I()} to prevent double-escaping. For "json",
parameters are automatically "unboxed" (i.e. length 1 vectors are
converted to scalars). To preserve a length 1 vector as a vector,
wrap in \code{I()}. For "raw", either a character or raw vector. You'll
need to make sure to set the \code{\link[httr:content_type]{content_type()}} yourself.}

\item{user_agent}{A user agent string, prepared by \code{\link[httr:user_agent]{httr::user_agent()}}. When
in doubt, a client package should have an internal function that extends
\code{gargle_user_agent()} by prepending its return value with the client
package's name and version.}
}
\value{
Object of class \code{response} from \link[httr:httr-package]{httr::httr}.
}
\description{
Intended primarily for internal use in client packages that provide
high-level wrappers for users. \code{request_make()} does relatively little:
\itemize{
\item Calls an HTTP method.
\item Adds a user agent.
\item Enforces \code{"json"} as the default for \code{encode}. This differs from httr's
default behaviour, but aligns better with Google APIs.
}

Typically the input is created with \code{\link[=request_build]{request_build()}} and the output is
processed with \code{\link[=response_process]{response_process()}}.
}
\examples{
\dontrun{
req <- gargle::request_build(
  method = "GET",
  path = "path/to/the/resource",
  token = "PRETEND_I_AM_TOKEN"
)
gargle::request_make(req)
}
}
\seealso{
Other requests and responses: 
\code{\link{request_develop}()},
\code{\link{response_process}()}
}
\concept{requests and responses}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/field_mask.R
\name{field_mask}
\alias{field_mask}
\title{Generate a field mask}
\usage{
field_mask(x)
}
\arguments{
\item{x}{A named R list, where the requirement for names applies at all
levels, i.e. recursively.}
}
\value{
A Google API field mask, as a string.
}
\description{
Many Google API requests take a field mask, via a \code{fields} parameter, in the
URL and/or in the body. \code{field_mask()} generates such a field mask from an R
list, typically a list that is destined to be part of the body of a request
that writes or updates a resource. \code{field_mask()} is designed to help in the
common case where the attributes you wish to modify are exactly the ones
represented in the object. It is possible to use a "larger" field mask, that
is either less specific or that explicitly includes other attributes, in
which case the attributes covered by the mask but absent from the object are
reset to default values. This is not exactly the use case \code{field_mask()} is
designed for, but its output could still be useful as a first step in
constructing such a mask.
}
\examples{
x <- list(sheetId = 1234, title = "my_favorite_worksheet")
field_mask(x)

x <- list(
  userEnteredFormat = list(
    backgroundColor = list(
      red = 159 / 255, green = 183 / 255, blue = 196 / 255
    )
  )
)
field_mask(x)

x <- list(
  sheetId = 1234,
  gridProperties = list(rowCount = 5, columnCount = 3)
)
field_mask(x)
}
\seealso{
The documentation for the \href{https://protobuf.dev/reference/protobuf/google.protobuf/#json-encoding-of-field-masks}{JSON encoding of a Protocol Buffers FieldMask}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/credentials_app_default.R
\name{credentials_app_default}
\alias{credentials_app_default}
\title{Load Application Default Credentials}
\usage{
credentials_app_default(scopes = NULL, ..., subject = NULL)
}
\arguments{
\item{scopes}{A character vector of scopes to request. Pick from those listed
at \url{https://developers.google.com/identity/protocols/oauth2/scopes}.

For certain token flows, the
\code{"https://www.googleapis.com/auth/userinfo.email"} scope is unconditionally
included. This grants permission to retrieve the email address associated
with a token; gargle uses this to index cached OAuth tokens. This grants no
permission to view or send email and is generally considered a low-value
scope.}

\item{...}{Additional arguments passed to all credential functions.}

\item{subject}{An optional subject claim. Specify this if you wish to use the
service account represented by \code{path} to impersonate the \code{subject}, who is
a normal user. Before this can work, an administrator must grant the service
account domain-wide authority. Identify the user to impersonate via their
email, e.g. \code{subject = "user@example.com"}. Note that gargle automatically
adds the non-sensitive \code{"https://www.googleapis.com/auth/userinfo.email"}
scope, so this scope must be enabled for the service account, along with
any other \code{scopes} being requested.}
}
\value{
An \code{\link[httr:Token-class]{httr::TokenServiceAccount}}, a \code{\link{WifToken}},
an \code{\link[httr:Token-class]{httr::Token2.0}} or \code{NULL}.
}
\description{
Loads credentials from a file identified via a search strategy known as
Application Default Credentials (ADC). The hope is to make auth "just work"
for someone working on Google-provided infrastructure or who has used Google
tooling to get started, such as the \href{https://docs.cloud.google.com/sdk/gcloud}{\code{gcloud} command line tool}.

A sequence of paths is consulted, which we describe here, with some abuse of
notation. ALL_CAPS represents the value of an environment variable and \verb{\%||\%}
is used in the spirit of a \href{https://en.wikipedia.org/wiki/Null_coalescing_operator}{null coalescing operator}.

\if{html}{\out{<div class="sourceCode">}}\preformatted{GOOGLE_APPLICATION_CREDENTIALS
CLOUDSDK_CONFIG/application_default_credentials.json
# on Windows:
(APPDATA \%||\% SystemDrive \%||\% C:)\\gcloud\\application_default_credentials.json
# on not-Windows:
~/.config/gcloud/application_default_credentials.json
}\if{html}{\out{</div>}}

If the above search successfully identifies a JSON file, it is parsed and
ingested as a service account, an external account ("workload identity
federation"), or a user account. Literally, if the JSON describes a service
account, we call \code{\link[=credentials_service_account]{credentials_service_account()}} and if it describes an
external account, we call \code{\link[=credentials_external_account]{credentials_external_account()}}.
}
\examples{
\dontrun{
credentials_app_default()
}
}
\seealso{
\itemize{
\item \url{https://docs.cloud.google.com/docs/authentication}
\item \url{https://docs.cloud.google.com/sdk/docs}
}

Other credential functions: 
\code{\link{credentials_byo_oauth2}()},
\code{\link{credentials_external_account}()},
\code{\link{credentials_gce}()},
\code{\link{credentials_service_account}()},
\code{\link{credentials_user_oauth2}()},
\code{\link{token_fetch}()}
}
\concept{credential functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/oauth-cache.R
\name{gargle_oauth_sitrep}
\alias{gargle_oauth_sitrep}
\title{OAuth token situation report}
\usage{
gargle_oauth_sitrep(cache = NULL)
}
\arguments{
\item{cache}{Specifies the OAuth token cache. Defaults to the option named
\code{"gargle_oauth_cache"}, retrieved via \code{\link[=gargle_oauth_cache]{gargle_oauth_cache()}}.}
}
\value{
A data frame with one row per cached token, invisibly. Note this data
frame may contain more columns than it seems, e.g. the \code{filepath} column
isn't printed by default.
}
\description{
Get a human-oriented overview of the existing gargle OAuth tokens:
\itemize{
\item Filepath of the current cache
\item Number of tokens found there
\item Compact summary of the associated
\itemize{
\item Email = Google identity
\item OAuth client (actually, just its nickname)
\item Scopes
\item Hash (actually, just the first 7 characters)
Mostly useful for the development of gargle and client packages.
}
}
}
\examples{
gargle_oauth_sitrep()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gargle_oauth_client.R
\name{gargle_client}
\alias{gargle_client}
\title{OAuth client for demonstration purposes}
\usage{
gargle_client(type = NULL)
}
\arguments{
\item{type}{Specifies the type of OAuth client. The valid values are a subset
of possible Google client types and reflect the key used to describe the
client in its JSON representation:
\itemize{
\item \code{"installed"} is associated with a "Desktop app"
\item \code{"web"} is associated with a "Web application"
}}
}
\value{
An OAuth client, produced by \code{\link[=gargle_oauth_client]{gargle_oauth_client()}}, invisibly.
}
\description{
Invisibly returns an instance of
\code{\link[=gargle_oauth_client]{gargle_oauth_client}} that can be used to test drive
gargle before obtaining your own client ID and secret. This OAuth client may
be deleted or rotated at any time. There are no guarantees about which APIs
are enabled. DO NOT USE THIS IN A PACKAGE or for anything other than
interactive, small-scale experimentation.

You can get your own OAuth client ID and secret, without these limitations.
See the \code{vignette("get-api-credentials")} for more details.
}
\examples{
\dontrun{
gargle_client()
}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gargle_api_key.R
\name{gargle_api_key}
\alias{gargle_api_key}
\title{API key for demonstration purposes}
\usage{
gargle_api_key()
}
\description{
Some APIs accept requests for public resources, in which case
the request must be sent with an API key in lieu of a token. This function
provides an API key for limited use in prototyping and for testing and
documentation of gargle itself. This key may be deleted or rotated at any
time. There are no guarantees about which APIs are enabled. DO NOT USE THIS
IN A PACKAGE or for anything other than interactive, small-scale
experimentation.

You can get your own API key, without these limitations. See the \href{https://gargle.r-lib.org/articles/get-api-credentials.html}{How to get your own API credentials}
vignette for more details.
}
\examples{
gargle_api_key()
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/credentials_external_account.R
\name{WifToken}
\alias{WifToken}
\title{Token for use with workload identity federation}
\description{
Token for use with workload identity federation

Token for use with workload identity federation
}
\details{
Not intended for direct use. See \code{\link[=credentials_external_account]{credentials_external_account()}} instead.
}
\keyword{internal}
\section{Super classes}{
\code{\link[httr:Token]{httr::Token}} -> \code{\link[httr:Token2.0]{httr::Token2.0}} -> \code{WifToken}
}
\section{Methods}{
\subsection{Public methods}{
\itemize{
\item \href{#method-WifToken-new}{\code{WifToken$new()}}
\item \href{#method-WifToken-init_credentials}{\code{WifToken$init_credentials()}}
\item \href{#method-WifToken-refresh}{\code{WifToken$refresh()}}
\item \href{#method-WifToken-format}{\code{WifToken$format()}}
\item \href{#method-WifToken-print}{\code{WifToken$print()}}
\item \href{#method-WifToken-can_refresh}{\code{WifToken$can_refresh()}}
\item \href{#method-WifToken-cache}{\code{WifToken$cache()}}
\item \href{#method-WifToken-load_from_cache}{\code{WifToken$load_from_cache()}}
\item \href{#method-WifToken-validate}{\code{WifToken$validate()}}
\item \href{#method-WifToken-revoke}{\code{WifToken$revoke()}}
\item \href{#method-WifToken-clone}{\code{WifToken$clone()}}
}
}
\if{html}{\out{
<details open><summary>Inherited methods</summary>
<ul>
<li><span class="pkg-link" data-pkg="httr" data-topic="Token" data-id="hash"><a href='../../httr/html/Token.html#method-Token-hash'><code>httr::Token$hash()</code></a></span></li>
<li><span class="pkg-link" data-pkg="httr" data-topic="Token2.0" data-id="sign"><a href='../../httr/html/Token2.0.html#method-Token2.0-sign'><code>httr::Token2.0$sign()</code></a></span></li>
</ul>
</details>
}}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-WifToken-new"></a>}}
\if{latex}{\out{\hypertarget{method-WifToken-new}{}}}
\subsection{Method \code{new()}}{
Get a token via workload identity federation
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{WifToken$new(params = list())}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{params}}{A list of parameters for \code{init_oauth_external_account()}.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
A WifToken.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-WifToken-init_credentials"></a>}}
\if{latex}{\out{\hypertarget{method-WifToken-init_credentials}{}}}
\subsection{Method \code{init_credentials()}}{
Enact the actual token exchange for workload identity
federation.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{WifToken$init_credentials()}\if{html}{\out{</div>}}
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-WifToken-refresh"></a>}}
\if{latex}{\out{\hypertarget{method-WifToken-refresh}{}}}
\subsection{Method \code{refresh()}}{
Refreshes the token, which means re-doing the entire token
flow in this case.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{WifToken$refresh()}\if{html}{\out{</div>}}
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-WifToken-format"></a>}}
\if{latex}{\out{\hypertarget{method-WifToken-format}{}}}
\subsection{Method \code{format()}}{
Format a \code{\link[=WifToken]{WifToken()}}.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{WifToken$format(...)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{...}}{Not used.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-WifToken-print"></a>}}
\if{latex}{\out{\hypertarget{method-WifToken-print}{}}}
\subsection{Method \code{print()}}{
Print a \code{\link[=WifToken]{WifToken()}}.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{WifToken$print(...)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{...}}{Not used.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-WifToken-can_refresh"></a>}}
\if{latex}{\out{\hypertarget{method-WifToken-can_refresh}{}}}
\subsection{Method \code{can_refresh()}}{
Placeholder implementation of required method. Returns \code{TRUE}.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{WifToken$can_refresh()}\if{html}{\out{</div>}}
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-WifToken-cache"></a>}}
\if{latex}{\out{\hypertarget{method-WifToken-cache}{}}}
\subsection{Method \code{cache()}}{
Placeholder implementation of required method. Returns self.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{WifToken$cache()}\if{html}{\out{</div>}}
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-WifToken-load_from_cache"></a>}}
\if{latex}{\out{\hypertarget{method-WifToken-load_from_cache}{}}}
\subsection{Method \code{load_from_cache()}}{
Placeholder implementation of required method. Returns self.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{WifToken$load_from_cache()}\if{html}{\out{</div>}}
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-WifToken-validate"></a>}}
\if{latex}{\out{\hypertarget{method-WifToken-validate}{}}}
\subsection{Method \code{validate()}}{
Placeholder implementation of required method.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{WifToken$validate()}\if{html}{\out{</div>}}
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-WifToken-revoke"></a>}}
\if{latex}{\out{\hypertarget{method-WifToken-revoke}{}}}
\subsection{Method \code{revoke()}}{
Placeholder implementation of required method.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{WifToken$revoke()}\if{html}{\out{</div>}}
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-WifToken-clone"></a>}}
\if{latex}{\out{\hypertarget{method-WifToken-clone}{}}}
\subsection{Method \code{clone()}}{
The objects of this class are cloneable with this method.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{WifToken$clone(deep = FALSE)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{deep}}{Whether to make a deep clone.}
}
\if{html}{\out{</div>}}
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils-ui.R
\name{gargle_map_cli}
\alias{gargle_map_cli}
\title{Map a cli-styled template over an object}
\usage{
gargle_map_cli(x, ...)
}
\description{
For internal use in gargle, googledrive, and googlesheets4 (for now).
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gargle-package.R
\docType{package}
\name{gargle-package}
\alias{gargle}
\alias{gargle-package}
\title{gargle: Utilities for Working with Google APIs}
\description{
Provides utilities for working with Google APIs \url{https://developers.google.com/apis-explorer}. This includes functions and classes for handling common credential types and for preparing, executing, and processing HTTP requests.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://gargle.r-lib.org}
  \item \url{https://github.com/r-lib/gargle}
  \item Report bugs at \url{https://github.com/r-lib/gargle/issues}
}

}
\author{
\strong{Maintainer}: Jennifer Bryan \email{jenny@posit.co} (\href{https://orcid.org/0000-0002-6983-2759}{ORCID})

Authors:
\itemize{
  \item Craig Citro \email{craigcitro@google.com}
  \item Hadley Wickham \email{hadley@posit.co} (\href{https://orcid.org/0000-0003-4757-117X}{ORCID})
}

Other contributors:
\itemize{
  \item Google Inc [copyright holder]
  \item Posit Software, PBC [copyright holder, funder]
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cred_funs.R
\name{cred_funs}
\alias{cred_funs}
\alias{cred_funs_list}
\alias{cred_funs_add}
\alias{cred_funs_set}
\alias{cred_funs_clear}
\alias{cred_funs_list_default}
\alias{cred_funs_set_default}
\alias{local_cred_funs}
\alias{with_cred_funs}
\title{Credential function registry}
\usage{
cred_funs_list()

cred_funs_add(...)

cred_funs_set(funs, ls = deprecated())

cred_funs_clear()

cred_funs_list_default()

cred_funs_set_default()

local_cred_funs(
  funs = cred_funs_list_default(),
  action = c("replace", "modify"),
  .local_envir = caller_env()
)

with_cred_funs(
  funs = cred_funs_list_default(),
  code,
  action = c("replace", "modify")
)
}
\arguments{
\item{...}{<\code{\link[rlang:dyn-dots]{dynamic-dots}}> One or more credential
functions, in \code{name = value} form. Each credential function is subject to a
superficial check that it at least "smells like" a credential function: its
first argument must be named \code{scopes}, and its signature must include
\code{...}. To remove a credential function, you can use a specification like
\code{name = NULL}.}

\item{funs}{A named list of credential functions.}

\item{ls}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} This argument has been renamed
to \code{funs}.}

\item{action}{Whether to use \code{funs} to replace or modify the registry with
funs:
\itemize{
\item \code{"replace"} does \code{cred_funs_set(funs)}
\item \code{"modify"} does \code{cred_funs_add(!!!funs)}
}}

\item{.local_envir}{The environment to use for scoping. Defaults to current
execution environment.}

\item{code}{Code to run with temporary credential function registry.}
}
\value{
A list of credential functions or \code{NULL}.
}
\description{
Functions to query or manipulate the registry of credential functions
consulted by \code{\link[=token_fetch]{token_fetch()}}.
}
\section{Functions}{
\itemize{
\item \code{cred_funs_list()}: Get the list of registered credential functions.

\item \code{cred_funs_add()}: Register one or more new credential fetching functions.
Function(s) are added to the \emph{front} of the list. So:
\itemize{
\item "First registered, last tried."
\item "Last registered, first tried."
}

Can also be used to \emph{remove} a function from the registry.

\item \code{cred_funs_set()}: Register a list of credential fetching functions.

\item \code{cred_funs_clear()}: Clear the credential function registry.

\item \code{cred_funs_list_default()}: Return the default list of credential functions.

\item \code{cred_funs_set_default()}: Reset the registry to the gargle default.

\item \code{local_cred_funs()}: Modify the credential function registry in the current
scope. It is an example of the \verb{local_*()} functions in \pkg{withr}.

\item \code{with_cred_funs()}: Evaluate \code{code} with a temporarily modified credential
function registry. It is an example of the \verb{with_*()} functions in
\pkg{withr}.

}}
\examples{
names(cred_funs_list())

creds_one <- function(scopes, ...) {}

cred_funs_add(one = creds_one)
cred_funs_add(two = creds_one, three = creds_one)
names(cred_funs_list())

cred_funs_add(two = NULL)
names(cred_funs_list())

# restore the default list
cred_funs_set_default()

# remove one specific credential fetcher
cred_funs_add(credentials_gce = NULL)
names(cred_funs_list())

# force the use of one specific credential fetcher
cred_funs_set(list(credentials_user_oauth2 = credentials_user_oauth2))
names(cred_funs_list())

# restore the default list
cred_funs_set_default()

# run some code with a temporary change to the registry
# creds_one ONLY
with_cred_funs(
  list(one = creds_one),
  names(cred_funs_list())
)
# add creds_one to the list
with_cred_funs(
  list(one = creds_one),
  names(cred_funs_list()),
  action = "modify"
)
# remove credentials_gce
with_cred_funs(
  list(credentials_gce = NULL),
  names(cred_funs_list()),
  action = "modify"
)
}
\seealso{
\code{\link[=token_fetch]{token_fetch()}}, which is where the registry is actually used.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gargle_oauth_client.R
\name{gargle_oauth_client_from_json}
\alias{gargle_oauth_client_from_json}
\alias{gargle_oauth_client}
\title{Create an OAuth client for Google}
\usage{
gargle_oauth_client_from_json(path, name = NULL)

gargle_oauth_client(
  id,
  secret,
  redirect_uris = NULL,
  type = c("installed", "web"),
  name = hash(id)
)
}
\arguments{
\item{path}{JSON downloaded from \href{https://console.cloud.google.com}{Google Cloud Console}, containing a client id and
secret, in one of the forms supported for the \code{txt} argument of
\code{\link[jsonlite:fromJSON]{jsonlite::fromJSON()}} (typically, a file path or JSON string).}

\item{name}{A label for this specific client, presumably the same name used
to label it in Google Cloud Console. Unfortunately there is no way to
make that true programmatically, i.e. the JSON representation does not
contain this information.}

\item{id}{Client ID}

\item{secret}{Client secret}

\item{redirect_uris}{Where your application listens for the response from
Google's authorization server. If you didn't configure this specifically
when creating the client (which is only possible for clients of the "web"
type), you can leave this unspecified.}

\item{type}{Specifies the type of OAuth client. The valid values are a subset
of possible Google client types and reflect the key used to describe the
client in its JSON representation:
\itemize{
\item \code{"installed"} is associated with a "Desktop app"
\item \code{"web"} is associated with a "Web application"
}}
}
\value{
An OAuth client: An S3 list with class \code{gargle_oauth_client}. For
backwards compatibility reasons, this currently also inherits from the httr
S3 class \code{oauth_app}, but that is a temporary measure. An instance of
\code{gargle_oauth_client} stores more information than httr's \code{oauth_app}, such
as the OAuth client's type ("web" or "installed").

There are some redundant fields in this object during the httr-to-httr2
transition period. The legacy fields \code{appname} and \code{key} repeat the
information in the future-facing fields \code{name} and (client) \code{id}. Prefer
\code{name} and \code{id} to \code{appname} and \code{key} in downstream code. Prefer the
constructors \code{gargle_oauth_client_from_json()} and \code{gargle_oauth_client()}
to \code{\link[httr:oauth_app]{httr::oauth_app()}} and \code{\link[=oauth_app_from_json]{oauth_app_from_json()}}.
}
\description{
A \code{gargle_oauth_client} consists of:
\itemize{
\item A type. gargle only supports the "Desktop app" and "Web application" client
types. Different types are associated with different OAuth flows.
\item A client ID and secret.
\item Optionally, one or more redirect URIs.
\item A name. This is really a human-facing label. Or, rather, it can be used
that way, but the default is just a hash. We recommend using the same name
here as the name used to label the client ID in the \href{https://console.cloud.google.com}{Google Cloud Platform Console}.
}

A \code{gargle_oauth_client} is an adaptation of httr's \code{\link[httr:oauth_app]{httr::oauth_app()}} (currently)
and httr2's \code{oauth_client()} (which gargle will migrate to in the future).
}
\examples{
\dontrun{
gargle_oauth_client_from_json(
  path = "/path/to/the/JSON/you/downloaded/from/gcp/console.json",
  name = "my-nifty-oauth-client"
)
}

gargle_oauth_client(
  id = "some_long_id",
  secret = "ssshhhhh_its_a_secret",
  name = "my-nifty-oauth-client"
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/credentials_service_account.R
\name{credentials_service_account}
\alias{credentials_service_account}
\title{Load a service account token}
\usage{
credentials_service_account(scopes = NULL, path = "", ..., subject = NULL)
}
\arguments{
\item{scopes}{A character vector of scopes to request. Pick from those listed
at \url{https://developers.google.com/identity/protocols/oauth2/scopes}.

For certain token flows, the
\code{"https://www.googleapis.com/auth/userinfo.email"} scope is unconditionally
included. This grants permission to retrieve the email address associated
with a token; gargle uses this to index cached OAuth tokens. This grants no
permission to view or send email and is generally considered a low-value
scope.}

\item{path}{JSON identifying the service account, in one of the forms
supported for the \code{txt} argument of \code{\link[jsonlite:fromJSON]{jsonlite::fromJSON()}} (typically, a
file path or JSON string).}

\item{...}{Additional arguments passed to all credential functions.}

\item{subject}{An optional subject claim. Specify this if you wish to use the
service account represented by \code{path} to impersonate the \code{subject}, who is
a normal user. Before this can work, an administrator must grant the service
account domain-wide authority. Identify the user to impersonate via their
email, e.g. \code{subject = "user@example.com"}. Note that gargle automatically
adds the non-sensitive \code{"https://www.googleapis.com/auth/userinfo.email"}
scope, so this scope must be enabled for the service account, along with
any other \code{scopes} being requested.}
}
\value{
An \code{\link[httr:Token-class]{httr::TokenServiceAccount}} or \code{NULL}.
}
\description{
Load a service account token
}
\details{
Note that fetching a token for a service account requires a
reasonably accurate system clock. For more information, see the
\code{vignette("how-gargle-gets-tokens")}.
}
\examples{
\dontrun{
token <- credentials_service_account(
  scopes = "https://www.googleapis.com/auth/userinfo.email",
  path = "/path/to/your/service-account.json"
)
}
}
\seealso{
Additional reading on delegation of domain-wide authority:
\itemize{
\item \url{https://developers.google.com/identity/protocols/oauth2/service-account#delegatingauthority}
}

Other credential functions: 
\code{\link{credentials_app_default}()},
\code{\link{credentials_byo_oauth2}()},
\code{\link{credentials_external_account}()},
\code{\link{credentials_gce}()},
\code{\link{credentials_user_oauth2}()},
\code{\link{token_fetch}()}
}
\concept{credential functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Gargle-class.R
\name{gargle2.0_token}
\alias{gargle2.0_token}
\title{Generate a gargle token}
\usage{
gargle2.0_token(
  email = gargle_oauth_email(),
  client = gargle_client(),
  package = "gargle",
  scope = NULL,
  use_oob = gargle_oob_default(),
  credentials = NULL,
  cache = if (is.null(credentials)) gargle_oauth_cache() else FALSE,
  ...,
  app = deprecated()
)
}
\arguments{
\item{email}{Optional. If specified, \code{email} can take several different
forms:
\itemize{
\item \code{"jane@gmail.com"}, i.e. an actual email address. This allows the user to
target a specific Google identity. If specified, this is used for token
lookup, i.e. to determine if a suitable token is already available in the
cache. If no such token is found, \code{email} is used to pre-select the targeted
Google identity in the OAuth chooser. (Note, however, that the email
associated with a token when it's cached is always determined from the token
itself, never from this argument).
\item \code{"*@example.com"}, i.e. a domain-only glob pattern. This can be helpful if
you need code that "just works" for both \code{alice@example.com} and
\code{bob@example.com}.
\item \code{TRUE} means that you are approving email auto-discovery. If exactly one
matching token is found in the cache, it will be used.
\item \code{FALSE} or \code{NA} mean that you want to ignore the token cache and force a
new OAuth dance in the browser.
}

Defaults to the option named \code{"gargle_oauth_email"}, retrieved by
\code{\link[=gargle_oauth_email]{gargle_oauth_email()}} (unless a wrapper package implements different
default behavior).}

\item{client}{A Google OAuth client, preferably constructed via
\code{\link[=gargle_oauth_client_from_json]{gargle_oauth_client_from_json()}}, which returns an instance of
\code{gargle_oauth_client}. For backwards compatibility, for a limited time,
gargle will still accept an "OAuth app" created with \code{\link[httr:oauth_app]{httr::oauth_app()}}.}

\item{package}{Name of the package requesting a token. Used in messages.}

\item{scope}{A character vector of scopes to request.}

\item{use_oob}{Whether to use out-of-band authentication (or, perhaps, a
variant implemented by gargle and known as "pseudo-OOB") when first
acquiring the token. Defaults to the value returned by
\code{\link[=gargle_oob_default]{gargle_oob_default()}}. Note that (pseudo-)OOB auth only affects
the initial OAuth dance. If we retrieve (and possibly refresh) a
cached token, \code{use_oob} has no effect.

If the OAuth client is provided implicitly by a wrapper package, its type
probably defaults to the value returned by
\code{\link[=gargle_oauth_client_type]{gargle_oauth_client_type()}}. You can take control of the client
type by setting \code{options(gargle_oauth_client_type = "web")} or
\code{options(gargle_oauth_client_type = "installed")}.}

\item{credentials}{Advanced use only: allows you to completely customise
token generation.}

\item{cache}{Specifies the OAuth token cache. Defaults to the option named
\code{"gargle_oauth_cache"}, retrieved via \code{\link[=gargle_oauth_cache]{gargle_oauth_cache()}}.}

\item{...}{Absorbs arguments intended for use by other credential functions.
Not used.}

\item{app}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Replaced by the \code{client}
argument.}
}
\value{
An object of class \link{Gargle2.0}, either new or loaded from the cache.
}
\description{
Constructor function for objects of class \link{Gargle2.0}.
}
\examples{
\dontrun{
gargle2.0_token()
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/aaa.R
\name{gargle_last_response}
\alias{gargle_last_response}
\alias{gargle_last_content}
\title{Access the last response}
\usage{
gargle_last_response()

gargle_last_content()
}
\value{
\itemize{
\item \code{gargle_last_response()} returns the most recent \code{\link[httr:response]{httr::response()}}
object.
\item \code{gargle_last_content()} returns the parsed JSON content from the most
recent response or an empty list if unavailable.
}
}
\description{
These functions give access to the most recent response processed by
\code{\link[=response_process]{response_process(..., remember = TRUE)}}. They can be
useful for \emph{post mortem} analysis of puzzling or failed API interactions.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gargle_oauth_client.R
\name{oauth_app_from_json}
\alias{oauth_app_from_json}
\alias{gargle_app}
\title{Create an OAuth app from JSON}
\usage{
oauth_app_from_json(path, appname = NULL)

gargle_app()
}
\arguments{
\item{path}{JSON downloaded from \href{https://console.cloud.google.com}{Google Cloud Console}, containing a client id and
secret, in one of the forms supported for the \code{txt} argument of
\code{\link[jsonlite:fromJSON]{jsonlite::fromJSON()}} (typically, a file path or JSON string).}

\item{appname}{name of the application.  This is not used for OAuth, but is
used to make it easier to identify different applications.}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}}

\code{oauth_app_from_json()} is being replaced with
\code{\link[=gargle_oauth_client_from_json]{gargle_oauth_client_from_json()}}, in light of the new
\code{gargle_oauth_client} class. Now \code{oauth_app_from_json()} potentially warns
about this deprecation and immediately passes its inputs through to
\code{\link[=gargle_oauth_client_from_json]{gargle_oauth_client_from_json()}}.

\code{gargle_app()} is being replaced with \code{\link[=gargle_client]{gargle_client()}}.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/token-info.R
\name{token-info}
\alias{token-info}
\alias{token_userinfo}
\alias{token_email}
\alias{token_tokeninfo}
\title{Get info from a token}
\usage{
token_userinfo(token)

token_email(token)

token_tokeninfo(token)
}
\arguments{
\item{token}{A token with class \link[httr:Token-class]{Token2.0} or an object of
httr's class \code{request}, i.e. a token that has been prepared with
\code{\link[httr:config]{httr::config()}} and has a \link[httr:Token-class]{Token2.0} in the
\code{auth_token} component.}
}
\value{
A list containing:
\itemize{
\item \code{token_userinfo()}: user info
\item \code{token_email()}: user's email (obtained from a call to \code{token_userinfo()})
\item \code{token_tokeninfo()}: token info
}
}
\description{
These functions send the \code{token} to Google endpoints that return info about a
token or a user.
}
\details{
It's hard to say exactly what info will be returned by the "userinfo"
endpoint targetted by \code{token_userinfo()}. It depends on the token's scopes.
Where possible, OAuth2 tokens obtained via the gargle package include the
\verb{https://www.googleapis.com/auth/userinfo.email} scope, which guarantees we
can learn the email associated with the token. If the token has the
\verb{https://www.googleapis.com/auth/userinfo.profile} scope, there will be even
more information available. But for a token with unknown or arbitrary scopes,
we can't make any promises about what information will be returned.
}
\examples{
\dontrun{
# with service account token
t <- token_fetch(
  scopes = "https://www.googleapis.com/auth/drive",
  path   = "path/to/service/account/token/blah-blah-blah.json"
)
# or with an OAuth token
t <- token_fetch(
  scopes = "https://www.googleapis.com/auth/drive",
  email  = "janedoe@example.com"
)
token_userinfo(t)
token_email(t)
tokens_tokeninfo(t)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/AuthState-class.R
\name{init_AuthState}
\alias{init_AuthState}
\title{Create an AuthState}
\usage{
init_AuthState(
  package = NA_character_,
  client = NULL,
  api_key = NULL,
  auth_active = TRUE,
  cred = NULL,
  app = deprecated()
)
}
\arguments{
\item{package}{Package name, an optional string. It is recommended to record
the name of the package whose auth state is being managed. Ultimately, this
may be used in some downstream messaging.}

\item{client}{A Google OAuth client, preferably constructed via
\code{\link[=gargle_oauth_client_from_json]{gargle_oauth_client_from_json()}}, which returns an instance of
\code{gargle_oauth_client}. For backwards compatibility, for a limited time,
gargle will still accept an "OAuth app" created with \code{\link[httr:oauth_app]{httr::oauth_app()}}.}

\item{api_key}{Optional. API key (a string). Some APIs accept unauthorized,
"token-free" requests for public resources, but only if the request
includes an API key.}

\item{auth_active}{Logical. \code{TRUE} means requests should include a token (and
probably not an API key). \code{FALSE} means requests should include an API key
(and probably not a token).}

\item{cred}{Credentials. Typically populated indirectly via \code{\link[=token_fetch]{token_fetch()}}.}

\item{app}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Replaced by the \code{client}
argument.}
}
\value{
An object of class \link{AuthState}.
}
\description{
Constructor function for objects of class \link{AuthState}.
}
\examples{
my_client <- gargle_oauth_client(
  id = "some_long_client_id",
  secret = "ssshhhhh_its_a_secret",
  name = "my-nifty-oauth-client"
)

init_AuthState(
  package = "my_package",
  client = my_client,
  api_key = "api_key_api_key_api_key",
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/credentials_gce.R
\name{gce_access_token}
\alias{gce_access_token}
\title{Fetch access token for a service account on GCE}
\usage{
gce_access_token(
  scopes = "https://www.googleapis.com/auth/cloud-platform",
  service_account = "default"
)
}
\arguments{
\item{scopes}{A character vector of scopes to request. Pick from those listed
at \url{https://developers.google.com/identity/protocols/oauth2/scopes}.

For certain token flows, the
\code{"https://www.googleapis.com/auth/userinfo.email"} scope is unconditionally
included. This grants permission to retrieve the email address associated
with a token; gargle uses this to index cached OAuth tokens. This grants no
permission to view or send email and is generally considered a low-value
scope.}

\item{service_account}{Name of the GCE service account to use.}
}
\description{
Fetch access token for a service account on GCE
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/token_fetch.R
\name{token_fetch}
\alias{token_fetch}
\title{Fetch a token for the given scopes}
\usage{
token_fetch(scopes = NULL, ...)
}
\arguments{
\item{scopes}{A character vector of scopes to request. Pick from those listed
at \url{https://developers.google.com/identity/protocols/oauth2/scopes}.

For certain token flows, the
\code{"https://www.googleapis.com/auth/userinfo.email"} scope is unconditionally
included. This grants permission to retrieve the email address associated
with a token; gargle uses this to index cached OAuth tokens. This grants no
permission to view or send email and is generally considered a low-value
scope.}

\item{...}{Additional arguments passed to all credential functions.}
}
\value{
An \code{\link[httr:Token-class]{httr::Token}} (often an instance of something
that inherits from \code{httr::Token}) or \code{NULL}.
}
\description{
This is a rather magical function that calls a series of concrete
credential-fetching functions, each wrapped in a \code{tryCatch()}.
\code{token_fetch()} keeps trying until it succeeds or there are no more functions
to try. See the \code{vignette("how-gargle-gets-tokens")} for a full description
of \code{token_fetch()}.
}
\examples{
\dontrun{
token_fetch(scopes = "https://www.googleapis.com/auth/userinfo.email")
}
}
\seealso{
\code{\link[=cred_funs_list]{cred_funs_list()}} reveals the current registry of
credential-fetching functions, in order.

Other credential functions: 
\code{\link{credentials_app_default}()},
\code{\link{credentials_byo_oauth2}()},
\code{\link{credentials_external_account}()},
\code{\link{credentials_gce}()},
\code{\link{credentials_service_account}()},
\code{\link{credentials_user_oauth2}()}
}
\concept{credential functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/credentials_gce.R
\name{gce_instance_service_accounts}
\alias{gce_instance_service_accounts}
\title{List all service accounts available on this GCE instance}
\usage{
gce_instance_service_accounts()
}
\value{
A data frame, where each row is a service account. Due to aliasing,
there is no guarantee that each row represents a distinct service account.
}
\description{
List all service accounts available on this GCE instance
}
\examples{
\dontshow{if (gargle:::is_gce()) withAutoprint(\{ # examplesIf}
credentials_gce()
\dontshow{\}) # examplesIf}
}
\seealso{
The return value is built from a recursive query of the so-called
"directory" of the instance's service accounts as documented in
\url{https://cloud.google.com/compute/docs/metadata/default-metadata-values#vm_instance_metadata}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/response_process.R
\name{response_process}
\alias{response_process}
\alias{response_as_json}
\alias{gargle_error_message}
\title{Process a Google API response}
\usage{
response_process(
  resp,
  error_message = gargle_error_message,
  error_class = NULL,
  remember = TRUE,
  call = caller_env()
)

response_as_json(resp, call = caller_env())

gargle_error_message(resp, call = caller_env())
}
\arguments{
\item{resp}{Object of class \code{response} from \link[httr:httr-package]{httr::httr}.}

\item{error_message}{Function that produces an informative error message from
the primary input, \code{resp}. It must return a character vector.}

\item{error_class}{Optional character vector of error classes to add to the
condition object. These classes are prepended to gargle's default classes
(\code{"gargle_error_request_failed"} and \code{"http_error_{status_code}"}).}

\item{remember}{Whether to remember the most recently processed response.}

\item{call}{The execution environment of a currently running
function, e.g. \code{call = caller_env()}. The corresponding function
call is retrieved and mentioned in error messages as the source
of the error.

You only need to supply \code{call} when throwing a condition from a
helper function which wouldn't be relevant to mention in the
message.

Can also be \code{NULL} or a \link[rlang:topic-defuse]{defused function call} to
respectively not display any call or hard-code a code to display.

For more information about error calls, see \ifelse{html}{\link[rlang:topic-error-call]{Including function calls in error messages}}{\link[rlang:topic-error-call]{Including function calls in error messages}}.}
}
\value{
The content of the request, as a list. An HTTP status code of 204 (No
content) is a special case returning \code{TRUE}.
}
\description{
\code{response_process()} is intended primarily for internal use in client
packages that provide high-level wrappers for users. Typically applied as the
final step in this sequence of calls:
\itemize{
\item Request prepared with \code{\link[=request_build]{request_build()}}.
\item Request made with \code{\link[=request_make]{request_make()}}.
\item Response processed with \code{response_process()}.
}

All that's needed for a successful request is to parse the JSON extracted via
\code{httr::content()}. Therefore, the main point of \code{response_process()} is to
handle less happy outcomes:
\itemize{
\item Status codes in the 400s (client error) and 500s (server error). The
structure of the error payload varies across Google APIs and we try to
create a useful message for all variants we know about.
\item Non-JSON content type, such as HTML.
\item Status code in the 100s (information) or 300s (redirection). These are
unexpected.
}

If \code{response_process()} results in an error, a redacted version of the \code{resp}
input is returned in the condition (auth tokens are removed).
}
\details{
When \code{remember = TRUE} (the default), gargle stores the most recently seen
response internally, for \emph{post hoc} examination. The stored response is
literally just the most recent \code{resp} input, but with auth tokens redacted.
It can be accessed via \code{\link[=gargle_last_response]{gargle_last_response()}}. A companion function
\code{\link[=gargle_last_content]{gargle_last_content()}} returns just the parsed content, which is probably
the most useful form for \emph{post mortem} analysis.

The \code{response_as_json()} helper is exported only as an aid to maintainers who
wish to use their own \code{error_message} function, instead of gargle's built-in
\code{gargle_error_message()}. When implementing a custom \code{error_message}
function, call \code{response_as_json()} immediately on the input in order to
inherit gargle's handling of non-JSON input.
}
\examples{
\dontrun{
# get an OAuth2 token with 'userinfo.email' scope
token <- token_fetch(scopes = "https://www.googleapis.com/auth/userinfo.email")

# see the email associated with this token
req <- gargle::request_build(
  method = "GET",
  path = "v1/userinfo",
  token = token,
  base_url = "https://openidconnect.googleapis.com"
)
resp <- gargle::request_make(req)
response_process(resp)

# make a bad request (this token has incorrect scope)
req <- gargle::request_build(
  method = "GET",
  path = "fitness/v1/users/{userId}/dataSources",
  token = token,
  params = list(userId = 12345)
)
resp <- gargle::request_make(req)
response_process(resp)
}
}
\seealso{
Other requests and responses: 
\code{\link{request_develop}()},
\code{\link{request_make}()}
}
\concept{requests and responses}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/request_retry.R
\name{request_retry}
\alias{request_retry}
\title{Make a Google API request, repeatedly}
\usage{
request_retry(..., max_tries_total = 5, max_total_wait_time_in_seconds = 100)
}
\arguments{
\item{...}{Passed along to \code{\link[=request_make]{request_make()}}.}

\item{max_tries_total}{Maximum number of tries.}

\item{max_total_wait_time_in_seconds}{Total seconds we are willing to
dedicate to waiting, summed across all tries. This is a technical upper
bound and actual cumulative waiting will be less.}
}
\value{
Object of class \code{response} from \link[httr:httr-package]{httr::httr}.
}
\description{
Intended primarily for internal use in client packages that provide
high-level wrappers for users. It is a drop-in substitute for
\code{\link[=request_make]{request_make()}} that also has the ability to retry the request. Codes that
are considered retryable: 408, 429, 500, 502, 503.
}
\details{
Consider an example where we are willing to make a request up to 5 times.

\if{html}{\out{<div class="sourceCode">}}\preformatted{try  1  2    3        4                5
     |--|----|--------|----------------|
wait  1   2      3           4
}\if{html}{\out{</div>}}

There will be up to 5 - 1 = 4 waits and we generally want the waiting period
to get longer, in an exponential way. Such schemes are called exponential
backoff. \code{request_retry()} implements exponential backoff with "full jitter",
where each waiting time is generated from a uniform distribution, where the
interval of support grows exponentially. A common alternative is "equal
jitter", which adds some noise to fixed, exponentially increasing waiting
times.

Either way our waiting times are based on a geometric series, which, by
convention, is usually written in terms of powers of 2:

\if{html}{\out{<div class="sourceCode">}}\preformatted{b, 2b, 4b, 8b, ...
  = b * 2^0, b * 2^1, b * 2^2, b * 2^3, ...
}\if{html}{\out{</div>}}

The terms in this series require knowledge of \code{b}, the so-called exponential
base, and many retry functions and libraries require the user to specify
this. But most users find it easier to declare the total amount of waiting
time they can tolerate for one request. Therefore \code{request_retry()} asks for
that instead and solves for \code{b} internally. This is inspired by the Opnieuw
Python library for retries. Opnieuw's interface is designed to eliminate
uncertainty around:
\itemize{
\item Units: Is this thing given in seconds? minutes? milliseconds?
\item Ambiguity around how things are counted: Are we starting at 0 or 1?
Are we counting tries or just the retries?
\item Non-intuitive required inputs, e.g., the exponential base.
}

Let \emph{n} be the total number of tries we're willing to make (the argument
\code{max_tries_total}) and let \emph{W} be the total amount of seconds we're willing
to dedicate to making and retrying this request (the argument
\code{max_total_wait_time_in_seconds}). Here's how we determine \emph{b}:

\if{html}{\out{<div class="sourceCode">}}\preformatted{sum_\{i=0\}^(n - 1) b * 2^i = W
b * sum_\{i=0\}^(n - 1) 2^i = W
       b * ( (2 ^ n) - 1) = W
                        b = W / ( (2 ^ n) - 1)
}\if{html}{\out{</div>}}
}
\section{Special cases}{

\code{request_retry()} departs from exponential backoff in three special cases:
\itemize{
\item It actually implements \emph{truncated} exponential backoff. There is a floor
and a ceiling on random wait times.
\item \code{Retry-After} header: If the response has a header named \code{Retry-After}
(case-insensitive), it is assumed to provide a non-negative integer
indicating the number of seconds to wait. If present, we wait this many
seconds and do not generate a random waiting time. (In theory, this header
can alternatively provide a datetime after which to retry, but we have no
first-hand experience with this variant for a Google API.)
\item Sheets API quota exhaustion: In the course of googlesheets4 development,
we've grown very familiar with the \verb{429 RESOURCE_EXHAUSTED} error. As of
2023-04-15, the Sheets API v4 has a limit of 300 requests per minute per
project and 60 requests per minute per user per project. Limits for reads
and writes are tracked separately. In our experience, the "60 (read or
write) requests per minute per user" limit is the one you hit most often.
If we detect this specific failure, the first wait time is a bit more than
one minute, then we revert to exponential backoff.
}
}

\examples{
\dontrun{
req <- gargle::request_build(
  method = "GET",
  path = "path/to/the/resource",
  token = "PRETEND_I_AM_TOKEN"
)
gargle::request_retry(req)
}
}
\seealso{
\itemize{
\item \url{https://aws.amazon.com/blogs/architecture/exponential-backoff-and-jitter/}
\item \verb{https://www.channable.com/tech/opnieuw-a-simple-and-intuitive-retrying-library-for-python}
\item \url{https://github.com/channable/opnieuw}
\item \url{https://docs.cloud.google.com/storage/docs/retry-strategy}
\item \url{https://www.rfc-editor.org/rfc/rfc7231#section-7.1.3}
\item \url{https://developers.google.com/sheets/api/limits}
\item \url{https://googleapis.dev/python/google-api-core/latest/retry.html}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/credentials_gce.R
\name{credentials_gce}
\alias{credentials_gce}
\title{Get a token from the Google metadata server}
\usage{
credentials_gce(
  scopes = "https://www.googleapis.com/auth/cloud-platform",
  service_account = "default",
  ...
)
}
\arguments{
\item{scopes}{A character vector of scopes to request. Pick from those listed
at \url{https://developers.google.com/identity/protocols/oauth2/scopes}.

For certain token flows, the
\code{"https://www.googleapis.com/auth/userinfo.email"} scope is unconditionally
included. This grants permission to retrieve the email address associated
with a token; gargle uses this to index cached OAuth tokens. This grants no
permission to view or send email and is generally considered a low-value
scope.}

\item{service_account}{Name of the GCE service account to use.}

\item{...}{Additional arguments passed to all credential functions.}
}
\value{
A \code{\link[=GceToken]{GceToken()}} or \code{NULL}.
}
\description{
If your code is running on Google Cloud, we can often obtain a token for an
attached service account directly from a metadata server. This is more secure
than working with an explicit a service account key, as
\code{\link[=credentials_service_account]{credentials_service_account()}} does, and is the preferred method of auth for
workloads running on Google Cloud.

The most straightforward scenario is when you are working in a VM on Google
Compute Engine and it's OK to use the default service account. This should
"just work" automatically.

\code{credentials_gce()} supports other use cases (such as GKE Workload Identity),
but may require some explicit setup, such as:
\itemize{
\item Create a service account, grant it appropriate scopes(s) and IAM roles,
attach it to the target resource. This prep work happens outside of R, e.g.,
in the Google Cloud Console. On the R side, provide the email address of this
appropriately configured service account via \code{service_account}.
\item Specify details for constructing the root URL of the metadata service:
\itemize{
\item The logical option \code{"gargle.gce.use_ip"}. If undefined, this defaults to
\code{FALSE}.
\item The environment variable \code{GCE_METADATA_URL} is consulted when
\code{"gargle.gce.use_ip"} is \code{FALSE}. If undefined, the default is
\code{metadata.google.internal}.
\item The environment variable \code{GCE_METADATA_IP} is consulted when
\code{"gargle.gce.use_ip"} is \code{TRUE}. If undefined, the default is
\verb{169.254.169.254}.
}
\item Change (presumably increase) the timeout for requests to the metadata
server via the \code{"gargle.gce.timeout"} global option. This timeout is given in
seconds and is set to a value (strategy, really) that often works well in
practice. However, in some cases it may be necessary to increase the timeout
with code such as:
}

\if{html}{\out{<div class="sourceCode r">}}\preformatted{options(gargle.gce.timeout = 3)
}\if{html}{\out{</div>}}

For details on specific use cases, such as Google Kubernetes Engine (GKE),
see \code{vignette("non-interactive-auth")}.
}
\examples{
\dontrun{
credentials_gce()
}
}
\seealso{
A related auth flow that can be used on certain non-Google cloud
providers is workload identity federation, which is implemented in
\code{\link[=credentials_external_account]{credentials_external_account()}}.

\url{https://docs.cloud.google.com/compute/docs/access/service-accounts}

\url{https://docs.cloud.google.com/iam/docs/best-practices-service-accounts}

How to attach a service account to a resource:
\url{https://cloud.google.com/iam/docs/impersonating-service-accounts#attaching-to-resources}

\url{https://docs.cloud.google.com/kubernetes-engine/docs/concepts/workload-identity}

\url{https://docs.cloud.google.com/kubernetes-engine/docs/how-to/workload-identity}

\url{https://docs.cloud.google.com/compute/docs/metadata/overview}

Other credential functions: 
\code{\link{credentials_app_default}()},
\code{\link{credentials_byo_oauth2}()},
\code{\link{credentials_external_account}()},
\code{\link{credentials_service_account}()},
\code{\link{credentials_user_oauth2}()},
\code{\link{token_fetch}()}
}
\concept{credential functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils-ui.R
\name{bulletize}
\alias{bulletize}
\title{Abbreviate a bullet list neatly}
\usage{
bulletize(x, bullet = "*", n_show = 5, n_fudge = 2)
}
\description{
For internal use in gargle, googledrive, and googlesheets4 (for now).
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/credentials_byo_oauth2.R
\name{credentials_byo_oauth2}
\alias{credentials_byo_oauth2}
\title{Load a user-provided token}
\usage{
credentials_byo_oauth2(scopes = NULL, token, ...)
}
\arguments{
\item{scopes}{A character vector of scopes to request. Pick from those listed
at \url{https://developers.google.com/identity/protocols/oauth2/scopes}.

For certain token flows, the
\code{"https://www.googleapis.com/auth/userinfo.email"} scope is unconditionally
included. This grants permission to retrieve the email address associated
with a token; gargle uses this to index cached OAuth tokens. This grants no
permission to view or send email and is generally considered a low-value
scope.}

\item{token}{A token with class \link[httr:Token-class]{Token2.0} or an object of
httr's class \code{request}, i.e. a token that has been prepared with
\code{\link[httr:config]{httr::config()}} and has a \link[httr:Token-class]{Token2.0} in the
\code{auth_token} component.}

\item{...}{Additional arguments passed to all credential functions.}
}
\value{
An \link[httr:Token-class]{Token2.0}.
}
\description{
This function is designed to pass its \code{token} input through, after doing a
few checks and some light processing:
\itemize{
\item If \code{token} has class \code{request}, i.e. it is a token that has been prepared
with \code{\link[httr:config]{httr::config()}}, the \code{auth_token} component is extracted. For
example, such input could be returned by \code{googledrive::drive_token()}
or \code{bigrquery::bq_token()}.
\item If \code{token} is an instance of \code{Gargle2.0} (so: a gargle-obtained user
token), checks that it appears to be a Google OAuth token, based on its
embedded \code{oauth_endpoint}. Refreshes the token, if it's refreshable.
\item Returns the \code{token}.
}

There is no point in providing \code{scopes}. They are ignored because the
\code{scopes} associated with the token have already been baked in to the token
itself and gargle does not support incremental authorization. The main point
of \code{credentials_byo_oauth2()} is to allow \code{token_fetch()} (and packages that
wrap it) to accommodate a "bring your own token" workflow.

This also makes it possible to obtain a token with one package and then
register it for use with another package. For example, the default scope
requested by googledrive is also sufficient for operations available in
googlesheets4. You could use a shared token like so:

\if{html}{\out{<div class="sourceCode">}}\preformatted{library(googledrive)
library(googlesheets4)
drive_auth(email = "jane_doe@example.com")
gs4_auth(token = drive_token())
# work with both packages freely now, with the same identity
}\if{html}{\out{</div>}}
}
\examples{
\dontrun{
# assume `my_token` is a Token2.0 object returned by a function such as
# credentials_user_oauth2()
credentials_byo_oauth2(token = my_token)
}
}
\seealso{
Other credential functions: 
\code{\link{credentials_app_default}()},
\code{\link{credentials_external_account}()},
\code{\link{credentials_gce}()},
\code{\link{credentials_service_account}()},
\code{\link{credentials_user_oauth2}()},
\code{\link{token_fetch}()}
}
\concept{credential functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/request_develop.R
\name{request_develop}
\alias{request_develop}
\alias{request_build}
\title{Build a Google API request}
\usage{
request_develop(
  endpoint,
  params = list(),
  base_url = "https://www.googleapis.com"
)

request_build(
  method = "GET",
  path = "",
  params = list(),
  body = list(),
  token = NULL,
  key = NULL,
  base_url = "https://www.googleapis.com"
)
}
\arguments{
\item{endpoint}{List of information about the target endpoint or, in
Google's vocabulary, the target "method". Presumably prepared from the
\href{https://developers.google.com/discovery/v1/getting_started#background-resources}{Discovery Document}
for the target API.}

\item{params}{Named list. Values destined for URL substitution, the query,
or, for \code{request_develop()} only, the body. For \code{request_build()}, body
parameters must be passed via the \code{body} argument.}

\item{base_url}{Character.}

\item{method}{Character. An HTTP verb, such as \code{GET} or \code{POST}.}

\item{path}{Character. Path to the resource, not including the API's
\code{base_url}. Examples: \code{drive/v3/about} or \code{drive/v3/files/{fileId}}. The
\code{path} can be a template, i.e. it can include variables inside curly
brackets, such as \code{{fileId}} in the example. Such variables are substituted
by \code{request_build()}, using named parameters found in \code{params}.}

\item{body}{List. Values to send in the API request body.}

\item{token}{Token, ready for inclusion in a request, i.e. prepared with
\code{\link[httr:config]{httr::config()}}.}

\item{key}{API key. Needed for requests that don't contain a token. For more,
see Google's document Credentials, access, security, and identity
(\verb{https://support.google.com/googleapi/answer/6158857?hl=en&ref_topic=7013279}).
A key can be passed as a named component of \code{params}, but note that the
formal argument \code{key} will clobber it, if non-\code{NULL}.}
}
\value{
\code{request_develop()}: \code{list()} with components \code{method}, \code{path}, \code{params},
\code{body}, and \code{base_url}.

\code{request_build()}: \code{list()} with components \code{method}, \code{path}
(post-substitution), \code{query} (the input \code{params} not used in URL
substitution), \code{body}, \code{token}, \code{url} (the full URL, post-substitution,
including the query).
}
\description{
Intended primarily for internal use in client packages that provide
high-level wrappers for users. The \code{vignette("request-helper-functions")}
describes how one might use these functions inside a wrapper package.
}
\section{\code{request_develop()}}{


Combines user input (\code{params}) with information about an API endpoint.
\code{endpoint} should contain these components:
\itemize{
\item \code{path}: See documentation for argument.
\item \code{method}: See documentation for argument.
\item \code{parameters}: Compared with \code{params} supplied by user. An error is
thrown if user-supplied \code{params} aren't named in
\code{endpoint$parameters} or if user fails to supply all required
parameters. In the return value, body parameters are separated from
those destined for path substitution or the query.
}

The return value is typically used as input to \code{request_build()}.
}

\section{\code{request_build()}}{


Builds a request, in a purely mechanical sense. This function does nothing
specific to any particular Google API or endpoint.
\itemize{
\item Use with the output of \code{request_develop()} or with hand-crafted input.
\item \code{params} are used for variable substitution in \code{path}. Leftover \code{params}
that are not bound by the \code{path} template automatically become HTTP
query parameters.
\item Adds an API key to the query iff \code{token = NULL} and removes the API key
otherwise. Client packages should generally pass their own API key in, but
note that \code{\link[=gargle_api_key]{gargle_api_key()}} is available for small-scale experimentation.
}

See \code{googledrive::generate_request()} for an example of usage in a client
package. googledrive has an internal list of selected endpoints, derived from
the Drive API Discovery Document
(\verb{https://www.googleapis.com/discovery/v1/apis/drive/v3/rest}),
exposed via \code{googledrive::drive_endpoints()}. An element from such a list is
the expected input for \code{endpoint}. \code{googledrive::generate_request()} is a
wrapper around \code{request_develop()} and \code{request_build()} that inserts a
googledrive-managed API key and some logic about Team Drives. All user-facing
functions use \code{googledrive::generate_request()} under the hood.
}

\examples{
\dontrun{
## Example with a prepared endpoint
ept <- googledrive::drive_endpoints("drive.files.update")[[1]]
req <- request_develop(
  ept,
  params = list(
    fileId = "abc",
    addParents = "123",
    description = "Exciting File"
  )
)
req

req <- request_build(
  method = req$method,
  path = req$path,
  params = req$params,
  body = req$body,
  token = "PRETEND_I_AM_A_TOKEN"
)
req

## Example with no previous knowledge of the endpoint
## List a file's comments
## https://developers.google.com/drive/v3/reference/comments/list
req <- request_build(
  method = "GET",
  path = "drive/v3/files/{fileId}/comments",
  params = list(
    fileId = "your-file-id-goes-here",
    fields = "*"
  ),
  token = "PRETEND_I_AM_A_TOKEN"
)
req

# Example with no previous knowledge of the endpoint and no token
# use an API key for which the Places API is enabled!
API_KEY <- "1234567890"

# get restaurants close to a location in Vancouver, BC
req <- request_build(
  method = "GET",
  path = "maps/api/place/nearbysearch/json",
  params = list(
    location = "49.268682,-123.167117",
    radius = 100,
    type = "restaurant"
  ),
  key = API_KEY,
  base_url = "https://maps.googleapis.com"
)
resp <- request_make(req)
out <- response_process(resp)
vapply(out$results, function(x) x$name, character(1))
}
}
\seealso{
Other requests and responses: 
\code{\link{request_make}()},
\code{\link{response_process}()}
}
\concept{requests and responses}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gargle-package.R, R/utils-ui.R
\name{gargle_options}
\alias{gargle_options}
\alias{gargle_oauth_email}
\alias{gargle_oob_default}
\alias{gargle_oauth_cache}
\alias{gargle_oauth_client_type}
\alias{gargle_verbosity}
\alias{local_gargle_verbosity}
\alias{with_gargle_verbosity}
\title{Options consulted by gargle}
\usage{
gargle_oauth_email()

gargle_oob_default()

gargle_oauth_cache()

gargle_oauth_client_type()

gargle_verbosity()

local_gargle_verbosity(level, env = caller_env())

with_gargle_verbosity(level, code)
}
\arguments{
\item{level}{Verbosity level: "debug" > "info" > "silent"}

\item{env}{The environment to use for scoping}

\item{code}{Code to execute with specified verbosity level}
}
\description{
Wrapper functions around options consulted by gargle, which provide:
\itemize{
\item A place to hang documentation.
\item The mechanism for setting a default.
}

If the built-in defaults don't suit you, set one or more of these options.
Typically, this is done in the \code{.Rprofile} startup file, with code along
these lines:

\if{html}{\out{<div class="sourceCode">}}\preformatted{options(
  gargle_oauth_email = "jane@example.com",
  gargle_oauth_cache = "/path/to/folder/that/does/not/sync/to/cloud"
)
}\if{html}{\out{</div>}}
}
\section{\code{gargle_oauth_email}}{

\code{gargle_oauth_email()} returns the option named "gargle_oauth_email", which
is undefined by default. If set, this option should be one of:
\itemize{
\item An actual email address corresponding to your preferred Google identity.
Example:\code{janedoe@gmail.com}.
\item A glob pattern that indicates your preferred Google domain.
Example:\verb{*@example.com}.
\item \code{TRUE} to allow email and OAuth token auto-discovery, if exactly one
suitable token is found in the cache.
\item \code{FALSE} or \code{NA} to force the OAuth dance in the browser.
}
}

\section{\code{gargle_oob_default}}{

\code{gargle_oob_default()} returns \code{TRUE} unconditionally on RStudio Server,
Posit Workbench, Posit Cloud, or Google Colaboratory, since it is not
possible to launch a local web server in these contexts. In this case, for
the final step of the OAuth dance, the user is redirected to a specific URL
where they must copy a code and paste it back into the R session.

In all other contexts, \code{gargle_oob_default()} consults the option named
\code{"gargle_oob_default"}, then the option named \code{"httr_oob_default"}, and
eventually defaults to \code{FALSE}.

"oob" stands for out-of-band. Read more about out-of-band authentication in
the vignette \code{vignette("auth-from-web")}.
}

\section{\code{gargle_oauth_cache}}{

\code{gargle_oauth_cache()} returns the option named "gargle_oauth_cache",
defaulting to \code{NA}. If defined, the option must be set to a logical value or
a string. \code{TRUE} means to cache using the default user-level cache file,
\verb{~/.R/gargle/gargle-oauth}, \code{FALSE} means don't cache, and \code{NA} means to
guess using some sensible heuristics.
}

\section{\code{gargle_oauth_client_type}}{

\code{gargle_oauth_client_type()} returns the option named
"gargle_oauth_client_type", if defined. If defined, the option must be either
"installed" or "web". If the option is not defined, the function returns:
\itemize{
\item "web" on RStudio Server, Posit Workbench, Posit Cloud, or Google
Colaboratory
\item "installed" otherwise
}

Primarily intended to help infer the most suitable OAuth client type when a
user is relying on a built-in client, such as the tidyverse client used by
packages like bigrquery, googledrive, and googlesheets4.
}

\section{\code{gargle_verbosity}}{

\code{gargle_verbosity()} returns the option named "gargle_verbosity", which
determines gargle's verbosity. There are three possible values, inspired by
the logging levels of log4j:
\itemize{
\item "debug": Fine-grained information helpful when debugging, e.g. figuring out
how \code{token_fetch()} is working through the registry of credential
functions. Previously, this was activated by setting an option named
"gargle_quiet" to \code{FALSE}.
\item "info" (default): High-level information that a typical user needs to see.
Since typical gargle usage is always indirect, i.e. gargle is called by
another package, gargle itself is very quiet. There are very few messages
emitted when \code{gargle_verbosity = "info"}.
\item "silent": No messages at all. However, warnings or errors are still thrown
normally.
}
}

\examples{
gargle_oauth_email()
gargle_oob_default()
gargle_oauth_cache()
gargle_oauth_client_type()
gargle_verbosity()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/credentials_service_account.R
\name{check_is_service_account}
\alias{check_is_service_account}
\title{Check for a service account}
\usage{
check_is_service_account(path, hint, call = caller_env())
}
\arguments{
\item{path}{JSON identifying the service account, in one of the forms
supported for the \code{txt} argument of \code{\link[jsonlite:fromJSON]{jsonlite::fromJSON()}} (typically, a
file path or JSON string).}

\item{hint}{The relevant function to call for configuring an OAuth client.}

\item{call}{The execution environment of a currently running
function, e.g. \code{call = caller_env()}. The corresponding function
call is retrieved and mentioned in error messages as the source
of the error.

You only need to supply \code{call} when throwing a condition from a
helper function which wouldn't be relevant to mention in the
message.

Can also be \code{NULL} or a \link[rlang:topic-defuse]{defused function call} to
respectively not display any call or hard-code a code to display.

For more information about error calls, see \ifelse{html}{\link[rlang:topic-error-call]{Including function calls in error messages}}{\link[rlang:topic-error-call]{Including function calls in error messages}}.}
}
\value{
Nothing. Exists purely to throw an error.
}
\description{
This pre-checks information provided to a high-level, user-facing auth
function, such as \code{googledrive::drive_auth()}, before passing the user's
input along to \code{\link[=token_fetch]{token_fetch()}}, which is designed to silently swallow errors.
Some users are confused about the difference between an OAuth client and a
service account and they provide the (path to the) JSON for one, when the
other is what's actually expected.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gargle_api_key.R, R/gargle_oauth_client.R
\name{internal-assets}
\alias{internal-assets}
\alias{tidyverse_api_key}
\alias{tidyverse_client}
\alias{tidyverse_app}
\title{Assets for internal use}
\usage{
tidyverse_api_key()

tidyverse_client(type = NULL)

tidyverse_app()
}
\description{
Assets for use inside specific packages maintained by the tidyverse team.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/credentials_gce.R
\name{GceToken}
\alias{GceToken}
\title{Token for use on Google Compute Engine instances}
\description{
Token for use on Google Compute Engine instances

Token for use on Google Compute Engine instances
}
\details{
This class uses the metadata service available on GCE VMs to fetch access
tokens. Not intended for direct use. See \code{\link[=credentials_gce]{credentials_gce()}} instead.
}
\keyword{internal}
\section{Super classes}{
\code{\link[httr:Token]{httr::Token}} -> \code{\link[httr:Token2.0]{httr::Token2.0}} -> \code{GceToken}
}
\section{Methods}{
\subsection{Public methods}{
\itemize{
\item \href{#method-GceToken-new}{\code{GceToken$new()}}
\item \href{#method-GceToken-init_credentials}{\code{GceToken$init_credentials()}}
\item \href{#method-GceToken-refresh}{\code{GceToken$refresh()}}
\item \href{#method-GceToken-can_refresh}{\code{GceToken$can_refresh()}}
\item \href{#method-GceToken-format}{\code{GceToken$format()}}
\item \href{#method-GceToken-print}{\code{GceToken$print()}}
\item \href{#method-GceToken-cache}{\code{GceToken$cache()}}
\item \href{#method-GceToken-load_from_cache}{\code{GceToken$load_from_cache()}}
\item \href{#method-GceToken-revoke}{\code{GceToken$revoke()}}
\item \href{#method-GceToken-validate}{\code{GceToken$validate()}}
\item \href{#method-GceToken-clone}{\code{GceToken$clone()}}
}
}
\if{html}{\out{
<details open><summary>Inherited methods</summary>
<ul>
<li><span class="pkg-link" data-pkg="httr" data-topic="Token" data-id="hash"><a href='../../httr/html/Token.html#method-Token-hash'><code>httr::Token$hash()</code></a></span></li>
<li><span class="pkg-link" data-pkg="httr" data-topic="Token2.0" data-id="sign"><a href='../../httr/html/Token2.0.html#method-Token2.0-sign'><code>httr::Token2.0$sign()</code></a></span></li>
</ul>
</details>
}}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-GceToken-new"></a>}}
\if{latex}{\out{\hypertarget{method-GceToken-new}{}}}
\subsection{Method \code{new()}}{
Get an access for a GCE service account.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{GceToken$new(params)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{params}}{A list of parameters for \code{fetch_gce_access_token()}.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
A GceToken.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-GceToken-init_credentials"></a>}}
\if{latex}{\out{\hypertarget{method-GceToken-init_credentials}{}}}
\subsection{Method \code{init_credentials()}}{
Request an access token.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{GceToken$init_credentials()}\if{html}{\out{</div>}}
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-GceToken-refresh"></a>}}
\if{latex}{\out{\hypertarget{method-GceToken-refresh}{}}}
\subsection{Method \code{refresh()}}{
Refreshes the token. In this case, that just means "ask again
for an access token".
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{GceToken$refresh()}\if{html}{\out{</div>}}
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-GceToken-can_refresh"></a>}}
\if{latex}{\out{\hypertarget{method-GceToken-can_refresh}{}}}
\subsection{Method \code{can_refresh()}}{
Placeholder implementation of required method. Returns \code{TRUE}.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{GceToken$can_refresh()}\if{html}{\out{</div>}}
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-GceToken-format"></a>}}
\if{latex}{\out{\hypertarget{method-GceToken-format}{}}}
\subsection{Method \code{format()}}{
Format a \code{\link[=GceToken]{GceToken()}}.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{GceToken$format(...)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{...}}{Not used.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-GceToken-print"></a>}}
\if{latex}{\out{\hypertarget{method-GceToken-print}{}}}
\subsection{Method \code{print()}}{
Print a \code{\link[=GceToken]{GceToken()}}.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{GceToken$print(...)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{...}}{Not used.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-GceToken-cache"></a>}}
\if{latex}{\out{\hypertarget{method-GceToken-cache}{}}}
\subsection{Method \code{cache()}}{
Placeholder implementation of required method.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{GceToken$cache()}\if{html}{\out{</div>}}
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-GceToken-load_from_cache"></a>}}
\if{latex}{\out{\hypertarget{method-GceToken-load_from_cache}{}}}
\subsection{Method \code{load_from_cache()}}{
Placeholder implementation of required method.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{GceToken$load_from_cache()}\if{html}{\out{</div>}}
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-GceToken-revoke"></a>}}
\if{latex}{\out{\hypertarget{method-GceToken-revoke}{}}}
\subsection{Method \code{revoke()}}{
Placeholder implementation of required method.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{GceToken$revoke()}\if{html}{\out{</div>}}
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-GceToken-validate"></a>}}
\if{latex}{\out{\hypertarget{method-GceToken-validate}{}}}
\subsection{Method \code{validate()}}{
Placeholder implementation of required method
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{GceToken$validate()}\if{html}{\out{</div>}}
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-GceToken-clone"></a>}}
\if{latex}{\out{\hypertarget{method-GceToken-clone}{}}}
\subsection{Method \code{clone()}}{
The objects of this class are cloneable with this method.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{GceToken$clone(deep = FALSE)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{deep}}{Whether to make a deep clone.}
}
\if{html}{\out{</div>}}
}
}
}
